#include "MAPL_Generic.h"

!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 910.1     !
!-------------------------------------------------------------------------
!BOP
!
! !MODULE: CARMAchem_GridCompMod - The Community Aerosol and Radiation Model
!                                  for Atmospheres
!
! !INTERFACE:
!
   MODULE CARMAchem_GridCompMod
!
! !USES:
!
   USE ESMF
   USE MAPL_Mod
   USE MAPL_GenericMod
   USE Chem_Mod 	                        ! Chemistry Base Class

   USE CARMA_GridCompMod                        ! ESMF parent component
   USE CARMA_UtilMod
   USE Chem_UtilMod
   USE m_inpak90	                        ! Resource file management

!  CARMA Specific Methods
   use carma_precision_mod 
   use carma_constants_mod 
   use carma_enums_mod 
   use carma_types_mod 
   use carmaelement_mod
   use carmagroup_mod
   use carmastate_mod
   use carma_mod

   IMPLICIT NONE
   PRIVATE
!
! !PUBLIC MEMBER FUNCTIONS:

   PUBLIC SetServices
!
! !DESCRIPTION: 
!
!  {\tt CARMAchem_GridComp} is an ESMF gridded component for the Community
!  Aerosol and Radiation Model for Atmospheres aerosol and cloud
!  microphysics packages.
!
!  Developed for GEOS-5 release Eros-beta7p6 and later.
!
! !REVISION HISTORY:
!
!  31Jul2006  da Silva  Created the GMI stub.
!  11Dec2007  Nielsen   Real code for Eros-beta7p17.
!  18May2009  Colarco   Developed based on GMIchem_GridComp.F90
!
!EOP
!-------------------------------------------------------------------------

  TYPE CARMAchem_State
     PRIVATE
     TYPE(CARMA_GridComp), POINTER     :: gcCARMA   => null()
     TYPE(CARMA_Registry), POINTER     :: CARMAReg  => null()
     TYPE(Chem_Array), POINTER         :: qa(:)     => null()
  END TYPE CARMAchem_State

  TYPE CARMAchem_WRAP
     TYPE (CARMAchem_State), pointer :: PTR => null()
  END TYPE CARMAchem_WRAP

CONTAINS


!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 910.1     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE: SetServices --- Sets IRF services for CARMA Grid Component
!
! !INTERFACE:

   SUBROUTINE SetServices ( GC, RC )

! !ARGUMENTS:

    type(ESMF_GridComp)                :: GC  ! gridded component
    integer, intent(OUT)               :: RC  ! return code

! !DESCRIPTION: Sets Initialize, Run and Finalize services. 
!
! !REVISION HISTORY:
!
!  31Jul2006  da Silva  First crack.
!  18May2009  Colarco   Adapted for CARMA
!
!EOP
!-------------------------------------------------------------------------

!   ErrLog Variables
!   ----------------
    character(len=ESMF_MAXSTR)      :: IAm = 'SetServices'
    integer                         :: STATUS
    character(len=ESMF_MAXSTR)      :: COMP_NAME

!   Local derived type aliases
!   --------------------------
    type (ESMF_Config)                :: CF
    type (CARMAchem_State), pointer   :: state   ! internal, that is
    type (CARMAchem_wrap)             :: wrap
    type(CARMA_Registry), pointer     :: r

    integer                           :: n, i_XX, j_XX, i, j, iq
    character(len=ESMF_MAXSTR)        :: binstr

!                              ------------

!   Get my name and set-up traceback handle
!   ---------------------------------------
    call ESMF_GridCompGet( GC, NAME=COMP_NAME, RC=STATUS )
    VERIFY_(STATUS)
    Iam = TRIM(COMP_NAME) // '::' // TRIM(Iam)

!   Wrap internal state for storing in GC; rename legacyState
!   -------------------------------------
    allocate ( state, stat=STATUS )
    VERIFY_(STATUS)
    wrap%ptr => state
 
!   Start by loading the CARMA Registry
!   -----------------------------------
    allocate ( state%CARMAReg )
    call registry_ ( state%CARMAReg )
    VERIFY_(STATUS)
    call registry_print_ ( state%CARMAReg )

    r => state%CARMAReg   ! short hand

!                       ------------------------
!                       ESMF Functional Services
!                       ------------------------

!   Set the Initialize, Run, Finalize entry points
!   ----------------------------------------------
     if ( r%doing_CARMA ) then

        if (MAPL_AM_I_ROOT()) print *, trim(Iam)//': ACTIVE'

        call MAPL_GridCompSetEntryPoint ( GC, ESMF_METHOD_INITIALIZE,  Initialize_, &
             RC=STATUS)
        VERIFY_(STATUS)
        
        call MAPL_GridCompSetEntryPoint ( GC,  ESMF_METHOD_RUN,  Run_,        &
             RC=STATUS)
        VERIFY_(STATUS)
        
        call MAPL_GridCompSetEntryPoint ( GC,  ESMF_METHOD_FINALIZE,  Finalize_,  &
             RC=STATUS)
        VERIFY_(STATUS)
        
!       Store internal state in GC
!       --------------------------
        call ESMF_UserCompSetInternalState ( GC, 'CARMA_state', wrap, STATUS )
        VERIFY_(STATUS)

     else

        if (MAPL_AM_I_ROOT()) &
        print *, trim(Iam)//': NOT ACTIVE, defaulting to Generic No-op stubs'

     endif
  
!                         ------------------
!                         GEOS Data Services
!                         ------------------
!
! !IMPORT STATE:
#include "CARMA_ImportSpec___.h"

! !INTERNAL STATE:

!
!  NOTES: 
!    1)  vtitle as it stands is as the CF definition of long name.
!        I may need to create a "standard name" in chemReg and pass
!        this to GEOS Generic
!    2)  Host model MUST provide convective transport as well
!
!    Convention for tracer names is CARMA::groupname::elemname::XXX
!    where XXX is the bin number.
!    Will need to add space for the NGAS array
!
     if ( r%doing_CARMA ) then

!     Add particle tracers
      do j = 1, r%NELEM
       do i = 1, r%NBIN
        write(binstr,'(i3)') i
        binstr = adjustl(binstr)
        if(i .lt. 10)  binstr = '0'//binstr
        if(i .lt. 100) binstr = '0'//binstr

        iq = i + (j-1)*r%NBIN
        r%vname(iq) = trim(r%groupname(r%igroup(j)))// '::' &
                   // trim(r%elemname(j))// '::'            &
                   // trim(binstr)
        call MAPL_AddInternalSpec(GC,                              &
               SHORT_NAME  = trim(COMP_NAME)// '::'                &
                          // trim(r%groupname(r%igroup(j)))// '::' &
                          // trim(r%elemname(j))// '::'            &
                          // trim(binstr),                         &
               LONG_NAME   = trim(COMP_NAME)// '::'                &
                          // trim(r%groupname(r%igroup(j)))// '::' &
                          // trim(r%elemname(j))// '::'            &
                          // trim(binstr),                         &
               UNITS       = 'kg/kg',                              &     ! placeholder
               FRIENDLYTO  = 'DYNAMICS:TURBULENCE',          &
               DIMS        = MAPL_DimsHorzVert,                    &
               VLOCATION   = MAPL_VLocationCenter,     RC=STATUS  )
          VERIFY_(STATUS)

       end do
      end do
       
!     Add gas tracers
      do j = 1, r%NGAS

        iq = j + r%NBIN*r%NELEM
        r%vname(iq) = trim(r%gasname(j))

        call MAPL_AddInternalSpec(GC,                              &
               SHORT_NAME  = trim(COMP_NAME)// '::'                &
                          // trim(r%gasname(j)),                   &
               LONG_NAME   = trim(COMP_NAME)// '::'                &
                          // trim(r%gasname(j)),                   &
               UNITS       = 'kg/kg',                              &     ! placeholder
               FRIENDLYTO  = 'DYNAMICS:TURBULENCE',                &
               DIMS        = MAPL_DimsHorzVert,                    &
               VLOCATION   = MAPL_VLocationCenter,     RC=STATUS  )
          VERIFY_(STATUS)

      end do

!     Non-advected tracers
!     Add the old temperature
      iq = r%NBIN*r%NELEM + r%NGAS + 1
      r%vname(iq) = 't_old'

      call MAPL_AddInternalSpec(GC,                              &
             SHORT_NAME  = trim(COMP_NAME)// '::'                &
                        // trim(r%vname(iq)),                    &
             LONG_NAME   = trim(COMP_NAME)// '::'                &
                        // trim(r%vname(iq)),                    &
             UNITS       = 'K',                                  &     ! placeholder
             ADD2EXPORT      = .TRUE.,                           &
             DIMS        = MAPL_DimsHorzVert,                    &
             VLOCATION   = MAPL_VLocationCenter,     RC=STATUS  )
      VERIFY_(STATUS)


!     Add the prior time step gas tracers
      do j = 1, r%NGAS

        iq = j + r%NBIN*r%NELEM + r%NGAS + 1
        r%vname(iq) = trim(r%gasname(j))//'_old'

        call MAPL_AddInternalSpec(GC,                              &
               SHORT_NAME  = trim(COMP_NAME)// '::'                &
                          // trim(r%vname(iq)),                     &
               LONG_NAME   = trim(COMP_NAME)// '::'                &
                          // trim(r%vname(iq)),                     &
               UNITS       = 'kg/kg',                              &     ! placeholder
               ADD2EXPORT      = .TRUE.,                           &
               DIMS        = MAPL_DimsHorzVert,                    &
               VLOCATION   = MAPL_VLocationCenter,     RC=STATUS  )
          VERIFY_(STATUS)

      end do
       
!     Add the prior time step saturation wrt liquid
      do j = 1, r%NGAS

        iq = j + r%NBIN*r%NELEM + r%NGAS + r%NGAS + 1
        r%vname(iq) = 'satliq_'//trim(r%gasname(j))//'_old'
        call MAPL_AddInternalSpec(GC,                              &
               SHORT_NAME  = trim(COMP_NAME)// '::'                &
                          // trim(r%vname(iq)),                     &
               LONG_NAME   = trim(COMP_NAME)// '::'                &
                          // trim(r%vname(iq)),                     &
               UNITS       = '1',                                  &     ! placeholder
               ADD2EXPORT      = .TRUE.,                           &
               DIMS        = MAPL_DimsHorzVert,                    &
               VLOCATION   = MAPL_VLocationCenter,     RC=STATUS  )
          VERIFY_(STATUS)

      end do
       
!     Add the prior time step saturation wrt ice
      do j = 1, r%NGAS

        iq = j + r%NBIN*r%NELEM + r%NGAS + r%NGAS + r%NGAS + 1
        r%vname(iq) = 'satice_'//trim(r%gasname(j))//'_old'

        call MAPL_AddInternalSpec(GC,                              &
               SHORT_NAME  = trim(COMP_NAME)// '::'                &
                          // trim(r%vname(iq)),                     &
               LONG_NAME   = trim(COMP_NAME)// '::'                &
                          // trim(r%vname(iq)),                     &
               UNITS       = '1',                                  &     ! placeholder
               ADD2EXPORT      = .TRUE.,                           &
               DIMS        = MAPL_DimsHorzVert,                    &
               VLOCATION   = MAPL_VLocationCenter,     RC=STATUS  )
          VERIFY_(STATUS)

      end do
       
     endif

! !EXPORT STATE:
!   This bundle is needed by radiation - It will contain the 
!   basically the same as the internal state
!   --------------------------------------------------------
    call MAPL_AddExportSpec(GC,                                 &
        SHORT_NAME         = 'AERO',                              &
        LONG_NAME          = 'aerosol_mass_mixing_ratios',        &
        UNITS              = 'kg kg-1',                           &
        DIMS               = MAPL_DimsHorzVert,                   &
        VLOCATION          = MAPL_VLocationCenter,                &
        DATATYPE           = MAPL_BundleItem,                     &
                                                       RC=STATUS  )
    VERIFY_(STATUS)

!   This bundle is not currently filled in by CARMA, just a place holder
!   for symmetry with GOCART
    call MAPL_AddExportSpec(GC,                                   &
        SHORT_NAME         = 'AERO_DP',                           &
        LONG_NAME          = 'aerosol_deposition',                &
        UNITS              = 'kg m-2 s-1',                        &
        DIMS               = MAPL_DimsHorzOnly,                   &
        DATATYPE           = MAPL_BundleItem,                     &
                                                       RC=STATUS  )
   VERIFY_(STATUS)

!
#include "CARMA_ExportSpec___.h"

!   Generic Set Services
!   --------------------
    call MAPL_GenericSetServices ( GC, RC=STATUS )
    VERIFY_(STATUS)

!   All done
!   --------

    RETURN_(ESMF_SUCCESS)

  END SUBROUTINE SetServices


!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  Initialize_ --- Initialize CARMA
!
! !INTERFACE:
!

   SUBROUTINE Initialize_ ( gc, import, export, clock, rc )

! !USES:

   implicit NONE

! !INPUT PARAMETERS:

   type(ESMF_Clock),  intent(inout) :: clock      ! The clock

! !OUTPUT PARAMETERS:

   type(ESMF_GridComp), intent(inout)  :: gc      ! Grid Component
   type(ESMF_State), intent(inout) :: import     ! Import State
   type(ESMF_State), intent(inout) :: export     ! Export State
   integer, intent(out) ::  rc                    ! Error return code:
                                                  !  0 - all is well
                                                  !  1 - 

! !DESCRIPTION: This is a simple ESMF wrapper.
!
! !REVISION HISTORY:
!
!  27Feb2005 da Silva  First crack.
!
!EOP
!-------------------------------------------------------------------------


!  ErrLog Variables
!  ----------------
   character(len=ESMF_MAXSTR)      :: IAm = 'Initialize_'
   integer                         :: STATUS
   character(len=ESMF_MAXSTR)      :: COMP_NAME

   type(CARMA_GridComp), pointer   :: gcCARMA     ! Grid Component
   integer                         :: nymd, nhms  ! time of day
   real                            :: cdt         ! chemistry timestep (secs)

   type(ESMF_Grid)                 :: grid        
 
   integer                         :: i1=1, i2, ig=0, im  ! dist grid indices
   integer                         :: j1=1, j2, jg=0, jm  ! dist grid indices
   integer                         :: km, nq              ! dist grid indices
   integer                         :: n, dims(3), l

   type(Chem_Array), pointer       :: qa(:)	   ! array of pointers
   type(MAPL_MetaComp), pointer    :: ggState	   ! GEOS Generic State
   type (ESMF_State)               :: internal
   type(MAPL_VarSpec), pointer     :: InternalSpec(:)
   type(ESMF_Field)                :: field
   type(ESMF_FieldBundle)          :: bundle

   character(len=ESMF_MAXSTR)      :: short_name
   type(CARMA_Registry), pointer   :: reg => null()
   integer :: nCARMAbegin, nCARMAend, ibin, ielem, igroup, igas
   real    :: fscav
   
   integer :: i, j, k, iq, istart, iend
   real, parameter :: rad2deg = 180. / MAPL_PI

!  Declare pointers to IMPORT/EXPORT/INTERNAL states 
!  -------------------------------------------------
#  include "CARMA_DeclarePointer___.h"

!  Get my name and set-up traceback handle
!  ---------------------------------------
   call ESMF_GridCompGet( GC, NAME=COMP_NAME, RC=STATUS )
   VERIFY_(STATUS)
   Iam = trim(COMP_NAME) // '::Initialize_'

!  Get my internal MAPL_Generic state
!  -----------------------------------
   call MAPL_GetObjectFromGC ( GC, ggState, RC=STATUS)
   VERIFY_(STATUS)

!   Initialize GEOS Generic
!   ------------------------
    call MAPL_GenericInitialize ( gc, import, export, clock,  RC=STATUS )
    VERIFY_(STATUS)

!  Get parameters from gc and clock
!  --------------------------------
   call extract_ ( gc, clock, gcCARMA, qa, nymd, nhms, cdt, STATUS )
   VERIFY_(STATUS)

!  Get the grid
!  ------------
   call ESMF_GridCompGet ( GC, grid=grid, rc=STATUS)
   VERIFY_(STATUS)

   call MAPL_GridGet ( grid, globalCellCountPerDim=DIMS, RC=STATUS)
   VERIFY_(STATUS)

   gcCARMA%im = dims(1)
   gcCARMA%jm = dims(2)

   call ESMF_GridGet(GRID, localDE=0, &
        staggerloc=ESMF_STAGGERLOC_CENTER, &
        computationalCount=DIMS, RC=STATUS)
   VERIFY_(STATUS)

   gcCARMA%grid = grid
   reg => gcCARMA%carmaReg

!  Get pointers to IMPORT/EXPORT/INTERNAL states 
!  ---------------------------------------------
#  include "CARMA_GetPointer___.h"

!  Associate the Internal State fields with our legacy state 
!  ---------------------------------------------------------
   call MAPL_Get ( ggSTATE, INTERNALSPEC=InternalSpec, &
                                        INTERNAL_ESMF_STATE=internal, &
                                        LONS=gcCARMA%LONS, &
                                        LATS=gcCARMA%LATS, &
                                        RC=STATUS  )
   VERIFY_(STATUS)

! Convert radians to degrees
  gcCARMA%lons = gcCARMA%lons * rad2deg
  gcCARMA%lats = gcCARMA%lats * rad2deg

! Local sizes of three dimensions
!--------------------------------
   gcCARMA%i2 = dims(1)
   gcCARMA%j2 = dims(2)
   gcCARMA%km = dims(3)

! PRC: Should do it this way, but it crashes...
!   call ESMF_GridGetCoord(grid, fptr=lons, coordDim=1, localDE=0, rc=STATUS)
!   VERIFY_(STATUS)
!   call ESMF_GridGetCoord(grid, fptr=lats, coordDim=2, localDE=0, rc=STATUS)
!   VERIFY_(STATUS)

!   Initialize the tracer array
!   ---------------------------
    ASSERT_ ( size(InternalSpec) == reg%nq )

    do L = 1, size(InternalSpec)

      call MAPL_VarSpecGet ( InternalSpec(L),          &
                             SHORT_NAME = short_name,  &
                             RC=STATUS )
      VERIFY_(STATUS)

      call MAPL_GetPointer ( internal, NAME=short_name, ptr=qa(L)%data3d, &
                             rc = STATUS )
      VERIFY_(STATUS)

    end do

!   Bootstrapping option
!   --------------------
!   For sub-stepping of gases we need to provide the prior time-step
!   temperature, gas mmr, and saturation ratios for liquid and ice.
!   If these are not set from the carma_internal_rst we need to 
!   initialize.

!   Prior time-step temperature
    iq = reg%NBIN*reg%NELEM + reg%NGAS + 1
    if(qa(iq)%data3d(gcCARMA%i2,gcCARMA%j2,gcCARMA%km) <= 0.) qa(iq)%data3d = -1.

!   Prior time-step gases
    istart = reg%NBIN*reg%NELEM + reg%NGAS + 1 + 1
    iend   = istart + reg%NGAS - 1
    do iq = istart, iend
     if(qa(iq)%data3d(gcCARMA%i2,gcCARMA%j2,gcCARMA%km) <= 0.) qa(iq)%data3d = qa(iq-reg%NGAS-1)%data3d
    enddo

!   Prior time-step saturation wrt liquid
    istart = reg%NBIN*reg%NELEM + reg%NGAS + 1 + reg%NGAS + 1
    iend   = istart + reg%NGAS - 1
    do iq = istart, iend
     if(qa(iq)%data3d(gcCARMA%i2,gcCARMA%j2,gcCARMA%km) <= 0.) qa(iq)%data3d = -1.
    enddo

!   Prior time-step saturation wrt ice
    istart = reg%NBIN*reg%NELEM + reg%NGAS + 1 + reg%NGAS + reg%NGAS + 1
    iend   = istart + reg%NGAS - 1
    do iq = istart, iend
     if(qa(iq)%data3d(gcCARMA%i2,gcCARMA%j2,gcCARMA%km) <= 0.) qa(iq)%data3d = -1.
    enddo

!   Call initialize
!   ---------------
    call CARMA_GridCompInitialize ( gcCARMA, import, export, nymd, nhms, cdt, &
                                    STATUS )
    VERIFY_(STATUS)

#ifdef PRINT_STATES

   if (MAPL_AM_I_ROOT()) then
       print *, trim(Iam)//': IMPORT   State during Initialize():'
       call ESMF_StatePrint ( import  )
       print *, trim(Iam)//': EXPORT   State during Initialize():' 
       call ESMF_StatePrint ( export  )
    end if

#endif

!   Fill in the scavenging attribute
!   --------------------------------
    reg => gcCARMA%carmaReg
    nCARMABegin =  1
    nCARMAEnd   =  reg%nq
    do ielem = 1, reg%NELEM
     igroup = reg%igroup(ielem)
     do ibin = 1, reg%NBIN
      n = nCARMAbegin + (ielem-1)*reg%NBIN + ibin - 1

      call MAPL_VarSpecGet ( InternalSpec(n),          &
                             SHORT_NAME = short_name,  &
                             RC=STATUS )
      VERIFY_(STATUS)

      call ESMF_StateGet ( internal,                   &
                           short_name,                 &
                           field, rc=STATUS )
      VERIFY_(STATUS)

      fscav = reg%fscav(igroup)
      call ESMF_AttributeSet (field, NAME  = "ScavengingFractionPerKm", &
                              VALUE = fscav,                            &
                              RC = STATUS )
      VERIFY_(STATUS)

      call ESMF_AttributeGet (field, NAME  = "ScavengingFractionPerKm", &
                              VALUE = fscav,                  &
                              RC = STATUS )
      VERIFY_(STATUS)

     end do
    end do

!   Get the Mie tables (GEOS-5 like Mie tables)
!   -------------------------------------------
    call CARMA_GetMieTables(gcCARMA, rc)
    if(rc /= 0) then
     if(MAPL_AM_I_ROOT()) print *, 'CARMA: Failed reading Mie tables'
     RETURN_(ESMF_FAILURE)
    endif

!   Fill the AERO bundle - For now we add all concentration elements
!   --------------------
    call ESMF_StateGet(export, 'AERO', bundle, __RC__ )

    do ielem = 1, reg%NELEM
     igroup = reg%igroup(ielem)
     if(ielem /= gcCARMA%carma%f_group(igroup)%f_ienconc ) cycle
     do ibin = 1, reg%NBIN
      n = nCARMAbegin + (ielem-1)*reg%NBIN + ibin - 1
      call ESMF_StateGet ( INTERNAL,                     &
                           trim(COMP_NAME) // '::'//     &
                           trim(reg%vname(n)),           &
                           FIELD, __RC__ )
      call MAPL_FieldBundleAdd ( BUNDLE, FIELD, __RC__ )
     end do
    end do

#ifdef PRINT_STATES

   if (MAPL_AM_I_ROOT()) then
       print *, trim(Iam)//': AERO Bundle during Initialize():' 
       call ESMF_FieldBundlePrint ( bundle )
   end if

#endif


    RETURN_(ESMF_SUCCESS)

   END SUBROUTINE Initialize_


!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  Run_ --- Runs CARMA
!
! !INTERFACE:
!

   SUBROUTINE Run_ ( gc, import, export, clock, rc )

! !USES:

  implicit NONE

! !INPUT PARAMETERS:

   type(ESMF_Clock),  intent(inout) :: clock      ! The clock

! !OUTPUT PARAMETERS:

   type(ESMF_GridComp), intent(inout)  :: gc      ! Grid Component
   type(ESMF_State), intent(inout) :: import     ! Import State
   type(ESMF_State), intent(inout) :: export     ! Export State
   integer, intent(out) ::  rc                    ! Error return code:
                                                  !  0 - all is well
                                                  !  1 - 

! !DESCRIPTION: This is a simple ESMF wrapper.
!
! !REVISION HISTORY:
!
!  27Feb2005 da Silva  First crack.
!
!EOP
!-------------------------------------------------------------------------

!  ErrLog Variables
!  ----------------
   character(len=ESMF_MAXSTR)      :: IAm = 'Run_'
   integer                         :: STATUS
   character(len=ESMF_MAXSTR)      :: COMP_NAME

   type(CARMA_GridComp), pointer   :: gcCARMA     ! Grid Component
   integer                         :: nymd, nhms  ! time
   real                            :: cdt         ! chemistry timestep (secs)
   type(Chem_Array), pointer       :: qa(:)
   integer                         :: i1=1, i2, ig=0, im  ! dist grid indices
   integer                         :: j1=1, j2, jg=0, jm  ! dist grid indices
   integer                         :: km, nq              ! dist grid indices
   integer                         :: k, n, dims(3), l, ijl, iq
   real                            :: qmin, qmax

   type(ESMF_Config)               :: CF
   type(ESMF_Grid)                 :: grid
   type(ESMF_Time)                 :: TIME

   type(ESMF_State)                :: internal
   type(MAPL_VarSpec), pointer     :: InternalSpec(:)


   real, pointer, dimension(:,:)   :: LATS
   real, pointer, dimension(:,:)   :: LONS
   type (MAPL_SunOrbit)            :: ORBIT


!  Declare pointers to IMPORT/EXPORT/INTERNAL states 
!  -------------------------------------------------
#  include "CARMA_DeclarePointer___.h"

!  Get my name and set-up traceback handle
!  ---------------------------------------
   call ESMF_GridCompGet( GC, NAME=COMP_NAME, __RC__)
   Iam = trim(COMP_NAME) // '::Run_'

!  Get pointers to IMPORT/EXPORT/INTERNAL states 
!  ---------------------------------------------
#  include "CARMA_GetPointer___.h"

!  Get parameters from gc and clock
!  --------------------------------
   call extract_ ( gc, clock, gcCARMA, qa, nymd, nhms, cdt, STATUS )
   VERIFY_(STATUS)

!  Run
!  ---
   call CARMA_Emissions  ( gcCARMA, qa, import, export, nymd, nhms, &
                           cdt, STATUS )
   VERIFY_(STATUS)

   call CARMA_GridCompRun ( gcCARMA, qa, import, export, nymd, nhms, &
                            cdt, STATUS )
   VERIFY_(STATUS)

   call CARMA_DryDeposition  ( gcCARMA, qa, import, export, nymd, nhms, &
                               cdt, STATUS )
   VERIFY_(STATUS)

   call CARMA_WetRemoval  ( gcCARMA, qa, import, export, nymd, nhms, &
                            cdt, STATUS )
   VERIFY_(STATUS)

   call CARMA_Convection  ( gcCARMA, qa, import, export, nymd, nhms, &
                            cdt, STATUS )
   VERIFY_(STATUS)

   call CARMA_ComputeDiags  ( gcCARMA, qa, import, export, nymd, nhms, &
                              cdt, STATUS )
   VERIFY_(STATUS)

   RETURN_(ESMF_SUCCESS)

   END SUBROUTINE Run_

!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  Finalize_ --- Finalize CARMA
!
! !INTERFACE:
!

   SUBROUTINE Finalize_ ( gc, import, export, clock, rc )

! !USES:

  implicit NONE

! !INPUT PARAMETERS:

   type(ESMF_Clock),  intent(inout) :: clock      ! The clock

! !OUTPUT PARAMETERS:

   type(ESMF_GridComp), intent(inout)  :: gc      ! Grid Component
   type(Chem_Array), pointer           :: qa(:)
   type(ESMF_State), intent(inout) :: import     ! Import State
   type(ESMF_State), intent(inout) :: export     ! Export State
   integer, intent(out) ::  rc                    ! Error return code:
                                                  !  0 - all is well
                                                  !  1 - 

! !DESCRIPTION: This is a simple ESMF wrapper.
!
! !REVISION HISTORY:
!
!  27Feb2005 da Silva  First crack.
!
!EOP
!-------------------------------------------------------------------------


!  ErrLog Variables
!  ----------------
   character(len=ESMF_MAXSTR)      :: IAm = 'Finalize_'
   integer                         :: STATUS
   character(len=ESMF_MAXSTR)      :: COMP_NAME

   type(CARMA_GridComp), pointer   :: gcCARMA     ! Grid Component
   integer                         :: nymd, nhms  ! time
   real                            :: cdt         ! chemistry timestep (secs)

    type(CARMAchem_state), pointer     :: state

!  Get my name and set-up traceback handle
!  ---------------------------------------
   call ESMF_GridCompGet( GC, NAME=COMP_NAME, RC=STATUS )
   VERIFY_(STATUS)
   Iam = trim(COMP_NAME) // 'Finalize_'

!  Get ESMF parameters from gc and clock
!  -------------------------------------
   call extract_ ( gc, clock, gcCARMA, qa, nymd, nhms, cdt, STATUS, &
                   state = state )
   VERIFY_(STATUS)

!  Call ESMF version
!  -----------------
   call CARMA_GridCompFinalize ( gcCARMA, import, export, &
                                 nymd, nhms, cdt, STATUS )
   VERIFY_(STATUS)

!  Finalize MAPL Generic.  Atanas says, "Do not deallocate foreign objects."
!  -------------------------------------------------------------------------
   call MAPL_GenericFinalize ( gc, import, export, clock,  RC=STATUS )
   VERIFY_(STATUS)

!  Destroy Mie Tables
!  ------------------
   call CARMA_DestroyMieTables(gcCARMA, rc)

!  Destroy emissions
!  -----------------
   if(associated(gcCARMA%dust_source))       deallocate( gcCARMA%dust_source, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%blackcarbon_emissions)) deallocate( gcCARMA%blackcarbon_emissions, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%smoke_source))      deallocate( gcCARMA%smoke_source, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%smoke_altitude))    deallocate( gcCARMA%smoke_altitude, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%so2biomass_src_))   deallocate( gcCARMA%so2biomass_src_, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%so2biomass_src))    deallocate( gcCARMA%so2biomass_src, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%so2anthro_l1_src))  deallocate( gcCARMA%so2anthro_l1_src, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%so2anthro_l2_src))  deallocate( gcCARMA%so2anthro_l2_src, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%so2ship_src))       deallocate( gcCARMA%so2ship_src, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%so4ship_src))       deallocate( gcCARMA%so4ship_src, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%aircraft_fuel_src)) deallocate( gcCARMA%aircraft_fuel_src, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%dmso_conc))         deallocate( gcCARMA%dmso_conc, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%vLat))              deallocate( gcCARMA%vLat, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%vLon))              deallocate( gcCARMA%vLon, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%vSO2))              deallocate( gcCARMA%vSO2, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%vElev))             deallocate( gcCARMA%vElev, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%vCloud))            deallocate( gcCARMA%vCloud, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%oh_conc))           deallocate( gcCARMA%oh_conc, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%no3_mr))            deallocate( gcCARMA%no3_mr, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%h2o2_mr))           deallocate( gcCARMA%h2o2_mr, stat=STATUS)
   VERIFY_(STATUS)
   if(associated(gcCARMA%h2o2_int))          deallocate( gcCARMA%h2o2_int, stat=STATUS)
   VERIFY_(STATUS)

!  Destroy Legacy state
!  --------------------
   call registry_destroy_ (state%CARMAreg)
   deallocate ( state%CARMAreg, state%qa, state%gcCARMA, stat = STATUS )
   VERIFY_(STATUS)

   RETURN_(ESMF_SUCCESS)

   END SUBROUTINE Finalize_

!.......................................................................
    subroutine extract_ ( gc, clock, gcCARMA, qa, nymd, nhms, cdt, &
                          rc, state )

    type(ESMF_GridComp), intent(INout)  :: gc
    type(ESMF_Clock), intent(in)     :: clock
    type(CARMA_GridComp), pointer    :: gcCARMA
    type(Chem_Array), pointer        :: qa(:)
    integer, intent(out)             :: nymd, nhms
    real, intent(out)                :: cdt
    integer, intent(out)             :: rc
    type(CARMAchem_state), pointer, optional   :: state


    type(CARMAchem_state), pointer    :: myState

!   ErrLog Variables
!   ----------------
    character(len=ESMF_MAXSTR)      :: IAm
    integer                         :: STATUS
    character(len=ESMF_MAXSTR)      :: COMP_NAME

    type(ESMF_Time)      :: TIME
    type(ESMF_Config)    :: CF
    type(CARMAchem_Wrap) :: wrap
    integer              :: IYR, IMM, IDD, IHR, IMN, ISC


!   Get my name and set-up traceback handle
!   ---------------------------------------
    call ESMF_GridCompGet( GC, NAME=COMP_NAME, RC=STATUS )
    VERIFY_(STATUS)
    Iam = trim(COMP_NAME) // '::' // 'extract_'

    rc = 0

!   Get my internal state
!   ---------------------
    call ESMF_UserCompGetInternalState(gc, 'CARMA_state', WRAP, STATUS)
    VERIFY_(STATUS)
    myState => wrap%ptr
    if ( present(state) ) then
         state => wrap%ptr
    end if

    if ( .not. associated(myState%gcCARMA) ) then
         allocate ( myState%gcCARMA, stat=STATUS )
         VERIFY_(STATUS)
    end if

    if ( .not. associated(myState%CARMAreg) ) then
         allocate ( myState%CARMAreg, stat=STATUS )
         VERIFY_(STATUS)
    end if

    if ( .not. associated(myState%qa) ) then
         allocate ( myState%qa(myState%CARMAreg%nq), stat=STATUS )
         VERIFY_(STATUS)
    end if

    gcCARMA => myState%gcCARMA
    gcCARMA%CARMAreg => myState%CARMAreg
    qa => myState%qa

!   Get the configuration
!   ---------------------
    call ESMF_GridCompGet ( GC, CONFIG = CF, RC=STATUS )
    VERIFY_(STATUS)

!   Get time step
!   -------------
    call ESMF_ConfigGetAttribute ( CF, cdt, Label="RUN_DT:", RC=STATUS )
    VERIFY_(STATUS)

!   Need code to extract nymd(20050205), nhms(120000) from clock
!   ------------------------------------------

    call ESMF_ClockGet(CLOCK,currTIME=TIME,rc=STATUS)
    VERIFY_(STATUS)

    call ESMF_TimeGet(TIME ,YY=IYR, MM=IMM, DD=IDD, H=IHR, M=IMN, S=ISC, rc=STATUS)
    VERIFY_(STATUS)

    call MAPL_PackTime(NYMD,IYR,IMM,IDD)
    call MAPL_PackTime(NHMS,IHR,IMN,ISC)

    RETURN_(ESMF_SUCCESS)

   end subroutine extract_

!.......................................................................
   subroutine registry_ (r)

   type(CARMA_Registry), pointer   :: r
   CHARACTER(LEN=255) :: string
   integer :: ios, ier(20), i, j, n, rc

!  Load resource file
!  ------------------
   CALL I90_loadf ( TRIM(r%rcfilen), ier(1) )
   IF ( ier(1) .NE. 0 ) THEN
    CALL final_(10)
    RETURN
   END IF
   ier(1)=0

!  Particle/Gas/Radiation structure
!  --------------------------------
!  NBIN
   call i90_label ( 'NBIN:', ier(1) )
   r%NBIN = i90_gint ( ier(2) )
   if ( any(ier(1:2) /= 0) ) then
      call final_(20)
      return
   end if

!  NGROUP
   call i90_label ( 'NGROUP:', ier(1) )
   r%NGROUP = i90_gint ( ier(2) )
   if ( any(ier(1:2) /= 0) ) then
      call final_(20)
      return
   end if

!  NELEM
   call i90_label ( 'NELEM:', ier(1) )
   r%NELEM = i90_gint ( ier(2) )
   if ( any(ier(1:2) /= 0) ) then
      call final_(20)
      return
   end if

!  NGAS
   call i90_label ( 'NGAS:', ier(1) )
   r%NGAS = i90_gint ( ier(2) )
   if ( any(ier(1:2) /= 0) ) then
      call final_(20)
      return
   end if

!  NSOLUTE
   call i90_label ( 'NSOLUTE:', ier(1) )
   r%NSOLUTE = i90_gint ( ier(2) )
   if ( any(ier(1:2) /= 0) ) then
      call final_(20)
      return
   end if

!  NWAVE
   call i90_label ( 'NWAVE:', ier(1) )
   r%NWAVE = i90_gint ( ier(2) )
   if ( any(ier(1:2) /= 0) ) then
      call final_(20)
      return
   end if

!  Group Characteristics
!  ---------------------
   allocate ( r%rmrat(r%NGROUP), &
              r%rmin(r%NGROUP), &
              r%fscav(r%NGROUP), &
              r%irhswell(r%NGROUP), &
              r%irhswcomp(r%NGROUP), &
              r%groupname(r%NGROUP), stat=ios )
   if ( ios /= 0) then
    call final_(100)
    return
   endif

   call i90_label ( 'RMRAT:', ier(1) )
   do j = 1, r%NGROUP
    r%rmrat(j) = i90_gfloat(ier(i+1))
   end do
   if(any(ier(1:r%NGROUP+1) /= 0)) then
    call final_(101)
    return
   endif
    
   call i90_label ( 'RMIN:', ier(1) )
   do j = 1, r%NGROUP
    r%rmin(j) = i90_gfloat(ier(i+1))
   end do
   if(any(ier(1:r%NGROUP+1) /= 0)) then
    call final_(101)
    return
   endif
    
   call i90_label ( 'FSCAV:', ier(1) )
   do j = 1, r%NGROUP
    r%fscav(j) = i90_gfloat(ier(i+1))
   end do
   if(any(ier(1:r%NGROUP+1) /= 0)) then
    call final_(101)
    return
   endif
    
   call i90_label ( 'IRHSWELL:', ier(1) )
   do j = 1, r%NGROUP
    r%irhswell(j) = i90_gint(ier(i+1))
   end do
   if(any(ier(1:r%NGROUP+1) /= 0)) then
    call final_(101)
    return
   endif
    
   call i90_label ( 'IRHSWCOMP:', ier(1) )
   do j = 1, r%NGROUP
    r%irhswcomp(j) = i90_gint(ier(i+1))
   end do
   if(any(ier(1:r%NGROUP+1) /= 0)) then
    call final_(101)
    return
   endif
    
   call i90_label ( 'GROUPNAME:', ier(1) )
   do j = 1, r%NGROUP
    call i90_gtoken(string, ier(i+1))
    r%groupname(j) = trim(string)
   end do
   if(any(ier(1:r%NGROUP+1) /= 0)) then
    call final_(101)
    return
   endif
    

!  Element Characteristics
!  -----------------------
   allocate ( r%rhop(r%NELEM), r%igroup(r%NELEM), r%itype(r%NELEM), &
              r%elemname(r%NELEM), r%icomposition(r%NELEM), stat=ios )
   if ( ios /= 0) then
    call final_(100)
    return
   endif

   call i90_label ( 'IGROUP:', ier(1) )
   do i = 1, r%NELEM
    r%igroup(i) = i90_gint(ier(i+1))
   end do
   if(any(ier(1:r%NELEM+1) /= 0)) then
    call final_(101)
    return
   endif
    
   call i90_label ( 'ICOMPOSITION:', ier(1) )
   do i = 1, r%NELEM
    r%icomposition(i) = i90_gint(ier(i+1))
   end do
   if(any(ier(1:r%NELEM+1) /= 0)) then
    call final_(101)
    return
   endif
    
   call i90_label ( 'ITYPE:', ier(1) )
   do i = 1, r%NELEM
    r%itype(i) = i90_gint(ier(i+1))
   end do
   if(any(ier(1:r%NELEM+1) /= 0)) then
    call final_(101)
    return
   endif
    
   call i90_label ( 'RHOP:', ier(1) )
   do i = 1, r%NELEM
    r%rhop(i) = i90_gfloat(ier(i+1))
   end do
   if(any(ier(1:r%NELEM+1) /= 0)) then
    call final_(101)
    return
   endif
    
   call i90_label ( 'ELEMNAME:', ier(1) )
   do i = 1, r%NELEM
    call i90_gtoken(string, ier(i+1))
    r%elemname(i) = trim(string)
   end do
   if(any(ier(1:r%NELEM+1) /= 0)) then
    call final_(101)
    return
   endif

!  Gas Characteristics
!  -------------------
   if(r%NGAS .gt. 0) then
    allocate ( r%gasname(r%NGAS), r%igcomp(r%NGAS), &
               r%igvapreq(r%NGAS), stat = ios )
    if ( ios /= 0) then
     call final_(200)
     return
    endif

    call i90_label ( 'GASNAME:', ier(1) )
    do i = 1, r%NGAS
     call i90_gtoken(string, ier(i+1))
     r%gasname(i) = trim(string)
    end do
    if(any(ier(1:r%NGAS+1) /= 0)) then
     call final_(201)
     return
    endif

    call i90_label ( 'IGCOMP:', ier(1) )
    do i = 1, r%NGAS
     r%igcomp(i) = i90_gint(ier(i+1))
    end do
    if(any(ier(1:r%NGAS+1) /= 0)) then
     call final_(202)
     return
    endif

    call i90_label ( 'IGVAPREQ:', ier(1) )
    do i = 1, r%NGAS
     r%igvapreq(i) = i90_gint(ier(i+1))
    end do
    if(any(ier(1:r%NGAS+1) /= 0)) then
     call final_(203)
     return
    endif

  endif



!  Microphysical process flags and timesteps
!  -----------------------------------------
!  Note that the labels are not required to be present
   call i90_label ( 'DO_COAG:', ios )
   if(ios .eq. 0) r%do_coag = i90_gint( ier(1) )
   call i90_label ( 'DO_GROW:', ios )
   if(ios .eq. 0) r%do_grow = i90_gint( ier(2) )
   call i90_label ( 'DO_SUBSTEP:', ios )
   if(ios .eq. 0) r%do_substep = i90_gint( ier(3) )
   call i90_label ( 'DO_THERMO:', ios )
   if(ios .eq. 0) r%do_thermo = i90_gint( ier(4) )
   call i90_label ( 'DO_VDIFF:', ios )
   if(ios .eq. 0) r%do_vdiff = i90_gint( ier(5) )
   call i90_label ( 'DO_VTRAN:', ios )
   if(ios .eq. 0) r%do_vtran = i90_gint( ier(6) )
   call i90_label ( 'VF_CONST:', ios )
   if(ios .eq. 0) r%vf_const = i90_gfloat( ier(7) )
   call i90_label ( 'MINSUBSTEPS:', ios )
   if(ios .eq. 0) r%minsubsteps = i90_gint( ier(8) )
   call i90_label ( 'MAXSUBSTEPS:', ios )
   if(ios .eq. 0) r%maxsubsteps = i90_gint( ier(9) )
   call i90_label ( 'MAXRETRIES:', ios )
   if(ios .eq. 0) r%maxretries = i90_gint( ier(10) )
   call i90_label ( 'CONMAX:', ios )
   if(ios .eq. 0) r%conmax = i90_gfloat( ier(11) )
   if(any(ier(1:11) /= 0)) then
    call final_(102)
    return
   endif

!  Species specific
!  ----------------
!  Dust
   call i90_label ( 'dust_source_filename:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( string, ier(1) )
      r%dust_source_filename = trim(string)
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if

!  Black Carbon
   call i90_label ( 'bcbiomass_srcfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( string, ier(1) )
      r%bc_bb_srcfilen = trim(string)
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if
   call i90_label ( 'bcbiofuel_srcfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( string, ier(1) )
      r%bc_bf_srcfilen = trim(string)
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if
   call i90_label ( 'bcanthro_l1_srcfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( string, ier(1) )
      r%ebcant1_srcfilen = trim(string)
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if
   call i90_label ( 'bcanthro_l2_srcfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( string, ier(1) )
      r%ebcant2_srcfilen = trim(string)
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if
   call i90_label ( 'bcship_srcfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( string, ier(1) )
      r%bc_ship_srcfilen = trim(string)
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if

!  Smoke
   call i90_label ( 'smoke_source_filename:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( string, ier(1) )
      r%smoke_source_filename = trim(string)
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if

   call i90_label ( 'smoke_altitude_filename:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( string, ier(1) )
      r%smoke_altitude_filename = trim(string)
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if

!  Sulfate
   call i90_label ( 'so2biomass_srcfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( r%so2bb_srcfilen, ier(1) )
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if

   call i90_label ( 'so2volcano_srcfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( r%so2volcano_srcfilen, ier(1) )
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if

   call i90_label ( 'so2anthro_l1_srcfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( r%so2anthro_l1_srcfilen, ier(1) )
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if

   call i90_label ( 'so2anthro_l2_srcfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( r%so2anthro_l2_srcfilen, ier(1) )
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if

   call i90_label ( 'so2ship_srcfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( r%so2ship_srcfilen, ier(1) )
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if

   call i90_label ( 'so4ship_srcfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( r%so4ship_srcfilen, ier(1) )
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if

   call i90_label ( 'aircraft_fuel_srcfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( r%aircraft_fuel_srcfilen, ier(1) )
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if

   call i90_label ( 'dmso_concfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( r%dmso_concfilen, ier(1) )
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if

   call i90_label ( 'oh_concfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( r%oh_concfilen, ier(1) )
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if

   call i90_label ( 'no3_mrfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( r%no3_mrfilen, ier(1) )
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if

   call i90_label ( 'h2o2_mrfilen:', ier(1) )
   if ( ier(1) /= 0 ) then
      call final_(30)
      return
   else
      call i90_gtoken ( r%h2o2_mrfilen, ier(1) )
      if ( ier(1) /= 0 ) then
         call final_(40)
         return
      end if
   end if

!  Fraction of anthropogenic emissions to SO4
!  ---------------
   call i90_label ( 'so4anthropogenic_fraction:', ier(1) )
   r%fSO4ant = i90_gfloat ( ier(2) )
   if ( any(ier(1:2) /= 0) ) then
      call final_(50)
      return
   end if


!  Biomass Burning Emission Factor
!  ---------------
   call i90_label ( 'so2biomass_emission_factor:', ier(1) )
   r%eBiomassBurning = i90_gfloat ( ier(2) )
   if ( any(ier(1:2) /= 0) ) then
      call final_(50)
      return
   end if


!  Aircraft Fuel Emission Factor
!  ---------------
   call i90_label ( 'aircraft_fuel_emission_factor:', ier(1) )
   r%eAircraftFuel = i90_gfloat ( ier(2) )
   if ( any(ier(1:2) /= 0) ) then
      call final_(50)
      return
   end if

!  Mie Tables
!  ----------
!   Set the number of channels to calculate over
!   --------------------------------------------
    call i90_label ( 'n_channels:', ios )
    if ( ios /= 0 ) then
     call final_(60)
    else
     r%nchannels = i90_gint ( ios )
     if ( ios /= 0 ) call final_(61)
    end if

!   Set the number of moments
!   -------------------------
    call i90_label ( 'n_moments:', ios )
    if ( ios /= 0 ) then
     r%nmoments = 0
    else
     r%nmoments = i90_gint ( ios )
     if ( ios /= 0 ) call final_(62)
    end if

!   Set the channels to calculate over
!   ----------------------------------
    allocate( r%channels(r%nchannels), stat = ios )
    call i90_label ( 'r_channels:', ios )
    if ( ios /= 0 ) then
     call final_(63)
    else
     do n = 1, r%nchannels
      r%channels(n) = i90_gfloat ( ios )
      if ( ios /= 0 ) call final_(64)
     enddo
    end if

    call i90_label ( 'DU_OPTICS:', ios )
     if ( ios /= 0 ) then
      call final_(65)
     else
      call i90_gtoken ( r%du_optics_file, ios )
      if ( ios /= 0 ) call final_(66)
     end if

    call i90_label ( 'SS_OPTICS:', ios )
     if ( ios /= 0 ) then
      call final_(67)
     else
      call i90_gtoken ( r%ss_optics_file, ios )
      if ( ios /= 0 ) call final_(68)
     end if

    call i90_label ( 'BC_OPTICS:', ios )
     if ( ios /= 0 ) then
      call final_(67)
     else
      call i90_gtoken ( r%bc_optics_file, ios )
      if ( ios /= 0 ) call final_(68)
     end if

    call i90_label ( 'SM_OPTICS:', ios )
     if ( ios /= 0 ) then
      call final_(69)
     else
      call i90_gtoken ( r%sm_optics_file, ios )
      if ( ios /= 0 ) call final_(70)
     end if


   r%nq = r%NBIN*r%NELEM + r%NGAS + r%NGAS + r%NGAS + r%NGAS + 1

   if(r%nq .gt. 0) r%doing_CARMA = .true.

   if(r%doing_CARMA) allocate(r%vname(r%nq), stat=ios)
   if(ios /= 0) call final_(100)

   CALL I90_release()

   end subroutine registry_





   subroutine registry_destroy_ (r)

   type(CARMA_Registry), pointer   :: r
   integer :: ios

!  Group Characteristics
!  ---------------------
   deallocate ( r%rmrat, &
                r%rmin, &
                r%fscav, &
                r%irhswell, &
                r%irhswcomp, &
                r%groupname, stat=ios )
   if ( ios /= 0) then
    call final_(100)
    return
   endif

!  Element Characteristics
!  -----------------------
   deallocate ( r%rhop, r%igroup, r%itype, &
                r%elemname, r%icomposition, stat=ios )
   if ( ios /= 0) then
    call final_(100)
    return
   endif

!  Gas Characteristics
!  -------------------
   if(r%NGAS .gt. 0) then
    deallocate ( r%gasname, r%igcomp, &
                 r%igvapreq, stat = ios )
    if ( ios /= 0) then
     call final_(200)
     return
    endif
   endif

!   Mie Tables
!   ----------
    deallocate( r%channels, stat = ios )
    if ( ios /= 0) then
     call final_(300)
     return
    endif

    if(r%doing_CARMA) deallocate(r%vname, stat=ios)
    if(ios /= 0) then
     call final_(400)
     return
    endif

   end subroutine registry_destroy_





!  print the CARMA registry (for checking)
   subroutine registry_print_ (r)

   type(CARMA_Registry), pointer   :: r
   CHARACTER(LEN=255) :: string
   integer :: ios, ier(20), i, j, rc

   if(MAPL_AM_I_ROOT()) then
    print *, 'CARMAchem_GridCompMod: registry_print_'
    print *, 'NBIN:    ', r%NBIN
    print *, 'NGROUP:  ', r%NGROUP
    print *, 'NELEM:   ', r%NELEM
    print *, 'NGAS:    ', r%NGAS
    print *, 'NSOLUTE: ', r%NSOLUTE
    print *, 'NWAVE:   ', r%NWAVE

!   Group Characteristics
!   -----------------------
    if(r%NGROUP .gt. 0) then
     do j = 1, r%NGROUP
      print *, 'GROUP (',j,'): GROUPNAME = ',trim(r%groupname(j)), &
               ', RMIN = ',r%rmin(j),', RMRAT = ',r%rmrat(j), &
               ', FSCAV = ',r%fscav(j),', IRHSWELL = ', r%irhswell(j), &
               ', IRHSWCOMP = ',r%irhswcomp(j)
     enddo
    endif 

!   Element Characteristics
!   -----------------------
    if(r%NELEM .gt. 0) then
     do j = 1, r%NELEM
      print *, 'ELEMENT (',j,'): ELEMNAME = ',trim(r%elemname(j)), &
               ', IGROUP = ',r%igroup(j),', RHOP = ',r%rhop(j)
     enddo
    endif 

!   Gas Characteristics
!   -------------------
    if(r%NGAS .gt. 0) then
     do j = 1, r%NGAS
      print *, 'GAS (',j,'): GASNAME = ',trim(r%gasname(j)), &
               ', IGCOMP = ',r%igcomp(j),', IGVAPREQ = ',r%igvapreq(j)
     enddo
    endif 

!   Microphysical process flags and timesteps
!   -----------------------------------------
    print *, 'DO_COAG:     ', r%do_coag
    print *, 'DO_GROW:     ', r%do_grow
    print *, 'DO_SUBSTEP:  ', r%do_substep
    print *, 'DO_THERMO:   ', r%do_thermo
    print *, 'DO_VDIFF:    ', r%do_vdiff
    print *, 'DO_VTRAN:    ', r%do_vtran
    print *, 'VF_CONST:    ', r%vf_const
    print *, 'MINSUBSTEPS: ', r%minsubsteps
    print *, 'MAXSUBSTEPS: ', r%maxsubsteps
    print *, 'MAXRETRIES:  ', r%maxretries
    print *, 'CONMAX:      ', r%conmax


   endif

   end subroutine registry_print_





   SUBROUTINE final_(ierr)
   INTEGER :: ios, ierr, rc
   CALL I90_release()
   rc = ierr
   END SUBROUTINE final_




 END MODULE CARMAchem_GridCompMod
