#include "MAPL_Exceptions.h"
#include "MAPL_Generic.h"

!=============================================================================
!BOP

! !MODULE: GEOS_ChemGridCompMod -- Parent Aerosol/Chemistry Component

! !INTERFACE:

module GEOS_ChemGridCompMod

! !USES:

  use ESMF
  use MAPL_Mod
  use Chem_Mod
  use Chem_UtilMod

  use GEOS_ChemEnvGridCompMod,  only : EChemSetServices   => SetServices
  use GOCART_GridCompMod,       only : AChemSetServices   => SetServices
  use StratChem_GridCompMod,    only : SChemSetServices   => SetServices
  use GMIchem_GridCompMod,      only : GChemSetServices   => SetServices
  use CARMAchem_GridCompMod,    only : CChemSetServices   => SetServices
  use GEOSCHEMchem_GridCompMod, only : GCChemSetServices  => SetServices
  use MATRIXchem_GridCompMod,   only : MChemSetServices   => SetServices
  use MAMchem_GridCompMod,      only : MAMChemSetServices => SetServices
  use GEOS_PChemGridCompMod,    only : PChemSetServices   => SetServices
  use GEOS_AChemGridCompMod,    only : GAChemSetServices  => SetServices
  use GAAS_GridCompMod,         only : GAASSetServices    => SetServices
  use H2O_GridCompMod,          only : H2OSetServices     => SetServices
  use TR_GridCompMod,           only : TRSetServices      => SetServices

  implicit none
  private

! !PUBLIC MEMBER FUNCTIONS:

  public SetServices

!                                             -----------
 
! Private state
! -------------
  TYPE GEOS_ChemGridComp
     PRIVATE
     type(ESMF_Config), pointer :: CF  ! Private Config
     LOGICAL :: enable_PCHEM
     LOGICAL :: enable_GEOSACHEM
     LOGICAL :: enable_GOCART
     LOGICAL :: enable_GOCARTdata
     LOGICAL :: enable_GAAS
     LOGICAL :: enable_H2O
     LOGICAL :: enable_STRATCHEM
     LOGICAL :: enable_GMICHEM
     LOGICAL :: enable_CARMA
     LOGICAL :: enable_GEOSCHEM
     LOGICAL :: enable_MATRIX
     LOGICAL :: enable_MAM
     LOGICAL :: enable_MAMdata
     LOGICAL :: enable_TR
     INTEGER :: AERO_PROVIDER
     INTEGER :: RATS_PROVIDER          ! WARNING: May be multiple RATS_PROVIDERs 
  END TYPE GEOS_ChemGridComp

! Hook for the ESMF
! -----------------
  TYPE GEOS_ChemGridComp_Wrap
     TYPE (GEOS_ChemGridComp), pointer :: PTR => null()
  END TYPE GEOS_ChemGridComp_Wrap

!=============================================================================

! !DESCRIPTION: This gridded component (GC) combines 
 
!EOP

! IMPORTANT: If adding a new component, make sure to update private function GetProvider_()
! -----------------------------------------------------------------------------------------
  integer ::      CHEMENV = -1
  integer ::        PCHEM = -1
  integer ::    GEOSACHEM = -1
  integer ::       GOCART = -1
  integer ::   GOCARTdata = -1
  integer ::         GAAS = -1
  integer ::          H2O = -1
  integer ::    STRATCHEM = -1
  integer ::      GMICHEM = -1
  integer ::        CARMA = -1
  integer ::     GEOSCHEM = -1
  integer ::       MATRIX = -1
  integer ::          MAM = -1
  integer ::      MAMdata = -1
  integer ::           TR = -1

contains

!BOP

! !IROUTINE: SetServices -- Sets ESMF services for this component

! !INTERFACE:

    subroutine SetServices ( GC, RC )

! !ARGUMENTS:

    type(ESMF_GridComp), intent(INOUT) :: GC  ! gridded component
    integer,             intent(  OUT) :: RC  ! return code

! !DESCRIPTION:  The SetServices for the Chemistry GC needs to register its
!   Initialize and Run.  It uses the MAPL\_Generic construct for defining 
!   state specs and couplings among its children.  In addition, it creates the   
!   children GCs and runs their respective SetServices.

!EOP

!=============================================================================
!
! ErrLog Variables

    __Iam__('SetServices')
    character(len=ESMF_MAXSTR) :: COMP_NAME

! Locals

   type (ESMF_GridComp), pointer :: GCS(:)

    integer                    :: I, RATS_PROVIDER, AERO_PROVIDER
    type (ESMF_Config), target :: CF, myCF

    integer                    :: n, id
    
    INTEGER, PARAMETER         :: numRATs = 8
    INTEGER                    :: RATsProviderNumber(numRATs)
    CHARACTER(LEN=ESMF_MAXSTR) :: RATsProviderName(numRATs)
    CHARACTER(LEN=ESMF_MAXSTR) :: specieName(numRATs) = (/ "OX    ", "O3    ", "O3PPMV", "CH4   ", &
                                                           "N2O   ", "CFC11 ", "CFC12 ", "HCFC22"/)
    CHARACTER(LEN=ESMF_MAXSTR) :: providerName
    CHARACTER(LEN=ESMF_MAXSTR) :: shortName
    CHARACTER(LEN=ESMF_MAXSTR) :: str

!   Private state
!   -------------
    type (GEOS_ChemGridComp), pointer :: myState   ! private, that is
    type (GEOS_ChemGridComp_Wrap)     :: wrap
    type(Chem_Registry)               :: chemReg

!=============================================================================

! Begin...

! Get my name and set-up traceback handle
! ---------------------------------------
    call ESMF_GridCompGet( GC, NAME=COMP_NAME, CONFIG=CF, __RC__ )
    Iam = TRIM(COMP_NAME) // '::' //TRIM(Iam)

!   Wrap internal state for storing in GC; rename legacyState
!   -------------------------------------
    allocate ( myState, stat=STATUS )
    VERIFY_(STATUS)
    wrap%ptr => myState

!   Load the Chemistry Registry
!   ---------------------------
    chemReg = Chem_RegistryCreate ( STATUS )
    VERIFY_(STATUS)

! Register services for this component
! ------------------------------------
    call MAPL_GridCompSetEntryPoint ( GC, ESMF_METHOD_INITIALIZE, Init, __RC__ )
    call MAPL_GridCompSetEntryPoint ( GC, ESMF_METHOD_RUN,  Run,  __RC__ )

!   Store private state in GC
!   -------------------------
    call ESMF_UserCompSetInternalState ( GC, 'GEOSchem_GridComp_State', wrap, STATUS )
    VERIFY_(STATUS)
  
! Choose children to birth and which children not to conceive
! -----------------------------------------------------------
    myCF = ESMF_ConfigCreate(__RC__)
    call ESMF_ConfigLoadFile ( myCF, 'GEOS_ChemGridComp.rc', __RC__ )
    call ESMF_ConfigGetAttribute(myCF,      myState%enable_PCHEM, Default=.FALSE., Label="ENABLE_PCHEM:",       __RC__ )
    call ESMF_ConfigGetAttribute(myCF,  myState%enable_GEOSACHEM, Default=.FALSE., Label="ENABLE_GEOSACHEM:",   __RC__ )
    call ESMF_ConfigGetAttribute(myCF,     myState%enable_GOCART, Default=.FALSE., Label="ENABLE_GOCART:",      __RC__ )
    call ESMF_ConfigGetAttribute(myCF, myState%enable_GOCARTdata, Default=.FALSE., Label="ENABLE_GOCART_DATA:", __RC__ )
    call ESMF_ConfigGetAttribute(myCF,       myState%enable_GAAS, Default=.FALSE., Label="ENABLE_GAAS:",        __RC__ )
    call ESMF_ConfigGetAttribute(myCF,        myState%enable_H2O, Default=.FALSE., Label="ENABLE_H2O:",         __RC__ )
    call ESMF_ConfigGetAttribute(myCF,  myState%enable_STRATCHEM, Default=.FALSE., Label="ENABLE_STRATCHEM:",   __RC__ )
    call ESMF_ConfigGetAttribute(myCF,    myState%enable_GMICHEM, Default=.FALSE., Label="ENABLE_GMICHEM:",     __RC__ )
    call ESMF_ConfigGetAttribute(myCF,      myState%enable_CARMA, Default=.FALSE., Label="ENABLE_CARMA:",       __RC__ )
    call ESMF_ConfigGetAttribute(myCF,   myState%enable_GEOSCHEM, Default=.FALSE., Label="ENABLE_GEOSCHEM:",    __RC__ )
    call ESMF_ConfigGetAttribute(myCF,     myState%enable_MATRIX, Default=.FALSE., Label="ENABLE_MATRIX:",      __RC__ )
    call ESMF_ConfigGetAttribute(myCF,        myState%enable_MAM, Default=.FALSE., Label="ENABLE_MAM:",         __RC__ )
    call ESMF_ConfigGetAttribute(myCF,    myState%enable_MAMdata, Default=.FALSE., Label="ENABLE_MAM_DATA:",    __RC__ )
    call ESMF_ConfigGetAttribute(myCF,         myState%enable_TR, Default=.FALSE., Label="ENABLE_TR:",          __RC__ )


    myState%CF => myCF ! save for later

! Sanity checks:
! --------------
    if (myState%enable_GAAS) then
       ASSERT_(myState%enable_GOCART)
    end if

! Create children's gridded components and invoke their SetServices
! -----------------------------------------------------------------
    CHEMENV = MAPL_AddChild(GC, NAME='CHEMENV', SS=EChemSetServices, __RC__)

    if (     myState%enable_PCHEM)       PCHEM = MAPL_AddChild(GC, NAME=       'PCHEM', SS=PChemSetServices,   __RC__)
    if ( myState%enable_GEOSACHEM)   GEOSACHEM = MAPL_AddChild(GC, NAME=   'GEOSACHEM', SS=GAChemSetServices,  __RC__)
    if (    myState%enable_GOCART)      GOCART = MAPL_AddChild(GC, NAME=      'GOCART', SS=AChemSetServices,   __RC__)
    if (myState%enable_GOCARTdata)  GOCARTdata = MAPL_AddChild(GC, NAME= 'GOCART.data', SS=AChemSetServices,   __RC__)
    if (      myState%enable_GAAS)        GAAS = MAPL_AddChild(GC, NAME=        'GAAS', SS=GAASSetServices,    __RC__)
    if (       myState%enable_H2O)         H2O = MAPL_AddChild(GC, NAME=         'H2O', SS=H2OSetServices,     __RC__)
    if ( myState%enable_STRATCHEM)   STRATCHEM = MAPL_AddChild(GC, NAME=   'STRATCHEM', SS=SChemSetServices,   __RC__)
    if (   myState%enable_GMICHEM)     GMICHEM = MAPL_AddChild(GC, NAME=     'GMICHEM', SS=GChemSetServices,   __RC__)
    if (     myState%enable_CARMA)       CARMA = MAPL_AddChild(GC, NAME=       'CARMA', SS=CChemSetServices,   __RC__)
    if (  myState%enable_GEOSCHEM)    GEOSCHEM = MAPL_AddChild(GC, NAME='GEOSCHEMCHEM', SS=GCChemSetServices,  __RC__)
    if (    myState%enable_MATRIX)      MATRIX = MAPL_AddChild(GC, NAME=      'MATRIX', SS=MChemSetServices,   __RC__)
    if (       myState%enable_MAM)         MAM = MAPL_AddChild(GC, NAME=         'MAM', SS=MAMChemSetServices, __RC__)
    if (   myState%enable_MAMdata)     MAMdata = MAPL_AddChild(GC, NAME=    'MAM.data', SS=MAMChemSetServices, __RC__)
    if (        myState%enable_TR)          TR = MAPL_AddChild(GC, NAME=          'TR', SS=TRSetServices,      __RC__)


! A container for the friendly tracers
! ------------------------------------
    call MAPL_AddInternalSpec(GC,                                  &
         SHORT_NAME = 'CHEM_TRACERS',                              &
         LONG_NAME  = 'chemistry_friendly_tracers',                &
         UNITS      = 'X',                                         &
         DATATYPE   = MAPL_BundleItem,                             &
         __RC__  )

! Radiatively Active Tracers (RATs).  If a RATS_PROVIDER is not
! specified in the AGCM.rc, then the provider defaults to PCHEM.
! --------------------------------------------------------------
  call GetProvider_(CF, Label='RATS_PROVIDER:', ID=RATS_PROVIDER, Name=providerName, Default='PCHEM', __RC__)

  RATsProviderNumber(1:numRATs) = RATS_PROVIDER
  RATsProviderName(1:numRATs)   = trim(providerName)

! Assume the H2O Tendency is available from the above provider
! ------------------------------------------------------------
  CALL MAPL_AddExportSpec ( GC, SHORT_NAME = 'H2O_TEND', &
                            CHILD_ID = RATS_PROVIDER, __RC__ )

! Priority for first three RATs, OX, O3 and O3PPMV, goes to the ANALYSIS_OX_PROVIDER.
! -----------------------------------------------------------------------------------
  call GetProvider_(CF, Label='ANALYSIS_OX_PROVIDER:', ID=i, Name=providerName, Default=TRIM(RATsProviderName(1)), __RC__)

  RATsProviderNumber(1:3) = i
  RATsProviderName(1:3)   = trim(providerName)

! Option to set RAT CH4 distinct from others (e.g. GOCART) -  RAT_CH4_PROVIDER
! -----------------------------------------------------------------------------------
  call GetProvider_(CF, Label='RAT_CH4_PROVIDER:', ID=i, Name=providerName, Default=TRIM(RATsProviderName(4)), __RC__)

  RATsProviderNumber(4) = i
  RATsProviderName(4)   = trim(providerName)

! Add export specs for the RATs ...
! ---------------------------------
  DO i = 1, numRATs
   CALL MAPL_AddExportSpec( GC, SHORT_NAME = TRIM(specieName(i)), &
                            CHILD_ID = RATsProviderNumber(i), __RC__ )
  END DO

  IF(MAPL_AM_I_ROOT()) THEN
   PRINT *," "
   PRINT *, TRIM(Iam)//": RATs Provider List" 
    DO i = 1, numRATs
     PRINT *,"  "//TRIM(specieName(i))//": "//TRIM(RATsProviderName(i))
    END DO
  END IF

! Choose the ozone mixing ratio tendency from same child as OX.
! -------------------------------------------------------------
  CALL MAPL_AddExportSpec ( GC, SHORT_NAME = 'OX_TEND', &
                            CHILD_ID = RATsProviderNumber(1), __RC__ )

! Aerosol for radiation.  If an AERO_PROVIDER is not specified
! in the AGCM.tmpl, then the provider defaults to GOCART.data.
! -----------------------------------------------------------

  call ESMF_ConfigGetAttribute(CF, providerName, Default='GOCART.data', &
                               Label="AERO_PROVIDER:", __RC__ )

  str = trim(providerName)
  call ESMF_StringLowerCase(str, __RC__)

  if (trim(str) == 'none') then
      AERO_PROVIDER = -1

      call MAPL_AddExportSpec(GC,                                        &
                              SHORT_NAME = 'AERO',                       &
                              LONG_NAME  = 'aerosol_mass_mixing_ratios', &
                              UNITS      = 'kg kg-1',                    &
                              DIMS       = MAPL_DimsHorzVert,            &
                              VLOCATION  = MAPL_VLocationCenter,         &
                              DATATYPE   = MAPL_StateItem, __RC__ )

      call MAPL_AddExportSpec(GC,                                &
                              SHORT_NAME = 'AERO_DP',            &
                              LONG_NAME  = 'aerosol_deposition', &
                              UNITS      = 'kg m-2 s-1',         &
                              DIMS       = MAPL_DimsHorzOnly,    &
                              DATATYPE   = MAPL_BundleItem, __RC__) 
  else
      call GetProvider_(CF, Label='AERO_PROVIDER:', ID=AERO_PROVIDER, Name=providerName, Default='GOCART.data', __RC__)

!     Add export specs for aerosols and aerosol deposition
!     ----------------------------------------------------
      call MAPL_AddExportSpec ( GC, SHORT_NAME = 'AERO',    &
                                CHILD_ID = AERO_PROVIDER, __RC__  )

      call MAPL_AddExportSpec ( GC, SHORT_NAME = 'AERO_DP', &
                                CHILD_ID = AERO_PROVIDER, __RC__  )
  end if

! Save this information in private state for later use. 
! WARNING: Dangerous if there is more than one RATS_PROVIDER
! ----------------------------------------------------------
  myState%RATS_PROVIDER = RATS_PROVIDER
  myState%AERO_PROVIDER = AERO_PROVIDER


  IF(MAPL_AM_I_ROOT()) THEN
   PRINT *," "
   PRINT *,TRIM(Iam)//":"
   PRINT *," AERO Provider is ", TRIM(providerName)
   PRINT *," "
  END IF

! Connectivities between Children
! -------------------------------
  IF(myState%enable_GOCART) then
     CALL MAPL_AddConnectivity ( GC, &
          SHORT_NAME  = (/'AIRDENS    ','AIRDENS_DRY', 'DELP       ', 'TPREC      ', 'CN_PRCP    ', 'NCN_PRCP   '/), &
          DST_ID = GOCART, SRC_ID = CHEMENV, __RC__  )
  ENDIF

  IF(myState%enable_GOCARTdata) then
     CALL MAPL_AddConnectivity ( GC, &
          SHORT_NAME  = (/'DELP       '/), &
          DST_ID = GOCARTdata, SRC_ID = CHEMENV, __RC__  )
  ENDIF

  IF(myState%enable_GAAS) then
     CALL MAPL_AddConnectivity ( GC, &
          SHORT_NAME  = (/ 'AIRDENS ', 'DELP    ' /), &
          DST_ID = GAAS, SRC_ID = CHEMENV, __RC__  )
          IF(myState%enable_GOCART) then
              CALL MAPL_AddConnectivity ( GC, &
                   SRC_NAME  = (/ 'GOCART::du001   ', 'GOCART::du002   ', 'GOCART::du003   ', 'GOCART::du004   ', 'GOCART::du005   ', &
                                  'GOCART::ss001   ', 'GOCART::ss002   ', 'GOCART::ss003   ', 'GOCART::ss004   ', 'GOCART::ss005   ', &
                                  'GOCART::OCphobic', 'GOCART::OCphilic', 'GOCART::BCphobic', 'GOCART::BCphilic', &
                                  'GOCART::SO4     ' /), & 
                   DST_NAME  = (/ 'du001   ', 'du002   ', 'du003   ', 'du004   ', 'du005   ', &
                                  'ss001   ', 'ss002   ', 'ss003   ', 'ss004   ', 'ss005   ', &
                                  'OCphobic', 'OCphilic', 'BCphobic', 'BCphilic', &
                                  'SO4     ' /), &
                   DST_ID = GAAS, SRC_ID = GOCART, __RC__  )
          ELSE
              __raise__(MAPL_RC_ERROR,"Cannot have GAAS enabled without GOCART")
          ENDIF
  ENDIF

  IF(myState%enable_CARMA) then
      CALL MAPL_AddConnectivity ( GC, &
           SHORT_NAME  = (/'AIRDENS ', 'TPREC   ', 'CN_PRCP ', 'NCN_PRCP'/), &
           DST_ID = CARMA, SRC_ID = CHEMENV, __RC__  )
  ENDIF

  IF(myState%enable_STRATCHEM) then
     CALL MAPL_AddConnectivity ( GC, &
          SHORT_NAME  = (/'AIRDENS', 'DELP   ' /), &
          DST_ID = STRATCHEM, SRC_ID = CHEMENV, __RC__  )
  ENDIF

  IF(myState%enable_GMICHEM) then
     CALL MAPL_AddConnectivity ( GC, &
          SHORT_NAME  = (/'AIRDENS', 'DELP   ' /), &
          DST_ID = GMICHEM, SRC_ID = CHEMENV, __RC__  )
  ENDIF

  IF(myState%enable_MATRIX) then
     CALL MAPL_AddConnectivity ( GC, &
          SHORT_NAME  = (/'AIRDENS', 'DELP   '/), &
          DST_ID = MATRIX, SRC_ID = CHEMENV, __RC__  )
  ENDIF

  IF(myState%enable_MAM) then
     CALL MAPL_AddConnectivity ( GC, &
          SHORT_NAME  = (/'AIRDENS ', 'DELP    ', 'TPREC   ', 'CN_PRCP ', 'NCN_PRCP'/), &
          DST_ID = MAM, SRC_ID = CHEMENV, __RC__  )
  ENDIF

  IF(myState%enable_GEOSACHEM) then
     CALL MAPL_AddConnectivity ( GC, &
          SHORT_NAME  = (/'AIRDENS ', 'DELP    ', 'TPREC   ', 'CN_PRCP ', 'NCN_PRCP'/), &
          DST_ID = GEOSACHEM, SRC_ID = CHEMENV, __RC__  )
  ENDIF

  IF(myState%enable_TR) then
     CALL MAPL_AddConnectivity ( GC, &
          SHORT_NAME  = (/'AIRDENS', 'DELP   '/), &
          DST_ID = TR, SRC_ID = CHEMENV, __RC__  )
  ENDIF

  IF(myState%enable_GEOSCHEM) then
     CALL MAPL_AddConnectivity ( GC, &
          SHORT_NAME  = (/'AIRDENS', 'DELP   '/), &
          DST_ID = GEOSCHEM, SRC_ID = CHEMENV, __RC__  )
  ENDIF

! Ozone mole fraction needed by GOCART for
! CFC-12 photolysis.  For GMICHEM case, see below.
! ------------------------------------------------
  IF(myState%enable_GOCART .AND. myState%enable_PCHEM) then
   CALL MAPL_AddConnectivity ( GC, &
        SRC_NAME  = (/"OX"/), &
        DST_NAME  = (/"O3"/), &
        DST_ID = GOCART, SRC_ID = PCHEM, __RC__  )
  ENDIF

  IF(myState%enable_GOCART .AND. myState%enable_STRATCHEM) then
   CALL MAPL_AddConnectivity ( GC, &
        SRC_NAME  = (/"O3CHEM"/), &
        DST_NAME  = (/"O3"/), &
        DST_ID = GOCART, SRC_ID = STRATCHEM, __RC__  )
  ENDIF

  IF(myState%enable_GEOSACHEM .AND. myState%enable_PCHEM) then
   CALL MAPL_AddConnectivity ( GC, &
        SRC_NAME  = (/"O3"/), &
        DST_NAME  = (/"O3"/), &
        DST_ID = GEOSACHEM, SRC_ID = PCHEM, __RC__  )
  ENDIF

  IF(myState%enable_MAM .AND. myState%enable_GEOSACHEM) then
   CALL MAPL_AddConnectivity ( GC, &
        SRC_NAME  = (/'SO2    ', 'H2SO4  ', 'NH3    ', 'SOAG   ', 'pSO4_aq', 'pNH4_aq'/), &
        DST_NAME  = (/'SO2    ', 'H2SO4  ', 'NH3    ', 'SOA_GAS', 'pSO4_aq', 'pNH4_aq'/), &
        DST_ID = MAM, SRC_ID = GEOSACHEM, __RC__  )
  ENDIF

  IF(myState%enable_MATRIX .AND. myState%enable_GEOSACHEM) then
   CALL MAPL_AddConnectivity ( GC, &
        SRC_NAME  = (/'SO2    ', 'H2SO4  ', 'NH3    ', 'SOAG   ', 'pSO4_aq', 'pNH4_aq'/), &
        DST_NAME  = (/'SO2    ', 'H2SO4  ', 'NH3    ', 'SOA_GAS', 'pSO4_aq', 'pNH4_aq'/), &
        DST_ID = MATRIX, SRC_ID = GEOSACHEM, __RC__  )
  ENDIF

! GOCART <=> StratChem coupling ...
! ---------------------------------
  IF(chemReg%doing_GOCART .AND. chemReg%doing_SC) then
     CALL MAPL_AddConnectivity ( GC, &
          SHORT_NAME = (/ "GOCART::DMS", "GOCART::SO2", "GOCART::SO4", "GOCART::MSA" /), &
          DST_ID = STRATCHEM, SRC_ID = GOCART, __RC__)
     CALL MAPL_AddConnectivity ( GC, &
          SHORT_NAME = (/"GOCART::DMSv", "GOCART::SO2v", "GOCART::SO4v", "GOCART::MSAv" /), &
          DST_ID = STRATCHEM, SRC_ID = GOCART, __RC__)
  END IF

! GOCART.data <=> GMICHEM coupling ...
! ------------------------------------
  IF(myState%enable_GMICHEM .AND. TRIM(providerName) == "GOCART.data") THEN
   CALL MAPL_AddConnectivity ( GC, SHORT_NAME=(/ "AERO" /), DST_ID=GMICHEM, SRC_ID=GOCARTdata, __RC__)
  END IF

! GOCART <=> GMICHEM coupling ...
! -------------------------------
  IF(myState%enable_GMICHEM .AND. TRIM(providerName) == "GOCART") THEN

! GOCART connections to GMICHEM
! -----------------------------
   IF(chemReg%doing_BC) &
   CALL MAPL_AddConnectivity ( GC, &
     SHORT_NAME  = (/ "GOCART::BCphobic", "GOCART::BCphilic" /), &
     DST_ID=GMICHEM, SRC_ID=GOCART, __RC__)

   IF(chemReg%doing_DU) &
   CALL MAPL_AddConnectivity ( GC, &
     SHORT_NAME  = (/ "GOCART::du001", "GOCART::du002", "GOCART::du003", &
                      "GOCART::du004", "GOCART::du005" /), &
     DST_ID=GMICHEM, SRC_ID=GOCART, __RC__)

   IF(chemReg%doing_OC) &
   CALL MAPL_AddConnectivity ( GC, &
     SHORT_NAME  = (/ "GOCART::OCphobic", "GOCART::OCphilic" /), &
     DST_ID=GMICHEM, SRC_ID=GOCART, __RC__)

   IF(chemReg%doing_SS) &
   CALL MAPL_AddConnectivity ( GC, &
     SHORT_NAME  = (/ "GOCART::ss001", "GOCART::ss002", "GOCART::ss003", &
                      "GOCART::ss004", "GOCART::ss005" /), &
     DST_ID=GMICHEM, SRC_ID=GOCART, __RC__)

   IF(chemReg%doing_SU) &
   CALL MAPL_AddConnectivity ( GC, &
     SHORT_NAME  = (/ "GOCART::DMS", "GOCART::SO2", &
                      "GOCART::SO4", "GOCART::MSA" /), &
     DST_ID=GMICHEM, SRC_ID=GOCART, __RC__)

! GMICHEM connections to GOCART
! -----------------------------
  IF(myState%enable_GOCART) THEN

! ... For GOCART::SU
! ------------------
   CALL MAPL_AddConnectivity ( GC, &
    SHORT_NAME  = (/ "OH  ", "CH4 ", "H2O2", "NO3 " /), &
    DST_ID=GOCART, SRC_ID=GMICHEM, __RC__)

! ... For GOCART::CFC
! -------------------
   CALL MAPL_AddConnectivity ( GC, &
    SRC_NAME  = (/"OX"/), &
    DST_NAME  = (/"O3"/), &
    DST_ID=GOCART, SRC_ID=GMICHEM, __RC__)
  
  END IF

! GMICHEM connections to MAM
! -----------------------------
   IF(myState%enable_MAM) &
   CALL MAPL_AddConnectivity ( GC, &
    SHORT_NAME  = (/ "ALK4", "CH4 ", "C2H6", "C3H6", "C3H8", "ISOP" /), &
    DST_ID=MAM, SRC_ID=GMICHEM, __RC__)

! GMICHEM connections to GEOSACHEM
! --------------------------------
   IF(myState%enable_GEOSACHEM) THEN

   CALL MAPL_AddConnectivity ( GC, &
    SHORT_NAME  = (/ "OH  ", "H2O2", "NO3 " /), &
    DST_ID=GEOSACHEM, SRC_ID=GMICHEM, __RC__)

   CALL MAPL_AddConnectivity ( GC, &
    SRC_NAME  = (/"OX"/), &
    DST_NAME  = (/"O3"/), &
    DST_ID=GEOSACHEM, SRC_ID=GMICHEM, __RC__)

   END IF

  END IF


! PCHEM connections to GEOSCHEM
! -----------------------------
  IF(myState%enable_PCHEM .AND. myState%enable_GEOSCHEM) THEN

   CALL MAPL_AddConnectivity ( GC, &
    SHORT_NAME  = (/ "TO3" /), &
    DST_ID=GEOSCHEM, SRC_ID=PCHEM, __RC__)

  END IF


! Finally, set the services
! -------------------------
  call MAPL_GenericSetServices ( GC, __RC__ )

  RETURN_(ESMF_SUCCESS)
  
  end subroutine SetServices


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!BOP
! !IROUTINE: Initialize -- Initialized method for composite Aero-Chemistry

! !INTERFACE:

  subroutine Init ( GC, IMPORT, EXPORT, CLOCK, RC )

! !ARGUMENTS:

  type(ESMF_GridComp), intent(inout) :: GC     ! Gridded component 
  type(ESMF_State),    intent(inout) :: IMPORT ! Import state
  type(ESMF_State),    intent(inout) :: EXPORT ! Export state
  type(ESMF_Clock),    intent(inout) :: CLOCK  ! The clock
  integer, optional,   intent(  out) :: RC     ! Error code

! !DESCRIPTION: The Initialize method of the Chemistry Composite Gridded 
!  Component. It acts as a driver for the initializtion of the children.

!EOP

! ErrLog Variables

  __Iam__('Init')
  character(len=ESMF_MAXSTR)           :: COMP_NAME

! Local derived type aliases

   type (MAPL_MetaComp),       pointer :: MAPL
   type (ESMF_GridComp),       pointer :: GCS(:)
   type (ESMF_State)                   :: INTERNAL
   type (ESMF_State),          pointer :: GEX(:)
   type (ESMF_FieldBundle)             :: fBUNDLE
   type (ESMF_State)                   :: AERO
   type (ESMF_State)                   :: AERO_ACI
   type (ESMF_Config)                  :: CF, myCF

!   Private state
!   -------------
    type (GEOS_ChemGridComp), pointer  :: myState   ! private, that is
    type (GEOS_ChemGridComp_Wrap)      :: wrap

!=============================================================================
 
! Begin... 

!   Get the target components name and set-up traceback handle.
!   -----------------------------------------------------------
    call ESMF_GridCompGet ( GC, name=COMP_NAME, CONFIG=CF, __RC__ )
    Iam = trim(COMP_NAME) // "::" // trim(Iam)

!   Get my internal state
!   ---------------------
    call ESMF_UserCompGetInternalState(gc, 'GEOSchem_GridComp_State', WRAP, STATUS)
    VERIFY_(STATUS)
    myState => wrap%ptr

!   Call GenericInitialize for every Child
!   --------------------------------------
    call MAPL_GenericInitialize ( GC, IMPORT, EXPORT, CLOCK, __RC__ )

!   Get my MAPL_Generic state
!   --------------------------
    call MAPL_GetObjectFromGC ( GC, MAPL, __RC__ )

!   Get children and their im/ex states from my generic state.
!   ----------------------------------------------------------
    call MAPL_Get ( MAPL, GCS=GCS, GEX=GEX,              &
                    INTERNAL_ESMF_STATE=INTERNAL, __RC__ )

!   Fill in INTERNAL friendly bundle: CHEM_TRACERS
!   VERY IMPORTANT: Only the RATS provider can make OX friendly
!                   to ANALYSIS.
!   -----------------------------------------------------------
    call ESMF_StateGet (INTERNAL, 'CHEM_TRACERS', fBUNDLE, __RC__)
    call MAPL_GridCompGetFriendlies(GCS,                     &
                                         (/ "ANALYSIS  ",    &
                                            "TURBULENCE",    &
                                            "DYNAMICS  ",    &
                                            "MOIST     " /), &
                                           fBUNDLE, __RC__ )

!   AERO State for AERO_PROVIDER set to NONE
!   ----------------------------------------
    if (myState%AERO_PROVIDER < 0) then
        ! Radiation will not call the aerosol optics method 
        ! unless this attribute is explicitly set to true.
        call ESMF_StateGet(EXPORT, 'AERO', AERO, __RC__)
        call ESMF_AttributeSet(AERO, name='implements_aerosol_optics_method', value=.false., __RC__)
    end if

#ifdef PRINT_STATES

!   Print what my states are
!   ------------------------
    if ( MAPL_am_I_root() ) then

       print *,  trim(Iam)//": IMPORT State" 
                                             call ESMF_StatePrint ( IMPORT )
       print *,  trim(Iam)//": INTERNAL State" 
                                             call ESMF_StatePrint ( INTERNAL )
       print *,  trim(Iam)//": EXPORT State" 
                                             call ESMF_StatePrint ( EXPORT )

       print *,  trim(Iam)//": AERO State (EXPORT)"
                                             call ESMF_StateGet   ( EXPORT, 'AERO', AERO, __RC__ ) 
                                             call ESMF_StatePrint ( AERO, nestedFlag=.false., __RC__ )
       print *,  trim(Iam)//": AERO State (PROVIDER)",  myState%AERO_PROVIDER
                                         if (myState%AERO_PROVIDER > 0) then
                                             call ESMF_StateGet   ( GEX(myState%AERO_PROVIDER), 'AERO', AERO, __RC__ )
                                             call ESMF_StatePrint ( AERO, nestedFlag=.false., __RC__ )
                                         end if

       print *,  trim(Iam)//": AERO_ACI State (EXPORT)"
                                             call ESMF_StateGet   ( EXPORT, 'AERO_ACI', AERO_ACI, __RC__ ) 
                                             call ESMF_StatePrint ( AERO_ACI, nestedFlag=.false., __RC__ )
       print *,  trim(Iam)//": AERO State (PROVIDER)",  myState%AERO_PROVIDER
                                         if (myState%AERO_PROVIDER > 0) then
                                             call ESMF_StateGet   ( GEX(myState%AERO_PROVIDER), 'AERO_ACI', AERO_ACI, __RC__ )
                                             call ESMF_StatePrint ( AERO_ACI, nestedFlag=.false., __RC__ )
                                         end if

       print *,  trim(Iam)//": Friendly Tracers (INTERNAL)"
                                             call ESMF_FieldBundlePrint ( fBUNDLE )
    end if

#endif

!   All Done
!   --------
    RETURN_(ESMF_SUCCESS)

  end subroutine Init


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! !IROUTINE: Run -- Run method for the composite Physics Gridded Component

! !INTERFACE:

  subroutine Run ( GC, IMPORT, EXPORT, CLOCK, RC )

! !ARGUMENTS:

  type(ESMF_GridComp), intent(inout) :: GC     ! Gridded component 
  type(ESMF_State),    intent(inout) :: IMPORT ! Import state
  type(ESMF_State),    intent(inout) :: EXPORT ! Export state
  type(ESMF_Clock),    intent(inout) :: CLOCK  ! The clock
  integer, optional,   intent(  out) :: RC     ! Error code

! !DESCRIPTION: The Run method of the Chemistry Composite Gridded Component.
!  It acts as a driver for the initializtion of the children.

!EOP

! ErrLog Variables

  __Iam__('Run')
  character(len=ESMF_MAXSTR)           :: COMP_NAME

! Local derived type aliases

  type (MAPL_MetaComp),     pointer  :: MAPL
  type (ESMF_Alarm)                  :: ALARM

!=============================================================================

    INTEGER, PARAMETER :: numRATs = 8
    CHARACTER(LEN=ESMF_MAXSTR) :: specieName(numRats) = (/"OX    ", "O3    ", "CH4   ", "N2O   ", &
                                                          "CFC11 ", "CFC12 ", "HCFC22", "O3PPMV"/)
    integer :: i, im, jm, km, ijm
    real, pointer :: rat(:,:,:)
    real :: qmax, qmin

! Begin... 

! Get parameters from generic state. The RUNALARM is used to control
!  the calling of the full chemistry
!-------------------------------------------------------------------

   call MAPL_GetObjectFromGC ( GC, MAPL, RC=STATUS)
   VERIFY_(STATUS)
   call MAPL_Get(MAPL,              &
         RUNALARM            = ALARM,            &
                                       RC=STATUS )
   VERIFY_(STATUS)


   if ( ESMF_AlarmIsRinging   (ALARM, RC=STATUS) ) then
      call ESMF_AlarmRingerOff(ALARM, RC=STATUS)
      VERIFY_(STATUS)

!   Get the target components name and set-up traceback handle.
!   -----------------------------------------------------------
    call ESMF_GridCompGet ( GC, name=COMP_NAME, __RC__ )
    Iam = trim(COMP_NAME) // "::" // trim(Iam)

!   Call Run for every Child
!   -------------------------
    call MAPL_GenericRun ( GC, IMPORT, EXPORT, CLOCK,  __RC__)

!   Check contents of RATS
!   ----------------------
#ifdef DEBUG
    if ( MAPL_AM_I_ROOT() ) then
       print *, '--------------------------------------------------------'
    end if
    do i = 1, numRATS
       rat => null()
       call MAPL_GetPointer ( EXPORT, rat, trim(specieName(i)), __RC__)
       if ( .not. associated(rat) ) then
          print *, '******** RAT with null pointer: ', trim(specieName(i))
       else
          im = ubound(rat,1)
          jm = ubound(rat,2)
          km = ubound(rat,3)
          ijm = im*jm
          call pmaxmin(specieName(i), rat(:,:,:), qmin, qmax, ijm, km, 1. )
       end if
    end do
#endif

   endif

!   All Done
!   --------
    RETURN_(ESMF_SUCCESS)

 end subroutine Run

!-----------------------------------------------------------------------

     subroutine GetProvider_ ( CF, Label, ID, Name, Default, RC )
!
!    Returns provider name as per resource file.
!
     type (ESMF_Config), intent(inout) :: CF
     character(len=*),   intent(in)    :: Label
     character(len=*),   intent(in)    :: Default
     character(len=*),   intent(out)   :: Name
     integer, intent(out)              :: ID
     integer, intent(out)              :: RC

! ErrLog Variables

     __Iam__('GetProvider_')

     character(len=ESMF_MAXSTR)        :: message
     character(len=ESMF_MAXSTR)        :: providerName


     call ESMF_ConfigGetAttribute(CF, providerName, Default=trim(default), &
                                  Label=Label, __RC__)

     ID   = -1 
     name = trim(providerName)

     select case ( trim(name) )

           case ('PCHEM')
                                    ID = PCHEM
           case ('GOCART.data')
                                    ID = GOCARTdata
           case ('GOCART')
                                    ID = GOCART
           case ('GAAS')
                                    ID = GAAS
           case ('H2O')
                                    ID = H2O
           case ('STRATCHEM')
                                    ID = STRATCHEM
           case ('GMICHEM')
                                    ID = GMICHEM
           case ('CARMA')
                                    ID = CARMA
           case ('GEOSCHEMCHEM')
                                    ID = GEOSCHEM
           case ('MATRIX')
                                    ID = MATRIX
           case ('MAM')
                                    ID = MAM
           case ('MAM.data')
                                    ID = MAMdata
           case ('TR')
                                    ID = TR
           case DEFAULT

              message = "unknown provider "//trim(name)
              __raise__(MAPL_RC_ERROR, message)

     end select
     
     if ( ID < 0 ) then
        message = "Component "//trim(name)//" is NOT enabled; cannot specify it for "//trim(label)
        __raise__(MAPL_RC_ERROR, message)
     end if

     RC = ESMF_SUCCESS

   end subroutine GetProvider_

end module GEOS_ChemGridCompMod


