#include "MAPL_Generic.h"

!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 900.3     !
!-------------------------------------------------------------------------
!BOP
!
! !MODULE:  OH_GridCompMod --- OH Parameterization Grid Component Class
!
! !INTERFACE:
!

   MODULE  OH_Para_GridCompMod

! !USES:

   USE ESMF
   USE MAPL_Mod

   USE Chem_Mod              ! Chemistry Base Class
   USE Chem_StateMod         ! Chemistry State
   USE Chem_UtilMod          ! I/O

   USE m_inpak90             ! Resource file management
   USE m_die, ONLY: die
   USE ohParameterizationMethod_mod
   USE solarZenithAngle_mod
   USE computeReflectivity_mod

   IMPLICIT NONE

! !PUBLIC TYPES:
!
   PRIVATE
   PUBLIC  OH_GridComp       ! OH object 

!
! !PUBLIC MEMBER FUNCTIONS:
!

   PUBLIC  OH_GridCompSetServices
   PUBLIC  OH_GridCompInitialize
   PUBLIC  OH_GridCompRun
   PUBLIC  OH_GridCompFinalize

#     include "CMN_SIZE"       ! Size parameters
#     include "CMN"            ! STT
#     include "CMN_CO"         ! CO arrays
#     include "CMN_CO_BUDGET"  ! FMOL_CO
#     include "CMN_OH"
!
!
! !DESCRIPTION:
!
!  This module implements the OH Parameterization Grid Component. 
!
! !REVISION HISTORY:
!
!  06September2011 Kouatchou  First crack.
!  27Oct2015       Manyin     Included updates from Yasin Elshorbany and
!                             adapted for Heracles.
!
!EOP
!-------------------------------------------------------------------------

! MANYIN- Notes (things to change/test in future)
!   LPAUSE should be 36
!   o3up should be 0.0 at level km
!   i_OH+6 (DryAirMass24h) never gets used in the current GETINFO module
!   Current year is always used in Acquire_GMIspecies, but
!           2001 is always used in Acquire_OHratios

  TYPE OH_GridComp

       CHARACTER(LEN=255) :: name            ! generic name of the package
       CHARACTER(LEN=255) :: iname           ! instance name
       CHARACTER(LEN=255) :: rcfilen         ! resource file name
       CHARACTER(LEN=255) :: OHFileName      ! OH concentration file name
       CHARACTER(LEN=255) :: GMI_infile_name ! GMI file name
       CHARACTER(LEN=255) :: Qscl_infile_name ! Q scalar file name
       CHARACTER(LEN=255) :: gmiSpeciesName(20)

       INTEGER :: instance                   ! instance number
       INTEGER :: nymd_oh
       INTEGER :: BCnymd   
       INTEGER :: BCnymd_24h
       INTEGER :: BCnymdCO_24h
       INTEGER :: BCnymdCH4_24h
       INTEGER :: BCnymd_month
       INTEGER :: BCnymd_monthOH

       LOGICAL :: DBG

       REAL*8, POINTER :: Temp24h(:,:,:)          ! 24h average temperature                              bottom-up
       REAL*8, POINTER :: Press24h(:,:,:)         ! 24h average pressure (hPa, bottom of gridbox)        bottom-up
     ! REAL*8, POINTER :: Tropp24h(:,:)           ! 24h average tropopause pressure
     ! REAL*8, POINTER :: CloudFlux24h(:,:,:)     ! 24h average cloud flux
       REAL*8, POINTER :: SpecHum24h(:,:,:)       ! 24h average specific humidity                        bottom-up
       REAL*8, POINTER :: DryAirMass24h(:,:,:)    ! 24h average dry air mass                                       top-down
     ! REAL*8, POINTER :: optDepth24h(:,:,:)      ! 24h average optical depth

      REAL*8, POINTER :: TempAcc(:,:,:)          ! Accumulation of temperature                           bottom-up
      REAL*8, POINTER :: PressAcc(:,:,:)         ! Accumulation of pressure (hPa)                        bottom-up
    ! REAL*8, POINTER :: TroppAcc(:,:)           ! Accumulation of tropopause pressure
    ! REAL*8, POINTER :: CloudFluxAcc(:,:,:)     ! Accumulation of cloud flux
      REAL*8, POINTER :: SpecHumAcc(:,:,:)       ! Accumulation of specific humidity                     bottom-up
      REAL*8, POINTER :: DryAirMassAcc(:,:,:)    ! Accumulation of dry air mass                                    top-down
    ! REAL*8, POINTER :: optDepthAcc(:,:,:)      ! Accumulation of optical depth

      REAL*8, POINTER :: Ravga(:,:,:)            !                                                       bottom-up
      REAL*8, POINTER :: Ravgb(:,:,:)            !                                                       bottom-up

      INTEGER, POINTER :: LPAUSE(:,:)            ! tropopause level

      REAL*8, POINTER :: troppresIn(:,:)         ! tropopause pressure read from file
      REAL*8, POINTER :: qIn(:,:,:)              ! GMI water vapor pressure read from file        unused
      REAL*8, POINTER :: CH4_24h(:,:,:)          ! 24h average of CH4                                    bottom-up
      REAL*8, POINTER :: CO_24h (:,:,:)          ! 24h average of CO                                     bottom-up
      REAL*8, POINTER :: CH4qMassAcc(:,:,:)      ! 24h average of CH4                                              top-down
      REAL*8, POINTER :: COqMassAcc(:,:,:)       ! 24h average of CO                                               top-down
!     REAL*8, POINTER :: CH4_Acc(:,:,:)          ! Accumulation of CH4
!     REAL*8, POINTER :: CO_Acc (:,:,:)          ! Accumulation of CO
      REAL, ALLOCATABLE :: rxnRateRatio(:,:,:)   ! ratio to adjust for updated  reaction rate constants  bottom-up


      type(t_OHparam) :: OHparam
  END TYPE OH_GridComp

  REAL,    PARAMETER :: mwtAir = 28.97
  REAL,    PARAMETER :: mwtWat = 18.016
  REAL,    PARAMETER :: mwtCO  = 28.01
  REAL,    PARAMETER :: mwtCH4 = 16.043
  REAL,    PARAMETER :: Pa2hPa = 0.01
  REAL,    PARAMETER :: secPerDay = 86400.00
  INTEGER, PARAMETER :: negativeOne = -1

  INTEGER :: use_offline_fields=0

!-------------------------------------------------------------------------
CONTAINS

!------------------------------------------------------------------------
!     NASA/GSFC, Atmospheric Chemistry and Dynamics Lab, Code 614       !
!------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  OH_GridCompSetServices --- Set Services for OH_Para_GridComp
!
! !INTERFACE:
!
   subroutine OH_GridCompSetServices( gc, chemReg, rc)

! !USES:

  IMPLICIT NONE

! !INPUT PARAMETERS:

! !OUTPUT PARAMETERS:

   TYPE(ESMF_GridComp), INTENT(INOUT) :: gc
   TYPE(Chem_Registry), INTENT(INOUT) :: chemReg
   INTEGER,             INTENT(OUT  ) :: rc

! !DESCRIPTION: Set Services for the OH Grid Component.
!
! !REVISION HISTORY:
!
!  23Nov2015  Manyin    First version
!
!EOP
!-------------------------------------------------------------------------

   type (MAPL_MetaComp),  pointer   :: MAPL

   integer            :: status
   character(len=ESMF_MAXSTR) :: Iam

   Iam = "OH_GridCompSetServices"

!  Get my MAPL_Generic state
!  -------------------------
   call MAPL_GetObjectFromGC ( GC, MAPL, __RC__ )

!  Do we use the ECCOH offline fields?
!  -----------------------------------
   call MAPL_GetResource ( MAPL, use_offline_fields, Label="ECCOH_USE_OFFLINE_METFIELDS:", DEFAULT=0, __RC__ )

!  If using offline fields, MAPL_AddImportSpec() has already been called by Aero_GridCompSetServices
!  for fields needed in OH code: T, TV, Q, FCLD, TAUCLW, TAUCLI, ALBVF

!  if ( use_offline_fields==1 ) then
!     IF(MAPL_AM_I_ROOT()) PRINT *, TRIM(Iam)//': Using OFFLINE (file-based) met fields for ECCOH'
!  else
!     IF(MAPL_AM_I_ROOT()) PRINT *, TRIM(Iam)//': Using ONLINE met fields for ECCOH (default)'
!  end if


   RETURN_(ESMF_SUCCESS)

   end subroutine OH_GridCompSetServices


!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 900.3     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  OH_GridCompInitialize --- Initialize OH_GridComp
!
! !INTERFACE:
!

      subroutine OH_GridCompInitialize ( this, w_c, impChem, expChem, &
                                      nymd, nhms, cdt, rc )
! !USES:

     IMPLICIT NONE

! !INPUT PARAMETERS:
      INTEGER, INTENT(IN) :: nymd, nhms        ! time
      REAL,    INTENT(IN) :: cdt               ! chemical timestep (secs)
      TYPE(Chem_Bundle), intent(in) :: w_c     ! Chemical tracer fields      
!
! !OUTPUT PARAMETERS:

      TYPE(OH_GridComp), INTENT(INOUT) :: this   ! Grid Component
      TYPE(ESMF_State), INTENT(INOUT)  :: impChem  ! Import State
      TYPE(ESMF_State), INTENT(INOUT)  :: expChem  ! Export State
      INTEGER, INTENT(OUT) ::  rc                  ! Error return code:
                                                   !  0 - all is well
                                                   !  1 - 
!
! !DESCRIPTION: 
! Initializes the OH Parameterization Grid Component. 
!
! !REVISION HISTORY:
!
!  24June2010 Nielsen  First crack.
!
!EOP
!-------------------------------------------------------------------------

      CHARACTER(LEN=*), PARAMETER :: myname = 'OH_GridCompInitialize'
      CHARACTER(LEN=255) :: name
   
      INTEGER :: ios, j, i, ic, n
      INTEGER, ALLOCATABLE :: ier(:)

      INTEGER :: JDAY, curMonth
      INTEGER :: i1, i2, im, j1, j2, jm, km
      INTEGER :: nbeg, nend, begTime, nTimes, incSecs

      this%rcfilen = 'OH_GridComp.rc'
      this%name = 'GEOS-5/GOCART OH Parameterization Package'

      IF(MAPL_AM_I_ROOT()) PRINT *,myname,": Begin..."

!  Initialize local variables
!  --------------------------

      rc = 0
      i1 = w_c%grid%i1
      i2 = w_c%grid%i2
      im = w_c%grid%im

      j1 = w_c%grid%j1
      j2 = w_c%grid%j2
      jm = w_c%grid%jm

      km = w_c%grid%km

      !------------------------------------------------
      ! i_OH+0: the only OH bin
      ! i_OH+1: 24 hour average of CO
      ! i_OH+2: 24 hour average of CH4
      ! i_OH+3: 24 hour average of temperature
      ! i_OH+4: 24 hour average of atmospheric pressure
      ! i_OH+5: 24 hour average of specific humidity
      ! i_OH+6: 24 hour average of dry air mass     -> ONLY WAS USED IN AN OLD VERSION OF GETINFO
      ! j_OH = i_OH + 6
      !------------------------------------------------

      nbeg  = w_c%reg%i_OH
      nend  = w_c%reg%j_OH

!  It requires 7 bins
!  -----------------
      if ( (nend - nbeg) /= 6 ) then
         IF(MAPL_AM_I_ROOT()) PRINT *,myname,": Not enough OH related bins"
         rc = 10
         return
      end if

      this%BCnymd         = negativeOne
      this%BCnymd_24h     = negativeOne
      this%BCnymdCO_24h   = negativeOne
      this%BCnymdCH4_24h  = negativeOne
      this%BCnymd_month   = negativeOne
      this%BCnymd_monthOH = negativeOne
      allocate(ier(199))
      ier(:) = 0

!  Load resource file
!  ------------------
      CALL I90_loadf ( TRIM(this%rcfilen), ier(1) )
      if ( ier(1) .NE. 0 ) then
         rc = 11
         return
      end if

!  Parse resource file
!  -------------------
      CALL I90_label ( 'GMI_infile_name:', ier(3) )
      CALL I90_gtoken( this%GMI_infile_name, ier(4) )

! only needed if we are calling adjustH2OMER
!     CALL I90_label ( 'Qscale_infile_name:', ier(5) )
!     CALL I90_gtoken( this%Qscl_infile_name, ier(6) )

      IF ( ANY(ier(:) /= 0)) THEN
!        PRINT*, TRIM(myname),': Failed to parse GMI or Qscale infile name.'
         PRINT*, TRIM(myname),': Failed to parse GMI infile name.'
         rc = 14
         RETURN
      END IF
      ier(:)=0

      ! GMI species names for the BBIJ array:
      ! MANYIN-- note: only NFIELDS2 (12) species are currently allocated
      ! MANYIN         see paramCode/include/CMN_CO
      this%gmiSpeciesName(1)  = 'NO'
      this%gmiSpeciesName(2)  = 'NO2'
      this%gmiSpeciesName(3)  = 'N2O5'
      this%gmiSpeciesName(4)  = 'NO3'
      this%gmiSpeciesName(5)  = 'HNO2'
      this%gmiSpeciesName(6)  = 'HNO4'
      this%gmiSpeciesName(7)  = 'ALK4'         ! alkanes
      this%gmiSpeciesName(8)  = 'ISOP'         ! Isoprene
      this%gmiSpeciesName(9)  = 'ACET'         ! Acetone
      this%gmiSpeciesName(10) = 'PRPE'         ! Propene
      this%gmiSpeciesName(11) = 'C3H8'         ! Propane
      this%gmiSpeciesName(12) = 'C2H6'         ! Ethane
      this%gmiSpeciesName(13) = 'O3' 
      this%gmiSpeciesName(14) = 'CO'
      this%gmiSpeciesName(15) = 'CH4'
      this%gmiSpeciesName(16) = 'TRPPS'
      this%gmiSpeciesName(17) = 'METWATER'


      allocate(       this%CO_24h(i1:i2,j1:j2,1:km), STAT=ier(10))
      allocate(      this%CH4_24h(i1:i2,j1:j2,1:km), STAT=ier(11))
      allocate(      this%Temp24h(i1:i2,j1:j2,1:km), STAT=ier(12))
      allocate(     this%Press24h(i1:i2,j1:j2,1:km), STAT=ier(13))
      allocate(   this%SpecHum24h(i1:i2,j1:j2,1:km), STAT=ier(14))
      allocate(this%DryAirMass24h(i1:i2,j1:j2,1:km), STAT=ier(15))
      allocate( this%rxnRateRatio(i1:i2,j1:j2,1:km), STAT=ier(16))
      allocate(   this%COqMassAcc(i1:i2,j1:j2,1:km), STAT=ier(17))
      allocate(  this%CH4qMassAcc(i1:i2,j1:j2,1:km), STAT=ier(18)) 

      IF ( ANY(ier(:) /= 0)) THEN
         PRINT*, TRIM(myname),': Failed to allocate work space.'
         rc = 12
         RETURN
      END IF
      ier(:) = 0

      this%CO_24h        = 0.0d0
      this%CH4_24h       = 0.0d0
      this%Temp24h       = 0.0d0
      this%Press24h      = 0.0d0
      this%SpecHum24h    = 0.0d0
      this%DryAirMass24h = 0.0d0
      this%rxnRateRatio  =       1.0d0
      this%COqMassAcc    = 0.0d0
      this%CH4qMassAcc   = 0.0d0

!     allocate(       this%CO_Acc(i1:i2,j1:j2,km), STAT=ier(20))
!     allocate(      this%CH4_Acc(i1:i2,j1:j2,km), STAT=ier(21))
      allocate(      this%TempAcc(i1:i2,j1:j2,km), STAT=ier(22))
      allocate(     this%PressAcc(i1:i2,j1:j2,km), STAT=ier(23))
      allocate(   this%SpecHumAcc(i1:i2,j1:j2,km), STAT=ier(24))
      allocate(this%DryAirMassAcc(i1:i2,j1:j2,km), STAT=ier(25))

      IF ( ANY(ier(:) /= 0)) THEN
         PRINT*, TRIM(myname),': Failed to allocate work space.'
         rc = 13
         RETURN
      END IF
      ier(:) = 0

!     this%CO_Acc        = 0.0d0
!     this%CH4_Acc       = 0.0d0
      this%TempAcc       = 0.0d0
      this%PressAcc      = 0.0d0
      this%SpecHumAcc    = 0.0d0
      this%DryAirMassAcc = 0.0d0

      allocate( this%Ravga(i1:i2, j1:j2, km), this%Ravgb(i1:i2, j1:j2, km), STAT=ier(26))
      allocate(this%LPAUSE(i1:i2, j1:j2), this%troppresIn(i1:i2, j1:j2), this%qIn(i1:i2,j1:j2,km), STAT=ier(27))

      IF ( ANY(ier(:) /= 0)) THEN
         PRINT*, TRIM(myname),': Failed to allocate space.'
         rc = 15
         RETURN
      END IF
      ier(:) = 0
   
      this%Ravga = 0.0d0
      this%Ravgb = 0.0d0

          this%LPAUSE(i1:i2, j1:j2)     = 0.0d0
      this%troppresIn(i1:i2, j1:j2)     = 0.0d0
             this%qIn(i1:i2, j1:j2, : ) = 0.0d0

      CALL initializeOHparam(this%OHparam, i2, j2, km, this%rcfilen)

      DEALLOCATE(ier)

      RETURN

 end subroutine OH_GridCompInitialize

!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 900.3     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  OH_GridCompRun --- Run OH_GridComp
!
! !INTERFACE:
!

   subroutine OH_GridCompRun ( this, w_c, impChem, expChem, &
                                      nymd, nhms, cdt, rc )

! !USES:

  IMPLICIT NONE

! !INPUT PARAMETERS:

   TYPE(Chem_Bundle), intent(in) :: w_c        ! Chemical tracer fields      
   INTEGER, INTENT(IN) :: nymd, nhms           ! time
   REAL,    INTENT(IN) :: cdt                  ! chemical timestep (secs)


! !OUTPUT PARAMETERS:

   TYPE(OH_GridComp), INTENT(INOUT) :: this   ! Grid Component
   TYPE(ESMF_State), INTENT(INOUT)  :: impChem  ! Import State
   TYPE(ESMF_State), INTENT(INOUT)  :: expChem  ! Export State
   INTEGER, INTENT(OUT) ::  rc                  ! Error return code:
                                                !  0 - all is well
                                                !  1 - 

! !DESCRIPTION: Runs the OH Parameterization Grid Component. Multiple instance
!               version.
!
! !REVISION HISTORY:
!
!  24June2010 Nielsen  First crack.
!
!EOP
!-------------------------------------------------------------------------

      INTEGER, ALLOCATABLE :: ier(:)
      CHARACTER(LEN=*), PARAMETER :: myname = 'OH_GridCompRun'
      CHARACTER(LEN=*), PARAMETER :: Iam = myname

!  Input fields from fvGCM
!  -----------------------

      REAL, POINTER, DIMENSION(:,:)   ::  tropp     => null()
      REAL, POINTER, DIMENSION(:,:)   ::  albedo    => null()
      REAL, POINTER, DIMENSION(:,:)   ::  albedoVF  => null()
      REAL, POINTER, DIMENSION(:,:)   ::  albedoVR  => null()
      REAL, POINTER, DIMENSION(:,:)   ::  albedoNF  => null()
      REAL, POINTER, DIMENSION(:,:)   ::  albedoNR  => null()      
      REAL, POINTER, DIMENSION(:,:)   ::  cellArea  => null()

!  All are  top-down:
      REAL, POINTER, DIMENSION(:,:,:) ::  T          => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  TV         => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exO3up     => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exOHVMR    => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exNOy      => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exvarOut   => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exALK4     => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exIsoprene => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exAcetone  => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exPropane  => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exPropene  => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exEthane   => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exOzone    => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exSpecHum  => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exCO       => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exCH4      => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exPress    => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exTemp     => null()
      REAL, POINTER, DIMENSION(:,:)   ::  exTRPPS    => null()
      REAL, POINTER, DIMENSION(:,:)   ::  exAlbedo   => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exRavga    => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exRavgb    => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exRLAT     => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  exDEC      => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  ple        => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  zle        => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  rhoWet     => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  rhoDry     => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  specHum    => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  taucli     => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  tauclw     => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  fcld       => null()

      REAL, ALLOCATABLE :: pe(:,:,:),p(:,:,:),ndDry(:,:,:),ndWet(:,:,:)   !   top-down
      REAL, ALLOCATABLE :: mw_moist_air(:,:,:)                  !   bottom-up

      REAL*8,  ALLOCATABLE :: scaleO3    (:,:)
      REAL*8,  ALLOCATABLE :: solarZenithAngle    (:,:)
      REAL*8,  ALLOCATABLE :: o3up       (:,:,:)                !   bottom-up
      REAL*8,  ALLOCATABLE :: toms_climO3(:,:,:)      !  never allocated or referenced
      REAL*8,  ALLOCATABLE :: STT        (:,:,:,:)    !  ppbv       bottom-up
      REAL*8,  ALLOCATABLE :: Qppmv      (:,:,:)      !             bottom-up
      REAL*8,  ALLOCATABLE :: OHparamNum (:,:,:)      !             bottom-up
      REAL*8,  ALLOCATABLE :: CH4VAR     (:,:,:)      !             bottom-up
      REAL*8,  ALLOCATABLE :: RLON       (:,:)   ! Array of longitudes
      REAL*8,  ALLOCATABLE :: RLAT       (:,:)   ! Array of latitudes
      REAL*8,  ALLOCATABLE :: BBIJ       (:, :, :,:)  !             bottom-up
      REAL*8,  ALLOCATABLE :: gmiOH(:,:,:) !-sas                    bottom-up
      REAL*8,  ALLOCATABLE :: mass(:,:,:)             !  of dry air           top-down
      REAL*8,  ALLOCATABLE :: gridBoxThickness(:,:,:) !                       top-down
      REAL*8,  ALLOCATABLE :: PL(:,:,:)               !  never used          (top-down)
      REAL*8,  ALLOCATABLE :: optDepth(:,:,:)         !             bottom-up
      REAL*8,  ALLOCATABLE :: ALBD(:,:)
      REAL*8,  ALLOCATABLE :: tropPress(:,:)
      REAL,    ALLOCATABLE :: var3d(:,:,:)
      REAL,    ALLOCATABLE :: var2d(:,:)

      INTEGER :: i1, i2, im, j1, j2, jm, km, ios, iXj
      INTEGER :: i, j, k, kReverse, n, nbeg, nend, ic
      INTEGER :: hms

      INTEGER :: JDAY, curMonth
      INTEGER :: cymd, dymd, ix, jx
      REAL*8  :: time_sec, offset_sec
      integer :: STATUS
      REAL    :: radToDeg, degToRad, pi
      REAL    :: qmin, qmax, h20factor

     
      pi       = 4.00*ATAN(1.00)
      degToRad = pi/180.00
      radToDeg = 180.00/pi

      radToDeg = 57.2957795

      IF(MAPL_AM_I_ROOT()) PRINT *,myname,": begin...."

      !  Initialize local variables
      !  --------------------------
      rc = 0
      i1 = w_c%grid%i1
      i2 = w_c%grid%i2
      im = w_c%grid%im
   
      j1 = w_c%grid%j1
      j2 = w_c%grid%j2
      jm = w_c%grid%jm
      km = w_c%grid%km
   
      iXj = ( i2 - i1 + 1 ) * ( j2 - j1 + 1 )

      allocate(ier(128))
      ier(:) = 0

      nbeg  = w_c%reg%i_OH
      nend  = w_c%reg%j_OH

      !  It requires 7 bins
      !  -----------------
      if ( (nend - nbeg) /= 6 ) then
         IF(MAPL_AM_I_ROOT()) PRINT *,myname,": Must have 7 OH bins"
         rc = 20
         return 
      end if

      !Get imports

      IF ( use_offline_fields==1 ) THEN
!        IF(MAPL_AM_I_ROOT()) PRINT *,TRIM(Iam)//": using OFFLINE fields"
         CALL MAPL_GetPointer(impChem, T,        'ECCOH_T',      __RC__ )    ! top-down
         CALL MAPL_GetPointer(impChem, TV,       'ECCOH_TV',     __RC__ )    ! top-down
         CALL MAPL_GetPointer(impChem, specHum,  'ECCOH_Q',      __RC__ )    ! top-down
         CALL MAPL_GetPointer(impChem, fcld,     'ECCOH_FCLD',   __RC__ )    ! top-down
         CALL MAPL_GetPointer(impChem, tauclw,   'ECCOH_TAUCLW', __RC__ )    ! top-down
         CALL MAPL_GetPointer(impChem, taucli,   'ECCOH_TAUCLI', __RC__ )    ! top-down
         CALL MAPL_GetPointer(impChem, albedoVF, 'ECCOH_ALBVF',  __RC__ )
      ELSE
!        IF(MAPL_AM_I_ROOT()) PRINT *,TRIM(Iam)//": using ONLINE fields"
         CALL MAPL_GetPointer(impChem, T,        'T',            __RC__ )    ! top-down
         CALL MAPL_GetPointer(impChem, TV,       'TV',           __RC__ )    ! top-down
         CALL MAPL_GetPointer(impChem, specHum,  'Q',            __RC__ )    ! top-down
         CALL MAPL_GetPointer(impChem, fcld,     'FCLD',         __RC__ )    ! top-down
         CALL MAPL_GetPointer(impChem, tauclw,   'TAUCLW',       __RC__ )    ! top-down
         CALL MAPL_GetPointer(impChem, taucli,   'TAUCLI',       __RC__ )    ! top-down
         CALL MAPL_GetPointer(impChem, albedoVF, 'ALBVF',        __RC__ )
      END IF


      CALL MAPL_GetPointer(impChem,    rhoWet,     'AIRDENS',    __RC__ )    ! top-down
      CALL MAPL_GetPointer(impChem,    rhoDry, 'AIRDENS_DRY',    __RC__ )    ! top-down
      CALL MAPL_GetPointer(impChem,       zle,         'ZLE',    __RC__ )    ! top-down
      CALL MAPL_GetPointer(impChem,       ple,         'PLE',    __RC__ )    ! top-down
      CALL MAPL_GetPointer(impChem,     tropp,       'TROPP',    __RC__ )  
      CALL MAPL_GetPointer(impChem,  cellArea,        'AREA',    __RC__ )  
      CALL MAPL_GetPointer(impChem,  albedoVR,       'ALBVR',    __RC__ )  
      CALL MAPL_GetPointer(impChem,  albedoNF,       'ALBNF',    __RC__ )  
      CALL MAPL_GetPointer(impChem,  albedoNR,       'ALBNR',    __RC__ )  


      ! Reserve Some local work space
      !------------------------------
      allocate(            RLON(i1:i2,j1:j2),STAT=ier(11))
      allocate(            RLAT(i1:i2,j1:j2),STAT=ier(12))
      allocate(            ALBD(i1:i2,j1:j2),STAT=ier(13))
      allocate(       tropPress(i1:i2,j1:j2),STAT=ier(14))
      allocate(solarZenithAngle(i1:i2,j1:j2),STAT=ier(15))

      allocate(            o3up(i1:i2,j1:j2,1:km),STAT=ier(16))
      allocate(            mass(i1:i2,j1:j2,1:km),STAT=ier(17))
      allocate(        optDepth(i1:i2,j1:j2,1:km),STAT=ier(18))
      allocate(gridBoxThickness(i1:i2,j1:j2,1:km),STAT=ier(19))
      allocate(           var3d(i1:i2,j1:j2,1:km),STAT=ier(20))
      allocate(           var2d(i1:i2,j1:j2     ),STAT=ier(21))
      allocate(              PL(i1:i2,j1:j2,1:km         ),STAT=ier(22))
      allocate(              pe(i1:i2,j1:j2,1:km+1       ),STAT=ier(23))
      allocate(               p(i1:i2,j1:j2,1:km         ),STAT=ier(24))
      allocate(             STT(i1:i2,j1:j2,1:km,NNPAR   ),STAT=ier(25))
      allocate(            BBIJ(i1:i2,j1:j2,1:km,NFIELDS2),STAT=ier(26))
      allocate(          CH4VAR(i1:i2,j1:j2,1:km         ),STAT=ier(27))
      allocate(           Qppmv(i1:i2,j1:j2,1:km         ),STAT=ier(28))
      allocate(           ndDry(i1:i2,j1:j2,1:km         ),STAT=ier(29))
      allocate(           ndWet(i1:i2,j1:j2,1:km         ),STAT=ier(30))
      allocate(    mw_moist_air(i1:i2,j1:j2,1:km         ),STAT=ier(31))
      allocate(      OHparamNum(i1:i2,j1:j2,1:km         ),STAT=ier(32))
      allocate(           gmiOH(i1:i2,j1:j2,1:km         ),STAT=ier(33)) !sas

      IF ( ANY(ier(:) /= 0)) THEN
         PRINT*, TRIM(myname),': Failed to allocate work space.'
         rc = 21
         RETURN
      END IF
      ier(:) = 0

      ! Extract latitude and longitude - convert to degrees ! EY
      RLON(:,:) = w_c%grid%lon(:,:)*radToDeg 
      RLAT(:,:) = w_c%grid%lat(:,:)*radToDeg 

   !  Layer interface pressures
   !  -------------------------
      pe(i1:i2,j1:j2,1)=w_c%grid%ptop

      DO k=2,km+1
         pe(i1:i2,j1:j2,k)=pe(i1:i2,j1:j2,k-1)+w_c%delp(i1:i2,j1:j2,k-1)
      END DO

   !  Layer mean pressures  (Pa)
   !  --------------------
      DO k=1,km
         p(i1:i2,j1:j2,k)=(pe(i1:i2,j1:j2,k)+pe(i1:i2,j1:j2,k+1))*0.50
      END DO

   !    Dry air number density  (molec/m3)
   !  ------------------------
      ndDry(:,:,1:km) = rhoDry(:,:,1:km)*MAPL_AVOGAD/MAPL_AIRMW

   !  Moist air number density  (molec/m3)
   !  ------------------------
      ndWet = (MAPL_AVOGAD * p) / (MAPL_RUNIV * TV)

   !  Molecular Weight of Moist Air  (kg/mol)
   !  -----------------------------
      mw_moist_air(:,:,1:km) = rhoWet(:,:,km:1:-1)*MAPL_AVOGAD/ndWet(:,:,km:1:-1)

 
   !  Units for OH: w_c%qa(nbeg)%data3D(i1:i2,j1:j2,1:km) is always in terms of molec/cm3


   !  Cell depth (in meters)
   !  ---------------------
   !  (Note: instead we could compute from ZLE)
      gridBoxThickness(:,:,:) = w_c%delp(:,:,:)/(rhoWet(:,:,:)*MAPL_GRAV)
      
      ! Read monthly mean tropPress from GMI file
      CALL Acquire_GMItrpps(rc)
     ! tropPress(i1:i2,j1:j2)       =  tropp(i1:i2,j1:j2)*Pa2hPa

!      PL(:,:,:)  = 0.5*(ple(:,:,1:km)+ple(:,:,0:km-1))*Pa2hPa
      ALBD(i1:i2,j1:j2) = albedoVF(i1:i2,j1:j2) !+albedoNF(i1:i2,j1:j2))/2.0

      ! get OH ratios for aerosol correction
      CALL Acquire_OHratios(rc)

      ! OHparamGrid Comp can run on its own, or together with CO and/or CH4
      ! If running CO Grid Comp, monitor 24 hr average CO 
      if ( w_c%reg%doing_CO ) then
        CALL MonitorCO_24h(rc)
      end if

      ! If running CH4 Grid Comp, monitor 24 hr average CH4 
      if ( w_c%reg%doing_CH4) then
        CALL MonitorCH4_24h(rc)
      end if

      ! get 24 hour average fields from the CCM (such as T,P, water vapor)
      CALL MonitorCCM_Fields24h(rc)

      ! MANYIN: Note the reversing in the vertical, from top-down to bottom-up
            this%Temp24h(:,:,1:km) = w_c%qa(w_c%reg%i_OH+3)%data3d(:,:,km:1:-1)
           this%Press24h(:,:,1:km) = w_c%qa(w_c%reg%i_OH+4)%data3d(:,:,km:1:-1)
         this%specHum24h(:,:,1:km) = w_c%qa(w_c%reg%i_OH+5)%data3d(:,:,km:1:-1)
      this%DryAirMass24h(:,:,1:km) = w_c%qa(w_c%reg%i_OH+6)%data3d(:,:,km:1:-1)

      !-------------------------------------------------
      ! Determine the tropopause level for each grid box
      !-------------------------------------------------

      this%LPAUSE = 0   

!     DO k = km,2,-1
!        DO jx = j1, j2
!           DO ix = i1, i2
!              IF ((tropPress(ix,jx) .LE. PL(ix,jx,k)) .AND. &
!                  (tropPress(ix,jx) .GT. PL(ix,jx,k-1))) THEN
!                 this%LPAUSE(ix,jx) = km - k + 1
!              END IF
!           END DO
!        END DO
!     END DO

      !CALL calcTroppLevel()
      ! Use a constant value of ~72 hPa for tropopause; the correct OH in the
      ! stratosphere within the CH4 and CO code
      ! MANYIN:  This should be changed from 37 to 36 (which is ~ 66.6 - 78.5 hPa)
      this%LPAUSE = 37.

      call calcRxnRateFix()

      JDAY     = JulianDay(nymd)
      curMonth = MOD(nymd, 10000) / 100

      offset_sec = 0.0d0
      time_sec   = convertTimeToSeconds(nhms)

!     solarZenithAngle(i1:i2,j1:j2) = computeSolarZenithAngle (JDAY, time_sec, &
!                                     offset_sec, RLAT, RLON, i1, i2, j1, j2)
      solarZenithAngle(i1:i2,j1:j2) = acos(w_c%cosz(i1:i2,j1:j2))*radToDeg    

      CALL avgrefl(optDepth, solarZenithAngle, this%LPAUSE, this%Ravga, this%Ravgb, &
                            nhms, i2, j2, km)

      OHparamNum(:,:,:) = -999

      !===============================
      ! Do calculations every 24 hours
      !===============================
      IF (nhms == 0) THEN

         CALL Acquire_GMIspecies(rc)

        ! if not doing CO, use CO from input file (already in ppbv)    
        !else convert CO from mol/mol (GEOS-5) to ppbv in parameterization 
        if  ( .NOT. w_c%reg%doing_CO ) then
           STT(:,:,1:km,1) = BBIJ(:,:,:,9) 
        else
           STT(:,:,1:km,1) = this%CO_24h(i1:i2,j1:j2,1:km)*1.0d9   
        end if

        ! Same for CH4
        if  ( .NOT. w_c%reg%doing_CH4 ) then
            CH4VAR(:,:,1:km) = BBIJ(:,:,:,10) 
        else
            CH4VAR(:,:,1:km)= this%CH4_24h(i1:i2,j1:j2,1:km)*1.0d9    
        endif

        ! convert Spec Humidity (kg H20/kg air) to ppmv 
        Qppmv(:,:,1:km) = this%SpecHum24h(:,:,1:km)*mw_moist_air(:,:,1:km)/mwtWat*1.0d6 

        ! test with water vapor from GMI on GMI levels run (in ppmv)
        ! Qppmv(:,:,1:km) = this%qIn(:,:,1:km)*1.0d6 

        ! H2O adjustment to account for H2O bias in CCM in comparison with AIRS	
! yelshorbany, oct, 27, 2014, change the "print" statement below to OFF and CALL adjustH2OMER()
        IF (MAPL_AM_I_ROOT()) print *, 'H2O adjustment TURNED OFF'
!       IF (MAPL_AM_I_ROOT()) print *, 'H2O adjustment TURNED ON'
!	CALL adjustH20() !sas turned this back on
!       CALL adjustH2OMER() !sas version

        CALL runOHparam(this%OHparam, curMonth, ALBD, this%LPAUSE,    &
                 Qppmv, this%Press24h, this%Temp24h, &
                 this%Ravga, this%Ravgb, this%DryAirMass24h,    &
                 STT, BBIJ, o3up, scaleO3, toms_climO3, &
                 CH4VAR, JDAY, RLON, RLAT,OHparamNum)
         ! runOH param returns OH in molec/cm-3 
         ! Reverse coordinates to pass into GEOS-5 environment
         IF (MAPL_AM_I_ROOT())  print *, ' After runOHparam: maxval BOH ', maxval(this%OHparam%BOH(:,:,:))

         ! Apply reaction rate ratio adjustment, still on GMI levels
         
         this%OHparam%BOH(:,:,:)= this%rxnRateRatio(:,:,:)*this%OHparam%BOH(:,:,:)

         this%BCnymd_monthOH = curMonth

         ! Set BOH to 0 above the tropopause
         ! now set it to fullchem value instead of 0 -sas
         ! also do it above 300hPa, not just above tropopause -sas
         DO jx = j1, j2
            DO ix = i1, i2
       !        WHERE(this%Press24h(ix,jx,:) < this%troppresIn(ix,jx)) this%OHparam%BOH(ix,jx,:)=0.
       !        WHERE(this%Press24h(ix,jx,:) < this%troppresIn(ix,jx)) this%OHparam%BOH(ix,jx,:)= gmiOH(ix,jx,:) !-sas
                WHERE(this%Press24h(ix,jx,:) < 300) this%OHparam%BOH(ix,jx,:)= gmiOH(ix,jx,:) !sas
            END DO
         END DO

         ! reverse levels into GEOS-5 format 
         ! MANYIN - OH is stored in data3D and used by CH4 and CO; it is not transported, and during the next timestep it will be overwritten
         w_c%qa(nbeg)%data3D(:,:,1:km) = this%OHparam%BOH(:,:,km:1:-1)  

         ! Ouput input variables for debugging
         call getExportStateVariables(rc)

         this%Ravga = 0.0d0
         this%Ravgb = 0.0d0

      ENDIF
  
      CALL pmaxmin('OH: OH Conc  ',       w_c%qa(nbeg)%data3D(i1:i2,j1:j2,1:km), qmin, qmax, iXj, km,  1. )

  !    CALL pmaxmin('fulchem OH: OH Conc  ',       gmiOH(i1:i1,j1:j1,1:km), qmin, qmax, iXj, km, 1. ) !sas


!  Housekeeping
!  ------------

      DEALLOCATE(ndDry, ndWet, o3up, STT, CH4VAR, Qppmv,RLON, RLAT, BBIJ, var3d, var2d, ALBD,PL, pe,p,OHparamNum)
      DEALLOCATE(mass, gridBoxThickness, solarZenithAngle, tropPress, optDepth, mw_moist_air)
      DEALLOCATE (gmiOH) !-sas

   
      DEALLOCATE(ier)

      RETURN

!---------------------------------------------------------------------------
   CONTAINS
!---------------------------------------------------------------------------
!BOP
      SUBROUTINE adjustH20()
! !DESCRIPTION:
! Adjust Qppmv based on CCM/T1R1 vs AIRS comparison ratios 
!
!EOP
!----------------------------------------------------------------------------
!BOC
        ! Adjust vapor based on bias wrt to AIRS data (2002-2009)
        !----------------------------------------------------------------------------
        ! ****** DJF months ********************************************************
        !----------------------------------------------------------------------------
        IF (curMonth .EQ. 1. .OR. curMonth .EQ. 2. .OR. curMonth .EQ. 12. ) THEN
            DO jx = j1, j2
               DO ix = i1, i2
                   DO k = 1,km
                       !----------------------------------------------------------------------------
                       ! NH  
                       !---------------------------------------------------------------------------- 
                       IF (RLAT(ix,jx) .GE. 0.) THEN
                           ! 800 hPa to 600 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 800 .AND. this%press24h(ix,jx,k) .GT. 600) THEN
                                Qppmv(ix,jx,k) = (0.8)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! 600 hPa to 400 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 600 .AND. this%press24h(ix,jx,k) .GT. 400) THEN
                                Qppmv(ix,jx,k) = (0.6)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! 400 hPa to 200 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 400 .AND. this%press24h(ix,jx,k) .GT. 200) THEN
                                Qppmv(ix,jx,k) = (0.4)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! above 200 hPa 	 
                           IF (this%press24h(ix,jx,k) .LT. 200) THEN
                                Qppmv(ix,jx,k) = (1-0.45)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                       ENDIF
                       !----------------------------------------------------------------------------
                       ! SH
                       !----------------------------------------------------------------------------  
                       IF (RLAT(ix,jx) .LT. 0.) THEN
                           ! 800 hPa to 300 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 800 .AND. this%press24h(ix,jx,k) .GT. 300) THEN
                                Qppmv(ix,jx,k) = (0.7)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! above 300 hPa 	 
                           IF (this%press24h(ix,jx,k) .LT. 300) THEN
                                Qppmv(ix,jx,k) = (1-0.45)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                       ENDIF       
                ! E141 run version 	
                IF (.FALSE.) THEN   ! NH >  60N	
                       IF (RLAT(ix,jx) .GT. 60.) THEN
                           ! 800 hPa to 500 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 800 .AND. this%press24h(ix,jx,k) .GT. 500) THEN
                                Qppmv(ix,jx,k) = (1.1)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! 500 hPa to 300 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 500 .AND. this%press24h(ix,jx,k) .GT. 300) THEN
                                Qppmv(ix,jx,k) = (1.2)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! above 300 hPa 	 
                           IF (this%press24h(ix,jx,k) .LT. 300) THEN
                                Qppmv(ix,jx,k) = (0.90)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                       ENDIF
                       ! NH from 30N to 60N 
                       IF (RLAT(ix,jx) .GE. 30 .AND. RLAT(ix,jx) .LT. 60) THEN
                           ! 800 hPa to 700 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 800 .AND. this%press24h(ix,jx,k) .GT. 700) THEN
                                Qppmv(ix,jx,k) = (0.85)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! 700 hPa to 400 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 700 .AND. this%press24h(ix,jx,k) .GT. 400) THEN
                                Qppmv(ix,jx,k) = (0.65)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! 400 hPa to 300 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 400 .AND. this%press24h(ix,jx,k) .GT. 300) THEN
                                Qppmv(ix,jx,k) = (0.85)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! above 300 hPa 	 
                           IF (this%press24h(ix,jx,k) .LT. 300) THEN
                                Qppmv(ix,jx,k) = (0.60)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                       ENDIF
                       ! from 30N to 30S 
                       IF (RLAT(ix,jx) .GE. -30 .AND. RLAT(ix,jx) .LT. 30) THEN
                           ! 800 hPa to 600 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 800 .AND. this%press24h(ix,jx,k) .GT. 600) THEN
                                Qppmv(ix,jx,k) = (0.75)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! 600 hPa to 500 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 600 .AND. this%press24h(ix,jx,k) .GT. 500) THEN
                                Qppmv(ix,jx,k) = (0.6)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! 500 hPa to 300 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 500 .AND. this%press24h(ix,jx,k) .GT. 300) THEN
                                Qppmv(ix,jx,k) = (0.20)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! above 300 hPa 	 
                           IF (this%press24h(ix,jx,k) .LT. 300) THEN
                                Qppmv(ix,jx,k) = (0.10)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                       ENDIF
                ENDIF
                        ! >60S no correction ratio = 1
                   END DO
                END DO
            END DO  
        ENDIF
        !----------------------------------------------------------------------------
        ! ****** MAM months ********************************************************
        !----------------------------------------------------------------------------
        IF (curMonth .EQ. 3. .OR. curMonth .EQ. 4. .OR. curMonth .EQ. 5. ) THEN
            DO jx = j1, j2
               DO ix = i1, i2
                   DO k = 1,km
                       !----------------------------------------------------------------------------
                       ! NH  
                       !---------------------------------------------------------------------------- 
                       IF (RLAT(ix,jx) .GE. 0.) THEN
                           ! 800 hPa to 500 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 800 .AND. this%press24h(ix,jx,k) .GT. 500) THEN
                                Qppmv(ix,jx,k) = (0.8)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! 500 hPa to 300 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 500 .AND. this%press24h(ix,jx,k) .GT. 300) THEN
                                Qppmv(ix,jx,k) = (0.6)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! 300 hPa to 200 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 300 .AND. this%press24h(ix,jx,k) .GT. 200) THEN
                                Qppmv(ix,jx,k) = (0.4)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! above 200 hPa 	 
                           IF (this%press24h(ix,jx,k) .LT. 200) THEN
                                Qppmv(ix,jx,k) = (0.2)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                       ENDIF
                       !----------------------------------------------------------------------------
                       ! SH
                       !----------------------------------------------------------------------------  
                       IF (RLAT(ix,jx) .LT. 0.) THEN
                           ! 800 hPa to 300 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 800 .AND. this%press24h(ix,jx,k) .GT. 300) THEN
                                Qppmv(ix,jx,k) = (0.7)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! above 300 hPa 	 
                           IF (this%press24h(ix,jx,k) .LT. 300) THEN
                                Qppmv(ix,jx,k) = (1-0.45)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                       ENDIF
                   END DO
                END DO
            END DO  
        ENDIF

        !----------------------------------------------------------------------------
        ! ****** JJA months ********************************************************
        !----------------------------------------------------------------------------
        IF (curMonth .GE. 6 .AND. curMonth .LE. 8) THEN
            DO jx = j1, j2
               DO ix = i1, i2
                   DO k = 1,km
                       !----------------------------------------------------------------------------
                       ! NH from O to 60N
                       !----------------------------------------------------------------------------	
                       IF (RLAT(ix,jx) .GE. 0 .AND. RLAT(ix,jx) .LT. 60) THEN
                           ! 700 hPa to 600 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 700 .AND. this%press24h(ix,jx,k) .GT. 600) THEN
                                Qppmv(ix,jx,k) = (1-0.30)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! 600 hPa to 500 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 600 .AND. this%press24h(ix,jx,k) .GT. 500) THEN
                                Qppmv(ix,jx,k) = (1-0.40)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! above 500 hPa 	 
                           IF (this%press24h(ix,jx,k) .LT. 500) THEN
                                Qppmv(ix,jx,k) = (1-0.60)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           
                       ENDIF
                   END DO
                END DO
            END DO  
        ENDIF
        !----------------------------------------------------------------------------
        ! ****** SON months ********************************************************
        !----------------------------------------------------------------------------
        IF (curMonth .EQ. 9. .OR. curMonth .EQ. 10. .OR. curMonth .EQ. 11. ) THEN
            DO jx = j1, j2
               DO ix = i1, i2
                   DO k = 1,km
                       !----------------------------------------------------------------------------
                       ! NH  
                       !---------------------------------------------------------------------------- 
                       IF (RLAT(ix,jx) .GE. 0.) THEN
                           ! 800 hPa to 500 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 800 .AND. this%press24h(ix,jx,k) .GT. 500) THEN
                                Qppmv(ix,jx,k) = (0.8)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! 500 hPa to 300 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 500 .AND. this%press24h(ix,jx,k) .GT. 300) THEN
                                Qppmv(ix,jx,k) = (0.6)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! 300 hPa to 200 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 300 .AND. this%press24h(ix,jx,k) .GT. 200) THEN
                                Qppmv(ix,jx,k) = (0.4)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! above 200 hPa 	 
                           IF (this%press24h(ix,jx,k) .LT. 200) THEN
                                Qppmv(ix,jx,k) = (0.2)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                       ENDIF
                       !----------------------------------------------------------------------------
                       ! SH
                       !----------------------------------------------------------------------------  
                       IF (RLAT(ix,jx) .LT. 0.) THEN
                           ! 800 hPa to 300 hPa level	 
                           IF (this%press24h(ix,jx,k) .LE. 800 .AND. this%press24h(ix,jx,k) .GT. 300) THEN
                                Qppmv(ix,jx,k) = (0.8)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                           ! above 300 hPa 	 
                           IF (this%press24h(ix,jx,k) .LT. 300) THEN
                                Qppmv(ix,jx,k) = (1-0.45)*this%SpecHum24h(ix,jx,k)*mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                           ENDIF
                       ENDIF
                   END DO
                END DO
            END DO  
        ENDIF

      
      END SUBROUTINE adjustH20
!EOC
!----------------------------------------------------------------------------

!---------------------------------------------------------------------------
!BOP
      SUBROUTINE adjustH2OMER()
! !DESCRIPTION:
! Adjust Qppmv based on CCM/ACCMIP vs MERRA comparison ratios 
!
!EOP
!----------------------------------------------------------------------------
!BOC
        ! Adjust vapor based on bias wrt to MERRA data (1997-2010)
        !----------------------------------------------------------------------------

        INTEGER :: nymd1, nhms1

        REAL, ALLOCATABLE :: Qscale(:,:,:)
        REAL, ALLOCATABLE :: Qtest(:,:,:)

        !  Initialize local variables
        !  --------------------------
 
        ! allocate local variables
        allocate( Qscale(i1:i2,j1:j2,1:7),STAT=ier(38))
        allocate( Qtest(i1:i2,j1:j2,1:km),STAT=ier(39))

        nymd1 = 2000*10000 + MOD ( nymd, 10000 )  ! assumes 2000
        nhms1 = 120000

        var3d = 0.0d0
        ! read in Qratio: note that reader will flip to make lowest pressure 
        ! 1st value of array
        CALL Chem_UtilMPread ( this%Qscl_infile_name, 'Qratio', nymd1, nhms1, &
                           i1, i2, 0, im, j1, j2, 0, jm, 7, &
                           var3d=Qscale, cyclic=.true., &
                           grid=w_c%grid_esmf )
 
        CALL pmaxmin('Qratio:  ', Qscale, qmin, qmax, iXj, 7,  1. )

        Qtest = Qppmv
        CALL pmaxmin('Qppmv orig:  ', Qtest(i1:i2,j1:j2,50), qmin, qmax, iXj, 1,  1. )

        DO jx = j1, j2
           DO ix = i1, i2
              DO k = 1,km
                 IF (this%press24h(ix,jx,k) .LE. 300 .AND. this%press24h(ix,jx,k) .GT. 200) THEN 
                    Qppmv(ix,jx,k) = this%SpecHum24h(ix,jx,k) * Qscale(ix,jx,1) * mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                 ENDIF
                 IF (this%press24h(ix,jx,k) .LE. 400 .AND. this%press24h(ix,jx,k) .GT. 300) THEN 
                    Qppmv(ix,jx,k) = this%SpecHum24h(ix,jx,k) * Qscale(ix,jx,2) * mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                 ENDIF
                 IF (this%press24h(ix,jx,k) .LE. 500 .AND. this%press24h(ix,jx,k) .GT. 400) THEN 
                    Qppmv(ix,jx,k) = this%SpecHum24h(ix,jx,k) * Qscale(ix,jx,3) * mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                 ENDIF
                 IF (this%press24h(ix,jx,k) .LE. 600 .AND. this%press24h(ix,jx,k) .GT. 500) THEN 
                    Qppmv(ix,jx,k) = this%SpecHum24h(ix,jx,k) * Qscale(ix,jx,4) * mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                 ENDIF
                 IF (this%press24h(ix,jx,k) .LE. 700 .AND. this%press24h(ix,jx,k) .GT. 600) THEN 
                    Qppmv(ix,jx,k) = this%SpecHum24h(ix,jx,k) * Qscale(ix,jx,5) * mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                 ENDIF
                 IF (this%press24h(ix,jx,k) .LE. 800 .AND. this%press24h(ix,jx,k) .GT. 700) THEN 
                    Qppmv(ix,jx,k) = this%SpecHum24h(ix,jx,k) * Qscale(ix,jx,6) * mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                 ENDIF
                 IF (this%press24h(ix,jx,k) .LE. 900 .AND. this%press24h(ix,jx,k) .GT. 800) THEN 
                    Qppmv(ix,jx,k) = this%SpecHum24h(ix,jx,k) * Qscale(ix,jx,7) * mw_moist_air(ix,jx,k)/mwtWat*1.0d6 
                 ENDIF

              END DO
           END DO
        END DO

        Qtest = Qppmv
        CALL pmaxmin('Qppmv scaled:  ', Qtest(i1:i2,j1:j2,50), qmin, qmax, iXj, 1,  1. )

        DEALLOCATE( Qscale, Qtest)

      END SUBROUTINE adjustH2OMER
!EOC
!----------------------------------------------------------------------------


!---------------------------------------------------------------------------
!BOP
      SUBROUTINE calcRxnRateFix()

      INTEGER, SAVE :: savedDay = negativeOne
      INTEGER       :: currDay, k1strat
      REAL,    ALLOCATABLE :: k1_new(:,:,:)           !   bottom-up
      REAL,    ALLOCATABLE :: k1_old(:,:,:)           !   bottom-up
      REAL,    ALLOCATABLE :: k2_new(:,:,:)           !   bottom-up
      REAL,    ALLOCATABLE :: k2_old(:,:,:)           !   bottom-up
      REAL,    ALLOCATABLE :: k3_new(:,:,:)           !   bottom-up
      REAL,    ALLOCATABLE :: k3_old(:,:,:)           !   bottom-up
      REAL,    ALLOCATABLE :: N2molec(:,:,:)          !   bottom-up
      REAL,    ALLOCATABLE :: O2molec(:,:,:)          !   bottom-up
      REAL,    ALLOCATABLE :: H2Omolec(:,:,:)         !   bottom-up
      REAL,    ALLOCATABLE :: ADENSITY(:,:,:)         !   bottom-up
      REAL,    ALLOCATABLE :: newarr(:,:,:)           !   bottom-up
      REAL,    ALLOCATABLE :: oldarr(:,:,:)           !   bottom-up
      REAL,    ALLOCATABLE :: oldarr2(:,:,:)          !   bottom-up

      INTEGER, ALLOCATABLE :: ier(:)
! !DESCRIPTION:
! Calculate BOH adjustment ratio based on new and old reaction rates
! Use monthly mean water vapor, and N2, O2 (based on AIR DENSITY)
!
!
!EOP
!----------------------------------------------------------------------------
!BOC
 CHARACTER(LEN=255) :: MyName
  INTEGER :: ic
  currDay = MOD(nymd,100)
  rc=0
  MyName="calcRateFix"
  allocate(ier(199))
  ier(:) = 0
  allocate(k1_new(i1:i2,j1:j2,1:km),STAT=ier(10))
  allocate(k2_new(i1:i2,j1:j2,1:km),STAT=ier(11))
  allocate(k3_new(i1:i2,j1:j2,1:km),STAT=ier(12))
  allocate(k1_old(i1:i2,j1:j2,1:km),STAT=ier(13))
  allocate(k2_old(i1:i2,j1:j2,1:km),STAT=ier(14))
  allocate(k3_old(i1:i2,j1:j2,1:km),STAT=ier(15))
  allocate(N2molec(i1:i2,j1:j2,1:km),STAT=ier(16))
  allocate(O2molec(i1:i2,j1:j2,1:km),STAT=ier(17))
  allocate(H2Omolec(i1:i2,j1:j2,1:km),STAT=ier(18))
  allocate(ADENSITY(i1:i2,j1:j2,1:km),STAT=ier(19))
  allocate(newarr(i1:i2,j1:j2,1:km),STAT=ier(20))
  allocate(oldarr(i1:i2,j1:j2,1:km),STAT=ier(21))
  allocate(oldarr2(i1:i2,j1:j2,1:km),STAT=ier(22))
      IF ( ANY(ier(:) /= 0)) THEN
         PRINT*, TRIM(myname),': Failed to allocate work space.'
         rc = 21
         RETURN
      END IF
      ier(:) = 0

! Execute updates at 0:00 UTC and/or on first pass
! ------------------------------------------------

   IF (currDay /= savedDay) THEN
      hms = 120000
      dymd = nymd
      var3d = 0.0d0



      !  reverse coordinates into GMI levels and convert number density molec/m3 to molec/cm3
      ADENSITY(:,:,1:km) = ndWet(:,:,km:1:-1)/(1E6)
      ! convert Spec Humidity (on GMI levels) (kg H20/kg air) to VMR and them to molec/cm3
      H2Omolec(:,:,:) = this%SpecHum24h(:,:,:)*mw_moist_air(:,:,:)/mwtWat*ADENSITY(:,:,:)
      N2molec(:,:,:) = ADENSITY*0.78
      O2molec(:,:,:) = ADENSITY*0.21

      ! Calculate the change due to change in reaction 
      !  TempIn is in Kelvin
      ! R1 : O1D + H2O -> 2OH 
      ! R2 : O1D + N2  -> O3P + N2
      ! R3 : O1D + O2 -> O + O2

      ! IF (	min (this%Temp24h(:,:,:)-273) .LE. 0) print *, 'Error; Check your Temperature Array in Restart'

      !Temperature on GMI levels already surface 1-> 72 (surface --> toa)
      ! rate constants in  units of cm3/molec/s
      ! JPL11
      k1_new(:,:,:) = (1.63E-10)*EXP(60./(this%Temp24h(:,:,:)))
      ! JPL2000
      k1_old(:,:,:) = (2.2E-10)*EXP(100/(this%Temp24h(:,:,:)))

      ! JPL11
      k2_new(:,:,:) =(2.15E-11)*EXP(110/(this%Temp24h(:,:,:)))
      !Atkins 1997
      k2_old(:,:,:) =  (2.8E-11)*EXP(107/(this%Temp24h(:,:,:)))

      !JPL11
      k3_new(:,:,:) =(3.3E-11)*EXP(55/(this%Temp24h(:,:,:)))
      !Atkins 1997
      k3_old(:,:,:) = (3.2E-11)*EXP(67/(this%Temp24h(:,:,:)))

      newarr(:,:,:) = 2*(H2Omolec(:,:,:)*k1_new(:,:,:))/ &
         (H2Omolec(:,:,:)*(k1_new(:,:,:))+N2molec(:,:,:)*(k2_new(:,:,:))+O2molec(:,:,:)*k3_new(:,:,:))
      
      oldarr2(:,:,:) = (H2Omolec(:,:,:)*(k1_old(:,:,:))+N2molec(:,:,:)*(k2_old(:,:,:))+O2molec(:,:,:)*(k3_old(:,:,:)))
      
      oldarr(:,:,:) = 2*(H2Omolec(:,:,:)*k1_old(:,:,:))/oldarr2(:,:,:)
        
        
        ! Adjust for a O1D reaction rate changes
       DO jx = j1, j2
            DO ix = i1, i2
               DO k = 1,km
                    IF(oldarr2(ix,jx,k) == 0) THEN
                       oldarr(ix,jx,k) = newarr(ix,jx,k)
                    ELSE 
                       oldarr(ix,jx,k) = 2*(H2Omolec(ix,jx,k)*k1_old(ix,jx,k))/oldarr2(ix,jx,k)
                    END IF
                    IF(oldarr(ix,jx,k) == 0) THEN
                       this%rxnRateRatio(ix,jx,k) = 1.0
                    ELSE 
                       this%rxnRateRatio(ix,jx,k) = newarr(ix,jx,k)/oldarr(ix,jx,k)
                    END IF
            
               END DO
            END DO
         END DO

      CALL pmaxmin('Phot. Adjust Ratio', this%rxnRateRatio  , qmin, qmax, iXj,   km, 1. )
      
      DEALLOCATE(k1_new, k2_new, k3_new, k1_old, k2_old,k3_old,N2molec,O2molec, H2Omolec, ADENSITY)
      DEALLOCATE(newarr, oldarr)
      savedDay = currDay
      END IF
      DEALLOCATE(ier)

      END SUBROUTINE calcRxnRateFix
!EOC
!----------------------------------------------------------------------------
!---------------------------------------------------------------------------
!BOP
      SUBROUTINE calcTroppLevel()

      INTEGER, SAVE :: savedDay = negativeOne
      INTEGER       :: currDay, k1strat
      LOGICAL       :: tropo(im,jm,km)

!
! !DESCRIPTION:
! Determine the tropopause level of each grid box.
!
!EOP
!----------------------------------------------------------------------------
!BOC

!      currDay = nymd /1000000
       currDay = MOD(nymd, 100)
      IF (currDay /= savedDay) THEN

        this%LPAUSE = 0   
   
        tropo(:,:,:) = .FALSE.
         
        DO jx = j1, j2
               DO ix = i1, i2
                    WHERE(this%Press24h(ix,jx,:) >= this%troppresIn(ix,jx)) tropo(ix,jx,:)=.TRUE.
               END DO
        END DO
        ! itrop pres index
        k1strat = km
        
           DO jx = j1, j2
               DO ix = i1, i2
                   DO k = km,1, -1
                    IF (tropo(ix,jx,k)) EXIT
                   END DO
                   !this%LPAUSE(ix,jx) = k
                END DO
            END DO   
        
        
! Start from the top of atmosphere
         DO k = km,2, -1       
            DO jx = j1, j2
               DO ix = i1, i2
                 IF ((this%troppresIn(ix,jx) .GE. this%Press24h(ix,jx,k))) THEN 
                      this%LPAUSE(ix,jx) = k
                 END IF
               END DO
            END DO
         END DO
        
         savedDay = currDay
      END IF

      RETURN 

      END SUBROUTINE calcTroppLevel
!EOC
!BOP
!
! !ROUTINE:  Acquire_GMIspecies
!
! !INTERFACE:

  SUBROUTINE Acquire_GMIspecies(rc)

  INTEGER, INTENT(OUT) :: rc
!
! !DESCRIPTION:
! Reads the GMI species concentrations every 24 hours
! from an atmosphere with and without aerosols.
!
!EOP
!---------------------------------------------------------------------------

  CHARACTER(LEN=255) :: MyName
  INTEGER :: ic
  INTEGER :: nTimes, begTime, incSecs
  INTEGER :: file_ymd = 0

  rc=0
  MyName="Acquire_GMIspecies"

! Execute updates at 0:00 UTC and/or on first pass
! ------------------------------------------------
   IF(nhms == 0) THEN

    if(index(this%GMI_infile_name,'%') .gt. 0) then
       dymd = nymd
       hms  = 120000
    else
       call Chem_UtilGetTimeInfo( this%GMI_infile_name, file_ymd, begTime, nTimes, incSecs )
       dymd = (file_ymd/10000)*10000 + MOD( nymd, 10000 )
       hms  = 120000
    endif

    IF( MAPL_AM_I_ROOT() ) THEN
     PRINT *, ' '
     PRINT *,'OH_GridCompRun: Obtaining GMI species concentration from monthly means ...'
    END IF

      !=========================================
      ! Reading species concentrations from file
      !=========================================
      ! BBIJ stores species concentration for:
      ! 1. NOt             2. ALK4          3. Isoprene
      ! 4. Acetone         5. Propene       6. Propane
      ! 7. Ethane          8. O3            9. CO (not used)
      !10. CH4 (not used)
      BBIJ = 0.0d0
      DO ic = 1, 15                             ! 14 if reading CO too, 15 if CH4
      ! This proc reverses GMI levels into GEOS-5 levels
         var3d = 0.0d0
         CALL Chem_UtilMPread(this%GMI_infile_name, &
                TRIM(this%gmiSpeciesName(ic)), dymd, &
                hms, i1, i2, 0, im, j1, j2, 0, jm, km, &
                var3d=var3d, cyclic=.TRUE., grid = w_c%grid_esmf)

        ! Unreverse to get back to GMI and OH param level convention (1 at surface , 72 at top)
         DO k=1,km
            kReverse=km-k+1
            IF (ic .LE. 6) THEN
               BBIJ(:,:,k,1) = BBIJ(:,:,k,1) + var3d(:,:,kReverse)
               ! ADD in N2O5 the second time into the NOt calculation
               IF(ic .EQ. 3) THEN 
                 BBIJ(:,:,k,1) = BBIJ(:,:,k,1) + var3d(:,:,kReverse)
               ENDIF
            ELSE
               BBIJ(:,:,k,ic-5)=var3d(:,:,kReverse)
            END IF
         END DO
      END DO

      BBIJ(:,:,:, 1) = BBIJ(:,:,:, 1) * 1.D12     ! NOt  pptv
      BBIJ(:,:,:, 2) = BBIJ(:,:,:, 2) * 1.D12     ! ALK4 (alkanes pptv)
      BBIJ(:,:,:, 3) = BBIJ(:,:,:, 3) * 1.D12     ! ISOP (isoprene pptv)
      BBIJ(:,:,:, 4) = BBIJ(:,:,:, 4) * 1.D12     ! ACET (acetone pptv)
      BBIJ(:,:,:, 5) = BBIJ(:,:,:, 5) * 1.D12     ! PRPE (propene pptv)
      BBIJ(:,:,:, 6) = BBIJ(:,:,:, 6) * 1.D12     ! C3H8 (propane pptv)
      BBIJ(:,:,:, 7) = BBIJ(:,:,:, 7) * 1.D12     ! C2H6 (ethane pptv)
      BBIJ(:,:,:, 8) = BBIJ(:,:,:, 8) * 1.D9      ! O3 (ozone ppbv)
      BBIJ(:,:,:, 9) = BBIJ(:,:,:, 9) * 1.D9      ! CO (co ppbv)
      BBIJ(:,:,:,10) = BBIJ(:,:,:,10) * 1.D9      ! CH4 (ppbv)

      ! Also read in full-chem OH for use in the stratosphere -sas
      var3d = 0.0d0
      CALL Chem_UtilMPread(this%GMI_infile_name, &
           TRIM('OH'), dymd, &
           hms, i1, i2, 0 , im, j1, j2, 0, jm, km, &
           var3d=var3d, cyclic=.TRUE., grid = w_c%grid_esmf)
      ! convert OH from mol/mol to molec/cm3
      ! (assume that GMI output is volume mixing ratio with respect to dry air)
      var3d(:,:,:) = var3d(:,:,:) * ndDry(:,:,:) * 1.E-6
      ! Unreverse to get back to GMI and OH param level convention (1 at surface , 72 at top)
      DO k=1,km
         kReverse=km-k+1
         gmiOH(:,:,k)=var3d(:,:,kReverse) !molec/cm3
      END DO

      IF (MAPL_AM_I_ROOT())  print *, ' maxval gmiOH ', maxval(gmiOH(:,:,:))

      ! Determine the overhead O3 column
      !---------------------------------
      ! convert (ppbv) to ozone (DU) for each box
      ! O3 * air density * box thickness * conversion to DU
      ! ppbv *  kg/m3 * m * 1 mol air/0.02897 kg * 6.023e23 molec/mol air * 1DU/(2.69e20 molec/m2)

      ! The rhoWet just cancels out with the same term used in computing gridBoxThickness
      var3d(:,:,1:km) = BBIJ(:,:,:,8)*1.0d-9*rhoWet(:,:,km:1:-1)*gridBoxThickness(:,:,km:1:-1)/0.02897*6.023E23/(2.69E20)

! MANYIN  var3d and o3up are bottom-up
! old code [note that o3up(n) = SUM(var3d(n+1:km),  but o3up(km) == var3d(km) ]  (**)
      ! o3up(:,:,:) = 0.0d0
      ! o3up(:,:,km) = var3d(:,:,km)
      ! DO ic = 1, km-1
      !    DO k = ic+1, km
      !       o3up(:,:,ic) = o3up(:,:,ic) + var3d(:,:,k)
      !    END DO
      ! END DO

! MANYIN  instead do this, for 'overhead column including the current gridbox':
!     o3up(:,:,km) = var3d(:,:,km)
!     DO k = km-1,1,-1
!           o3up(:,:,k) = var3d(:,:,k) + o3up(:,:,k+1)
!     END DO
! MANYIN  or do this, for 'overhead column not including the current gridbox':
!     Note: the O3 above the top gridbox should not really be zero !
      o3up(:,:,km) = 0.0d0
      DO k = km-1,1,-1
            o3up(:,:,k) = var3d(:,:,k+1) + o3up(:,:,k+1)
      END DO

      ! (**)  Change the top layer to match earlier implementation; consider taking this out
      o3up(:,:,km) = var3d(:,:,km)

  END IF

  RETURN

  END SUBROUTINE Acquire_GMIspecies
!EOC
!------------------------------------------------------------------------------------
!BOP
!
! !ROUTINE:  Acquire_GMItrpps
!
! !INTERFACE:

  SUBROUTINE Acquire_GMItrpps(rc)

  INTEGER, INTENT(OUT) :: rc
!
! !DESCRIPTION:
! Reads the GMI trpps every 24 hours

!
!EOP
!---------------------------------------------------------------------------
 CHARACTER(LEN=255) :: MyName
  INTEGER :: ic, currMonth

  rc=0
  MyName="Acquire_GMItrpps"

! Execute updates at 0:00 UTC and/or on first pass
! ------------------------------------------------
   currMonth = MOD(nymd,10000) / 100

   IF (this%BCnymd_month /= currMonth) THEN
      !cymd = 2001*10000+MOD(nymd,10000)
      hms = 120000


      !dymd = cymd
      dymd = nymd

      ! Read in 2D field - monthly mean gmi 
      var2d = 0.0d0
      ic = 16
      CALL Chem_UtilMPread(this%GMI_infile_name, &
                TRIM(this%gmiSpeciesName(ic)), dymd, &
                hms, i1, i2, 0, im, j1, j2, 0, jm, 0, &
                var2d=var2d, cyclic=.TRUE., grid = w_c%grid_esmf)

       this%troppresIn(:,:) = var2d(:,:)


       ! Also read in GMI water vapor	
      ! Read in 3D field - monthly mean gmi 

      !dymd = nymd
      !var3d = 0.0d0
      !ic = 17
      !CALL Chem_UtilMPread(this%GMI_infile_name, &
      !          TRIM(this%gmiSpeciesName(ic)), dymd, &
      !          hms, i1, i2, 0, im, j1, j2, 0, jm, km, &
      !          var3d=var3d, cyclic=.TRUE., grid = w_c%grid_esmf)		
      ! This proc reversed vertical levels into GEOS5 grid; reverse back to GMI levels	
     !DO k=1,km
     !     kReverse=km-k+1
     !     this%qIn(:,:,k) = var3d(:,:,kReverse)
     !END DO       
       this%BCnymd_month = curMonth
    END IF

  RETURN

  END SUBROUTINE Acquire_GMItrpps
!EOC
!------------------------------------------------------------------------------------
!BOP
!
! !ROUTINE:  Acquire_OHratios
!
! !INTERFACE:

  SUBROUTINE Acquire_OHratios(rc)

  INTEGER, INTENT(OUT) :: rc
!
! !DESCRIPTION:
! Reads the array holding the percent reduction in OH
! from an atmosphere with and without aerosols.
!
!EOP
!---------------------------------------------------------------------------

  CHARACTER(LEN=255) :: MyName
  CHARACTER(LEN=  3) :: ratioID
  CHARACTER(LEN= 15) :: ratioName
  INTEGER :: ic
  REAL, ALLOCATABLE :: ratio2D(:,:)
  INTEGER :: nTimes, begTime, incSecs
  INTEGER :: file_ymd = 0

  rc=0
  MyName="Acquire_OHratios"

! Execute updates at 0:00 UTC and/or on first pass
! ------------------------------------------------
   IF(this%BCnymd /= nymd) THEN

    if(index(this%OHparam%allAerosolRatio_inFileName,'%') .gt. 0) then
       dymd = nymd
       hms  = 120000
    else
       call Chem_UtilGetTimeInfo( this%OHparam%allAerosolRatio_inFileName, file_ymd, begTime, nTimes, incSecs )
       dymd = (file_ymd/10000)*10000 + MOD( nymd, 10000 )
       hms  = 120000
    endif

    IF( MAPL_AM_I_ROOT() ) THEN
     PRINT *, ' '
     PRINT *,'OH_GridCompRun: Updating boundary conditions from monthly means ...'
    END IF

     ratioName = 'ohRatio'
 
     ! This proc reverses GMI levels into GEOS-5 levels
     CALL Chem_UtilMPread(trim(this%OHparam%allAerosolRatio_inFileName),    &
                      TRIM(ratioName), dymd, hms, i1, i2, 0, im, j1, j2, 0, jm, &
                      km, var3d=var3d, cyclic=.TRUE., grid = w_c%grid_esmf)
     ! Reverse levels back to GMI (& OH param) convention 1(at surface) , 72 (at top of atmos) 
     DO k=1,km
          kReverse=km-k+1
          this%OHparam%dustrat(:,:,k) = var3d(:,:,kReverse)
     END DO

    this%BCnymd = nymd
  END IF

  RETURN

  END SUBROUTINE Acquire_OHratios
!EOC
!------------------------------------------------------------------------------------
!BOP
!
! !ROUTINE:  MonitorCO_24h
!
! !INTERFACE:

  SUBROUTINE MonitorCO_24h(rc)

  IMPLICIT NONE

  INTEGER, INTENT(OUT) :: rc
!
! !DESCRIPTION:
!
! Obtain 24h averages of CO
!
!EOP
!---------------------------------------------------------------------------

  CHARACTER(LEN=255) :: Iam
  REAL :: avgFactor
  logical, save :: firstTime = .true.

  rc = 0
  Iam = "MonitorCO_24h"
  avgFactor = cdt/secPerDay

! Functions for startup and at change of day
! ------------------------------------------
  IF (this%BCnymdCO_24h /= nymd .AND. nhms == 0) THEN
 
     !===============================================
     ! If first time step, use values in restart file
     !===============================================

     IF (this%BCnymdCO_24h == -1) THEN
        this%CO_24h(:,:,1:km) = w_c%qa(w_c%reg%i_OH+1)%data3D(:,:,km:1:-1)
     ELSE
        ! Gets here on all changes of day, use previous hour mass average to compute
        ! Mass weighted Average , reverse levels to get variable ready for OHparameterization
        ! Manyin: Desired units for CO_24h - parts per part wrt Dry Air
        
        this%CO_24h(i1:i2,j1:j2,1:km) = this%COqMassAcc(i1:i2,j1:j2,km:1:-1)/this%DryAirMassAcc(i1:i2,j1:j2,km:1:-1)

        ! And store as restart value for CO, reverse back into GEOS-5
        w_c%qa(w_c%reg%i_OH+1)%data3D(:,:,:) = this%CO_24h(i1:i2,j1:j2,km:1:-1)
     END IF

! Reset indicator
! ---------------
    this%BCnymdCO_24h = nymd
!   this%CO_Acc(i1:i2,j1:j2,1:km) = 0.0d0
  END IF

  RETURN

  END SUBROUTINE MonitorCO_24h
!EOP
!---------------------------------------------------------------------------
!BOP
!
! !ROUTINE:  MonitorCH4_24h
!
! !INTERFACE:

  SUBROUTINE MonitorCH4_24h(rc)

  IMPLICIT NONE

  INTEGER, INTENT(OUT) :: rc
!
! !DESCRIPTION:
!
! Obtain 24h averages of CH4
!
!EOP
!---------------------------------------------------------------------------

  CHARACTER(LEN=255) :: Iam
  REAL :: avgFactor
  INTEGER :: j_CH4
  logical, save :: firstTime = .true.

   REAL, PARAMETER :: secPerDay = 86400.00

  rc = 0
  Iam = "MonitorCH4_24h"
  avgFactor = cdt/secPerDay

! Functions for startup and at change of day
! ------------------------------------------
  IF (this%BCnymdCH4_24h /= nymd .AND. nhms == 0) THEN
 
     !=============================================
     ! If first time step, use instantaneous values
     !=============================================

     IF (this%BCnymdCH4_24h == -1) THEN
        this%CH4_24h(:,:,1:km) = w_c%qa(w_c%reg%i_OH+2)%data3D(:,:,km:1:-1)
     ELSE
        ! Gets here on all changes of day, use previous hour mass average to compute     
        ! Mass weighted Average , reverse levels to get variable rady for OHparameterization 
        ! no need to multiply by sumFactor twice, it cancels out to 1    
        ! Manyin: Desired units for CH4_24h - parts per part wrt Dry Air
        
        this%CH4_24h(i1:i2,j1:j2,1:km) = this%CH4qMassAcc(i1:i2,j1:j2,km:1:-1)/this%DryAirMassAcc(i1:i2,j1:j2,km:1:-1)
        
        ! And store as restart value for CH4
        w_c%qa(w_c%reg%i_OH+2)%data3D(:,:,:) = this%CH4_24h(i1:i2,j1:j2,km:1:-1)
     END IF

! Reset indicator
! ---------------
    this%BCnymdCH4_24h = nymd
!   this%CH4_Acc(i1:i2,j1:j2,1:km) = 0.0d0
  END IF

  RETURN

  END SUBROUTINE MonitorCH4_24h
!EOP
!---------------------------------------------------------------------------
!BOP
!
! !ROUTINE:  MonitorCCM_Fields24h
!
! !INTERFACE:

      SUBROUTINE MonitorCCM_Fields24h(rc)

      IMPLICIT NONE

      INTEGER, INTENT(OUT) :: rc
!
! !DESCRIPTION:
!
! Obtain 24h averages of temperature, pressure, etc.
!
!EOP
!---------------------------------------------------------------------------

      CHARACTER(LEN=255) :: Iam
      REAL :: avgFactor, sumFactor
      INTEGER :: ndt, k
      logical, save :: firstTime = .TRUE.

      rc = 0
      Iam = "MonitorCCM_Fields24h"
      avgFactor = cdt/secPerDay
      sumFactor = 1.0E15

      DO k=1,km
         mass(:,:,k)=rhoDry(:,:,k)*cellArea(:,:)*(zle(:,:,k-1)-zle(:,:,k))  ! kg
      END DO

      optDepth(i1:i2,j1:j2,1:km)   = (taucli(i1:i2,j1:j2,km:1:-1) + &
                                      tauclw(i1:i2,j1:j2,km:1:-1))*fcld(i1:i2,j1:j2,km:1:-1)**1.5

      ! Functions for startup and at change of day
      ! ------------------------------------------
      IF (this%BCnymd_24h /= nymd .AND. nhms == 0) THEN
 
         IF (this%BCnymd_24h .NE. negativeOne) THEN
! MANYIN: Note vertical flip
            w_c%qa(w_c%reg%i_OH+3)%data3D(i1:i2,j1:j2,km:1:-1) =       this%TempAcc(i1:i2,j1:j2,1:km)
            w_c%qa(w_c%reg%i_OH+4)%data3D(i1:i2,j1:j2,km:1:-1) =      this%PressAcc(i1:i2,j1:j2,1:km)
            w_c%qa(w_c%reg%i_OH+5)%data3D(i1:i2,j1:j2,km:1:-1) =    this%SpecHumAcc(i1:i2,j1:j2,1:km)
            w_c%qa(w_c%reg%i_OH+6)%data3D(i1:i2,j1:j2, 1:km  ) = this%DryAirMassAcc(i1:i2,j1:j2,1:km)  ! MANYIN: both are top-down and
         END IF

         ! Reset indicator
         ! ---------------
         this%BCnymd_24h = nymd

               this%TempAcc(i1:i2,j1:j2,1:km) = 0.0d0
              this%PressAcc(i1:i2,j1:j2,1:km) = 0.0d0
            this%SpecHumAcc(i1:i2,j1:j2,1:km) = 0.0d0
         this%DryAirMassAcc(i1:i2,j1:j2,1:km) = 0.0d0
            this%COqMassAcc(i1:i2,j1:j2,1:km) = 0.0d0
           this%CH4qMassAcc(i1:i2,j1:j2,1:km) = 0.0d0

      END IF

      ! Keep running average for current day
      ! -----------------------------------------------

         this%TempAcc(:,:,1:km) =    this%TempAcc(:,:,1:km) +       T(:,:,km:1:-1) *          avgFactor
        this%PressAcc(:,:,1:km) =   this%PressAcc(:,:,1:km) +     ple(:,:,km:1:-1) * Pa2hPa * avgFactor
      this%SpecHumAcc(:,:,1:km) = this%SpecHumAcc(:,:,1:km) + specHum(:,:,km:1:-1) *          avgFactor

      !--------------------------------------------------------------				     
      ! Sum up mass for the current day; devide by a factor to keep the number smaller; then remultiply at the end 
      ! These quantities are on GEOS-5 vertical grid
      !--------------------------------------------------------------				     

      this%DryAirMassAcc(:,:,1:km) = &
      this%DryAirMassAcc(:,:,1:km) + mass(:,:,1:km)/sumFactor
    
      if ( w_c%reg%doing_CO) then !sas -if not doing_CO, skip next line
         ! Manyin - convert CO units:  MMR wrt moist  to  VMR wrt dry
         this%COqMassAcc(:,:,1:km) = &
         this%COqMassAcc(:,:,1:km) + &
             ( w_c%qa(w_c%reg%i_CO)%data3D(:,:,:)  * rhoWet(:,:,:) * (MAPL_AVOGAD/mwtCO)  / ndDry(:,:,:) ) *  &
                  mass(:,:,:)/sumFactor
      endif

      if ( w_c%reg%doing_CH4) then !sas -if not doing_CH4, skip next line
         ! Manyin - convert CH4 units:  MMR wrt moist  to  VMR wrt dry
         this%CH4qMassAcc(:,:,1:km) = &
         this%CH4qMassAcc(:,:,1:km) + &
             ( w_c%qa(w_c%reg%j_CH4)%data3D(:,:,:) * rhoWet(:,:,:) * (MAPL_AVOGAD/mwtCH4) / ndDry(:,:,:) ) *  &
                  mass(:,:,1:km)/sumFactor
      endif

      RETURN

      END SUBROUTINE MonitorCCM_Fields24h

 !---------------------------------------------------------------------------
! NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1, GEOS/DAS !
!---------------------------------------------------------------------------
!BOP
!
! !ROUTINE:  getExportStateVariables
!
! !INTERFACE:
!
      SUBROUTINE getExportStateVariables(rc)

      IMPLICIT NONE

      INTEGER, INTENT(OUT) :: rc
!
! !DESCRIPTION:
!
!  Find pointers to import and export states
!
!EOP
!---------------------------------------------------------------------------

      CHARACTER(LEN=255) :: Iam
      REAL :: qmin,qmax

      rc=0
      Iam="getExportStateVariables"

      CALL MAPL_GetPointer(expChem, exO3up,     'O3upIn',     ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exNOy,      'NOyIn',      ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exALK4,     'ALK4In',     ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exIsoprene, 'IsopreneIn', ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exAcetone,  'AcetoneIn',  ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exPropene,  'PropeneIn',  ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exPropane,  'PropaneIn',  ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exEthane,   'EthaneIn',   ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exOzone,    'OzoneIn',    ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exSpecHum,  'SpecHumIn',  ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exCO,       'COIn',       ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exCH4,      'CH4In',      ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exPress,    'PressIn',    ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exTemp,     'TempIn',     ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exTRPPS,    'TRPPSIn',    ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exAlbedo,   'AlbedoIn',   ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exRavga,    'RavgaIn',    ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exRavgb,    'RavgbIn',    ALLOC=.TRUE., __RC__)
      CALL MAPL_GetPointer(expChem, exvarOut,   'varOut',     ALLOC=.TRUE., RC=STATUS)
      VERIFY_(STATUS)
      

      ! flip these back to GEOS5 (toa-down) convention -sas
      IF (ASSOCIATED(exO3up))         exO3up(:,:,:) = O3up(:,:,km:1:-1)
      IF (ASSOCIATED(exNOy))           exNOy(:,:,:) = BBIJ(:,:,km:1:-1,1)
      IF (ASSOCIATED(exALK4))         exALK4(:,:,:) = BBIJ(:,:,km:1:-1,2)
      IF (ASSOCIATED(exIsoprene)) exIsoprene(:,:,:) = BBIJ(:,:,km:1:-1,3)
      IF (ASSOCIATED(exAcetone))   exAcetone(:,:,:) = BBIJ(:,:,km:1:-1,4)
      IF (ASSOCIATED(exPropene))   exPropene(:,:,:) = BBIJ(:,:,km:1:-1,5)      
      IF (ASSOCIATED(exPropane))   exPropane(:,:,:) = BBIJ(:,:,km:1:-1,6)
      IF (ASSOCIATED(exEthane))     exEthane(:,:,:) = BBIJ(:,:,km:1:-1,7)
      IF (ASSOCIATED(exOzone))       exOzone(:,:,:) = BBIJ(:,:,km:1:-1,8)
      IF (ASSOCIATED(exSpecHum))   exSpecHum(:,:,:) = Qppmv(i1:i2,j1:j2,km:1:-1)
      IF (ASSOCIATED(exCO))             exCO(:,:,:) = STT(:,:,km:1:-1,1)
      IF (ASSOCIATED(exCH4))           exCH4(:,:,:) = CH4VAR(:,:,km:1:-1)
      IF (ASSOCIATED(exPress))       exPress(:,:,:) = this%Press24h(i1:i2,j1:j2,km:1:-1)
      IF (ASSOCIATED(exTemp))         exTemp(:,:,:) = this%Temp24h(i1:i2,j1:j2,km:1:-1)
      IF (ASSOCIATED(exTRPPS))         exTRPPS(:,:) = this%LPAUSE(i1:i2,j1:j2)
      IF (ASSOCIATED(exAlbedo))       exAlbedo(:,:) = ALBD(i1:i2,j1:j2)
      IF (ASSOCIATED(exRavga))       exRavga(:,:,:) = this%Ravga(i1:i2,j1:j2,km:1:-1)
      IF (ASSOCIATED(exRavgb))       exRavgb(:,:,:) = this%Ravgb(i1:i2,j1:j2,km:1:-1)
      IF (ASSOCIATED(exvarOut))     exvarOut(:,:,:) = OHparamNum(:,:,km:1:-1)
      RETURN

      END SUBROUTINE getExportStateVariables     


 END SUBROUTINE OH_GridCompRun
!EOP
!---------------------------------------------------------------------------

!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 900.3     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  OH_GridCompFinalize --- Finalize OH_GridComp
!
! !INTERFACE:
!

   subroutine OH_GridCompFinalize ( this, w_c, impChem, expChem, &
                                      nymd, nhms, cdt, rc )

! !USES:

  IMPLICIT NONE

! !INPUT PARAMETERS:

   TYPE(Chem_Bundle), intent(in) :: w_c        ! Chemical tracer fields      
   INTEGER, INTENT(IN) :: nymd, nhms           ! time
   REAL,    INTENT(IN) :: cdt                  ! chemical timestep (secs)


! !OUTPUT PARAMETERS:

   TYPE(OH_GridComp), INTENT(INOUT) :: this   ! Grid Component
   TYPE(ESMF_State), INTENT(INOUT)  :: impChem  ! Import State
   TYPE(ESMF_State), INTENT(INOUT)  :: expChem  ! Export State
   INTEGER, INTENT(OUT) ::  rc                  ! Error return code:
                                                !  0 - all is well
                                                !  1 - 

! !DESCRIPTION: Finalizes the OH Parameterization Grid Component. Multiple instance
!               version.
!
! !REVISION HISTORY:
!
!  27Feb2008  da Silva  Introduced multiple instances
!
!EOP
!-------------------------------------------------------------------------

   CHARACTER(LEN=*), PARAMETER :: myname = 'OH_GridCompFinalize'
   INTEGER :: ios

   rc = 0
   DEALLOCATE ( this%Temp24h, this%Press24h, this%troppresIn,this%qIn, this%rxnRateRatio, &
                this%SpecHum24h, this%DryAirMass24h, this%LPAUSE, &
                this%CO_24h, this%CH4_24h, this%COqMassAcc, this%CH4qMassAcc, &
                this%TempAcc, this%PressAcc, this%SpecHumAcc, this%DryAirMassAcc, &
                this%Ravga ,  this%Ravgb, &
                STAT=ios )
   IF ( ios /= 0 ) rc = 1

!  DEALLOCATE ( this%CO_Acc, this%CH4_Acc, STAT=ios )
!  IF ( ios /= 0 ) rc = 1

   CALL finalizeOHparam(this%OHparam)

 end subroutine OH_GridCompFinalize
!EOC
!--------------------------------------------------------------------------
!BOP
      INTEGER FUNCTION JulianDay(nymd)
         INTEGER :: nymd
!
! !DESCRIPTION:
! Determines the Julian day: number between 1 and 365 (or 366).
         INTEGER :: ny, mm, dd
         INTEGER :: m, ds
         INTEGER :: days(12)
!EOP
!--------------------------------------------------------------------------
!BOC
         data days /31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31/

         ny = nymd / 10000
         mm = mod(nymd, 10000) / 100
         dd = mod(nymd,   100)

         ds = dd

         if ( mm .ne. 1) then
            do m=1, mm-1
               if ( m.eq.2  .and. leap_year(ny) ) then
                  ds = ds + 29
               else
                  ds = ds + days(m)
               endif
            enddo
         endif

         JulianDay = ds
         
      END FUNCTION JulianDay
!EOC
!--------------------------------------------------------------------------
!BOP
      function leap_year(ny)
!
! Determine if year ny is a leap year
!
! Author: S.-J. Lin
      implicit none
      logical leap_year
      integer ny
      integer ny00
!EOP
!--------------------------------------------------------------------------
!BOC

!
! No leap years prior to 0000
!
      parameter ( ny00 = 0000 )   ! The threshold for starting leap-year

      if( ny >= ny00 ) then
         if( mod(ny,100) == 0. .and. mod(ny,400) == 0. ) then
             leap_year = .true.
         elseif( mod(ny,4) == 0. .and. mod(ny,100) /= 0.  ) then
             leap_year = .true.
         else
             leap_year = .false.
         endif
      else
          leap_year = .false.
      endif

      return
    end function leap_year

!EOC
!--------------------------------------------------------------------------
!BOP
      FUNCTION convertTimeToSeconds (nhms) result(this_)
!
! !INPUT PARAMETERS
      INTEGER, intent(in) :: nhms
!
! !RETURNED VALUE
      INTEGER :: this_
!
! !DESCRIPTION:
! Converts the current time into seconds.
!
! !DEFINED PARAMETERS:
      REAL*8, PARAMETER :: SECPHR = 3600.0d0
      REAL*8, PARAMETER :: SECPMN =   60.0d0
      REAL*8, PARAMETER :: SECPDY = 24.0d0*SECPHR
!EOP
!------------------------------------------------------------------------------
!BOC
      this_ = (nhms / 10000) * SECPHR +              &
              (Mod (nhms, 10000) / 100) * SECPMN +   &
               Mod (nhms, 100)

      RETURN

      END FUNCTION convertTimeToSeconds
!EOC
!------------------------------------------------------------------------------
END MODULE  OH_Para_GridCompMod
