!**********************************************************************
      SUBROUTINE SKIP(DOskip,COLON,COLAT,COALT, &
                      INDVARA, INDVARB, INDVARC, INDVARD)
!**********************************************************************
      IMPLICIT NONE
#     include "CMN_OH"
      REAL*8 :: INDVARA(:,:,:,:)
      REAL*8 :: INDVARB(:,:,:,:)
      REAL*8 :: INDVARC(:,:,:,:)
      REAL*8 :: INDVARD(:,:,:,:)
      INTEGER DOskip,COLON,COLAT,COALT,NODOMOD,NODOHI
!**********************************************************************
!   Created by Bryan Duncan.
!**********************************************************************
! SR SKIP performs the following functions:
! 1) Low OH :  Use if statements to keep calling OH parameterization
!              when OH is very low.  About 1/5 of the boxes
!              are skipped here.
! 2) RESET  :  Reset the values of isoprene and NOx to keep from calling
!              "nonexistent" parameterizations.
! 3) Other Miscellaneous Functions
!
!**********************************************************************
! List of Variables & Arrays
!
! DOskip = Counter to skip parameterization. 
!        = 0 : use parameterization to get [OH]
!        = 1 : set to low value and skip parameterization
!        = 2 : use average [OH] (see SR readavgOH)
!        = 3 : use [OH] = 1x10^5 if isoprene > 600 pptv and skip parameterization
!        = 4 : use [OH] = 1x10^4 if isoprene > 3000 pptv and skip parameterization
!
            DOskip=0
!
! Dimensions of troposphere.  User specified values in CMN_OH.
!   COALT = position of box in vertical in troposphere
!   COLAT = position of box from pole to pole (following one longitude band)
!   COLON = position of box circling globe (following one latitude band)
!
! PARAMOH = parameterized [OH].
!
! Variables specific to box in question:
!
! OH_NOX    = NOx concentration
! OH_PRESS  = pressure
! OH_LAT    = latitude
! OH_ISOP   = isoprene concentration
! OH_LON    = longitude
! OH_MONTH = month (January-December = 1-12)
! OH_SEASON = (Northern Hemispheric) season
!        1 --> winter (Dec, Jan, Feb)
!        2 --> spring (Mar, Apr, May)
!        3 --> summer (Jun, Jul, Aug)
!        4 --> autumn (Sep, Oct, Nov)
!
!***************************************************
! 1) Miscellaneous Functions.
!***************************************************
! PRESS > 100 mb:  Don't calculate OH over 100 mb.

       IF (OH_PRESS.LT.PRESSES(3)) THEN
!
! Error Check.
!
         PRINT*,'************************************************'
         PRINT*,'WARNING MESSAGE - SR SKIP!'
         PRINT*,'The pressure for this box is less than 100 mb.'
         PRINT*,'No parameterization is designed for pressures'
         PRINT*,'less than 100 mb. Box is skipped and [OH] set'
         PRINT*,'to low value. '
         PRINT*,'LON BOX=',COLON
         PRINT*,'LAT BOX=',COLAT
         PRINT*,'ALT BOX=',COALT
         PRINT*,'************************************************'
!
         DOskip=1
!
         GOTO 320
!
       ENDIF
!
! NOX < 1 ppt:  Reset NOx to 1 ppt.
!
       IF (OH_NOX.LT.1.) THEN
         OH_NOX=1.1
         INDVARA(COLON,COLAT,COALT,3)=OH_NOX
         INDVARB(COLON,COLAT,COALT,3)=OH_NOX
         INDVARC(COLON,COLAT,COALT,3)=OH_NOX
         INDVARD(COLON,COLAT,COALT,3)=OH_NOX
       ENDIF
!
!***************************************************
! 2) LOW OH
!***************************************************
! SH (80-90S)
      IF (OH_LAT.LE.-80.) THEN
         COCOUNT=0
         DOskip=2
         GOTO 320
      ENDIF
! NH (80-90N)
      IF (OH_LAT.GE.80.) THEN
         COCOUNT=0
         DOskip=2
         GOTO 320
      ENDIF
! WIN NH (40-90N)
      IF (OH_LAT.GE.ALATS(6).AND.OH_SEASON.EQ.1) THEN
         COCOUNT=0
         DOskip=2
         GOTO 320
      ENDIF
! SUM & SPR SH (60-90S)
      IF (OH_LAT.LE.ALATS(2).AND.(OH_SEASON.EQ.3.OR.  &
     &   OH_SEASON.EQ.2)) THEN
         COCOUNT=0
         DOskip=2
         GOTO 320
      ENDIF
! AUT NH (60-90N)
      IF (OH_LAT.GE.ALATS(7).AND.OH_SEASON.EQ.4) THEN
         COCOUNT=0
         DOskip=2
         GOTO 320
      ENDIF
!
!***************************************************
! 3) RESET
!***************************************************
!***************************************************
! ISOPRENE
!***************************************************
! [NOx] < 50 ppt
!    [ISOP] > 600 ppt set [OH] = 1e10^5 and skip parameterization
!    [ISOP] > 3000 ppt set [OH] = 1e10^4 and skip parameterization
! Assumptions based on:
! Mickley, L. J., D. J. Jacob, and D. Rind,
! Uncertainty in preindustrial abundance of tropospheric ozone:
! Implications for radiative forcing calculations.
! In press Journal of Geophysical Research, 2000.
!
      IF ( OH_NOX.LT.50..AND.OH_PRESS.GE.PRESSES(1) ) THEN
       IF ( OH_ISOP.GE.600.) THEN
         DOskip=3
         GOTO 320
       ENDIF
       IF ( OH_ISOP.GE.3000.) THEN
         DOskip=4
         GOTO 320
       ENDIF
      ENDIF
!
! These statements ensure isoprene parameterizations not called
!    for domains with no isoprene parameterizations.
!    40-90N WIN, 60-90N SPR, 60-90N AUT, 60-90S All Seasons
!
      IF (OH_LAT.GE.ALATS(NLATS-1).AND.OH_SEASON.EQ.1) OH_ISOP=1.
      IF (OH_LAT.GE.ALATS(NLATS).AND.OH_SEASON.NE.3) OH_ISOP=1.
      IF (OH_LAT.LE.ALATS(2).AND.OH_SEASON.EQ.2) OH_ISOP=1.
      IF (OH_LAT.LE.ALATS(2).AND.OH_SEASON.EQ.3) OH_ISOP=1.
      IF (OH_LAT.LE.ALATS(1)) OH_ISOP=1.
!
!***************************************************
! These statements ensure the MODERATELY POLLUTED parameterizations
!    not called for domains with no parameterization.
!    Surface and Middle Troposphere.
!***************************************************
!
! For surface.
!
      IF ( OH_NOX.GT.CNOXS(1).AND.  &
     &  OH_NOX.LE.CNOXS(3).AND.OH_PRESS.GE.PRESSES(1)) THEN
!
         NODOMOD=0
!
       IF (OH_LAT.LE.ALATS(1)) NODOMOD=1
       IF (OH_LAT.LE.ALATS(2).AND.OH_LAT.GT.ALATS(1)) THEN
          IF (OH_SEASON.EQ.3.OR.OH_SEASON.EQ.2) NODOMOD=1
       ENDIF
       IF (OH_LAT.GT.ALATS(6).AND.OH_SEASON.EQ.1) NODOMOD=1
       IF (OH_LAT.GT.ALATS(7).AND.OH_SEASON.NE.3) NODOMOD=1
!
          IF (NODOMOD.NE.0) THEN
                OH_NOX=CNOXS(1)-10.
                INDVARA(COLON,COLAT,COALT,3)=OH_NOX
                INDVARB(COLON,COLAT,COALT,3)=OH_NOX
                INDVARC(COLON,COLAT,COALT,3)=OH_NOX
                INDVARD(COLON,COLAT,COALT,3)=OH_NOX
          ENDIF
!
      ENDIF
!
!******
!
! For Middle Troposphere.
!
      IF (OH_NOX.GT.CNOXS(1).AND.OH_PRESS.LT.PRESSES(1).AND. &
          OH_PRESS.GE.PRESSES(2)) THEN
!
         NODOMOD=0
!
       IF (OH_LAT.LE.ALATS(3)) NODOMOD=1
       IF (OH_LAT.LE.ALATS(4).AND.OH_LAT.GT.ALATS(3)) THEN
           IF (OH_SEASON.EQ.1) NODOMOD=1
           IF (OH_SEASON.EQ.2) NODOMOD=1
       ENDIF
       IF (OH_LAT.LE.ALATS(5).AND.OH_LAT.GT.ALATS(4)) THEN
           IF (OH_SEASON.EQ.3) NODOMOD=1
           IF (OH_SEASON.EQ.4) NODOMOD=1
       ENDIF
       IF (OH_LAT.GT.ALATS(6).AND.OH_SEASON.EQ.1) NODOMOD=1
       IF (OH_LAT.GT.ALATS(7)) NODOMOD=1
!
          IF (NODOMOD.NE.0) THEN
                OH_NOX=CNOXS(1)-10.
                INDVARA(COLON,COLAT,COALT,3)=OH_NOX
                INDVARB(COLON,COLAT,COALT,3)=OH_NOX
                INDVARC(COLON,COLAT,COALT,3)=OH_NOX
                INDVARD(COLON,COLAT,COALT,3)=OH_NOX
          ENDIF
!
! Reset boxes where NOx > 1000 in MT.
          IF (OH_NOX.GT.CNOXS(3)) THEN
                OH_NOX=CNOXS(3)-10.
                INDVARA(COLON,COLAT,COALT,3)=OH_NOX
                INDVARB(COLON,COLAT,COALT,3)=OH_NOX
                INDVARC(COLON,COLAT,COALT,3)=OH_NOX
                INDVARD(COLON,COLAT,COALT,3)=OH_NOX
          ENDIF
!
         GOTO 320
      ENDIF
!
!******
!
! For Upper Troposphere.
!
      IF (OH_NOX.GT.CNOXS(2).AND.OH_PRESS.LT.PRESSES(2)) THEN
          OH_NOX=CNOXS(2)-10.
          INDVARA(COLON,COLAT,COALT,3)=OH_NOX
          INDVARB(COLON,COLAT,COALT,3)=OH_NOX
          INDVARC(COLON,COLAT,COALT,3)=OH_NOX
          INDVARD(COLON,COLAT,COALT,3)=OH_NOX
         GOTO 320
      ENDIF
!
!***************************************************
! These statements ensure the E) HEAVILY POLLUTED parameterizations
!    not called for domains with no parameterization.
!    Very hi NOx associated with low OH, especially in wintertime.
!***************************************************
!
! Winter months:  high NOx=low OH.
!
      IF (OH_NOX.GT.CNOXS(3)) THEN
        IF (OH_LAT.GE.ALATS(7).AND.OH_SEASON.EQ.2) THEN
!bnd            PARAMOH=0.
            PARAMOH=1.e3
            DOskip=1
         GOTO 320
        ENDIF
        IF (OH_LAT.GE.ALATS(5).AND.OH_SEASON.EQ.1) THEN
!bnd            PARAMOH=0.
            PARAMOH=1.e3
            DOskip=1
         GOTO 320
        ENDIF
        IF (OH_LAT.LE.ALATS(3).AND.OH_SEASON.EQ.3) THEN
!bnd            PARAMOH=0.
            PARAMOH=1.e3
            DOskip=1
         GOTO 320
        ENDIF
      ENDIF
!
      IF (OH_NOX.GT.CNOXS(4)) THEN
!
! Assume hi NOx = low OH for areas without heavily polluted parameterizations.
!  
       IF (OH_LAT.LT.ALATS(3).OR.OH_LAT.GE.ALATS(7).OR.  &
     &    (OH_LAT.GE.ALATS(6).AND.  &
     &    OH_SEASON.LE.2).OR.(OH_LAT.GE.ALATS(5).AND.  &
     &    OH_SEASON.EQ.1)) THEN

         DOskip=1
         GOTO 320
!
       ELSE
!
! Reset NOx < 5 ppbv in regions with heavily polluted parameterizations.
!
          OH_NOX=CNOXS(4)-10.
          INDVARA(COLON,COLAT,COALT,3)=OH_NOX
          INDVARB(COLON,COLAT,COALT,3)=OH_NOX
          INDVARC(COLON,COLAT,COALT,3)=OH_NOX
          INDVARD(COLON,COLAT,COALT,3)=OH_NOX
!
       ENDIF
!
      ENDIF
!
! Make sure not to call nonexistent parameterization. 
!
      IF (OH_NOX.GT.CNOXS(3).AND.OH_NOX.LT.CNOXS(4)) THEN
!       
         NODOHI=0
!
       IF (OH_LAT.LE.ALATS(3))  NODOHI=1
       IF (OH_LAT.GT.ALATS(5).AND.OH_SEASON.EQ.1) NODOHI=1
       IF (OH_LAT.GT.ALATS(7))  NODOHI=1
       IF (OH_LAT.GT.ALATS(6).AND.OH_SEASON.EQ.2) NODOHI=1

       IF (NODOHI.NE.0) THEN
          OH_NOX=CNOXS(3)-10.
          INDVARA(COLON,COLAT,COALT,3)=OH_NOX
          INDVARB(COLON,COLAT,COALT,3)=OH_NOX
          INDVARC(COLON,COLAT,COALT,3)=OH_NOX
          INDVARD(COLON,COLAT,COALT,3)=OH_NOX
          GOTO 320
       ENDIF
!
      ENDIF
!
!***************************************************
!
 320  RETURN
      END
