#include "MAPL_Generic.h"

!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 900.3     !
!-------------------------------------------------------------------------
!BOP
!
! !MODULE:  SF6_GridCompMod --- SF6 Grid Component Class
!
! !INTERFACE:
!

   MODULE  SF6_GridCompMod

! !USES:

   USE ESMF
   USE MAPL_Mod

   USE Chem_Mod                   ! Chemistry Base Class
   USE Chem_StateMod              ! Chemistry State
   USE Chem_UtilMod               ! I/O

   USE m_inpak90                  ! Resource file management
   USE m_die, ONLY: die

   IMPLICIT NONE

! !PUBLIC TYPES:
!
   PRIVATE
   PUBLIC  SF6_GridComp       ! SF6 object 

!
! !PUBLIC MEMBER FUNCTIONS:
!

   PUBLIC  SF6_GridCompInitialize
   PUBLIC  SF6_GridCompRun
   PUBLIC  SF6_GridCompFinalize

!
       REAL, PARAMETER :: Pa2hPa = 0.01
       REAL, PARAMETER :: secPerDay = 86400.00
       REAL, PARAMETER :: mwtAir=28.97
       REAL, PARAMETER :: mwSF6=146.0564192
       REAL    :: qmin, qmax
       INTEGER, PARAMETER :: SECS_PER_HOUR    = 3600
       INTEGER, PARAMETER :: HRS_PER_DAY      =   24
       INTEGER, PARAMETER :: SECS_PER_DAY     = SECS_PER_HOUR*HRS_PER_DAY
       INTEGER, PARAMETER :: DAYS_PER_YEAR    =  365
       INTEGER, PARAMETER :: DAYS_PER_YEAR_LY =  366
!
! !DESCRIPTION:
!
!  This module implements the SF6 Grid Component. 
!
! !REVISION HISTORY:
!
!  06September2011 Kouatchou  First crack.
!
!  NOTE: As far as we know, this module has not been tested  - Elshorbany & Manyin
!
!EOP
!-------------------------------------------------------------------------

  TYPE SF6_GridComp

       CHARACTER(LEN=255) :: name            ! generic name of the package
       CHARACTER(LEN=255) :: iname           ! instance name
       CHARACTER(LEN=255) :: rcfilen         ! resource file name
       CHARACTER(LEN=255) :: SF6_AnnualEmissions_inFile
       CHARACTER(LEN=255) :: SF6_CFC115_loss_inFile 

       INTEGER :: instance                   ! instance number
       INTEGER :: nymd_SF6
       INTEGER :: BCmonth   
       INTEGER :: BCyear   

       LOGICAL :: DBG
       ! Variables coming from the input files
       REAL*8, POINTER    ::   SF6_j(:,:,:)     ! photolysis rate, top-down
       REAL*8, POINTER    ::   SF6_k(:,:,:)     ! rate constant    top-down
       REAL*8, POINTER    ::   E_SF6(:,:)       ! Emissions data
!      REAL*8, POINTER    ::  SF6_AD(:,:,:) ! unused
       REAL*8, POINTER    :: SF6_O1D(:,:,:)     ! top-down
  END TYPE SF6_GridComp

!-------------------------------------------------------------------------
CONTAINS
!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 900.3     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  SF6_GridCompInitialize --- Initialize SF6_GridComp
!
! !INTERFACE:
!

      subroutine SF6_GridCompInitialize ( gcSF6, w_c, impChem, expChem, &
                                      nymd, nhms, cdt, rc )
! !USES:

     IMPLICIT NONE

! !INPUT PARAMETERS:
      INTEGER, INTENT(IN) :: nymd, nhms        ! time
      REAL,    INTENT(IN) :: cdt               ! chemical timestep (secs)
      TYPE(Chem_Bundle), intent(in) :: w_c     ! Chemical tracer fields      
!
! !OUTPUT PARAMETERS:

      TYPE(SF6_GridComp), INTENT(INOUT) :: gcSF6   ! Grid Component
      TYPE(ESMF_State), INTENT(INOUT)  :: impChem  ! Import State
      TYPE(ESMF_State), INTENT(INOUT)  :: expChem  ! Export State
      INTEGER, INTENT(OUT) ::  rc                  ! Error return code:
                                                   !  0 - all is well
                                                   !  1 - 
!
! !DESCRIPTION: 
! Initializes the SF6 Grid Component. 
!
! !REVISION HISTORY:
!
!  24June2010 Nielsen  First crack.
!
!EOP
!-------------------------------------------------------------------------

      CHARACTER(LEN=*), PARAMETER :: myname = 'SF6_GridCompInitialize'
      CHARACTER(LEN=255) :: name
   
      INTEGER :: ios, j, i, ic, n
      INTEGER, ALLOCATABLE :: ier(:)

      INTEGER :: JDAY, curMonth
      INTEGER :: i1, i2, im, j1, j2, jm, km
      INTEGER :: nbeg, nend, begTime, nTimes, incSecs

      gcSF6%rcfilen = 'SF6_GridComp.rc'
      gcSF6%name = 'GEOS-5/GOCART SF6 Package'

      IF(MAPL_AM_I_ROOT()) PRINT *,myname,": Begin..."

!  Initialize local variables
!  --------------------------

      rc = 0
      i1 = w_c%grid%i1
      i2 = w_c%grid%i2
      im = w_c%grid%im

      j1 = w_c%grid%j1
      j2 = w_c%grid%j2
      jm = w_c%grid%jm

      km = w_c%grid%km

      nbeg  = w_c%reg%i_SF6
      nend  = w_c%reg%j_SF6

!  It requires 1 bin
!  -----------------
      if ( nbeg /= nend ) then
         IF(MAPL_AM_I_ROOT()) PRINT *,myname,": Must have only one SF6 bin"
         rc = 10
         return
      end if

      gcSF6%BCmonth  = -1
      gcSF6%BCyear   = -1

      allocate(ier(199))
      ier(:) = 0

!  Load resource file
!  ------------------
      CALL I90_loadf ( TRIM(gcSF6%rcfilen), ier(1) )
      if ( ier(1) .NE. 0 ) then
         rc = 11
         IF (MAPL_AM_I_ROOT()) PRINT *,myname,"Could not open:",TRIM(gcSF6%rcfilen)
         return
      end if

!  Parse resource file
!  -------------------
      CALL I90_label ( 'SF6_CFC115_loss_inFile:', ier(3) )
      CALL I90_gtoken( gcSF6%SF6_CFC115_loss_inFile, ier(4) )

      CALL I90_label ( 'SF6_AnnualEmissions_inFile:', ier(3) )
      CALL I90_gtoken( gcSF6%SF6_AnnualEmissions_inFile, ier(4) )

      ier(:)=0

      allocate(  gcSF6%SF6_j(  i1:i2,j1:j2,1:km), STAT=ier(10))
      allocate(  gcSF6%SF6_k(  i1:i2,j1:j2,1:km), STAT=ier(11))
      allocate(  gcSF6%E_SF6(  i1:i2,j1:j2     ), STAT=ier(12))
!     allocate(  gcSF6%SF6_AD( i1:i2,j1:j2,1:km), STAT=ier(13))
      allocate(  gcSF6%SF6_O1D(i1:i2,j1:j2,1:km), STAT=ier(14)) 

      IF ( ANY(ier(:) /= 0)) THEN
         PRINT*, TRIM(myname),': Failed to allocate work space.'
         rc = 12
         RETURN
      END IF
      ier(:) = 0

      DEALLOCATE(ier)

      RETURN

 end subroutine SF6_GridCompInitialize

!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 900.3     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  SF6_GridCompRun --- Run SF6_GridComp
!
! !INTERFACE:
!

   subroutine SF6_GridCompRun ( gcSF6, w_c, impChem, expChem, &
                                      nymd, nhms, cdt, rc )

! !USES:

  IMPLICIT NONE

! !INPUT PARAMETERS:

   TYPE(Chem_Bundle), intent(in) :: w_c        ! Chemical tracer fields      
   INTEGER, INTENT(IN) :: nymd, nhms           ! time
   REAL,    INTENT(IN) :: cdt                  ! chemical timestep (secs)


! !OUTPUT PARAMETERS:

   TYPE(SF6_GridComp), INTENT(INOUT) :: gcSF6   ! Grid Component
   TYPE(ESMF_State), INTENT(INOUT)  :: impChem  ! Import State
   TYPE(ESMF_State), INTENT(INOUT)  :: expChem  ! Export State
   INTEGER, INTENT(OUT) ::  rc                  ! Error return code:
                                                !  0 - all is well
                                                !  1 - 

! !DESCRIPTION: Runs the SF6 Grid Component. 
!
! !REVISION HISTORY:
!
!  24June2010 Nielsen  First crack.
!
!EOP
!-------------------------------------------------------------------------

      INTEGER, ALLOCATABLE :: ier(:)
      CHARACTER(LEN=*), PARAMETER :: myname = 'SF6_GridCompRun'
      CHARACTER(LEN=*), PARAMETER :: Iam = myname

!  Input fields from fvGCM
!  -----------------------

      REAL, POINTER, DIMENSION(:,:)   ::  tropp     => null()
      REAL, POINTER, DIMENSION(:,:)   ::  albedo    => null()
      REAL, POINTER, DIMENSION(:,:)   ::  cellArea  => null()

      REAL, POINTER, DIMENSION(:,:,:) ::  T        => null()
!      REAL, POINTER, DIMENSION(:,:,:) ::  pl       => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  ple      => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  zle      => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  rhoa     => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  specHum  => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  taucli   => null()
      REAL, POINTER, DIMENSION(:,:,:) ::  tauclw   => null()

      REAL, ALLOCATABLE :: pe(:,:,:),p(:,:,:)
      REAL, ALLOCATABLE :: rkoh(:,:,:)
      REAL, ALLOCATABLE :: airNumDensity(:,:,:)        !  top-down

      REAL*8,  ALLOCATABLE :: scaleO3    (:,:)
      REAL*8,  ALLOCATABLE :: mass(:,:,:)              !  top-down  (unused)
      REAL*8,  ALLOCATABLE :: gridBoxThickness(:,:,:)  !  bottom-up (unused)
      REAL*8,  ALLOCATABLE :: PL(:,:,:)                !  top-down  (unused)
      REAL*8,  ALLOCATABLE :: optDepth(:,:,:)          !  bottom-up (unused)
      REAL*8,  ALLOCATABLE :: ALBD(:,:)                !            (unused)
      REAL*8,  ALLOCATABLE :: tropPress(:,:)           !            (unused)
      REAL,    ALLOCATABLE :: var3d(:,:,:)
      REAL,    ALLOCATABLE :: var2d(:,:)
      REAL,    ALLOCATABLE :: cellDepth(:,:,:)         !  top-down
      REAL,    ALLOCATABLE :: cellVolume(:,:,:)        !  top-down  (unused)

      INTEGER :: i1, i2, im, j1, j2, jm, km, ios, iXj
      INTEGER :: i, j, k, kReverse, n, nbeg, nend, ic, it
      INTEGER :: hms

      INTEGER :: JDAY, curMonth
      INTEGER :: cymd, dymd, ix, jx
      REAL*8 :: time_sec, offset_sec
      REAL :: radToDeg, degToRad, pi

#define EXPORT   expChem
!#define iNAME    TRIM(gcSF6%iname)

#define SF6LS   SF6_loss

   integer :: STATUS

!#include "SF6_GetPointer___.h"
!#include "SF6_DeclarePointer___.h"
!       Local arrays referencing the Import/Export states
!       -------------------------------------------------
        real, pointer, dimension(:,:,:) :: SF6LS ! EXPORT: SF6 Chemical Loss


        call MAPL_GetPointer ( EXPORT, SF6LS,  'SF6LS', ALLOC=.TRUE., RC=STATUS )
        VERIFY_(STATUS)

      !IF(MAPL_AM_I_ROOT()) PRINT *,myname,": begin...."
     
      pi       = 4.00*ATAN(1.00)
      degToRad = pi/180.00
      radToDeg = 180.00/pi

      radToDeg = 57.2957795

      !  Initialize local variables
      !  --------------------------
      rc = 0
      i1 = w_c%grid%i1
      i2 = w_c%grid%i2
      im = w_c%grid%im
   
      j1 = w_c%grid%j1
      j2 = w_c%grid%j2
      jm = w_c%grid%jm
      km = w_c%grid%km
   
      iXj = ( i2 - i1 + 1 ) * ( j2 - j1 + 1 )

      allocate(ier(128))
      ier(:) = 0

      nbeg  = w_c%reg%i_SF6
      nend  = w_c%reg%j_SF6

      !  It requires 1 bin
      !  -----------------
      if ( nbeg /= nend ) then
         IF(MAPL_AM_I_ROOT()) PRINT *,myname,": Must have only one SF6 bin"
         rc = 20
         return 
      end if

      ! Reserve Some local work space
      !------------------------------
      allocate(            ALBD(i1:i2,j1:j2),STAT=ier(13))
      allocate(       tropPress(i1:i2,j1:j2),STAT=ier(14))
      allocate(           var2d(i1:i2,j1:j2),STAT=ier(15))

      allocate(            mass(i1:i2,j1:j2,1:km),STAT=ier(22))
      allocate(        optDepth(i1:i2,j1:j2,1:km),STAT=ier(23))
      allocate(gridBoxThickness(i1:i2,j1:j2,1:km),STAT=ier(24))
      allocate(           var3d(i1:i2,j1:j2,1:km),STAT=ier(25))
      allocate(              PL(i1:i2,j1:j2,1:km),STAT=ier(26))
      allocate(       cellDepth(i1:i2,j1:j2,1:km),STAT=ier(27))
      allocate(      cellVolume(i1:i2,j1:j2,1:km),STAT=ier(28))
      allocate(   airNumDensity(i1:i2,j1:j2,1:km),STAT=ier(29))


      IF ( ANY(ier(:) /= 0)) THEN
         PRINT*, TRIM(myname),': Failed to allocate work space.'
         rc = 21
         RETURN
      END IF
      ier(:) = 0

      JDAY     = JulianDay(nymd)
      curMonth = MOD(nymd, 10000) / 100

      offset_sec = 0.0d0
      time_sec   = convertTimeToSeconds(nhms)

      CALL Acquire_CFC115_loss(rc)

      CALL Acquire_AnnualEmissions(rc)

      CALL MonitorCCM_Fields(rc)

      nbeg  = w_c%reg%i_SF6
      nend  = w_c%reg%j_SF6

!  Convert SF6 from mole fraction to number density
!  ----------------------------------------------------
      w_c%qa(nbeg)%data3d(:,:,:) = &
      w_c%qa(nbeg)%data3d(:,:,:) * airNumDensity(:,:,:)
   
! Loss due to photolysis and O1D (units: molec/m3/s) for calculated for output      

      var3d(:,:,:) = gcSF6%SF6_j(:,:,:) +  &
                     gcSF6%SF6_k(:,:,:)*gcSF6%SF6_O1D(:,:,:)*airNumDensity(:,:,:)*1.0E-6

      SF6_loss(:,:,:) = &
      SF6_loss(:,:,:) + w_c%qa(nbeg)%data3d(:,:,:) * var3d(:,:,:)

! Update the SF6 variable
     w_c%qa(nbeg)%data3d(:,:,:) =  &
     w_c%qa(nbeg)%data3d(:,:,:) * (1.00-cdt*(gcSF6%SF6_j(:,:,:) +  &
            gcSF6%SF6_k(:,:,:)*gcSF6%SF6_O1D(:,:,:)*airNumDensity(:,:,:)*1.0E-6)) 

!  SF6  Emission: mol m^-2 s^-1.  Convert to molec m^-3 s^-1,
!  multiply by dt, and add to the number density (molec/m3). 
!  -------------------------------------------------------------

      w_c%qa(nbeg)%data3d(:,:,km) = &
      w_c%qa(nbeg)%data3d(:,:,km) + (cdt*1.0E-03*MAPL_AVOGAD*gcSF6%E_SF6(:,:)/cellDepth(:,:,km))

!  Return to mole fraction
!  -----------------------
      w_c%qa(nbeg)%data3d(i1:i2,j1:j2,1:km) = &
      w_c%qa(nbeg)%data3d(i1:i2,j1:j2,1:km)/airNumDensity(i1:i2,j1:j2,1:km)

! output SF6 loss as mol/mol/s
      SF6_loss(:,:,:) = &
      SF6_loss(:,:,:) / airNumDensity(i1:i2,j1:j2,1:km) 


!  Housekeeping
!  ------------

      DEALLOCATE(ALBD, tropPress, var2d, mass, optDepth, gridBoxThickness, &
                 var3d, PL, cellDepth, cellVolume, airNumDensity )
   
      DEALLOCATE(ier)

      RETURN

!---------------------------------------------------------------------------
   CONTAINS
!---------------------------------------------------------------------------
!BOP
!
! !ROUTINE:  Acquire_CFC115_loss
!
! !INTERFACE:

  SUBROUTINE Acquire_CFC115_loss(rc)

  INTEGER, INTENT(OUT) :: rc
!
! !DESCRIPTION:
! Monthly reads the CFC115_loss variables.
!
!EOP
!---------------------------------------------------------------------------

  CHARACTER(LEN=255) :: MyName
  INTEGER :: ic

  rc=0
  MyName="Acquire_CFC115_loss"

! Execute updates at 0:00 UTC and/or on first pass
! ------------------------------------------------
   ! Read the file once a month
   IF (gcSF6%BCmonth /= (MOD(nymd,10000)/100) ) THEN
      cymd = 2001*10000+MOD(nymd,10000)
      hms = 120000

      dymd = cymd

      IF ( MAPL_AM_I_ROOT() ) THEN
         PRINT *, ' '
         PRINT *,'SF6_GridCompRun: Obtaining CFC115_loss variables ...'
      END IF

      CALL Chem_UtilMPread(gcSF6%SF6_CFC115_loss_inFile, &
                'j', dymd, &
                hms, i1, i2, 0, im, j1, j2, 0, jm, km, &
                var3d=var3d, cyclic=.TRUE., grid = w_c%grid_esmf)

      gcSF6%SF6_j(:,:,1:km) = var3d(:,:,:)  !km:1:-1

      CALL Chem_UtilMPread(gcSF6%SF6_CFC115_loss_inFile, &
                'k', dymd, &
                hms, i1, i2, 0, im, j1, j2, 0, jm, km, &
                var3d=var3d, cyclic=.TRUE., grid = w_c%grid_esmf)

      gcSF6%SF6_k(:,:,1:km) = var3d(:,:,:)  ! km:1:-1

      CALL Chem_UtilMPread(gcSF6%SF6_CFC115_loss_inFile, &
                'O1D', dymd, &
                hms, i1, i2, 0, im, j1, j2, 0, jm, km, &
                var3d=var3d, cyclic=.TRUE., grid = w_c%grid_esmf)

      gcSF6%SF6_O1D(:,:,1:km) = var3d(:,:,:)  ! km:1:-1

      CALL Chem_UtilMPread(gcSF6%SF6_CFC115_loss_inFile, &
                'AD', dymd, &
                hms, i1, i2, 0, im, j1, j2, 0, jm, km, &
                var3d=var3d, cyclic=.TRUE., grid = w_c%grid_esmf)

!     gcSF6%SF6_AD(:,:,1:km) = var3d(:,:,:)  ! km:1:-1

      gcSF6%BCmonth = MOD(nymd,10000) / 100
  END IF

  RETURN

  END SUBROUTINE Acquire_CFC115_loss
!EOC
!------------------------------------------------------------------------------------
!BOP
!
! !ROUTINE:  Acquire_AnnualEmissions
!
! !INTERFACE:

  SUBROUTINE Acquire_AnnualEmissions(rc)

  INTEGER, INTENT(OUT) :: rc
!
! !DESCRIPTION:
! Yearly reads the AnnualEmissions variable: mol SF6/m2/s
!
!EOP
!---------------------------------------------------------------------------

  CHARACTER(LEN=255) :: MyName
  INTEGER :: ic

  rc=0
  MyName="Acquire_AnnualEmissions"

! Execute updates at 0:00 UTC and/or on first pass
! ------------------------------------------------
   ! Read the file once a year
   IF (gcSF6%BCyear /= (nymd/10000)) THEN
      cymd = 2001*10000+MOD(nymd,10000)
      hms = 120000

     dymd = cymd

    IF( MAPL_AM_I_ROOT() ) THEN
     PRINT *, ' '
     PRINT *,'SF6_GridCompRun: Obtaining AnnualEmissions variables ...'
    END IF


      CALL Chem_UtilMPread(gcSF6%SF6_AnnualEmissions_inFile, &
                'E_SF6', dymd, &
                hms, i1, i2, 0, im, j1, j2, 0, jm, km, &
                var2d=var2d, cyclic=.TRUE., grid = w_c%grid_esmf)

      gcSF6%E_SF6(:,:) = var2d(:,:)

      gcSF6%BCyear = nymd/10000
  END IF

  RETURN

  END SUBROUTINE Acquire_AnnualEmissions
!EOC
!------------------------------------------------------------------------------------
!BOP
!
! !ROUTINE:  MonitorCCM_Fields
!
! !INTERFACE:

      SUBROUTINE MonitorCCM_Fields(rc)

      IMPLICIT NONE

      INTEGER, INTENT(OUT) :: rc
!
! !DESCRIPTION:
!
! Obtain 24h averages of temperature, pressure, etc.
!
!EOP
!---------------------------------------------------------------------------

      CHARACTER(LEN=255) :: Iam
      REAL :: avgFactor
      INTEGER :: ndt, k
      logical, save :: firstTime = .TRUE.

      rc = 0
      Iam = "MonitorCCM_Fields"
      avgFactor = cdt/secPerDay

      CALL MAPL_GetPointer(impChem,    rhoa, 'AIRDENS',  rc=ier(30))
      CALL MAPL_GetPointer(impChem, specHum,       'Q',  rc=ier(31))
      CALL MAPL_GetPointer(impChem,  albedo,   'ALBVF',  rc=ier(32))
      CALL MAPL_GetPointer(impChem,       T,       'T',  rc=ier(34))
      CALL MAPL_GetPointer(impChem,     zle,     'ZLE',  rc=ier(35))
      CALL MAPL_GetPointer(impChem,     ple,     'PLE',  rc=ier(36))
      CALL MAPL_GetPointer(impChem,   tropp,   'TROPP',  rc=ier(37))
      CALL MAPL_GetPointer(impChem,  taucli,  'TAUCLI',  rc=ier(38))
      CALL MAPL_GetPointer(impChem,  tauclw,  'TAUCLW',  rc=ier(39))
      CALL MAPL_GetPointer(impChem, cellArea,   'AREA',  rc=ier(40) ) 

      if ( ANY(ier(:) /= 0) ) then
         rc = 10
         PRINT *,TRIM(myname)//':'//TRIM(Iam),": Failed to obtain the import state"
         return
      end if

! MANYIN - should rhoa and airNumDesity share the same vertical orientation?
      ! Air number density (molec/m3)
      ! ------------------
      airNumDensity(i1:i2,j1:j2,1:km) = rhoa(i1:i2,j1:j2,1:km)*MAPL_AVOGAD/mwtAir

      !  Cell depth and volume
      !  ---------------------
! MANYIN - do  rhoa and w_c%delp  share the same vertical orientation?  YES  top-down
      DO k=1,km
       cellDepth(i1:i2,j1:j2,k) = w_c%delp(i1:i2,j1:j2,k)/(rhoa(i1:i2,j1:j2,k)*MAPL_GRAV)
       cellVolume(i1:i2,j1:j2,k) = cellArea(i1:i2,j1:j2)*cellDepth(i1:i2,j1:j2,k)
      END DO

! zle  TOP-DOWN
! mass TOP-DOWN
      DO k=1,km
         mass(:,:,k)=rhoa(:,:,k)*cellArea(:,:)* &                ! kg
                       (zle(:,:,k-1)-zle(:,:,k))
      END DO

! zle  TOP-DOWN
! gridBoxThickness  BOTTOM-UP
      gridBoxThickness(:,:,1:km) = zle(:,:,km-1:0:-1)-zle(:,:,km:1:-1)

! optDepth  BOTTOM-UP
! taucli  TOP-DOWN
! tauclw  TOP-DOWN
      optDepth(i1:i2,j1:j2,1:km)   = taucli(i1:i2,j1:j2,km:1:-1) &
                                   + tauclw(i1:i2,j1:j2,km:1:-1)

      tropPress(i1:i2,j1:j2)       =  tropp(i1:i2,j1:j2)*Pa2hPa

! ple TOP-DOWN
! PL  TOP-DOWN
      PL(:,:,:)  = 0.5*(ple(:,:,1:km)+ple(:,:,0:km-1))
      !PL(:,:,:)  = 0.5*(PLE(:,:,1:km)+PLE(:,:,0:km-1))*Pa2hPa

      ALBD(i1:i2,j1:j2) = albedo(i1:i2,j1:j2)


      RETURN

      END SUBROUTINE MonitorCCM_Fields

 END SUBROUTINE SF6_GridCompRun
!EOP
!---------------------------------------------------------------------------

!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 900.3     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  SF6_GridCompFinalize --- Initialize SF6_GridComp
!
! !INTERFACE:
!

   subroutine SF6_GridCompFinalize ( gcSF6, w_c, impChem, expChem, &
                                      nymd, nhms, cdt, rc )

! !USES:

  IMPLICIT NONE

! !INPUT PARAMETERS:

   TYPE(Chem_Bundle), intent(in) :: w_c         ! Chemical tracer fields      
   INTEGER, INTENT(IN) :: nymd, nhms            ! time
   REAL,    INTENT(IN) :: cdt                   ! chemical timestep (secs)


! !OUTPUT PARAMETERS:

   TYPE(SF6_GridComp), INTENT(INOUT) :: gcSF6   ! Grid Component
   TYPE(ESMF_State), INTENT(INOUT)  :: impChem  ! Import State
   TYPE(ESMF_State), INTENT(INOUT)  :: expChem  ! Export State
   INTEGER, INTENT(OUT) ::  rc                  ! Error return code:
                                                !  0 - all is well
                                                !  1 - 

! !DESCRIPTION: Finalizes the SF6 Grid Component.
!
! !REVISION HISTORY:
!
!  27Feb2008  da Silva  Introduced multiple instances
!
!EOP
!-------------------------------------------------------------------------

   CHARACTER(LEN=*), PARAMETER :: myname = 'SF6_GridCompFinalize'
   INTEGER :: ios

   rc = 0

   DEALLOCATE( gcSF6%SF6_j, gcSF6%SF6_k, gcSF6%E_SF6, gcSF6%SF6_O1D, STAT=ios )
   IF ( ios /= 0 ) rc = 1

!  DEALLOCATE( gcSF6%SF6_AD, STAT=ios )
!  IF ( ios /= 0 ) rc = 1

 end subroutine SF6_GridCompFinalize
!EOC
!--------------------------------------------------------------------------
!BOP
      INTEGER FUNCTION JulianDay(nymd)
         INTEGER :: nymd
!
! !DESCRIPTION:
! Determines the Julian day: number between 1 and 365 (or 366).
         INTEGER :: ny, mm, dd
         INTEGER :: m, ds
         INTEGER :: days(12)
!EOP
!--------------------------------------------------------------------------
!BOC
         data days /31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31/

         ny = nymd / 10000
         mm = mod(nymd, 10000) / 100
         dd = mod(nymd,   100)

         ds = dd

         if ( mm .ne. 1) then
            do m=1, mm-1
               if ( m.eq.2  .and. leap_year(ny) ) then
                  ds = ds + 29
               else
                  ds = ds + days(m)
               endif
            enddo
         endif

         JulianDay = ds
         
      END FUNCTION JulianDay
!EOC
!--------------------------------------------------------------------------
!BOP
      function leap_year(ny)
!
! Determine if year ny is a leap year
!
! Author: S.-J. Lin
      implicit none
      logical leap_year
      integer ny
      integer ny00
!EOP
!--------------------------------------------------------------------------
!BOC

!
! No leap years prior to 0000
!
      parameter ( ny00 = 0000 )   ! The threshold for starting leap-year

      if( ny >= ny00 ) then
         if( mod(ny,100) == 0. .and. mod(ny,400) == 0. ) then
             leap_year = .true.
         elseif( mod(ny,4) == 0. .and. mod(ny,100) /= 0.  ) then
             leap_year = .true.
         else
             leap_year = .false.
         endif
      else
          leap_year = .false.
      endif

      return
    end function leap_year

!EOC
!--------------------------------------------------------------------------
!BOP
      FUNCTION convertTimeToSeconds (nhms) result(this_)
!
! !INPUT PARAMETERS
      INTEGER, intent(in) :: nhms
!
! !RETURNED VALUE
      INTEGER :: this_
!
! !DESCRIPTION:
! Converts the current time into seconds.
!
! !DEFINED PARAMETERS:
      REAL*8, PARAMETER :: SECPHR = 3600.0d0
      REAL*8, PARAMETER :: SECPMN =   60.0d0
      REAL*8, PARAMETER :: SECPDY = 24.0d0*SECPHR
!EOP
!------------------------------------------------------------------------------
!BOC
      this_ = (nhms / 10000) * SECPHR +              &
              (Mod (nhms, 10000) / 100) * SECPMN +   &
               Mod (nhms, 100)

      RETURN

      END FUNCTION convertTimeToSeconds
!EOC
!------------------------------------------------------------------------------
END MODULE  SF6_GridCompMod
