#include "MAPL_Exceptions.h"

!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1    !
!-------------------------------------------------------------------------
!BOP
!
! !MODULE:  MAML_OpticsMod--- aerosol optics.
!
! !INTERFACE:
!

   module  MAML_OpticsMod

! !USES:

   use ESMF
   use MAPL_Mod

   use MAPL_SimpleBundleMod

   use MAM_BaseMod
   use MAM_ConstituentsDataMod
   use MAM_ComponentsDataMod

   use MAML_OpticsTableMod, only: MAML_OpticsTable

   implicit none


! !PUBLIC TYPES:

   private

!
! !PUBLIC MEMBER FUNCTIONS:
!
   public  MAML_OpticsInterpolate     ! Computes aerosol optical properties
   
!
! !DESCRIPTION:
!
!  This module computes aerosol optical properties.
!
! !REVISION HISTORY:
!
!  09May2013 Darmenov - Initial code.
!
!EOP
!-------------------------------------------------------------------------

contains

!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  MAML_OpticsCalculator --- Compute aerosol ext, sca, asy...
!
! !INTERFACE:
!

   subroutine MAML_OpticsInterpolate(mie_table, band, q, density, dgn_wet, delp, ext, sca, asy, nc, i1, i2, j1, j2, k1, k2, rc)

   implicit none

! !INPUT/OUTPUT PARAMETERS:

! !INPUT PARAMETERS:
   integer, intent(in)                               :: nc             ! number of aerosol components

   integer, intent(in)                               :: i1, i2         ! upper and lower bounds of spatial dims
   integer, intent(in)                               :: j1, j2         ! ...
   integer, intent(in)                               :: k1, k2         ! ...

   type(MAML_OpticsTable), intent(in)                :: mie_table
   integer, intent(in)                               :: band           ! band index
   real, dimension(nc,i1:i2,j1:j2,k1:k2), intent(in) :: q              ! mass mixing ratios of aerosol components (including absorbed water), 'kg kg-1'
   real, dimension(nc), intent(in)                   :: density        ! density of aerosol components, 'kg m-3'
   real, dimension(i1:i2,j1:j2,k1:k2), intent(in)    :: dgn_wet        ! wet size of number size distribution

   real, dimension(i1:i2,j1:j2,k1:k2), intent(in)    :: delp           ! 

! !OUTPUT PARAMETERS:
   real, dimension(i1:i2,j1:j2,k1:k2), intent(out)   :: ext, sca, asy  ! extinction, scattering and asymmetry parameter
   integer, optional, intent(out)                    :: rc             ! Error return code:
                                                                       !  0 - all is well
                                                                       !  1 -

! !DESCRIPTION:
!  Interpolates Mie lookup table and returns extinction, scattering and asymmetry parameter.
!  NOTE: It is important that the q and density arrays follow the same order of components 
!        as the mie_table.
!
! !REVISION HISTORY:
!
!  09May2013 Darmenov    API.
!
!EOP
!-------------------------------------------------------------------------

                           __Iam__("MAML_OpticsInterpolate")

   real(kind=8) :: Dgs_min, Dgs_max
   real(kind=8) :: logDgs_min, logDgs_max, logDgs
   real(kind=8) :: sigma, log_sigma
   real(kind=8) :: n_re_min, n_re_max
   real(kind=8) :: n_im_min, n_im_max
   real(kind=8) :: q_max

   real(kind=8) :: vol, vol_
   real(kind=8) :: n_re_mix, n_im_mix, n_re_mix_, n_im_mix_

   real(kind=8) :: x                   ! chebyshev polynomial parameter
   real(kind=8) :: tt, uu              ! bilinear interpolation and weights
   real(kind=8) :: w00, w01, w10, w11  ! ...

   real(kind=8), allocatable, dimension(:) :: T
   real(kind=8), allocatable, dimension(:) :: c_ext
   real(kind=8), allocatable, dimension(:) :: c_sca
   real(kind=8), allocatable, dimension(:) :: c_asy

   real(kind=8), allocatable, dimension(:) :: n_re
   real(kind=8), allocatable, dimension(:) :: n_im

   real(kind=8), allocatable, dimension(:) :: component_n_re
   real(kind=8), allocatable, dimension(:) :: component_n_im
   real(kind=8), allocatable, dimension(:) :: q_

   integer :: n_ac            ! number of aerosol components

   integer :: i_re, i_im      ! indexes

   integer :: i, j, k, n      ! loop counters

   real(kind=8)                            :: f                          ! multiplication factor
   real(kind=8), parameter                 :: density_water = 1000.0d0   ! kg m-3


   if (present(rc)) rc = MAM_NOT_IMPLEMENTED_ERROR


   ! basic dim-size check
   ASSERT_(mie_table%n_aerosol_components == nc)


   ! short hand and likely to improve data storage
   n_ac    = mie_table%n_aerosol_components
   Dgs_min = mie_table%Dgs_min
   Dgs_max = mie_table%Dgs_max
   sigma   = mie_table%sigma

   allocate(component_n_re(n_ac), __STAT__)
   allocate(component_n_im(n_ac), __STAT__)
   allocate(q_(n_ac), __STAT__)

   allocate(n_re(mie_table%n_refractive_index_re), __STAT__)
   allocate(n_im(mie_table%n_refractive_index_im), __STAT__)

   component_n_re = mie_table%component_refractive_index_re(band,:)
   component_n_im = mie_table%component_refractive_index_im(band,:)

   n_re     = mie_table%refractive_index_re(band, :)
   n_im     = mie_table%refractive_index_im(band, :)

   n_re_min = n_re(1) ! values are monotonically increasing
   n_re_max = n_re(mie_table%n_refractive_index_re)

   n_im_min = n_im(1) ! values are monotonically increasing
   n_im_max = n_im(mie_table%n_refractive_index_im)

   allocate(    T(mie_table%n_cheb), __STAT__)
   allocate(c_ext(mie_table%n_cheb), __STAT__)
   allocate(c_sca(mie_table%n_cheb), __STAT__)
   allocate(c_asy(mie_table%n_cheb), __STAT__)


   logDgs_min = log(Dgs_min)
   logDgs_max = log(Dgs_max)
   log_sigma  = log(sigma)

   ext = 0.0
   sca = 0.0
   asy = 0.0

   do k = k1, k2
       do j = j1, j2
           do i = i1, i2

               if (any(q(:,i,j,k) > 0.0)) then
                   ! normalize mass mixing ratios to avoid numerical issues 
                   q_max = maxval(q(:,i,j,k))
                   q_    = q(:,i,j,k) / q_max
                   vol   = sum(q_/density)

                   ! compute effective refractive index using volume mixing rule
                   n_re_mix = sum(component_n_re * q_/density) / vol
                   n_im_mix = sum(component_n_im * q_/density) / vol

                   vol = q_max * vol
               else
                   q_max = 0.0d0
                   vol   = 0.0d0
                   n_re_mix = component_n_re(1)
                   n_im_mix = component_n_im(1)
               end if   


               if (n_re_mix < n_re_min) then
                   n_re_mix = n_re_min + tiny(0.0)
               end if

               if (n_re_mix > n_re_max) then
                   n_re_mix = n_re_max - tiny(0.0)
               end if    
 
               if (n_im_mix < n_im_min) then
                   n_im_mix = n_im_min + tiny(0.0)
               end if

               if (n_im_mix > n_im_max) then
                   n_im_mix = n_im_max - tiny(0.0)
               end if

               i_re = locate(n_re_mix, n_re, mie_table%n_refractive_index_re)
               i_im = locate(n_im_mix, n_im, mie_table%n_refractive_index_im)

               if (i_re < 1                                 ) i_re = 1
               if (i_re > mie_table%n_refractive_index_re -1) i_re = mie_table%n_refractive_index_re -1
               if (i_im < 1                                 ) i_im = 1
               if (i_im > mie_table%n_refractive_index_im -1) i_im = mie_table%n_refractive_index_im -1


! bilinear interpolation: notes
!              t = (x - x1) / (x2 - x1)
!              u = (y - y1) / (y2 - y1)
!
!              w00 = (1.0d0 - t) * (1.0d0 - u), f00 = f1
!              w10 = (        t) * (1.0d0 - u), f10 = f2
!              w01 = (1.0d0 - t) * (        u), f01 = f4
!              w11 = (        t) * (        u), f11 = f3
!
!              bilinear_interpolation = (1.0d0 - t) * (1.0d0 - u) * f1 + &
!                                       (        t) * (1.0d0 - u) * f2 + &
!                                       (        t) * (        u) * f3 + &
!                                       (1.0d0 - t) * (        u) * f4
!
!              bilinear_interpolation = w00*f00 + w10*f10 + w01*f01 + w11*f11
!
!
!              mapping: 
!                       x -> re
!                       y -> im 

               tt = (    n_re_mix - n_re(i_re)) / &
                    (n_re(i_re+1) - n_re(i_re))

               uu = (    n_im_mix - n_im(i_im)) / &
                    (n_im(i_im+1) - n_im(i_im))


               ASSERT_ (tt >= 0 .and. tt <= 1.0)
               ASSERT_ (uu >= 0 .and. uu <= 1.0)

               w00 = (1.0d0 - tt) * (1.0d0 - uu)     ! weight for f(i_re+0, i_im+0)
               w10 = (        tt) * (1.0d0 - uu)     ! weight for f(i_re+1, i_im+0)
               w01 = (1.0d0 - tt) * (        uu)     ! weight for f(i_re+0, i_im+1)
               w11 = (        tt) * (        uu)     ! weight for f(i_re+1, i_im+1)

               T     = 0.0d0
               c_ext = 0.0d0
               c_sca = 0.0d0
               c_asy = 0.0d0

               do n = 1, mie_table%n_cheb
                   c_ext(n) = w00 * mie_table%c_ext(band, i_im  , i_re  , n) + &
                              w10 * mie_table%c_ext(band, i_im  , i_re+1, n) + &
                              w01 * mie_table%c_ext(band, i_im+1, i_re  , n) + &
                              w11 * mie_table%c_ext(band, i_im+1, i_re+1, n)
                   
                   c_sca(n) = w00 * mie_table%c_sca(band, i_im  , i_re  , n) + &
                              w10 * mie_table%c_sca(band, i_im  , i_re+1, n) + &
                              w01 * mie_table%c_sca(band, i_im+1, i_re  , n) + &
                              w11 * mie_table%c_sca(band, i_im+1, i_re+1, n)

                   c_asy(n) = w00 * mie_table%c_asy(band, i_im  , i_re  , n) + &
                              w10 * mie_table%c_asy(band, i_im  , i_re+1, n) + &
                              w01 * mie_table%c_asy(band, i_im+1, i_re  , n) + &
                              w11 * mie_table%c_asy(band, i_im+1, i_re+1, n)
               end do

               logDgs = log(dgn_wet(i,j,k)) + 2.0d0*log_sigma*log_sigma
               x = (2.0d0*logDgs - logDgs_max - logDgs_min) / (logDgs_max - logDgs_min)

               if (x > 1.0d0) then
                   x = 1.0d0 - tiny(0.0d0)
               end if

               if (x < -1.0d0) then
                   x = -1.0d0 + tiny(0.0d0)
               end if


               ! compute Chebyshev polynomials
               call cheb_poly(T, mie_table%n_cheb, x)

               ! extinction and scattering are unitless, i.e., AOT(i,j) = sum(ext(i,j,1:km)
               f = vol * density_water * delp(i,j,k) / MAPL_GRAV

               ext(i,j,k) = sum(c_ext * T) * f
               sca(i,j,k) = sum(c_sca * T) * f
               asy(i,j,k) = sum(c_asy * T)

               if (ext(i,j,k) < 0.0d0) then
                   ext(i,j,k) = tiny(0.0d0)
               end if

               if (sca(i,j,k) > ext(i,j,k)) then
                   sca(i,j,k) = ext(i,j,k)
               end if

           end do
       end do
   end do
 
   deallocate(component_n_re, component_n_im, __STAT__)
   deallocate(n_re, n_im, q_,                 __STAT__)
   deallocate(T, c_ext, c_sca, c_asy,         __STAT__)


   RETURN_(ESMF_SUCCESS)

  end subroutine MAML_OpticsInterpolate



!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  MAML_RefractiveIndex --- Computes refractive index of internaly mixed components.
!
! !INTERFACE:
!

   subroutine MAML_RefractiveIndex(q_mass,   &
                                   density,  &
                                   n_re,     &
                                   n_im,     &
                                   n_re_mix, &
                                   n_im_mix, &
                                   rc)

   implicit none

! !INPUT/OUTPUT PARAMETERS:

! !INPUT PARAMETERS:
   real, dimension(:), intent(in) :: q_mass      ! mass mixing ratios of components, including absorbed water
   real, dimension(:), intent(in) :: density     ! densities of components
   real, dimension(:), intent(in) :: n_re        ! real      part of refractive index 
   real, dimension(:), intent(in) :: n_im        ! imaginary part of refractive index


! !OUTPUT PARAMETERS:
   real, intent(out)              :: n_re_mix    ! real      part of effective refractive index
   real, intent(out)              :: n_im_mix    ! imaginary part of effective refractive index
   integer, optional, intent(out) :: rc          ! Error return code:
                                                 !  0 - all is well
                                                 !  1 - 

! !DESCRIPTION: Computes refractive index of internally mixed components using 
!               volume mixing rule.
!
!
! !REVISION HISTORY:
!
!  08May2014 Darmenov    API.
!
!EOP
!-------------------------------------------------------------------------

                           __Iam__("MAML_RefractiveIndex")

   ! local variables
   real :: vol         ! volume


   if (present(rc)) rc = MAM_NOT_IMPLEMENTED_ERROR


   vol = sum(q_mass / density)

   n_re_mix = sum(n_re_mix * q_mass / density) / vol
   n_im_mix = sum(n_im_mix * q_mass / density) / vol


   RETURN_(ESMF_SUCCESS)

  end subroutine MAML_RefractiveIndex




function locate(v, x, nx)

    !
    ! Returns the index of the largest element from the array x
    ! that is smaller than v.
    !
    ! Assumes that the array x is sorted in increasing order,
    ! and that min(x) <= v <= max(x)
    !

    implicit none

    integer :: locate

    real(kind=8), intent(in)                :: v
    real(kind=8), dimension(nx), intent(in) :: x
    integer, intent(in)                     :: nx

    ! local
    integer i
   
    do i = 1, nx
        if (x(i) > v) exit         
    end do
    
    locate = (i - 1)

    return
end function locate


function linear_interpolation(x1, x2, f1, f2, x)

    !
    ! Linear interpolation.
    !

    implicit none

    real(kind=8) :: linear_interpolation

    real(kind=8), intent(in) :: x1
    real(kind=8), intent(in) :: x2
    real(kind=8), intent(in) :: f1
    real(kind=8), intent(in) :: f2
    real(kind=8), intent(in) :: x

    linear_interpolation = f1 + ((x - x1) / (x2 - x1)) * (f2 - f1)
end function linear_interpolation


function bilinear_interpolation(x1, x2, y1, y2, f1, f2, f3, f4, x, y)
    
    ! 
    ! Bilinear interpolation.
    ! 

    implicit none

    real(kind=8) :: bilinear_interpolation

    real(kind=8), intent(in) :: x1
    real(kind=8), intent(in) :: x2
    real(kind=8), intent(in) :: y1
    real(kind=8), intent(in) :: y2
    real(kind=8), intent(in) :: f1
    real(kind=8), intent(in) :: f2
    real(kind=8), intent(in) :: f3
    real(kind=8), intent(in) :: f4
    real(kind=8), intent(in) :: x
    real(kind=8), intent(in) :: y

    ! local
    real(kind=8) :: t, u

    t = (x - x1) / (x2 - x1)
    u = (y - y1) / (y2 - y1)

    bilinear_interpolation = (1.0d0 - t) * (1.0d0 - u) * f1 + &
                             (        t) * (1.0d0 - u) * f2 + &
                             (1.0d0 - t) * (        u) * f4 + &
                             (        t) * (        u) * f3 

end function bilinear_interpolation

subroutine cheb_poly(T, n, x)

    !
    ! Returns the value of the first n Chebyshev 
    ! polynomials. 
    ! 
    ! Uses requrrence equation:
    !     T_n(x) = 2xT_(n-1)(x) - T_(n-2)(x)
    !     T_(n=1)(x) = 1.0
    !     T_(n=2)(x) = x
    !

    implicit none

    real(kind=8), dimension(n), intent(out) :: T
    integer, intent(in)                     :: n
    real(kind=8), intent(in)                :: x

    ! local 
    integer      ::  k
    real(kind=8) :: x2
    
    x2 = 2.0d0 * x

    if (n > 2) then
        T(1) = 1.0d0
        T(2) = x

        do k = 3, n
            T(k) = x2 * T(k-1) - T(k-2)
        end do
    else if (n == 2) then
        T(1) = 1.0d0
        T(2) = x 
    else if (n == 1) then
        T(1) = 1.0d0
    end if

    return
end subroutine cheb_poly


subroutine lookup(lut_refractive_index_re, & 
                  lut_refractive_index_im, &
                  lut_c_ext,               &
                  lut_c_sca,               &
                  lut_c_g,                 &
                  refractive_index_re,     & 
                  refractive_index_im,     &
                  x,                       &
                  ext,                     &
                  sca,                     &
                  g,                       &
                  n_lut_re,                &
                  n_lut_im,                &
                  n_lut_cheb,              &
                  n_lev,                   &
                  n_clm)

    !
    ! Returns 
    !

    implicit none

    integer, intent(in) :: n_lut_re
    integer, intent(in) :: n_lut_im
    integer, intent(in) :: n_lut_cheb

    integer, intent(in) :: n_lev
    integer, intent(in) :: n_clm

    real(kind=8), dimension(n_lut_re), intent(in) :: lut_refractive_index_re
    real(kind=8), dimension(n_lut_im), intent(in) :: lut_refractive_index_im

    real(kind=8), dimension(n_lut_cheb, n_lut_re, n_lut_im), intent(in) :: lut_c_ext
    real(kind=8), dimension(n_lut_cheb, n_lut_re, n_lut_im), intent(in) :: lut_c_sca
    real(kind=8), dimension(n_lut_cheb, n_lut_re, n_lut_im), intent(in) :: lut_c_g

    real(kind=8), dimension(n_clm, n_lev), intent(in) :: refractive_index_re
    real(kind=8), dimension(n_clm, n_lev), intent(in) :: refractive_index_im 
    real(kind=8), dimension(n_clm, n_lev), intent(in) :: x

    real(kind=8), dimension(n_clm, n_lev), intent(out) :: ext
    real(kind=8), dimension(n_clm, n_lev), intent(out) :: sca
    real(kind=8), dimension(n_clm, n_lev), intent(out) :: g


    ! local
    integer :: c, k, l
    integer :: i_re
    integer :: i_im
    real(kind=8), dimension(n_lut_cheb) :: T
    real(kind=8), dimension(n_lut_cheb) :: c_ext
    real(kind=8), dimension(n_lut_cheb) :: c_sca
    real(kind=8), dimension(n_lut_cheb) :: c_g


loop_levels:  do l = 1, n_lev
loop_columns: do c = 1, n_clm

    ! initialize 
    T     = 0.0d0
    c_ext = 0.0d0
    c_sca = 0.0d0
    c_g   = 0.0d0

    ! locate the indexes in refractive index plane
    i_re = locate(refractive_index_re(c,l), lut_refractive_index_re, n_lut_re)
    i_im = locate(refractive_index_im(c,l), lut_refractive_index_im, n_lut_im)

    ! interpolate in the rafractive index plane
    do k = 1, n_lut_cheb
        c_ext(k) = bilinear_interpolation(lut_refractive_index_re(i_re),   &
                                          lut_refractive_index_re(i_re+1), &
                                          lut_refractive_index_im(i_im),   &
                                          lut_refractive_index_im(i_im+1), &
                                          lut_c_ext(k,i_re  ,i_im  ),      &
                                          lut_c_ext(k,i_re+1,i_im  ),      &
                                          lut_c_ext(k,i_re+1,i_im+1),      &
                                          lut_c_ext(k,i_re  ,i_im+1),      &
                                          refractive_index_re(c,l),        &
                                          refractive_index_im(c,l))

        c_sca(k) = bilinear_interpolation(lut_refractive_index_re(i_re),   &
                                          lut_refractive_index_re(i_re+1), &
                                          lut_refractive_index_im(i_im),   &
                                          lut_refractive_index_im(i_im+1), &
                                          lut_c_sca(k,i_re  ,i_im  ),      &
                                          lut_c_sca(k,i_re+1,i_im  ),      &
                                          lut_c_sca(k,i_re+1,i_im+1),      &
                                          lut_c_sca(k,i_re  ,i_im+1),      &
                                          refractive_index_re(c,l),        &
                                          refractive_index_im(c,l))

        c_g(k)   = bilinear_interpolation(lut_refractive_index_re(i_re),   &
                                          lut_refractive_index_re(i_re+1), &
                                          lut_refractive_index_im(i_im),   &
                                          lut_refractive_index_im(i_im+1), &
                                          lut_c_g(k,i_re  ,i_im  ),        &
                                          lut_c_g(k,i_re+1,i_im  ),        &
                                          lut_c_g(k,i_re+1,i_im+1),        &
                                          lut_c_g(k,i_re  ,i_im+1),        &
                                          refractive_index_re(c,l),        &
                                          refractive_index_im(c,l))
    end do


    ! compute Chebyshev polunomials
    call cheb_poly(T, n_lut_cheb, x(c,l))

    ext(c,l) = sum(c_ext * T)
    sca(c,l) = sum(c_sca * T)
    g(c,l)   = sum(c_g   * T)

end do loop_columns
end do loop_levels

end subroutine lookup

end module MAML_OpticsMod
