#include "MAPL_Exceptions.h"
#include "MAPL_Generic.h"


!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1     !
!-------------------------------------------------------------------------
!BOP
!
! !MODULE: MAML_WetRemovalMod - Gravitational sedimentation/settling and dry
!                               deposition of aerosol particles and gases
!
! !INTERFACE:
!
   module MAML_WetRemovalMod
!
! !USES:
!
   use MAPL_Mod

   use MAPL_ConstantsMod

   use Chem_ArrayMod, only: Chem_Array
   use WetRemovalMod, only: WetRemovalGOCART


   implicit None
   private

!
! !PUBLIC MEMBER FUNCTIONS:

   public MAML_WetRemoval

!
! !PUBLIC PARAMETERS:

!
! !PRIVATE PARAMETERS:

   real, private, parameter :: g_E    = MAPL_GRAV     ! standard gravity,        'm s-2'

!
! !DESCRIPTION: 
!
!  {\tt MAML\_SettlingMod} provides a collection of methods for 
!  modeling graviational sedimentation/settling of aerosol particles
!
!
! !REVISION HISTORY:
!
!  29Nov2011  A. Darmenov  Initial version
!
!EOP
!-------------------------------------------------------------------------


   interface MAML_WetRemoval
       module procedure MAML_WetRemovalAerosol
!      module procedure MAML_WetRemovalGas
   end interface MAML_WetRemoval

   contains


!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE: MAML_WetRemovalAerosol --- 
!
! !INTERFACE:

   subroutine MAML_WetRemovalAerosol(qa, f_wet, delp, temperature, rhoa, dqcond, precc, precl, dt, flux, rc)

! !USES:

   implicit None


! !INPUT/OUTPUT PARAMETERS:
   real, target, dimension(:,:,:), intent(inout) :: qa    ! mixing ratio,     'kg kg-1' or '# kg-1'
   real, pointer, dimension(:,:),  intent(inout) :: flux  ! wet removal flux, 'kg m-2 s-1' or '# m-2 s-1'

! !INPUT PARAMETERS:
   real, intent(in)                      :: f_wet       ! large scale wet removal efficiency
   real, dimension(:,:,:), intent(in)    :: delp        ! pressure at model layer edges, 'Pa'
   real, dimension(:,:,:), intent(in)    :: temperature ! temperature
   real, dimension(:,:,:), intent(in)    :: rhoa
   real, dimension(:,:,:), intent(in)    :: dqcond
   real, dimension(:,:),   intent(in)    :: precc
   real, dimension(:,:),   intent(in)    :: precl
   real, intent(in)                      :: dt          ! model time step,              's'

! !OUTPUT PARAMETERS:
   integer, intent(out)                  :: rc          ! return code


! !DESCRIPTION: Wraps the WetRemovalGOCART() subroutine.
!
! !REVISION HISTORY:
!
!  10Dec2012  A. Darmenov
!
!EOP
!-------------------------------------------------------------------------
   
                  __Iam__('MAML_WetRemovalAerosol')


   integer             :: i1, i2, j1, j2, k1, km    ! indexes
   integer             :: i, j, k, kk, n            ! loop counters
   
   integer             :: LH
 
   real                :: F, B, BT                  ! temporary variables on cloud, freq.
   real                :: qmx, qd, A                ! temporary variables on moisture

   real                :: pls, pcv, pac             ! ls, cv, tot precip [mm day-1]
   real, dimension(:), allocatable :: qls, qcv      ! ls, cv portion dqcond [kg m-3 s-1]

   real, dimension(:,:,:), allocatable :: pdog
   real, allocatable   :: fd(:,:)                   ! flux across layers [kg m-2]
   real, allocatable   :: DC(:)                     ! scavenge change in mass mixing ratio

   real                :: effRemoval

   integer, parameter  :: nbins = 1                  ! work with a single bin tracer

   real, parameter     :: grav = g_E 

   !  Rain parameters from Liu et al.
   real, parameter     :: B0_ls = 1.0e-4
   real, parameter     :: F0_ls = 1.0
   real, parameter     :: XL_ls = 5.0e-4
   real, parameter     :: B0_cv = 1.5e-3
   real, parameter     :: F0_cv = 0.3
   real, parameter     :: XL_cv = 2.0e-3

!  Duration of rain: ls = model timestep, cv = 1800 s (<= cdt)
   real                :: Td_ls
   real, parameter     :: Td_cv = 1800.0


   i1 = lbound(qa, 1); i2 = ubound(qa, 1)
   j1 = lbound(qa, 2); j2 = ubound(qa, 2)
   k1 = lbound(qa, 3); km = ubound(qa, 3)
   
   effRemoval = f_wet

   rc = 0

!  Initialize local variables
!  --------------------------
   Td_ls = dt

   flux = 0.0

!  Allocate the dynamic arrays
   allocate(fd(km,nbins), __STAT__)
   allocate(dc(nbins),    __STAT__)

   allocate(qls(km),      __STAT__)         
   allocate(qcv(km),      __STAT__)

   allocate(pdog(i1:i2,j1:j2,km), __STAT__)

!  Accumulate the 3-dimensional arrays of rhoa and pdog
   pdog = delp / grav

!  Loop over spatial indices
   do j = j1, j2
    do i = i1, i2

!    Check for total precipitation amount
!    Assume no precip in column if precl+precc = 0
     pac = precl(i,j) + precc(i,j)
     if(pac .le. 0.) goto 100
     pls = precl(i,j)
     pcv = precc(i,j)

!    Initialize the precipitation fields
     qls(:)  = 0.
     qcv(:)  = 0.
     fd(:,:) = 0.

!    Find the highest model layer experiencing rainout.  Assumes no
!    scavenging if T < 258 K
     LH = 0
     do k = 1, km
      if(dqcond(i,j,k) .lt. 0. .and. temperature(i,j,k) .gt. 258.) then
       LH = k
       goto 15
      endif
     end do
 15  continue
     if(LH .lt. 1) goto 100

!    convert dqcond from kg water/kg air/s to kg water/m3/s and reverse
!    sign so that dqcond < 0. (positive precip) means qls and qcv > 0.
     do k = LH, km
      qls(k) = -dqcond(i,j,k)*pls/pac*rhoa(i,j,k)
!     qcv(k) = -dqcond(i,j,k)*pcv/pac*rhoa(i,j,k)
     end do

!    Loop over vertical to do the scavenging!
     do k = LH, km

!-----------------------------------------------------------------------------
!   (1) LARGE-SCALE RAINOUT:             
!       Tracer loss by rainout = TC0 * F * exp(-B*dt)
!         where B = precipitation frequency,
!               F = fraction of grid box covered by precipitating clouds.
!       We assume that tracer scavenged by rain is falling down to the
!       next level, where a fraction could be re-evaporated to gas phase
!       if Qls is less then 0 in that level.
!-----------------------------------------------------------------------------
      if (qls(k) .gt. 0.) then
       F  = F0_ls / (1. + F0_ls*B0_ls*XL_ls/(qls(k)*dt/Td_ls))
       B  = B0_ls/F0_ls +1./(F0_ls*XL_ls/qls(k)) 
       BT = B * Td_ls
       if (BT.gt.10.) BT = 10.               !< Avoid overflow >
!      Adjust du level:
       do n = 1, nbins
        effRemoval = f_wet
        DC(n) = qa(i,j,k) * F * effRemoval *(1.-exp(-BT))
        if (DC(n).lt.0.) DC(n) = 0.
        qa(i,j,k) = qa(i,j,k)-DC(n)
        if (qa(i,j,k) .lt. 1.0E-32) qa(i,j,k) = 1.0E-32
       end do
!      Flux down:  unit is kg m-2
!      Formulated in terms of production in the layer.  In the revaporation step
!      we consider possibly adding flux from above...
       do n = 1, nbins
        Fd(k,n) = DC(n)*pdog(i,j,k)
       end do

      end if                                    ! if Qls > 0  >>>

!-----------------------------------------------------------------------------
! * (2) LARGE-SCALE WASHOUT:
! *     Occurs when rain at this level is less than above.
!-----------------------------------------------------------------------------
      if(k .gt. LH .and. qls(k) .ge. 0.) then
       if(qls(k) .lt. qls(k-1)) then
!       Find a maximum F overhead until the level where Qls<0.
        Qmx   = 0.
        do kk = k-1,LH,-1
         if (Qls(kk).gt.0.) then
          Qmx = max(Qmx,Qls(kk))
         else
          goto 333
         end if
        end do

 333    continue
        F = F0_ls / (1. + F0_ls*B0_ls*XL_ls/(Qmx*dt/Td_ls))
        if (F.lt.0.01) F = 0.01
!-----------------------------------------------------------------------------
!  The following is to convert Q(k) from kgH2O/m3/sec to mm/sec in order
!  to use the Harvard formula.  Convert back to mixing ratio by multiplying
!  by rhoa.  Multiply by pdog gives kg/m2/s of precip.  Divide by density
!  of water (=1000 kg/m3) gives m/s of precip and multiply by 1000 gives
!  units of mm/s (omit the multiply and divide by 1000).
!-----------------------------------------------------------------------------

        Qd = Qmx /rhoa(i,j,k)*pdog(i,j,k)
        if (Qd.ge.50.) then
         B = 0.
        else
         B = Qd * 0.1
        end if
        BT = B * dt
        if (BT.gt.10.) BT = 10.

!       Adjust du level:
        do n = 1, nbins
         DC(n) = qa(i,j,k) * F * (1.-exp(-BT))
         if (DC(n).lt.0.) DC(n) = 0.
         qa(i,j,k) = qa(i,j,k)-DC(n)
         if (qa(i,j,k) .lt. 1.0E-32) & 
          qa(i,j,k) = 1.0E-32
         if( associated(flux) ) then
          flux(i,j) = flux(i,j)+DC(n)*pdog(i,j,k)/dt
         endif
        end do

       end if
      end if                                    ! if ls washout  >>>

!-----------------------------------------------------------------------------
!  (3) CONVECTIVE RAINOUT:
!      Tracer loss by rainout = dd0 * F * exp(-B*dt)
!        where B = precipitation frequency,
!              F = fraction of grid box covered by precipitating clouds.
!-----------------------------------------------------------------------------

      if (qcv(k) .gt. 0.) then
       F  = F0_cv / (1. + F0_cv*B0_cv*XL_cv/(Qcv(k)*dt/Td_cv))
       B  = B0_cv
       BT = B * Td_cv
       if (BT.gt.10.) BT = 10.               !< Avoid overflow >

!      Adjust du level: 
       do n = 1, nbins
        effRemoval = f_wet
        DC(n) = qa(i,j,k) * F * effRemoval * (1.-exp(-BT))
        if (DC(n).lt.0.) DC(n) = 0.
        qa(i,j,k) = qa(i,j,k)-DC(n)
        if (qa(i,j,k) .lt. 1.0E-32) qa(i,j,k) = 1.0E-32
       end do

!------  Flux down:  unit is kg. Including both ls and cv.
       do n = 1, nbins
        Fd(k,n) = Fd(k,n) + DC(n)*pdog(i,j,k)
       end do

      end if                                  ! if Qcv > 0   >>>

!-----------------------------------------------------------------------------
!  (4) CONVECTIVE WASHOUT:
!      Occurs when rain at this level is less than above.
!-----------------------------------------------------------------------------

      if (k.gt.LH .and. Qcv(k).ge.0.) then
       if (Qcv(k).lt.Qcv(k-1)) then
!-----  Find a maximum F overhead until the level where Qls<0.
        Qmx   = 0.
        do kk = k-1, LH, -1
         if (Qcv(kk).gt.0.) then
          Qmx = max(Qmx,Qcv(kk))
         else
          goto 444
         end if
        end do

 444    continue
        F = F0_cv / (1. + F0_cv*B0_cv*XL_cv/(Qmx*dt/Td_cv))
        if (F.lt.0.01) F = 0.01
!-----------------------------------------------------------------------------
!  The following is to convert Q(k) from kgH2O/m3/sec to mm/sec in order
!  to use the Harvard formula.  Convert back to mixing ratio by multiplying
!  by rhoa.  Multiply by pdog gives kg/m2/s of precip.  Divide by density
!  of water (=1000 kg/m3) gives m/s of precip and multiply by 1000 gives
!  units of mm/s (omit the multiply and divide by 1000).
!-----------------------------------------------------------------------------

        Qd = Qmx / rhoa(i,j,k)*pdog(i,j,k)
        if (Qd.ge.50.) then
         B = 0.
        else
         B = Qd * 0.1
        end if
        BT = B * dt
        if (BT.gt.10.) BT = 10.

!       Adjust du level:
        do n = 1, nbins
         DC(n) = qa(i,j,k) * F * (1.-exp(-BT))
         if (DC(n).lt.0.) DC(n) = 0.
         qa(i,j,k) = qa(i,j,k)-DC(n)
         if (qa(i,j,k) .lt. 1.0E-32) & 
          qa(i,j,k) = 1.0E-32
         if( associated(flux) ) then
          flux(i,j) = flux(i,j)+DC(n)*pdog(i,j,k)/dt
         endif
        end do

       end if
      end if                                    ! if cv washout  >>>

!-----------------------------------------------------------------------------
!  (5) RE-EVAPORATION.  Assume that SO2 is re-evaporated as SO4 since it
!      has been oxidized by H2O2 at the level above. 
!-----------------------------------------------------------------------------
!     Add in the flux from above, which will be subtracted if reevaporation occurs
      if(k .gt. LH) then
       do n = 1, nbins
        Fd(k,n) = Fd(k,n) + Fd(k-1,n)
       end do

!      Is there evaporation in the currect layer?
       if (-dqcond(i,j,k) .lt. 0.) then
!       Fraction evaporated = H2O(k)evap / H2O(next condensation level).
        if (-dqcond(i,j,k-1) .gt. 0.) then

          A =  abs(  dqcond(i,j,k) * pdog(i,j,k)    &
            /      ( dqcond(i,j,k-1) * pdog(i,j,k-1))  )
          if (A .gt. 1.) A = 1.

!         Adjust tracer in the level
          do n = 1, nbins
           DC(n) =  Fd(k-1,n) / pdog(i,j,k) * A
           qa(i,j,k) = qa(i,j,k) + DC(n)
           qa(i,j,k) = max(qa(i,j,k),1.e-32)
!          Adjust the flux out of the bottom of the layer
           Fd(k,n)  = Fd(k,n) - DC(n)*pdog(i,j,k)
          end do

        endif
       endif                                   ! if -moistq < 0
      endif
     end do  ! k

     do n = 1, nbins
      if( associated(flux) ) then
       flux = flux + Fd(km,n)/dt
      endif
     end do

 100 continue
    end do   ! i
   end do    ! j

   deallocate(fd,   __STAT__)
   deallocate(DC,   __STAT__)
   deallocate(qls,  __STAT__)         
   deallocate(qcv,  __STAT__)
   deallocate(pdog, __STAT__)

   end subroutine MAML_WetRemovalAerosol


   end module MAML_WetRemovalMod

