#include "MAPL_Generic.h"

!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 910.1     !
!-------------------------------------------------------------------------
!BOP
!
! !MODULE: MAMchem_GridCompMod - Implements MAM Chemistry
!
! !INTERFACE:
!
   module MAMchem_GridCompMod
!
! !USES:
!
   use ESMF
   use MAPL_Mod

   use MAPL_SimpleBundleMod

   use Chem_UtilMod,         only: Chem_UtilResVal

   use MAM3_DataMod
   use MAM7_DataMod

   use MAM_BaseMod

   use MAM_SizeMod,          only: MAM_DrySize, MAM_WetSize

   use MAM_SeasaltMod,       only: MAM_SS_Emission, MAM_SS_Diagnostics
   use MAM_DustMod,          only: MAM_DU_Emission, MAM_DU_Diagnostics
   use MAM_BlackCarbonMod,   only: MAM_BC_Emission
   use MAM_OrganicCarbonMod, only: MAM_OC_Emission

   use MAM_NucleationMod,    only: MAM_Nucleation
   use MAM_CoagulationMod,   only: MAM_Coagulation, MAM_CoagulationBimodal
   use MAM_DryRemovalMod,    only: MAM_DryRemoval

   use MAM_WetRemovalMod,    only: MAM_WetRemoval

   use MAML_OpticsTableMod,  only: MAML_OpticsTable,        &
                                   MAML_OpticsTableCreate,  &
                                   MAML_OpticsTableDestroy, &
                                   MAML_OpticsTableRead
   use MAML_OpticsMod,       only: MAML_OpticsInterpolate



#if (0)
   use CAM_BaseMod,          only: CAM_Initialize, CAM_CalculateSize
#endif

   implicit none
   private


   type(MAML_OpticsTable), save :: MAM7_MieTable(7)

   integer, parameter :: instanceComputational = 1
   integer, parameter :: instanceData          = 2


!
! !PUBLIC MEMBER FUNCTIONS:

   public SetServices
!
! !DESCRIPTION: 
!
!  {\tt MAMchem\_GridComp} is an ESMF gridded component implementing
!  the MAM aerosol microphysical processes.
!
!  Developed for GEOS-5 release Fortuna 2.0 and later.
!
! !REVISION HISTORY:
!
!  06Dec2009  da Silva     Created the MATRIX skeleton.
!  15Aug2011  A. Darmenov  Initial version of MAM
!
!EOP
!-------------------------------------------------------------------------

!  Legacy state
!  ------------
   type MAM_State
      private

      logical                     :: data_driven = .false.

      type(ESMF_Config)           :: CF           ! Private Config

      type(ESMF_Grid)             :: grid         ! Grid

      integer                     :: im_world     ! Horizontal dimensions - lon
      integer                     :: jm_world     ! Horizontal dimensions - lat

      type(MAPL_SimpleBundle)     :: qa           ! Interstitial aerosol species and absorbed water
      type(MAPL_SimpleBundle)     :: qc           ! Cloud-borne  aerosol species
      type(MAPL_SimpleBundle)     :: qg           ! Gas species 

      type(MAPL_SimpleBundle)     :: Da           ! Dry and 'wet' geometric mean diameter of interstitial aerosol number size distribution

      real                        :: dt           ! Model time step

      integer                     :: scheme_id    ! MAM7 or MAM3
      type(MAM_Scheme)            :: scheme       ! MAM scheme/configuration

      real                        :: femisSS      ! Seasalt emission tuning parameter
      real                        :: femisDU      ! Dust emission tuning parameter

      real                        :: pom_oc_ratio ! ratio of POM emissions to primary OC emissions

      logical                     :: dry_removal  ! turn on/off dry removal processes
      logical                     :: wet_removal  ! turn on/off wet removal processes
      logical                     :: nucleation   ! turn on/off nucleation process
      logical                     :: condensation ! turn on/off condensation process
      logical                     :: coagulation  ! turn on/off coagulation process

      real                        :: f_conv_scav_ait
      real                        :: f_conv_scav_acc
      real                        :: f_conv_scav_pcm
      real                        :: f_conv_scav_fss
      real                        :: f_conv_scav_css
      real                        :: f_conv_scav_fdu
      real                        :: f_conv_scav_cdu

      real                        :: f_wet_ait
      real                        :: f_wet_acc
      real                        :: f_wet_pcm
      real                        :: f_wet_fss
      real                        :: f_wet_css
      real                        :: f_wet_fdu
      real                        :: f_wet_cdu

      type(MAML_OpticsTable)      :: mie_ait
      type(MAML_OpticsTable)      :: mie_acc
      type(MAML_OpticsTable)      :: mie_pcm
      type(MAML_OpticsTable)      :: mie_fss
      type(MAML_OpticsTable)      :: mie_css
      type(MAML_OpticsTable)      :: mie_fdu
      type(MAML_OpticsTable)      :: mie_cdu

      logical                     :: verbose      ! verbosity flag
   end type MAM_State

!  Hook for the ESMF
!  -----------------
   type MAM_Wrap
      type (MAM_State), pointer :: PTR => null()
   end type MAM_Wrap


contains


!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 910.1     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE: SetServices --- Sets IRF services for the MAMchem Grid Component
!
! !INTERFACE:

   subroutine SetServices ( GC, RC )

! !ARGUMENTS:

    type(ESMF_GridComp), intent(INOUT) :: GC  ! gridded component
    integer, optional                  :: RC  ! return code

! !DESCRIPTION: Sets Initialize, Run and Finalize services. 
!
! !REVISION HISTORY:
!
!  1Dec2009  da Silva  First crack.
!
!EOP
!-------------------------------------------------------------------------

                            __Iam__('SetServices')

!   Local derived type aliases
!   --------------------------
    type (MAM_State), pointer  :: self   ! internal private state
    type (MAM_wrap)            :: wrap

    character(len=ESMF_MAXSTR) :: comp_name

!   Local variables
!   --------------------------
    character(len=ESMF_MAXSTR) :: scheme            ! name of MAM scheme/configuration

    character(len=MAM_MAXSTR)  :: field_name        ! field name
    character(len=MAM_MAXSTR)  :: field_long_name   ! field name
    character(len=MAM_MAXSTR)  :: mode_name         ! aerosol mode name
    character(len=MAM_MAXSTR)  :: mode_long_name    ! aerosol mode name
    character(len=MAM_MAXSTR)  :: species_name      ! aerosol species name
    character(len=MAM_MAXSTR)  :: attachment_state  ! attachment state of aerosols
    integer                    :: n_species         ! number of aerosol species
    integer                    :: m, s              ! mode and species indexes

!   local
!   -----
    type(ESMF_Config)          :: CF                ! Universal Config 
    character(len=1024)        :: mie_optics_file   ! MAM Mie optics table file


!   Get my name and set-up traceback handle
!   ---------------------------------------
    call ESMF_GridCompGet(GC, name=comp_name, __RC__)
    Iam = trim(comp_name) // '::' // trim(Iam)

!   Wrap internal state for storing in GC; rename legacyState
!   -------------------------------------
    allocate ( self, stat=STATUS )
    VERIFY_(STATUS)
    wrap%ptr => self


!   Is the component data driven
!   ----------------------------
    self%data_driven = isDataDrivenGC_(GC, __RC__)


!   Load private Config Attributes
!   ------------------------------
    self%CF = ESMF_ConfigCreate(__RC__)
    call ESMF_ConfigLoadFile ( self%CF, 'MAMchem_GridComp.rc', __RC__ )

    call ESMF_ConfigGetAttribute ( self%CF, self%verbose,      Label='verbose:',       default=.false.,  __RC__ )

    call ESMF_ConfigGetAttribute ( self%CF, scheme,            Label='scheme:',        default='MAM7' ,  __RC__ )

    call ESMF_ConfigGetAttribute ( self%CF, self%dry_removal,  Label='dry_removal:',   default=.true.,   __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%wet_removal,  Label='wet_removal:',   default=.true.,   __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%nucleation,   Label='nucleation:',    default=.true.,   __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%condensation, Label='condensation:',  default=.true.,   __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%coagulation,  Label='coagulation:',   default=.true.,   __RC__ )
    
!   call ESMF_ConfigGetAttribute ( self%CF, self%femisSS,      Label='seasalt_femis:', default=1.0,      __RC__ )
!   call ESMF_ConfigGetAttribute ( self%CF, self%femisDU,      Label='dust_femis:',    default=1.0,      __RC__ )

    call ESMF_ConfigGetAttribute ( self%CF, self%pom_oc_ratio, Label='pom_oc_ratio:',  default=1.4,      __RC__ )

    call ESMF_ConfigGetAttribute ( self%CF, self%f_conv_scav_ait, Label='f_conv_scav_ait:', default=0.0, __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%f_conv_scav_acc, Label='f_conv_scav_acc:', default=0.0, __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%f_conv_scav_pcm, Label='f_conv_scav_pcm:', default=0.0, __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%f_conv_scav_fss, Label='f_conv_scav_fss:', default=0.0, __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%f_conv_scav_css, Label='f_conv_scav_css:', default=0.0, __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%f_conv_scav_fdu, Label='f_conv_scav_fdu:', default=0.0, __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%f_conv_scav_cdu, Label='f_conv_scav_cdu:', default=0.0, __RC__ )

    call ESMF_ConfigGetAttribute ( self%CF, self%f_wet_ait, Label='f_wet_ait:', default=0.0, __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%f_wet_acc, Label='f_wet_acc:', default=0.0, __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%f_wet_pcm, Label='f_wet_pcm:', default=0.0, __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%f_wet_fss, Label='f_wet_fss:', default=0.0, __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%f_wet_css, Label='f_wet_css:', default=0.0, __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%f_wet_fdu, Label='f_wet_fdu:', default=0.0, __RC__ )
    call ESMF_ConfigGetAttribute ( self%CF, self%f_wet_cdu, Label='f_wet_cdu:', default=0.0, __RC__ )


!   Set the MAM model scheme
!   -------------------------
    call ESMF_StringUpperCase(scheme, __RC__)

    select case (scheme)
        case ('MAM7')
        self%scheme_id = MAM7_SCHEME

        case default
        __raise__ (MAM_UNKNOWN_SCHEME_ERROR, "Unsupported MAM scheme: " // trim(scheme))
    end select



!   Set the profiling timers
!   ------------------------
    call MAPL_TimerAdd(GC, name='RUN',                         __RC__)
    call MAPL_TimerAdd(GC, name='-EMISSIONS',                  __RC__)
    call MAPL_TimerAdd(GC, name='-MICROPHYSICS',               __RC__)
    call MAPL_TimerAdd(GC, name='--MICROPHYSICS_CONDENSATION', __RC__)
    call MAPL_TimerAdd(GC, name='--MICROPHYSICS_NUCLEATION',   __RC__)
    call MAPL_TimerAdd(GC, name='--MICROPHYSICS_COAGULATION',  __RC__)
    call MAPL_TimerAdd(GC, name='-REMOVAL',                    __RC__)
    call MAPL_TimerAdd(GC, name='--REMOVAL_DRY',               __RC__)
    call MAPL_TimerAdd(GC, name='--REMOVAL_WET',               __RC__)
    call MAPL_TimerAdd(GC, name='-HYGROSCOPIC_GROWTH',         __RC__)
    call MAPL_TimerAdd(GC, name='-DIAGNOSTICS',                __RC__)
    call MAPL_TimerAdd(GC, name='--DIAGNOSTICS_SEASALT',       __RC__)
    call MAPL_TimerAdd(GC, name='--DIAGNOSTICS_DUST',          __RC__)
    call MAPL_TimerAdd(GC, name='--DIAGNOSTICS_CIM',           __RC__)
    call MAPL_TimerAdd(GC, name='--DIAGNOSTICS_AOT',           __RC__)


!                       ------------------------
!                       ESMF Functional Services
!                       ------------------------

!   Set the Initialize, Run, Finalize entry points
!   ----------------------------------------------
    call MAPL_GridCompSetEntryPoint ( GC, ESMF_METHOD_INITIALIZE, Initialize_, __RC__ )
    call MAPL_GridCompSetEntryPoint ( GC, ESMF_METHOD_RUN,        Run_,        __RC__ )
    call MAPL_GridCompSetEntryPoint ( GC, ESMF_METHOD_FINALIZE,   Finalize_,   __RC__ )
        
!   Store internal state in GC
!   --------------------------
    call ESMF_UserCompSetInternalState ( GC, 'MAM_state', wrap, STATUS )
    VERIFY_(STATUS)
  
!                         ------------------
!                         MAPL Data Services
!                         ------------------

!BOP
!
! !IMPORT STATE:

#include "MAMchem_ImportSpec___.h"

! !INTERNAL STATE:
#define __MAMchem_InternalSpec_H__ 0
#if (__MAMchem_InternalSpec_H__)
#include "MAMchem_InternalSpec___.h"
#else
#define DECLARE_INTERNAL_STATE       type(ESMF_State) :: INTERNAL
#endif
! !EXTERNAL STATE:

#include "MAMchem_ExportSpec___.h"


!   Set MAM infrastructure
!   -----------------------------
    call MAM_SchemeInit(self%scheme, self%scheme_id, __RC__)


!   Add interstitial aerosols to the internal state
!   -----------------------------------------------
    attachment_state = 'interstitial'

    do m = 1, self%scheme%n_modes
        call MAM_AerosolModeGet(self%scheme%mode(m), name=mode_name, long_name=mode_long_name, n_species=n_species)

        ! number mixing ratio
        field_name = 'NUM_A_' // trim(mode_name)
        field_long_name = 'number of ' // trim(attachment_state) // ' aerosol particles in ' // trim(mode_long_name) // ' mode'

        call MAPL_AddInternalSpec(GC, SHORT_NAME = trim(field_name),      &
                                      LONG_NAME  = trim(field_long_name), &
                                      UNITS      = '#/kg',                &
                                      DIMS       = MAPL_DimsHorzVert,     &
                                      VLOCATION  = MAPL_VLocationCenter,  &
                                      FRIENDLYTO = 'DYNAMICS:TURBULENCE:MOIST', __RC__)

        ! mass mixing ratios
        do s = 1, n_species
            species_name = self%scheme%mode(m)%species(s)%name

            field_name = trim(species_name) // '_A_' // trim(mode_name)
            field_long_name = 'mass mixing ratio of ' // trim(attachment_state) // ' aerosol particles in ' // trim(mode_long_name) // ' mode'

            call MAPL_AddInternalSpec(GC, SHORT_NAME = trim(field_name),      &
                                          LONG_NAME  = trim(field_long_name), &
                                          UNITS      = 'kg kg-1',             &
                                          DIMS       = MAPL_DimsHorzVert,     &
                                          VLOCATION  = MAPL_VLocationCenter,  &
                                          FRIENDLYTO = 'DYNAMICS:TURBULENCE:MOIST', __RC__)
        end do

        ! absorbed water
        field_name = 'WTR_A_' // trim(mode_name)
        field_long_name = 'mass mixing ratio of absorbed water by ' // trim(attachment_state) // ' aerosol particles in ' // trim(mode_long_name) // ' mode'

        call MAPL_AddInternalSpec(GC, SHORT_NAME = trim(field_name),      &
                                      LONG_NAME  = trim(field_long_name), &
                                      UNITS      = 'kg kg-1',             &
                                      DIMS       = MAPL_DimsHorzVert,     &
                                      VLOCATION  = MAPL_VLocationCenter,  &
                                      ADD2EXPORT = .true.,                &
                                      RESTART    = .false.,         __RC__)

        ! dry size
        field_name = 'DGN_DRY_' // trim(mode_name)
        field_long_name = 'dry diameter of ' // trim(attachment_state) // ' aerosol particles in ' // trim(mode_long_name) // ' mode'

        call MAPL_AddInternalSpec(GC, SHORT_NAME = trim(field_name),      &
                                      LONG_NAME  = trim(field_long_name), &
                                      UNITS      = 'm',                   &
                                      DIMS       = MAPL_DimsHorzVert,     &
                                      VLOCATION  = MAPL_VLocationCenter,  &
                                      ADD2EXPORT = .true.,                &
                                      RESTART    = .false.,         __RC__)

        ! wet size
        field_name = 'DGN_WET_' // trim(mode_name)
        field_long_name = 'wet diameter of ' // trim(attachment_state) // ' aerosol particles in ' // trim(mode_long_name) // ' mode'

        call MAPL_AddInternalSpec(GC, SHORT_NAME = trim(field_name),      &
                                      LONG_NAME  = trim(field_long_name), &
                                      UNITS      = 'm',                   &
                                      DIMS       = MAPL_DimsHorzVert,     &
                                      VLOCATION  = MAPL_VLocationCenter,  &
                                      ADD2EXPORT = .true.,                &
                                      RESTART    = .false.,         __RC__)
    end do


!   Add cloud-borne aerosols to the internal state
!   ----------------------------------------------
    attachment_state = 'cloud-borne'

    do m = 1, self%scheme%n_modes
        call MAM_AerosolModeGet(self%scheme%mode(m), name=mode_name, long_name=mode_long_name, n_species=n_species)

        ! number mixing ratio
        field_name = 'NUM_C_' // trim(mode_name)
        field_long_name = 'number of ' // trim(attachment_state) // ' aerosol particles in ' // trim(mode_long_name) // ' mode'

        call MAPL_AddInternalSpec(GC, SHORT_NAME = trim(field_name),      &
                                      LONG_NAME  = trim(field_long_name), &
                                      UNITS      = '#/kg',                &
                                      DIMS       = MAPL_DimsHorzVert,     &
                                      VLOCATION  = MAPL_VLocationCenter,  &
                                      FRIENDLYTO = 'MOIST', __RC__)

        ! mass mixing ratios
        do s = 1, n_species
            species_name = self%scheme%mode(m)%species(s)%name

            field_name  = trim(species_name) // '_C_' // trim(mode_name)
            field_long_name = 'mass mixing ratio of ' // trim(attachment_state) // ' aerosol particles in ' // trim(mode_long_name) // ' mode'

            call MAPL_AddInternalSpec(GC, SHORT_NAME = trim(field_name),      &
                                          LONG_NAME  = trim(field_long_name), &
                                          UNITS      = 'kg kg-1',             &
                                          DIMS       = MAPL_DimsHorzVert,     &
                                          VLOCATION  = MAPL_VLocationCenter,  &
                                          FRIENDLYTO = 'MOIST', __RC__)
        end do
    end do

#if (0)
!   Add dry and wet aerosol sizes to the internal state
!   ---------------------------------------------------
    attachment_state = 'interstitial'

    do m = 1, self%scheme%n_modes
        call MAM_AerosolModeGet(self%scheme%mode(m), name=mode_name, long_name=mode_long_name, n_species=n_species)
      
        field_name = 'DGN_DRY_' // trim(mode_name)
        field_long_name = 'dry geometric mean diameter of ' // trim(attachment_state) // ' aerosol particles in ' // trim(mode_long_name) // ' mode'

        call MAPL_AddInternalSpec(GC, SHORT_NAME = trim(field_name),      &
                                      LONG_NAME  = trim(field_long_name), &
                                      UNITS      = 'm',                   &
                                      DIMS       = MAPL_DimsHorzVert,     &
                                      VLOCATION  = MAPL_VLocationCenter,  &
                                      RESTART    = .FALSE.,               &
                                      FRIENDLYTO = trim(COMP_NAME), __RC__)

        
        field_name = 'DGN_WET_' // trim(mode_name)
        field_long_name = 'wet size of ' // trim(attachment_state) // ' aerosol particles in ' // trim(mode_long_name) // ' mode'

        call MAPL_AddInternalSpec(GC, SHORT_NAME = trim(field_name),      &
                                      LONG_NAME  = trim(field_long_name), &
                                      UNITS      = 'm',                   &
                                      DIMS       = MAPL_DimsHorzVert,     &
                                      VLOCATION  = MAPL_VLocationCenter,  &
                                      RESTART    = .FALSE.,               &
                                      FRIENDLYTO = trim(COMP_NAME), __RC__)
    end do
#endif

!   This state is needed by radiation - It will contain 
!   aerosol number and mass mixing ratios, and aerosol optics
!   ---------------------------------------------------------------
    call MAPL_AddExportSpec(GC,                                   &
        SHORT_NAME         = 'AERO',                              &
        LONG_NAME          = 'aerosol_mixing_ratios',             &
        UNITS              = 'kg kg-1',                           &
        DIMS               = MAPL_DimsHorzVert,                   &
        VLOCATION          = MAPL_VLocationCenter,                &
        DATATYPE           = MAPL_StateItem,                __RC__)


!   This state is needed by MOIST - It will contain
!   aerosol number concentrations and aerosol activation properties
!   ---------------------------------------------------------------
    call MAPL_AddExportSpec(GC,                                   &
        SHORT_NAME         = 'AERO_ACI',                          &
        LONG_NAME          = 'aerosol_cloud_interaction',         &
        UNITS              = 'kg kg-1',                           &
        DIMS               = MAPL_DimsHorzVert,                   &
        VLOCATION          = MAPL_VLocationCenter,                &
        DATATYPE           = MAPL_StateItem,                      &
                                                       RC=STATUS  )
    VERIFY_(STATUS)


    
!   Create Mie tables for coupling with radiation
!   ---------------------------------------------
    call ESMF_GridCompGet(GC, config=CF, __RC__) ! read paths to Mie tables from global config

    call ESMF_ConfigGetAttribute(CF, mie_optics_file, label='MAM7_AIT_OPTICS:', __RC__)
    MAM7_MieTable(1) = MAML_OpticsTableCreate(trim(mie_optics_file), __RC__)
    call MAML_OpticsTableRead(MAM7_MieTable(1), __RC__)

    call ESMF_ConfigGetAttribute(CF, mie_optics_file, label='MAM7_ACC_OPTICS:', __RC__)
    MAM7_MieTable(2) = MAML_OpticsTableCreate(trim(mie_optics_file), __RC__)
    call MAML_OpticsTableRead(MAM7_MieTable(2), __RC__)

    call ESMF_ConfigGetAttribute(CF, mie_optics_file, label='MAM7_PCM_OPTICS:', __RC__)
    MAM7_MieTable(3) = MAML_OpticsTableCreate(trim(mie_optics_file), __RC__)
    call MAML_OpticsTableRead(MAM7_MieTable(3), __RC__)

    call ESMF_ConfigGetAttribute(CF, mie_optics_file, label='MAM7_FSS_OPTICS:', __RC__)
    MAM7_MieTable(4) = MAML_OpticsTableCreate(trim(mie_optics_file), __RC__)
    call MAML_OpticsTableRead(MAM7_MieTable(4), __RC__)

    call ESMF_ConfigGetAttribute(CF, mie_optics_file, label='MAM7_CSS_OPTICS:', __RC__)
    MAM7_MieTable(5) = MAML_OpticsTableCreate(trim(mie_optics_file), __RC__)
    call MAML_OpticsTableRead(MAM7_MieTable(5), __RC__)

    call ESMF_ConfigGetAttribute(CF, mie_optics_file, label='MAM7_FDU_OPTICS:', __RC__)
    MAM7_MieTable(6) = MAML_OpticsTableCreate(trim(mie_optics_file), __RC__)
    call MAML_OpticsTableRead(MAM7_MieTable(6), __RC__)

    call ESMF_ConfigGetAttribute(CF, mie_optics_file, label='MAM7_CDU_OPTICS:', __RC__)
    MAM7_MieTable(7) = MAML_OpticsTableCreate(trim(mie_optics_file), __RC__)
    call MAML_OpticsTableRead(MAM7_MieTable(7), __RC__)


!   This bundle is not filled in by MAM, just a place holder for now
!   ----------------------------------------------------------------
    call MAPL_AddExportSpec(GC, SHORT_NAME = 'AERO_DP',              &
                                LONG_NAME  = 'aerosol_deposition',   &
                                UNITS      = 'kg m-2 s-1',           &
                                DIMS       = MAPL_DimsHorzOnly,      &
                                DATATYPE   = MAPL_BundleItem,  __RC__)

!   Generic Set Services
!   --------------------
    call MAPL_GenericSetServices ( GC, __RC__ )

!   All done
!   --------

    RETURN_(ESMF_SUCCESS)

  end subroutine SetServices


!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  Initialize_ --- Initialize MAMchem
!
! !INTERFACE:
!

   subroutine Initialize_ ( GC, IMPORT, EXPORT, CLOCK, rc )

! !USES:

    implicit none

! !INPUT PARAMETERS:

    type(ESMF_Clock),  intent(inout)   :: CLOCK  ! The clock

! !OUTPUT PARAMETERS:

    type(ESMF_GridComp), intent(inout) :: GC     ! Grid Component
    type(ESMF_State), intent(inout)    :: IMPORT ! Import State
    type(ESMF_State), intent(inout)    :: EXPORT ! Export State
    integer, intent(out)               :: rc     ! Error return code:
                                                !  0 - all is well
                                                !  1 - 

! !DESCRIPTION: This is a simple ESMF wrapper.
!
! !REVISION HISTORY:
!
!  01Dec2009 da Silva  First crack.
!
!EOP
!-------------------------------------------------------------------------

                              __Iam__('Initialize_')

    type(MAM_state), pointer      :: self               ! Legacy state
    type(ESMF_Grid)               :: GRID               ! Grid
    type(ESMF_Config)             :: CF                 ! Universal Config 

    type(ESMF_State)              :: aero               ! 
    type(ESMF_FieldBundle)        :: aero_state_aerosols!
    logical                       :: implements_aerosol_optics
    type(ESMF_Field)              :: field              ! field
    character(len=MAM_MAXSTR)     :: field_name         ! field name

    type(ESMF_State)              :: aero_aci           !
    logical                       :: implements_aap_method

    integer                       :: im_World, jm_World ! Global 2D Dimensions
    integer                       :: im, jm, lm         ! 3D Dimensions

    integer, parameter            :: n_hres = 5         ! number of horizontal resolutions (a, b, c, d, e)
    real, dimension(n_hres)       :: f_hres             ! buffer for the resolution dependent factors
    integer                       :: n


    integer                       :: nymd, nhms         ! date, time
    real                          :: cdt                ! time step in secs

    character(len=1024)           :: var_names          ! comma separated names

    character(len=ESMF_MAXSTR)    :: comp_name          ! component's name

    character(len=MAM_MAXSTR)     :: mode_name          ! aerosol mode name
    character(len=MAM_MAXSTR)     :: species_name       ! aerosol species name
    integer                       :: n_species          ! number of aerosol species
    integer                       :: m, s               ! mode and species indexes

    real                          :: hygroscopicity     ! hygroscopicity of the aerosol species 
    real                          :: solubility         ! solubility of the aerosol specie

    real                          :: f_scav             ! globally uniform convective scavenging coefficient, km-1

    character(len=1024)           :: mie_optics_file

    real, parameter               :: f_scav_none = 0.0
    real, parameter               :: f_wet_none  = 0.0

    character(len=2), parameter   :: name_delimiter = ', '


!  Declare pointers to IMPORT/EXPORT/INTERNAL states 
!  -------------------------------------------------
#include "MAMchem_DeclarePointer___.h"
#if !(__MAMchem_InternalSpec_H__)
    type(MAPL_MetaComp), pointer  :: MetaComp
    type(ESMF_State)              :: INTERNAL
#endif
  
!   Get my name and set-up traceback handle
!   ---------------------------------------
    call ESMF_GridCompGet( GC, name=COMP_NAME, __RC__ )
    Iam = trim(comp_name) // '::' // trim(Iam)

!                               --------
    if (MAPL_AM_I_ROOT()) then
       write (*,*) trim(Iam)//': Starting...'
       write (*,*)
    end if


!   Initialize MAPL Generic
!   -----------------------
    call MAPL_GenericInitialize ( GC, IMPORT, EXPORT, clock,  __RC__ )

!   Get pointers to IMPORT/EXPORT/INTERNAL states 
!   ---------------------------------------------
#  include "MAMchem_GetPointer___.h"
#if !(__MAMchem_InternalSpec_H__)
    call MAPL_GetObjectFromGC ( GC, MetaComp, __RC__ )
    call MAPL_Get ( MetaComp, INTERNAL_ESMF_STATE=INTERNAL, __RC__)
#endif

!   Extract relevant runtime information
!   ------------------------------------
    call extract_ ( GC, CLOCK, self, GRID, CF, &
                    im_World, jm_World,        &
                    im, jm, lm,                &
                    nymd, nhms, cdt, __RC__ )

!   Aerosol Processes
!   -----------------
    if (self%verbose .and. MAPL_AM_I_ROOT()) then
       write (*,*) 'Aerosol processes:'
       call PrintProcessFlag('Dry removal ', self%dry_removal)
       call PrintProcessFlag('Wet removal ', self%wet_removal)

       call PrintProcessFlag('Condensation', self%condensation)
       call PrintProcessFlag('Nucleation  ', self%nucleation)
       call PrintProcessFlag('Coagulation ', self%coagulation)
       write (*,*)
    end if


!   Set the grid and dimensions
!   ---------------------------
    self%grid = GRID

    self%im_world = im_World
    self%jm_world = jm_World


!   Set the time step
!   -----------------
    self%dt = cdt


!   Set resolution dependent parameters
!   -----------------------------------
    call ESMF_ConfigFindLabel(self%CF, 'seasalt_femis:', __RC__)
    do n = 1, size(f_hres)
        call ESMF_ConfigGetAttribute(self%CF, f_hres(n), __RC__)
    end do
    self%femisSS = Chem_UtilResVal(self%im_world, self%jm_world, f_hres(:), STATUS)
    VERIFY_(STATUS)

   
    call ESMF_ConfigFindLabel(self%CF, 'dust_femis:', __RC__)
    do n = 1, size(f_hres)
        call ESMF_ConfigGetAttribute(self%CF, f_hres(n), __RC__)
    end do
    self%femisDU = Chem_UtilResVal(self%im_world, self%jm_world, f_hres(:), STATUS)
    VERIFY_(STATUS)

!   Bundle the interstitial and cloud-borne tracers
!   -------------------------------------------------
    var_names = ''
    do m = 1, self%scheme%n_modes
        call MAM_AerosolModeGet(self%scheme%mode(m), name=mode_name, n_species=n_species)

        ! number mixing ratio
        field_name = 'NUM_A_' // trim(mode_name)
        var_names  = trim(var_names) // trim(field_name) // trim(name_delimiter)

        ! absorbed water
        field_name = 'WTR_A_' // trim(mode_name)
        var_names  = trim(var_names) // trim(field_name) // trim(name_delimiter)

        ! mass mixing ratios
        do s = 1, n_species
            species_name = self%scheme%mode(m)%species(s)%name

            field_name  = trim(species_name) // '_A_' // trim(mode_name)
            var_names   = trim(var_names) // trim(field_name) // trim(name_delimiter)
        end do
    end do
    var_names = trim(var_names(1:len_trim(var_names)-1))

    self%qa = MAPL_SimpleBundleCreate(INTERNAL, name='MAM_INTERSTITIAL_AEROSOLS', &
                                                only_vars=var_names, __RC__)
    call MAPL_SimpleBundlePrint(self%qa)


    var_names = ''
    do m = 1, self%scheme%n_modes
        call MAM_AerosolModeGet(self%scheme%mode(m), name=mode_name, n_species=n_species)

        ! number mixing ratio
        field_name = 'NUM_C_' // trim(mode_name)
        var_names  = trim(var_names) // trim(field_name) // trim(name_delimiter)

        ! mass mixing ratios
        do s = 1, n_species
            species_name = self%scheme%mode(m)%species(s)%name

            field_name  = trim(species_name) // '_C_' // trim(mode_name)
            var_names   = trim(var_names) // trim(field_name) // trim(name_delimiter)
        end do
    end do
    var_names = trim(var_names(1:len_trim(var_names)-1))

    self%qc = MAPL_SimpleBundleCreate(INTERNAL, name='MAM_CLOUDBORNE_AEROSOLS', &
                                                only_vars=var_names, __RC__)
    call MAPL_SimpleBundlePrint(self%qc)


!   Bundle the dry and wet modal geometric diameters
!   -------------------------------------------------
    var_names = ''
    do m = 1, self%scheme%n_modes
        call MAM_AerosolModeGet(self%scheme%mode(m), name=mode_name, n_species=n_species)

        field_name = 'DGN_DRY_' // trim(mode_name)
        var_names  = trim(var_names) // trim(field_name) // trim(name_delimiter)

        field_name = 'DGN_WET_' // trim(mode_name)
        var_names  = trim(var_names) // trim(field_name) // trim(name_delimiter)
    end do
    var_names = trim(var_names(1:len_trim(var_names)-1))

    self%Da   = MAPL_SimpleBundleCreate(INTERNAL, name='MAM_INTERSTITIAL_AEROSOLS_DIAMETERS', &
                                                 only_vars=var_names, __RC__)
    call MAPL_SimpleBundlePrint(self%Da)

!   Bundle the gas tracers
!   -------------------------------------------------
    var_names = 'H2SO4, SO2, NH3, SOA_GAS'

    self%qg = MAPL_SimpleBundleCreate(IMPORT, name='MAM_GAS_SPECIES', &
                                              only_vars=var_names, __RC__)
    call MAPL_SimpleBundlePrint(self%qg)

   

!   Register/Initialize CAM chemistry and aerozols
!   ----------------------------------------------
!   lm = 72 ! extract_() returns lm = 0 !?   
!   call CAM_Initialize(levels=lm)  ! operate as a single column model by default


!   Fill the AERO State with the aerosol mixing ratios
!   ---------------------------------------------------
    call ESMF_StateGet(EXPORT, 'AERO', aero, __RC__)

    ! This attribute indicates if the aerosol optics method is implemented or not. 
    ! Radiation will not call the aerosol optics method unless this attribute is 
    ! explicitly set to true.

    implements_aerosol_optics = .true. 

    call ESMF_AttributeSet(aero, name  = 'implements_aerosol_optics_method', & 
                                 value = implements_aerosol_optics, __RC__)
  
    COUPLING_TO_RADIATION: if (implements_aerosol_optics) then

        aero_state_aerosols = ESMF_FieldBundleCreate(name="AEROSOLS", __RC__)
        call MAPL_StateAdd(aero, aero_state_aerosols, __RC__)

        do m = 1, self%scheme%n_modes
            call MAM_AerosolModeGet(self%scheme%mode(m), name=mode_name, n_species=n_species)

            ! interstitial aerosol tracers
            field_name = 'NUM_A_' // trim(mode_name)
            call ESMF_StateGet(INTERNAL, trim(field_name), field, __RC__)
            call ESMF_FieldBundleAdd(aero_state_aerosols, (/field/), __RC__)

            field_name = 'WTR_A_' // trim(mode_name)
            call ESMF_StateGet(INTERNAL, trim(field_name), field, __RC__)
            call ESMF_FieldBundleAdd(aero_state_aerosols, (/field/), __RC__)

            field_name = 'DGN_WET_' // trim(mode_name)
            call ESMF_StateGet(INTERNAL, trim(field_name), field, __RC__)
            call ESMF_FieldBundleAdd(aero_state_aerosols, (/field/), __RC__)

            do s = 1, n_species
                species_name = self%scheme%mode(m)%species(s)%name
                field_name  = trim(species_name) // '_A_' // trim(mode_name)
           
                call ESMF_StateGet(INTERNAL, trim(field_name), field, __RC__)
                call ESMF_FieldBundleAdd(aero_state_aerosols, (/field/), __RC__)
            end do
        end do 


        ! TODO
        ! MAM_MieTable = MAM_MieCreate(CF, __RC__)

        ! state of the atmosphere
        call ESMF_AttributeSet(aero, name='air_pressure_for_aerosol_optics',             value='PLE', __RC__)
        call ESMF_AttributeSet(aero, name='relative_humidity_for_aerosol_optics',        value='RH',  __RC__)
        call ESMF_AttributeSet(aero, name='cloud_area_fraction_for_aerosol_optics',      value='',    __RC__) ! 'cloud_area_fraction_in_atmosphere_layer_for_aerosol_optics'

        ! aerosol optics
        call ESMF_AttributeSet(aero, name='band_for_aerosol_optics',                     value=0,     __RC__)
        call ESMF_AttributeSet(aero, name='extinction_in_air_due_to_ambient_aerosol',    value='EXT', __RC__)
        call ESMF_AttributeSet(aero, name='single_scattering_albedo_of_ambient_aerosol', value='SSA', __RC__)
        call ESMF_AttributeSet(aero, name='asymmetry_parameter_of_ambient_aerosol',      value='ASY', __RC__)

        ! add PLE to Aero state
        call ESMF_AttributeGet(aero, name='air_pressure_for_aerosol_optics', value=field_name, __RC__)
        if (field_name /= '') then
            field = MAPL_FieldCreateEmpty(trim(field_name), self%grid, __RC__)

            call MAPL_FieldAllocCommit(field, dims=MAPL_DimsHorzVert, location=MAPL_VLocationEdge, typekind=MAPL_R4, hw=0, __RC__)
            call MAPL_StateAdd(aero, field, __RC__)
        end if

        ! add RH to Aero state
        call ESMF_AttributeGet(aero, name='relative_humidity_for_aerosol_optics', value=field_name, __RC__)
        if (field_name /= '') then
            field = MAPL_FieldCreateEmpty(trim(field_name), self%grid, __RC__)

            call MAPL_FieldAllocCommit(field, dims=MAPL_DimsHorzVert, location=MAPL_VLocationCenter, typekind=MAPL_R4, hw=0, __RC__)
            call MAPL_StateAdd(aero, field, __RC__)
        end if

        ! add EXT to Aero state
        call ESMF_AttributeGet(aero, name='extinction_in_air_due_to_ambient_aerosol', value=field_name, __RC__)
        if (field_name /= '') then 
            field = MAPL_FieldCreateEmpty(trim(field_name), self%grid, __RC__)

            call MAPL_FieldAllocCommit(field, dims=MAPL_DimsHorzVert, location=MAPL_VLocationCenter, typekind=MAPL_R4, hw=0, __RC__)
            call MAPL_StateAdd(aero, field, __RC__)
        end if

        ! add SSA to aero state
        call ESMF_AttributeGet(aero, name='single_scattering_albedo_of_ambient_aerosol', value=field_name, __RC__)
        if (field_name /= '') then
            field = MAPL_FieldCreateEmpty(trim(field_name), self%grid, __RC__)

            call MAPL_FieldAllocCommit(field, dims=MAPL_DimsHorzVert, location=MAPL_VLocationCenter, typekind=MAPL_R4, hw=0, __RC__)
            call MAPL_StateAdd(aero, field, __RC__)
        end if

        ! add ASY to aero state
        call ESMF_AttributeGet(aero, name='asymmetry_parameter_of_ambient_aerosol', value=field_name, RC=STATUS)
        if (field_name /= '') then
            field = MAPL_FieldCreateEmpty(trim(field_name), self%grid, __RC__)

            call MAPL_FieldAllocCommit(field, dims=MAPL_DimsHorzVert, location=MAPL_VLocationCenter, typekind=MAPL_R4, hw=0, __RC__)
            call MAPL_StateAdd(aero, field, __RC__)
        end if
       
        ! attach the aerosol optics method
        call ESMF_MethodAdd(aero, label='aerosol_optics', userRoutine=aerosol_optics, __RC__)

    end if COUPLING_TO_RADIATION


!   Fill the AERO State with the aerosol mixing ratios
!   ---------------------------------------------------
    call ESMF_StateGet(EXPORT, 'AERO_ACI', aero_aci, __RC__)

    ! This attribute indicates if the aerosol optics method is implemented or not. 
    ! Radiation will not call the aerosol optics method unless this attribute is 
    ! explicitly set to true.

    implements_aap_method = .false. 

    call ESMF_AttributeSet(aero_aci, name  = 'implements_aerosol_activation_properties_method', & 
                                     value = implements_aap_method, __RC__)
  
    COUPLING_TO_CLOUD_MICROPHYSICS: if (implements_aap_method) then

        ! TODO: Implement the aerosol activation propertires method

    end if COUPLING_TO_CLOUD_MICROPHYSICS



!   Fill the scavenging coefficients
!   --------------------------------
    do m = 1, self%scheme%n_modes
        call MAM_AerosolModeGet(self%scheme%mode(m), name=mode_name, n_species=n_species)

        select case (trim(mode_name))
            case (trim(MAM7_AITKEN_MODE_NAME))
                call MAM_AerosolModeSet(self%scheme%mode(m), f_conv_scavenging = self%f_conv_scav_ait, &
                                                             f_wet             = self%f_wet_ait)

            case (trim(MAM7_ACCUMULATION_MODE_NAME))
                call MAM_AerosolModeSet(self%scheme%mode(m), f_conv_scavenging = self%f_conv_scav_acc, &
                                                             f_wet             = self%f_wet_acc)

            case (trim(MAM7_PRIMARY_CARBON_MODE_NAME))
                call MAM_AerosolModeSet(self%scheme%mode(m), f_conv_scavenging = self%f_conv_scav_pcm, &
                                                             f_wet             = self%f_wet_pcm)

            case (trim(MAM7_FINE_SEASALT_MODE_NAME))
                call MAM_AerosolModeSet(self%scheme%mode(m), f_conv_scavenging = self%f_conv_scav_fss, &
                                                             f_wet             = self%f_wet_fss)

            case (trim(MAM7_FINE_DUST_MODE_NAME))
                call MAM_AerosolModeSet(self%scheme%mode(m), f_conv_scavenging = self%f_conv_scav_fdu, &
                                                             f_wet             = self%f_wet_fdu)

            case (trim(MAM7_COARSE_SEASALT_MODE_NAME))
                call MAM_AerosolModeSet(self%scheme%mode(m), f_conv_scavenging = self%f_conv_scav_css, &
                                                             f_wet             = self%f_wet_css)

            case (trim(MAM7_COARSE_DUST_MODE_NAME))
                call MAM_AerosolModeSet(self%scheme%mode(m), f_conv_scavenging = self%f_conv_scav_cdu, &
                                                             f_wet             = self%f_wet_cdu)

            case default
                call MAM_AerosolModeSet(self%scheme%mode(m), f_conv_scavenging = f_scav_none, f_wet = f_wet_none)
        end select
    end do

    do m = 1, self%scheme%n_modes
        call MAM_AerosolModeGet(self%scheme%mode(m), name=mode_name, n_species=n_species, f_conv_scavenging=f_scav)

        if (self%verbose .and. MAPL_AM_I_ROOT()) then
            print *, trim(Iam)//': Convective Scavenging Parameter of '//trim(mode_name)//' : ', f_scav
        end if
 
        ! interstitial aerosol tracers
        field_name = 'NUM_A_' // trim(mode_name)
        call ESMF_StateGet(INTERNAL, trim(field_name), field, __RC__)

        call ESMF_AttributeSet(field, NAME='ScavengingFractionPerKm', VALUE=f_scav, __RC__)

        do s = 1, n_species
            species_name = self%scheme%mode(m)%species(s)%name
            field_name  = trim(species_name) // '_A_' // trim(mode_name)
            call ESMF_StateGet(INTERNAL, trim(field_name), field, __RC__)

            call ESMF_AttributeSet(field, NAME='ScavengingFractionPerKm', VALUE=f_scav, __RC__)
        end do

        ! cloud-borne aerosol tracers
        field_name = 'NUM_C_' // trim(mode_name)
        call ESMF_StateGet(INTERNAL, trim(field_name), field, __RC__)
        call ESMF_AttributeSet(field, NAME='ScavengingFractionPerKm', VALUE=f_scav_none, __RC__)

        do s = 1, n_species
            species_name = self%scheme%mode(m)%species(s)%name
            field_name  = trim(species_name) // '_C_' // trim(mode_name)
            call ESMF_StateGet(INTERNAL, trim(field_name), field, __RC__)
            call ESMF_AttributeSet(field, NAME='ScavengingFractionPerKm', VALUE=f_scav_none, __RC__)
        end do
    end do

#ifdef PRINT_STATES
    if (MAPL_AM_I_ROOT()) then
        print *, trim(Iam)//': AERO State during Initialize():'
        call ESMF_StatePrint(aero, nestedFlag=.true., __RC__)
    end if
#endif

!   Create narrow-band Mie tables
!   -----------------------------
    call ESMF_ConfigGetAttribute ( self%CF, mie_optics_file, Label='narrowband_optics_ait:', default='', __RC__ )
    self%mie_ait = MAML_OpticsTableCreate(trim(mie_optics_file), __RC__)
    call MAML_OpticsTableRead(self%mie_ait, __RC__)

    call ESMF_ConfigGetAttribute ( self%CF, mie_optics_file, Label='narrowband_optics_acc:', default='', __RC__ )
    self%mie_acc = MAML_OpticsTableCreate(trim(mie_optics_file), __RC__)
    call MAML_OpticsTableRead(self%mie_acc, __RC__)

    call ESMF_ConfigGetAttribute ( self%CF, mie_optics_file, Label='narrowband_optics_pcm:', default='', __RC__ )
    self%mie_pcm = MAML_OpticsTableCreate(trim(mie_optics_file), __RC__)
    call MAML_OpticsTableRead(self%mie_pcm, __RC__)

    call ESMF_ConfigGetAttribute ( self%CF, mie_optics_file, Label='narrowband_optics_fss:', default='', __RC__ )
    self%mie_fss = MAML_OpticsTableCreate(trim(mie_optics_file), __RC__)
    call MAML_OpticsTableRead(self%mie_fss, __RC__)

    call ESMF_ConfigGetAttribute ( self%CF, mie_optics_file, Label='narrowband_optics_css:', default='', __RC__ )
    self%mie_css = MAML_OpticsTableCreate(trim(mie_optics_file), __RC__)
    call MAML_OpticsTableRead(self%mie_css, __RC__)

    call ESMF_ConfigGetAttribute ( self%CF, mie_optics_file, Label='narrowband_optics_fdu:', default='', __RC__ )
    self%mie_fdu = MAML_OpticsTableCreate(trim(mie_optics_file), __RC__)
    call MAML_OpticsTableRead(self%mie_fdu, __RC__)

    call ESMF_ConfigGetAttribute ( self%CF, mie_optics_file, Label='narrowband_optics_cdu:', default='', __RC__ )
    self%mie_cdu = MAML_OpticsTableCreate(trim(mie_optics_file), __RC__)
    call MAML_OpticsTableRead(self%mie_cdu, __RC__)


!   All done
!   --------
    RETURN_(ESMF_SUCCESS)

    contains
   
    subroutine PrintProcessFlag(name, state)
        implicit none        
        character(len=*), intent(in) :: name
        logical, intent(in)          :: state
       
        character(len=*), parameter  :: fmt_flag  = "(4X, A12, X, '-', X, A3)"

        if (state) then
            write (*, fmt_flag) trim(name), 'ON'
        else
            write (*, fmt_flag) trim(name), 'OFF'
        end if
    end subroutine PrintProcessFlag

   end subroutine Initialize_


!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  Run_ --- Runs MAMchem
!
! !INTERFACE:
!

   subroutine Run_ ( GC, IMPORT, EXPORT, CLOCK, rc )

! !USES:

    implicit none

! !INPUT PARAMETERS:

    type(ESMF_Clock),  intent(inout)   :: CLOCK  ! The clock

! !OUTPUT PARAMETERS:

    type(ESMF_GridComp), intent(inout) :: GC     ! Grid Component
    type(ESMF_State), intent(inout)    :: IMPORT ! Import State
    type(ESMF_State), intent(inout)    :: EXPORT ! Export State
    integer, intent(out)               :: rc     ! Error return code:
                                                !  0 - all is well
                                                !  1 - 

! !DESCRIPTION: This is a simple ESMF wrapper.
!
! !REVISION HISTORY:
!
!  27Feb2005 da Silva  First crack.
!
!EOP
!-------------------------------------------------------------------------

                              __Iam__('Run_')
   
    type(MAM_state), pointer      :: self               ! Legacy state
    type(ESMF_Grid)               :: GRID               ! Grid
    type(ESMF_Config)             :: CF                 ! Universal Config 

    type(MAPL_MetaComp), pointer  :: mgState            ! MAPL generic state
    type(ESMF_Alarm)              :: run_alarm
    logical                       :: run_alarm_ringing

    integer                       :: im_World, jm_World ! Global 2D Dimensions
    integer                       :: im, jm, lm         ! 3D Dimensions
    real(ESMF_KIND_R4), pointer   :: lons(:,:)          ! Longitudes
    real(ESMF_KIND_R4), pointer   :: lats(:,:)          ! Latitudes

    integer                       :: nymd, nhms         ! date, time
    real                          :: cdt                ! time step in secs

    character(len=ESMF_MAXSTR)    :: comp_name

!   Declare pointers to IMPORT/EXPORT/INTERNAL states 
!   -------------------------------------------------
#  include "MAMchem_DeclarePointer___.h"

#if !(__MAMchem_InternalSpec_H__)
    type(MAPL_MetaComp), pointer  :: MetaComp
    type(ESMF_State)              :: INTERNAL
#endif
  
!   Get my name and set-up traceback handle
!   ---------------------------------------
    call ESMF_GridCompGet( GC, name=comp_name, __RC__ )
    Iam = trim(comp_name) // '::' // trim(Iam)

!   Get pointers to IMPORT/EXPORT/INTERNAL states 
!   ---------------------------------------------
#  include "MAMchem_GetPointer___.h"
#if !(__MAMchem_InternalSpec_H__)
    call MAPL_GetObjectFromGC ( GC, MetaComp, __RC__ )
    call MAPL_Get ( MetaComp, INTERNAL_ESMF_STATE=INTERNAL, __RC__)
#endif

!   Get my internal MAPL_Generic state
!   -----------------------------------
    call MAPL_GetObjectFromGC(GC, mgState, __RC__)


    call MAPL_TimerOn(mgState, 'RUN', __RC__)


!   Get parameters from generic state
!   ----------------------------------
    call MAPL_Get(mgState, LONS=lons, &
                           LATS=lats, &
                           RunAlarm=run_alarm, __RC__)


!   If it is time, update MAM state
!   -------------------------------
    run_alarm_ringing = ESMF_AlarmIsRinging(run_alarm, __RC__)

    if (run_alarm_ringing) then
        call ESMF_AlarmRingerOff(run_alarm, __RC__)
    else
        RETURN_(ESMF_SUCCESS)
    endif


!   Extract relevant runtime information
!   ------------------------------------
    call extract_(GC, CLOCK, self, GRID, CF, &
                  im_World, jm_World,        &
                  im, jm, lm,                &
                  nymd, nhms, cdt, __RC__)


!   Force non-negative mixing ratios
!   --------------------------------
    do im = 1, self%qa%n3d
        where(self%qa%r3(im)%q < 0) self%qa%r3(im)%q = tiny(0.0)
    end do

    do im = 1, self%qc%n3d
        where(self%qc%r3(im)%q < 0) self%qc%r3(im)%q = tiny(0.0)
    end do

    do im = 1, self%qg%n3d
        where(self%qg%r3(im)%q < 0) self%qg%r3(im)%q = tiny(0.0)
    end do


    call MAPL_TimerOn(mgState, '-EMISSIONS', __RC__)

!   Seasalt emissions
!   -----------------
    call MAM_SS_Emission(self%scheme, import, export, self%qa, self%femisSS, self%dt, __RC__)

!   Dust emissions
!   -----------------
    call MAM_DU_Emission(self%scheme, import, export, self%qa, self%femisDU, self%dt, __RC__)

!   Black Carbon emissions
!   ----------------------
    call MAM_BC_Emission(self%scheme, import, export, self%qa, self%dt, __RC__)

!   Organic Carbon emissions
!   ----------------------
    call MAM_OC_Emission(self%scheme, import, export, self%qa, self%pom_oc_ratio, self%dt, __RC__)

    call MAPL_TimerOff(mgState, '-EMISSIONS', __RC__)

!   Partition aqueous phase production of SO4 and NH4
!   -------------------------------------------------
    call AqueousChemistry(self%scheme, import, export, self%qa, cdt, rc)


    call MAPL_TimerOn(mgState, '-MICROPHYSICS', __RC__)

!   Calculate dry size of the particles --- CAM interface
!   ------------------------------------------------------
!   call CAM_CalculateSize(self%scheme, self%qa, self%qc, self%Dg_dry, self%Dg_wet, delp, self%dt, rc)

  

!  
!   Gas-aerosol exchange (H2SO4, MSA, NH3 condensation)
!   ---------------------------------------------------
    call MAPL_TimerOn(mgState, '--MICROPHYSICS_CONDENSATION', __RC__)

    if (self%condensation) then
#if (0)
        call MAM_DrySize(self%scheme, import, export, self%qa, self%Da, __RC__)
        call MAM_WetSize(self%scheme, import, export, self%qa, self%Da, __RC__)

        call MAM_GasAerosolExchange(self%scheme, import, export, self%qa, self%Da, self%dt, __RC__)
#endif
    end if

    call MAPL_TimerOff(mgState, '--MICROPHYSICS_CONDENSATION', __RC__)

!  
!   Aerosol nucleation (new particle formation)
!   -------------------------------------------
    call MAPL_TimerOn(mgState, '--MICROPHYSICS_NUCLEATION', __RC__)

    if (self%nucleation) then
        call MAM_DrySize(self%scheme, import, export, self%qa, self%Da, __RC__)
        call MAM_WetSize(self%scheme, import, export, self%qa, self%Da, __RC__)

        call MAM_Nucleation(self%scheme, import, export, self%qa, self%qg, self%Da, self%dt, __RC__)
    end if

    call MAPL_TimerOff(mgState, '--MICROPHYSICS_NUCLEATION', __RC__)

!
!   Coagulation
!   -----------
    call MAPL_TimerOn(mgState, '--MICROPHYSICS_COAGULATION', __RC__)

    if (self%coagulation) then
        call MAM_DrySize(self%scheme, import, export, self%qa, self%Da, __RC__)
        call MAM_WetSize(self%scheme, import, export, self%qa, self%Da, __RC__)

!       call MAM_Coagulation(self%scheme, import, export, self%qa, self%Da, self%dt, __RC__)
        call MAM_CoagulationBimodal(self%scheme, import, export, self%qa, self%Da, self%dt, __RC__)
    end if

    call MAPL_TimerOff(mgState, '--MICROPHYSICS_COAGULATION', __RC__)
    call MAPL_TimerOff(mgState, '-MICROPHYSICS', __RC__)


    call MAPL_TimerOn(mgState, '-REMOVAL', __RC__)
!
!   Dry removal - gravitational settling and dry deposition
!   -------------------------------------------------------
    if (self%dry_removal) then
        call MAM_DrySize(self%scheme, import, export, self%qa, self%Da, __RC__)
        call MAM_WetSize(self%scheme, import, export, self%qa, self%Da, __RC__)

        call MAM_DryRemoval(self%scheme, import, export, self%qa, self%Da, self%dt, __RC__)
    end if

!
!   Wet removal - large scale wet scavenging
!   ----------------------------------------
    if (self%wet_removal) then
        call MAM_WetRemoval(self%scheme, import, export, self%qa, self%dt, __RC__)
    end if

    call MAPL_TimerOff(mgState, '-REMOVAL', __RC__)

!  
!   Update the aerosol size and absorbed water
!   ------------------------------------------
    call MAPL_TimerOn(mgState, '-HYGROSCOPIC_GROWTH', __RC__)
    call MAM_DrySize(self%scheme, import, export, self%qa, self%Da, __RC__)
    call MAM_WetSize(self%scheme, import, export, self%qa, self%Da, __RC__)
    call MAPL_TimerOff(mgState, '-HYGROSCOPIC_GROWTH', __RC__)


!   Diagnostics 
!   -----------------
!   NOTE : The order of which the processes are done will have 
!          some impact on the dignostic fields
!   -----------------------------------------------------------
    call MAPL_TimerOn(mgState, '-DIAGNOSTICS', __RC__)
    
    call MAPL_TimerOn(mgState,  '--DIAGNOSTICS_SEASALT', __RC__)
    call MAM_SS_Diagnostics(self%scheme, import, export, self%qa, self%dt, __RC__)
    call MAPL_TimerOff(mgState, '--DIAGNOSTICS_SEASALT', __RC__)

    call MAPL_TimerOn(mgState,  '--DIAGNOSTICS_DUST', __RC__)
    call MAM_DU_Diagnostics(self%scheme, import, export, self%qa, self%dt, __RC__)
    call MAPL_TimerOff(mgState, '--DIAGNOSTICS_DUST', __RC__)

    call MAPL_TimerOn(mgState,  '--DIAGNOSTICS_CIM', __RC__)
    call CIM_Diagnostics(self%scheme, import, export, self%qa, self%dt, __RC__)
    call MAPL_TimerOff(mgState, '--DIAGNOSTICS_CIM', __RC__)

    call MAPL_TimerOn(mgState,  '--DIAGNOSTICS_AOT', __RC__)
    call AOT_Diagnostics(self%scheme, import, export, self%qa, self%Da, self%mie_ait, &
                                                                        self%mie_acc, &
                                                                        self%mie_pcm, &
                                                                        self%mie_fss, &
                                                                        self%mie_css, &
                                                                        self%mie_fdu, &
                                                                        self%mie_cdu, &
                                                                        self%dt, __RC__)
    call MAPL_TimerOff(mgState, '--DIAGNOSTICS_AOT', __RC__)

    call MAPL_TimerOff(mgState, '-DIAGNOSTICS', __RC__)

    call MAPL_TimerOff(mgState, 'RUN', __RC__)

!   All done
!   --------
    RETURN_(ESMF_SUCCESS)

   end subroutine Run_


!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  Finalize_ --- Finalize MAMchem
!
! !INTERFACE:
!

   subroutine Finalize_ ( GC, IMPORT, EXPORT, CLOCK, rc )

! !USES:

  implicit none

! !INPUT PARAMETERS:

   type(ESMF_Clock),  intent(inout)   :: CLOCK  ! The clock

! !OUTPUT PARAMETERS:

   type(ESMF_GridComp), intent(inout) :: gc     ! Grid Component
   type(ESMF_State), intent(inout)    :: IMPORT ! Import State
   type(ESMF_State), intent(inout)    :: EXPORT ! Export State
   integer, intent(out)               :: rc     ! Error return code:
                                                !  0 - all is well
                                                !  1 - 

! !DESCRIPTION: This is a simple ESMF wrapper.
!
! !REVISION HISTORY:
!
!  01Dec2009  da Silva  First crack.
!
!EOP
!-------------------------------------------------------------------------

                              __Iam__('Finalize_')
   
    type(MAM_state), pointer      :: self               ! Legacy state
    type(ESMF_Grid)               :: GRID               ! Grid
    type(ESMF_Config)             :: CF                 ! Universal Config 

    integer                       :: im_World, jm_World ! Global 2D dimensions
    integer                       :: im, jm, lm         ! 3D Dimensions

    integer                       :: nymd, nhms         ! date, time
    real                          :: cdt                ! time step in secs

    character(len=ESMF_MAXSTR)    :: COMP_NAME

!  Get my name and set-up traceback handle
!  ---------------------------------------
   call ESMF_GridCompGet( GC, name=comp_name, __RC__ )
   Iam = trim(comp_name) // '::' // trim(Iam)

!  Finalize MAPL Generic
!  ---------------------
   call MAPL_GenericFinalize ( GC, IMPORT, EXPORT, CLOCK,  __RC__ )

!  Extract relevant runtime information
!  ------------------------------------
   call extract_ ( GC, CLOCK, self, GRID, CF, &
                   im_World, jm_World,        &
                   im, jm, lm,                &
                   nymd, nhms, cdt, __RC__ )


!  Delete the tracers bundles
!  ------------------------------------
   call MAPL_SimpleBundleDestroy(self%qa)
   call MAPL_SimpleBundleDestroy(self%qc)
   call MAPL_SimpleBundleDestroy(self%qg)

   call MAPL_SimpleBundleDestroy(self%Da)

!  Delete the broad-band optical tables
!  ------------------------------------
   call MAML_OpticsTableDestroy(MAM7_MieTable(1), __RC__)
   call MAML_OpticsTableDestroy(MAM7_MieTable(2), __RC__)
   call MAML_OpticsTableDestroy(MAM7_MieTable(3), __RC__)
   call MAML_OpticsTableDestroy(MAM7_MieTable(4), __RC__)
   call MAML_OpticsTableDestroy(MAM7_MieTable(5), __RC__)
   call MAML_OpticsTableDestroy(MAM7_MieTable(6), __RC__)
   call MAML_OpticsTableDestroy(MAM7_MieTable(7), __RC__)

!  Delete the narrow-band optical tables
!  ------------------------------------
   call MAML_OpticsTableDestroy(self%mie_ait, __RC__)
   call MAML_OpticsTableDestroy(self%mie_acc, __RC__)
   call MAML_OpticsTableDestroy(self%mie_pcm, __RC__)
   call MAML_OpticsTableDestroy(self%mie_fss, __RC__)
   call MAML_OpticsTableDestroy(self%mie_css, __RC__)
   call MAML_OpticsTableDestroy(self%mie_fdu, __RC__)
   call MAML_OpticsTableDestroy(self%mie_cdu, __RC__)

!  Delete the internal private state
!  ---------------------------------
   deallocate(self, __STAT__)


!  All done
!  --------
   RETURN_(ESMF_SUCCESS)

 end subroutine Finalize_

!.......................................................................

 subroutine extract_ (GC, CLOCK,             &
                          myState, GRID, CF, &
                          im_World, jm_World,&
                          im, jm, lm,        &
                          nymd, nhms,        &
                          cdt, rc)

    type(ESMF_GridComp), intent(inout) :: GC                 ! Grid Comp object
    type(ESMF_Clock), intent(in)       :: CLOCK              ! Clock

    type(MAM_state), pointer           :: myState            ! Legacy state
    type(ESMF_Grid),     intent(out)   :: GRID               ! Grid
    type(ESMF_Config),   intent(out)   :: CF                 ! Universal Config 

    integer, intent(out)               :: im_World, jm_World ! Global 2D Dimensions
    integer, intent(out)               :: im, jm, lm         ! 3D Dimensions

    integer, intent(out)               :: nymd, nhms         ! date, time
    real, intent(out)                  :: cdt                ! time step in secs
    integer, intent(out), optional     :: rc

!                            ---

    __Iam__('extract_')

    character(len=ESMF_MAXSTR)         :: comp_name

    type(MAPL_MetaComp), pointer       :: mgState      ! MAPL generic state
    type(MAM_Wrap)                     :: wrap

    integer, dimension(3)              :: dims

    type(ESMF_Alarm)                   :: run_alarm
    type(ESMF_TimeInterval)            :: ring_interval
    real(ESMF_KIND_R8)                 :: time_step

    type(ESMF_Time)                    :: time
    integer                            :: iyr, imm, idd, ihr, imn, isc


!   Get my name and set-up traceback handle
!   ---------------------------------------
    call ESMF_GridCompGet(GC, name=comp_name, __RC__)
    Iam = trim(comp_name) // '::' // trim(Iam)

    rc = 0

!   Get my internal MAPL_Generic state
!   -----------------------------------
    call MAPL_GetObjectFromGC(GC, mgState, __RC__)

!   Get my internal state
!   ---------------------
    call ESMF_UserCompGetInternalState(GC, 'MAM_state', wrap, STATUS)
    VERIFY_(STATUS)
    myState => wrap%ptr

!   Get the configuration
!   ---------------------
    call ESMF_GridCompGet(GC, config=CF, __RC__)

!   Get time step
!   -------------
    call MAPL_Get(mgState, RunAlarm=run_alarm, __RC__)
    call ESMF_AlarmGet(run_alarm, ringInterval=ring_interval, __RC__)

    call ESMF_TimeIntervalGet(ring_interval, s_r8=time_step, __RC__)
    cdt = real(time_step)

!   Extract time as simple integers from clock
!   ------------------------------------------
    call ESMF_ClockGet(CLOCK, currTime=time, __RC__)
    call ESMF_TimeGet(TIME, yy=iyr, mm=imm, dd=idd, h=ihr,   m=imn,  s=isc, __RC__)

    call MAPL_PackTime(nymd, iyr, imm, idd)
    call MAPL_PackTime(nhms, ihr, imn, isc)

!   Extract the ESMF Grid
!   ---------------------
    call ESMF_GridCompGet(GC, grid=GRID, __RC__)

!   Global dimensions
!   -----------------
    call MAPL_GridGet(GRID, globalCellCountPerDim=dims, __RC__)
    im_World = dims(1)
    jm_World = dims(2)

!   Local dimensions
!   ----------------
    call ESMF_GridGet(GRID, localDE=0, &
                            staggerloc=ESMF_STAGGERLOC_CENTER, &
                            computationalCount=dims, __RC__)
    im = dims(1)
    jm = dims(2)
    lm = dims(3)


    RETURN_(ESMF_SUCCESS)

 end subroutine extract_


!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  MAM_GetTracerName --- returns short and long names of a tracer
!
! !INTERFACE:
!

   subroutine MAM_GetFieldName (mode_short_name, mode_long_name, species, attachment_state, type, short_name, long_name, rc)

! !USES:

    implicit none

! !INPUT/OUTPUT PARAMETERS:

! !INPUT PARAMETERS:
    character(len=MAM_MAXSTR), intent(in)  :: mode_short_name     ! mode short name
    character(len=MAM_MAXSTR), intent(in)  :: mode_long_name      ! mode long name
    character(len=MAM_MAXSTR), intent(in)  :: species             ! species name/alias
    character(len=MAM_MAXSTR), intent(in)  :: attachment_state    ! attachment state = {'interstitial' | 'cloud-borne'}
    character(len=MAM_MAXSTR), intent(in)  :: type                ! tracer type = {'number' | 'mass'}


! !OUTPUT PARAMETERS:
    character(len=MAM_MAXSTR), intent(out) :: short_name          ! short name of tracer
    character(len=MAM_MAXSTR), intent(out) :: long_name           ! long name of tracer

    integer, intent(out)                   :: rc                  ! error return code:
                                                                  !    0 - all is well
                                                                  !    1 -
 
! !DESCRIPTION: This routine constructs short and long names of aerosol tracer 
!               in one of the MAM modes depending on its type and 
!               attachment state.
!
! !REVISION HISTORY:
!
!  6 May 2014    A. Darmenov    Initial implementation.
!
!EOP
!-------------------------------------------------------------------------

    character(len=*), parameter :: Iam = 'MAM_GetTracerName'

    integer :: STATUS

    character(len=3)          :: state
    character(len=MAM_MAXSTR) :: name, buff


!   Initialize local variables
!   --------------------------
    rc = ESMF_SUCCESS

    ASSERT_(attachment_state == 'interstitial' .or. attachment_state == 'cloud-borne')
    ASSERT_(type == 'number' .or. type == 'mass')
    

    if (type == 'number') then
        name = 'NUM'
        buff = 'number of '
    else
        name = trim(species)
        buff = 'mass mixing ratio of '
    end if

    if (attachment_state == 'interstitial') then
        state = '_A_'
    else
        state = '_C_'
    end if
            
   
    short_name = trim(name) // state // trim(mode_short_name)
    long_name  = buff // trim(attachment_state) // ' aerosol particles in ' // trim(mode_long_name) // ' mode'


    RETURN_(ESMF_SUCCESS)

 end subroutine MAM_GetFieldName


!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  AqueousChemistry --- Simplified treatment of aqueos chemistry
!
! !INTERFACE:
!

   subroutine AqueousChemistry (self, import, export, qa, cdt, rc)

! !USES:

    use Chem_ConstMod, only: grav

    implicit none

! !INPUT/OUTPUT PARAMETERS:
    type(MAPL_SimpleBundle), intent(inout) :: qa         ! interstitial aerosol tracer fields
    type(ESMF_State), intent(inout)        :: export     ! export fields

! !INPUT PARAMETERS:
    type(MAM_Scheme), intent(in)           :: self       ! MAM scheme/configuration

    type(ESMF_State), intent(inout)        :: import     ! import fields
    real,    intent(in)                    :: cdt        ! chemical timestep (secs)

! !OUTPUT PARAMETERS:
    integer, intent(out)                   :: rc         ! error return code:
                                                         !    0 - all is well
                                                         !    1 -
 
! !DESCRIPTION: This routine updates mass fields due to aqueos chemistry processes.
!
! !REVISION HISTORY:
!
!  2 June 2014    A. Darmenov    Initial implementation.
!
!EOP
!-------------------------------------------------------------------------

    character(len=*), parameter :: Iam = 'AqueousChemistry'

    integer :: STATUS
    integer :: i1, i2, j1, j2, k1, km, i, k, m, s
    integer :: ijl, ijkl

!   Input fields from fvGCM
!   -----------------------
    real, pointer, dimension(:,:,:) :: rhoa, ple, delp
    real, pointer, dimension(:,:,:) :: pSO4_aq, pNH4_aq

!   Exports - diagnostic fields
!   ---------------------------

!   Local
!   -----
    real, allocatable, dimension(:,:,:) :: num_aq       ! total number mixing ratio for modes affected by aqueos chemistry
    real, allocatable, dimension(:,:,:) :: f            ! fraction of aqueos production
    integer :: in_acc, in_fss, in_css, in_fdu, in_cdu   ! index of number mixing raio
    integer :: iq_acc, iq_fss, iq_css, iq_fdu, iq_cdu   ! index of mass mixing raio

!   Parameters 
!   ----------
    real, parameter :: mw_air = 28.97            ! molar mass of dry air, g mol-1
    real, parameter :: mw_SO4 = 96.07            ! molar mass of sulfate, g mol-1
    real, parameter :: mw_NH4 = 28.97            ! molar mass of dry air, g mol-1


!   Initialize local variables
!   --------------------------
    rc = ESMF_SUCCESS

    ASSERT_(self%id == MAM7_SCHEME) 
 

!   Get Imports
!   --------------
    call MAPL_GetPointer(import, rhoa,    'AIRDENS',   __RC__)
    call MAPL_GetPointer(import, ple,     'PLE',       __RC__)
    call MAPL_GetPointer(import, delp,    'DELP',      __RC__)

    call MAPL_GetPointer(import, pSO4_aq, 'pSO4_aq',   __RC__)
    call MAPL_GetPointer(import, pNH4_aq, 'pNH4_aq',   __RC__)

    
    if ((.not. associated(pSO4_aq)) .or. (.not. associated(pSO4_aq))) then
        print *, 'Skipping MAM::AqueousChemistry()'
        RETURN_(ESMF_SUCCESS)
    end if


!   Get Exports
!   -----------
    

!   Local dimensions
!   ----------------
    i1 = lbound(rhoa, 1)
    i2 = ubound(rhoa, 1)
    j1 = lbound(rhoa, 2)
    j2 = ubound(rhoa, 2)
    k1 = lbound(rhoa, 3)
    km = ubound(rhoa, 3)

    ijl  = (i2 - i1 + 1) * (j2 - j1 + 1)
    ijkl = ijl * km

#ifdef DEBUG
    call write_parallel(trim(Iam) // '::DEBUG::Indexes:')
    call write_parallel((/i1, i2/), format='(("i1, i2 = ", (X2I3)))')
    call write_parallel((/j1, j2/), format='(("j1, j2 = ", (XI3)))')
    call write_parallel((/k1, km/), format='(("k1, k2 = ", (XI3)))')

    call write_parallel(trim(Iam) // '::DEBUG::Inputs:')
    call write_parallel(self%id,    format='(("model = ", (I5)))')
#endif


    in_acc = MAPL_SimpleBundleGetIndex(qa, 'NUM_A_ACC' , 3, __RC__)
    in_fss = MAPL_SimpleBundleGetIndex(qa, 'NUM_A_FSS' , 3, __RC__)
    in_css = MAPL_SimpleBundleGetIndex(qa, 'NUM_A_CSS' , 3, __RC__)
    in_fdu = MAPL_SimpleBundleGetIndex(qa, 'NUM_A_FDU' , 3, __RC__)
    in_cdu = MAPL_SimpleBundleGetIndex(qa, 'NUM_A_CDU' , 3, __RC__)

    iq_acc = MAPL_SimpleBundleGetIndex(qa, 'SU_A_ACC'  , 3, __RC__)
    iq_fss = MAPL_SimpleBundleGetIndex(qa, 'SU_A_FSS'  , 3, __RC__)
    iq_css = MAPL_SimpleBundleGetIndex(qa, 'SU_A_CSS'  , 3, __RC__)
    iq_fdu = MAPL_SimpleBundleGetIndex(qa, 'SU_A_FDU'  , 3, __RC__)
    iq_cdu = MAPL_SimpleBundleGetIndex(qa, 'SU_A_CDU'  , 3, __RC__)

   
    allocate(num_aq(i1:i2,j1:j2,km), __STAT__)
    allocate(f(i1:i2,j1:j2,km), __STAT__)

    num_aq = 0.0
    f      = 0.0

    num_aq = ( qa%r3(in_acc)%q + &
               qa%r3(in_fss)%q + &
               qa%r3(in_css)%q + &
               qa%r3(in_fdu)%q + &
               qa%r3(in_cdu)%q )

    where (num_aq > tiny(0.0))
        f = 1.0/num_aq * (mw_SO4/mw_air) * pSO4_aq * cdt
    end where

#if (DEBUG)
    if (MAPL_AM_I_ROOT()) then
        print *, 'MAM:AqueousChemistry: su_acc = ', minval(qa%r3(iq_acc)%q), maxval(qa%r3(iq_acc)%q)
        print *, 'MAM:AqueousChemistry: su_fss = ', minval(qa%r3(iq_fss)%q), maxval(qa%r3(iq_fss)%q)
        print *, 'MAM:AqueousChemistry: su_css = ', minval(qa%r3(iq_css)%q), maxval(qa%r3(iq_css)%q)
        print *, 'MAM:AqueousChemistry: su_fdu = ', minval(qa%r3(iq_fdu)%q), maxval(qa%r3(iq_fdu)%q)
        print *, 'MAM:AqueousChemistry: su_cdu = ', minval(qa%r3(iq_cdu)%q), maxval(qa%r3(iq_cdu)%q)

        print *, 'SO4_aq   = ', minval(pSO4_aq), maxval(pSO4_aq)
        print *, 'd_su_acc = ', minval(f * qa%r3(in_acc)%q), maxval(f * qa%r3(in_acc)%q)
        print *, 'd_su_fss = ', minval(f * qa%r3(in_fss)%q), maxval(f * qa%r3(in_fss)%q)
        print *, 'd_su_css = ', minval(f * qa%r3(in_css)%q), maxval(f * qa%r3(in_css)%q)
        print *, 'd_su_fdu = ', minval(f * qa%r3(in_fdu)%q), maxval(f * qa%r3(in_fdu)%q)
        print *, 'd_su_cdu = ', minval(f * qa%r3(in_cdu)%q), maxval(f * qa%r3(in_cdu)%q)
    end if
#endif

    qa%r3(iq_acc)%q = qa%r3(iq_acc)%q + (f * qa%r3(in_acc)%q)
    qa%r3(iq_fss)%q = qa%r3(iq_fss)%q + (f * qa%r3(in_fss)%q)
    qa%r3(iq_css)%q = qa%r3(iq_css)%q + (f * qa%r3(in_css)%q)
    qa%r3(iq_fdu)%q = qa%r3(iq_fdu)%q + (f * qa%r3(in_fdu)%q)
    qa%r3(iq_cdu)%q = qa%r3(iq_cdu)%q + (f * qa%r3(in_cdu)%q)

    deallocate(num_aq, __STAT__)
    deallocate(f,      __STAT__)

    RETURN_(ESMF_SUCCESS)

 end subroutine AqueousChemistry


!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  CIM_Diagnostics --- Column integrated mass diagnostics
!
! !INTERFACE:
!

   subroutine CIM_Diagnostics (self, import, export, qa, cdt, rc)

! !USES:

    use Chem_ConstMod, only: grav

    implicit none

! !INPUT/OUTPUT PARAMETERS:
    type(MAPL_SimpleBundle), intent(inout) :: qa         ! interstitial aerosol tracer fields
    type(ESMF_State), intent(inout)        :: export     ! export fields

! !INPUT PARAMETERS:
    type(MAM_Scheme), intent(in)           :: self       ! MAM scheme/configuration

    type(ESMF_State), intent(inout)        :: import     ! import fields
    real,    intent(in)                    :: cdt        ! chemical timestep (secs)

! !OUTPUT PARAMETERS:
    integer, intent(out)                   :: rc         ! error return code:
                                                         !    0 - all is well
                                                         !    1 -
 
! !DESCRIPTION: This routine computes column integrated (dry) mass fields.
!
! !REVISION HISTORY:
!
!  8 Mar 2013    A. Darmenov    Initial implementation.
!
!EOP
!-------------------------------------------------------------------------

    character(len=*), parameter :: Iam = 'CIM_Diagnostics'

    integer :: STATUS
    integer :: i1, i2, j1, j2, k1, km, i, k, m, s
    integer :: ijl, ijkl

!   Mode parameters
!   ------------------------
    character(len=MAM_MAXSTR)     :: mode_name      ! aerosol mode name
    character(len=MAM_MAXSTR)     :: species_name   ! aerosol species name
    character(len=MAM_MAXSTR)     :: field_name     ! field name
    integer                       :: n_species      ! number of aerosol species


!   Input fields from fvGCM
!   -----------------------
    real, pointer, dimension(:,:,:) :: rhoa, ple, delp

!   Exports - diagnostic fields
!   ---------------------------
    real, pointer, dimension(:,:)   :: colmass      ! column integrated mass density, kg m-2


!   Initialize local variables
!   --------------------------
    rc = ESMF_SUCCESS

    ASSERT_(self%id == MAM7_SCHEME .or. self%id == MAM3_SCHEME) 
 

!   Get Imports
!   --------------
    call MAPL_GetPointer(import, rhoa,      'AIRDENS',   __RC__)
    call MAPL_GetPointer(import, ple,       'PLE',       __RC__)
    call MAPL_GetPointer(import, delp,      'DELP',      __RC__)

!   Get Exports
!   --------------


!   Local dimensions
!   ----------------
    i1 = lbound(rhoa, 1)
    i2 = ubound(rhoa, 1)
    j1 = lbound(rhoa, 2)
    j2 = ubound(rhoa, 2)
    k1 = lbound(rhoa, 3)
    km = ubound(rhoa, 3)

    ijl  = (i2 - i1 + 1) * (j2 - j1 + 1)
    ijkl = ijl * km

#ifdef DEBUG
    call write_parallel(trim(Iam) // '::DEBUG::Indexes:')
    call write_parallel((/i1, i2/), format='(("i1, i2 = ", (X2I3)))')
    call write_parallel((/j1, j2/), format='(("j1, j2 = ", (XI3)))')
    call write_parallel((/k1, km/), format='(("k1, k2 = ", (XI3)))')

    call write_parallel(trim(Iam) // '::DEBUG::Inputs:')
    call write_parallel(self%id,    format='(("model = ", (I5)))')
#endif


    do m = 1, self%n_modes
        call MAM_AerosolModeGet(self%mode(m), name=mode_name, n_species=n_species)

        call MAPL_GetPointer(export, colmass,  trim(mode_name) // 'CMASS',  __RC__)
        if (associated(colmass)) colmass = 0.0

        ! mass mixing ratios
        do s = 1, n_species
            species_name = self%mode(m)%species(s)%name

            field_name  = trim(species_name) // '_A_' // trim(mode_name)

            i = MAPL_SimpleBundleGetIndex(qa, trim(field_name), 3, __RC__)
           
            if (associated(colmass)) then
                do k = 1, km
                    colmass(:,:) = colmass(:,:) + qa%r3(i)%q(:,:,k) * delp(:,:,k)/grav
                end do
            end if
        end do
    end do

    RETURN_(ESMF_SUCCESS)

 end subroutine CIM_Diagnostics


!-------------------------------------------------------------------------
!     NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1     !
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  AOT_Diagnostics --- Aerosol Optical Thickness
!
! !INTERFACE:
!
   subroutine AOT_Diagnostics (self, import, export, qa, Da, mie_ait, mie_acc, mie_pcm, mie_fss, mie_css, mie_fdu, mie_cdu, cdt, rc)

! !USES:

    use Chem_ConstMod, only: grav
    use MAM_ComponentsDataMod, only : MAM_WATER_COMPONENT_DENSITY,        &
                                      MAM_SULFATE_COMPONENT_DENSITY,      &
                                      MAM_AMMONIUM_COMPONENT_DENSITY,     &
                                      MAM_BLACK_CARBON_COMPONENT_DENSITY, &
                                      MAM_DUST_COMPONENT_DENSITY,         &
                                      MAM_SEASALT_COMPONENT_DENSITY,      &
                                      MAM_SOA_COMPONENT_DENSITY,          &
                                      MAM_POM_COMPONENT_DENSITY

    implicit none

! !INPUT/OUTPUT PARAMETERS:
    type(MAPL_SimpleBundle), intent(inout) :: qa         ! interstitial aerosol tracer fields
    type(MAPL_SimpleBundle), intent(inout) :: Da         ! dry and wet sizes of interstital aerosols
    type(ESMF_State), intent(inout)        :: export     ! export fields

! !INPUT PARAMETERS:
    type(MAM_Scheme), intent(in)           :: self       ! MAM scheme/configuration

    type(ESMF_State), intent(inout)        :: import     ! import fields

    type(MAML_OpticsTable)                 :: mie_ait
    type(MAML_OpticsTable)                 :: mie_acc
    type(MAML_OpticsTable)                 :: mie_pcm
    type(MAML_OpticsTable)                 :: mie_fss
    type(MAML_OpticsTable)                 :: mie_css
    type(MAML_OpticsTable)                 :: mie_fdu
    type(MAML_OpticsTable)                 :: mie_cdu

    real,    intent(in)                    :: cdt        ! chemical timestep (secs)

! !OUTPUT PARAMETERS:
    integer, intent(out)                   :: rc         ! error return code:
                                                         !    0 - all is well
                                                         !    1 -
 
! !DESCRIPTION: This routine computes aerosol optical thickness.
!
! !REVISION HISTORY:
!
!  23 May 2014    A. Darmenov    Initial implementation.
!
!EOP
!-------------------------------------------------------------------------

!   Input fields from fvGCM
!   -----------------------
    real, pointer, dimension(:,:,:)         :: delp

!   Exports - diagnostic fields
!   ---------------------------
    real, pointer, dimension(:,:)           :: tot_ext, tot_sca         ! aerosol optical thickness, '1'
    real, pointer, dimension(:,:)           :: ait_ext, ait_sca
    real, pointer, dimension(:,:)           :: acc_ext, acc_sca
    real, pointer, dimension(:,:)           :: pcm_ext, pcm_sca
    real, pointer, dimension(:,:)           :: fss_ext, fss_sca
    real, pointer, dimension(:,:)           :: css_ext, css_sca
    real, pointer, dimension(:,:)           :: fdu_ext, fdu_sca
    real, pointer, dimension(:,:)           :: cdu_ext, cdu_sca

    real, dimension(:,:,:), pointer         :: rh

    real, dimension(:,:,:,:), allocatable   :: qa_

    integer                                 :: i_mmr
    integer                                 :: i_dwet

    real(kind=8), dimension(:,:,:), allocatable :: ext, sca, asy, ssa   ! total, (lon:,lat:,lev:,band:)

    real, dimension(:,:,:), allocatable     :: ext_, sca_, asy_         ! mode,  (lon:,lat:,lev:,band:)
    real, dimension(:),     allocatable     :: density
    integer                                 :: nc

    integer                                 :: i1, j1, i2, j2, k1, km
    integer                                 :: ijl, ijkl

    integer                                 :: band

  
    integer                                 :: STATUS
    character(len=ESMF_MAXSTR)              :: Iam



                                   Iam = 'AOD_Diagnostics'

!   Get Imports
!   --------------
    call MAPL_GetPointer(import, delp,    'DELP',      __RC__)

!   Get Exports
!   --------------
    call MAPL_GetPointer(export, tot_ext, 'TOTEXTTAU', __RC__)
    call MAPL_GetPointer(export, tot_sca, 'TOTSCATAU', __RC__)

    call MAPL_GetPointer(export, ait_ext, 'AITEXTTAU', __RC__)
    call MAPL_GetPointer(export, ait_sca, 'AITSCATAU', __RC__)

    call MAPL_GetPointer(export, acc_ext, 'ACCEXTTAU', __RC__)
    call MAPL_GetPointer(export, acc_sca, 'ACCSCATAU', __RC__)

    call MAPL_GetPointer(export, pcm_ext, 'PCMEXTTAU', __RC__)
    call MAPL_GetPointer(export, pcm_sca, 'PCMSCATAU', __RC__)

    call MAPL_GetPointer(export, fss_ext, 'FSSEXTTAU', __RC__)
    call MAPL_GetPointer(export, fss_sca, 'FSSSCATAU', __RC__)

    call MAPL_GetPointer(export, css_ext, 'CSSEXTTAU', __RC__)
    call MAPL_GetPointer(export, css_sca, 'CSSSCATAU', __RC__)

    call MAPL_GetPointer(export, fdu_ext, 'FDUEXTTAU', __RC__)
    call MAPL_GetPointer(export, fdu_sca, 'FDUSCATAU', __RC__)

    call MAPL_GetPointer(export, cdu_ext, 'CDUEXTTAU', __RC__)
    call MAPL_GetPointer(export, cdu_sca, 'CDUSCATAU', __RC__)

!   Local dimensions
!   ----------------
    i1 = lbound(delp, 1)
    i2 = ubound(delp, 1)
    j1 = lbound(delp, 2)
    j2 = ubound(delp, 2)
    k1 = lbound(delp, 3)
    km = ubound(delp, 3)

    ijl  = (i2 - i1 + 1) * (j2 - j1 + 1)
    ijkl = ijl * km


    ! Radiation band
    ! --------------
    band = 7   ! 550nm 

    ! Pressure at layer edges 
    ! ------------------------



  allocate(ext(i1:i2,j1:j2,km), &
           sca(i1:i2,j1:j2,km), &
           ssa(i1:i2,j1:j2,km), &
           asy(i1:i2,j1:j2,km), __STAT__)

  allocate(ext_(i1:i2,j1:j2,km), &
           sca_(i1:i2,j1:j2,km), &
           asy_(i1:i2,j1:j2,km), __STAT__)
  

  ext = 0.0
  sca = 0.0
  ssa = 0.0
  asy = 0.0


  ! compute ext, sca and ssa from aerosols in aitken mode; su, amm, soa, ss
  nc = 4 + 1
  allocate(qa_(nc,i1:i2,j1:j2,km), density(nc), __STAT__)
 
  qa_ = 0.0 
  density = 0.0

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'WTR_A_AIT', 3, __RC__)
  qa_(1,:,:,:) = qa%r3(i_mmr)%q
  density(1)  = MAM_WATER_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'SU_A_AIT',  3, __RC__)
  qa_(2,:,:,:) = qa%r3(i_mmr)%q
  density(2) = MAM_SULFATE_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'AMM_A_AIT', 3, __RC__)
  qa_(3,:,:,:) = qa%r3(i_mmr)%q
  density(3)  = MAM_AMMONIUM_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'SOA_A_AIT', 3, __RC__)
  qa_(4,:,:,:) = qa%r3(i_mmr)%q
  density(4)  = MAM_SOA_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'SS_A_AIT',  3, __RC__)
  qa_(5,:,:,:) = qa%r3(i_mmr)%q
  density(5)  = MAM_SEASALT_COMPONENT_DENSITY

  i_dwet = MAPL_SimpleBundleGetIndex(Da, 'DGN_WET_AIT', 3, __RC__)

  ext_ = 0.0 
  sca_ = 0.0 
  asy_ = 0.0

  call MAML_OpticsInterpolate(mie_ait, band, qa_, density, Da%r3(i_dwet)%q, delp, ext_, sca_, asy_, nc, i1, i2, j1, j2, 1, km, rc)

  if (associated(ait_ext)) ait_ext = sum(ext_, dim=3)
  if (associated(ait_sca)) ait_sca = sum(sca_, dim=3)

  ext = ext + ext_
  sca = sca + sca_
  asy = asy + sca_*asy_

  deallocate(qa_, density, __STAT__)


  ! compute ext, sca and ssa from aerosols in accumulation mode: su, amm, soa, pom, bc, ss
  nc = 6 + 1 
  allocate(qa_(nc,i1:i2,j1:j2,km), density(nc), __STAT__)
 
  qa_ = 0.0 
  density = 0.0

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'WTR_A_ACC', 3, __RC__)
  qa_(1,:,:,:) = qa%r3(i_mmr)%q
  density(1)  = MAM_WATER_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'SU_A_ACC',  3, __RC__)
  qa_(2,:,:,:) = qa%r3(i_mmr)%q
  density(2) = MAM_SULFATE_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'AMM_A_ACC', 3, __RC__)
  qa_(3,:,:,:) = qa%r3(i_mmr)%q
  density(3)  = MAM_AMMONIUM_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'SOA_A_ACC', 3, __RC__)
  qa_(4,:,:,:) = qa%r3(i_mmr)%q
  density(4)  = MAM_SOA_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'POM_A_ACC',  3, __RC__)
  qa_(5,:,:,:) = qa%r3(i_mmr)%q
  density(5)  = MAM_POM_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'BC_A_ACC',  3, __RC__)
  qa_(6,:,:,:) = qa%r3(i_mmr)%q
  density(6)  = MAM_BLACK_CARBON_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'SS_A_ACC',  3, __RC__)
  qa_(7,:,:,:) = qa%r3(i_mmr)%q
  density(7)  = MAM_SEASALT_COMPONENT_DENSITY

  i_dwet = MAPL_SimpleBundleGetIndex(Da, 'DGN_WET_ACC', 3, __RC__)

  ext_ = 0.0 
  sca_ = 0.0 
  asy_ = 0.0

  call MAML_OpticsInterpolate(mie_acc, band, qa_, density, Da%r3(i_dwet)%q, delp, ext_, sca_, asy_, nc, i1, i2, j1, j2, 1, km, rc)

  if (associated(acc_ext)) acc_ext = sum(ext_, dim=3)
  if (associated(acc_sca)) acc_sca = sum(sca_, dim=3)

  ext = ext + ext_
  sca = sca + sca_
  asy = asy + sca_*asy_

  deallocate(qa_, density, __STAT__)


  ! compute ext, sca and ssa from aerosols in primary carbon mode: pom, bc
  nc = 2 + 1
  allocate(qa_(nc,i1:i2,j1:j2,km), density(nc), __STAT__)
 
  qa_ = 0.0 
  density = 0.0

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'WTR_A_PCM', 3, __RC__)
  qa_(1,:,:,:) = qa%r3(i_mmr)%q
  density(1)  = MAM_WATER_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'POM_A_PCM',  3, __RC__)
  qa_(2,:,:,:) = qa%r3(i_mmr)%q
  density(2) = MAM_POM_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'BC_A_PCM', 3, __RC__)
  qa_(3,:,:,:) = qa%r3(i_mmr)%q
  density(3)  = MAM_BLACK_CARBON_COMPONENT_DENSITY

  i_dwet = MAPL_SimpleBundleGetIndex(Da, 'DGN_WET_PCM', 3, __RC__)

  ext_ = 0.0 
  sca_ = 0.0 
  asy_ = 0.0

  call MAML_OpticsInterpolate(mie_pcm, band, qa_, density, Da%r3(i_dwet)%q, delp, ext_, sca_, asy_, nc, i1, i2, j1, j2, 1, km, rc)

  if (associated(pcm_ext)) pcm_ext = sum(ext_, dim=3)
  if (associated(pcm_sca)) pcm_sca = sum(sca_, dim=3)

  ext = ext + ext_
  sca = sca + sca_
  asy = asy + sca_*asy_

  deallocate(qa_, density, __STAT__)


  ! compute ext, sca and ssa from aerosols in fine seasalt mode: su, amm, ss
  nc = 3 + 1
  allocate(qa_(nc,i1:i2,j1:j2,km), density(nc), __STAT__)
 
  qa_ = 0.0 
  density = 0.0

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'WTR_A_FSS', 3, __RC__)
  qa_(1,:,:,:) = qa%r3(i_mmr)%q
  density(1)  = MAM_WATER_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'SU_A_FSS',  3, __RC__)
  qa_(2,:,:,:) = qa%r3(i_mmr)%q
  density(2) = MAM_SULFATE_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'AMM_A_FSS', 3, __RC__)
  qa_(3,:,:,:) = qa%r3(i_mmr)%q
  density(3)  = MAM_AMMONIUM_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'SS_A_FSS', 3, __RC__)
  qa_(4,:,:,:) = qa%r3(i_mmr)%q
  density(4)  = MAM_SEASALT_COMPONENT_DENSITY

  i_dwet = MAPL_SimpleBundleGetIndex(Da, 'DGN_WET_FSS', 3, __RC__)

  ext_ = 0.0 
  sca_ = 0.0 
  asy_ = 0.0

  call MAML_OpticsInterpolate(mie_fss, band, qa_, density, Da%r3(i_dwet)%q, delp, ext_, sca_, asy_, nc, i1, i2, j1, j2, 1, km, rc)

  if (associated(fss_ext)) fss_ext = sum(ext_, dim=3)
  if (associated(fss_sca)) fss_sca = sum(sca_, dim=3)

  ext = ext + ext_
  sca = sca + sca_
  asy = asy + sca_*asy_

  deallocate(qa_, density, __STAT__)


  ! compute ext, sca and ssa from aerosols in coarse seasalt mode: su, amm, ss; lut = ('water', 'su', 'amm', 'ss')
  nc = 3 + 1
  allocate(qa_(nc,i1:i2,j1:j2,km), density(nc), __STAT__)
 
  qa_ = 0.0 
  density = 0.0

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'WTR_A_CSS', 3, __RC__)
  qa_(1,:,:,:) = qa%r3(i_mmr)%q
  density(1)  = MAM_WATER_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'SU_A_CSS',  3, __RC__)
  qa_(2,:,:,:) = qa%r3(i_mmr)%q
  density(2) = MAM_SULFATE_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'AMM_A_CSS', 3, __RC__)
  qa_(3,:,:,:) = qa%r3(i_mmr)%q
  density(3)  = MAM_AMMONIUM_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'SS_A_CSS', 3, __RC__)
  qa_(4,:,:,:) = qa%r3(i_mmr)%q
  density(4)  = MAM_SEASALT_COMPONENT_DENSITY

  i_dwet = MAPL_SimpleBundleGetIndex(Da, 'DGN_WET_CSS', 3, __RC__)

  ext_ = 0.0 
  sca_ = 0.0 
  asy_ = 0.0

  call MAML_OpticsInterpolate(mie_css, band, qa_, density, Da%r3(i_dwet)%q, delp, ext_, sca_, asy_, nc, i1, i2, j1, j2, 1, km, rc)

  if (associated(css_ext)) css_ext = sum(ext_, dim=3)
  if (associated(css_sca)) css_sca = sum(sca_, dim=3)

  ext = ext + ext_
  sca = sca + sca_
  asy = asy + sca_*asy_

  deallocate(qa_, density, __STAT__)


  ! compute ext, sca and ssa from aerosols in fine dust mode: su, amm, du; lut = ('water', 'su', 'amm', 'du')
  nc = 3 + 1
  allocate(qa_(nc,i1:i2,j1:j2,km), density(nc), __STAT__)
 
  qa_ = 0.0 
  density = 0.0

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'WTR_A_FDU', 3, __RC__)
  qa_(1,:,:,:) = qa%r3(i_mmr)%q
  density(1)  = MAM_WATER_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'SU_A_FDU',  3, __RC__)
  qa_(2,:,:,:) = qa%r3(i_mmr)%q
  density(2) = MAM_SULFATE_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'AMM_A_FDU', 3, __RC__)
  qa_(3,:,:,:) = qa%r3(i_mmr)%q
  density(3)  = MAM_AMMONIUM_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'DU_A_FDU', 3, __RC__)
  qa_(4,:,:,:) = qa%r3(i_mmr)%q
  density(4)  = MAM_DUST_COMPONENT_DENSITY

  i_dwet = MAPL_SimpleBundleGetIndex(Da, 'DGN_WET_FDU', 3, __RC__)

  ext_ = 0.0 
  sca_ = 0.0 
  asy_ = 0.0

  call MAML_OpticsInterpolate(mie_fdu, band, qa_, density, Da%r3(i_dwet)%q, delp, ext_, sca_, asy_, nc, i1, i2, j1, j2, 1, km, rc)

  if (associated(fdu_ext)) fdu_ext = sum(ext_, dim=3)
  if (associated(fdu_sca)) fdu_sca = sum(sca_, dim=3)

  ext = ext + ext_
  sca = sca + sca_
  asy = asy + sca_*asy_

  deallocate(qa_, density, __STAT__)


  ! compute ext, sca and ssa from aerosols in coarse dust mode: su, amm, du; lut = ('water', 'su', 'amm', 'du')
  nc = 3 + 1
  allocate(qa_(nc,i1:i2,j1:j2,km), density(nc), __STAT__)
 
  qa_ = 0.0 
  density = 0.0

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'WTR_A_CDU', 3, __RC__)
  qa_(1,:,:,:) = qa%r3(i_mmr)%q
  density(1)  = MAM_WATER_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'SU_A_CDU',  3, __RC__)
  qa_(2,:,:,:) = qa%r3(i_mmr)%q
  density(2) = MAM_SULFATE_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'AMM_A_CDU', 3, __RC__)
  qa_(3,:,:,:) = qa%r3(i_mmr)%q
  density(3)  = MAM_AMMONIUM_COMPONENT_DENSITY

  i_mmr = MAPL_SimpleBundleGetIndex(qa, 'DU_A_CDU', 3, __RC__)
  qa_(4,:,:,:) = qa%r3(i_mmr)%q
  density(4)  = MAM_DUST_COMPONENT_DENSITY

  i_dwet = MAPL_SimpleBundleGetIndex(Da, 'DGN_WET_CDU', 3, __RC__)

  ext_ = 0.0 
  sca_ = 0.0 
  asy_ = 0.0

  call MAML_OpticsInterpolate(mie_cdu, band, qa_, density, Da%r3(i_dwet)%q, delp, ext_, sca_, asy_, nc, i1, i2, j1, j2, 1, km, rc)

  if (associated(cdu_ext)) cdu_ext = sum(ext_, dim=3)
  if (associated(cdu_sca)) cdu_sca = sum(sca_, dim=3)

  ext = ext + ext_
  sca = sca + sca_
  asy = asy + sca_*asy_

  deallocate(qa_, density, __STAT__)


  ! total AOT
  if (associated(tot_ext)) tot_ext = sum(ext, dim=3)
  if (associated(tot_sca)) tot_sca = sum(sca, dim=3)


  deallocate(ext,  sca,  ssa,  asy, __STAT__)
  deallocate(ext_, sca_, asy_,      __STAT__)

  RETURN_(ESMF_SUCCESS)

 end subroutine AOT_Diagnostics


!! --


logical function isDataDrivenGC_(GC, rc)
   type(ESMF_GridComp), intent(inout) :: GC
   integer, intent(out)               :: rc
 
!  local 
   character(len=ESMF_MAXSTR)         :: IAm
   integer                            :: STATUS

   integer                            :: i
   character(len=ESMF_MAXSTR)         :: comp_name
   character(len=*), parameter        :: modifier = '.data'

   rc = ESMF_SUCCESS

   call ESMF_GridCompGet(GC, name=comp_name, __RC__)   
   i = index(trim(comp_name), trim(modifier), back=.true.)
 
   if (i > 0) then
       ! lets be strict
       if (comp_name(i:) == modifier) then
           isDataDrivenGC_ = .true.
       else
           isDataDrivenGC_ = .false.
       end if
   else
       isDataDrivenGC_ = .false.
   end if

   RETURN_(ESMF_SUCCESS)

end function isDataDrivenGC_



 subroutine aerosol_optics(state, rc)

  use MAM_ComponentsDataMod, only : MAM_WATER_COMPONENT_DENSITY,        &
                                    MAM_SULFATE_COMPONENT_DENSITY,      &
                                    MAM_AMMONIUM_COMPONENT_DENSITY,     &
                                    MAM_BLACK_CARBON_COMPONENT_DENSITY, &
                                    MAM_DUST_COMPONENT_DENSITY,         &
                                    MAM_SEASALT_COMPONENT_DENSITY,      &
                                    MAM_SOA_COMPONENT_DENSITY,          &
                                    MAM_POM_COMPONENT_DENSITY

  implicit none 



  

! Arguments
! ---------
  type(ESMF_State):: state
  integer, intent(out):: rc


! Local
! ---------
  type(ESMF_FieldBundle)                  :: aerosols

  real, dimension(:,:,:), pointer         :: ple
  real, dimension(:,:,:), pointer         :: rh
  real, dimension(:,:,:), pointer         :: var
  real, dimension(:,:,:), pointer         :: q
  real, dimension(:,:,:), pointer         :: dgn_wet

  real, dimension(:,:,:,:), allocatable   :: qa


  real, dimension(:,:,:), allocatable     :: dp

  character(len=ESMF_MAXSTR)              :: field_name

  real(kind=8), dimension(:,:,:), allocatable :: ext, sca, asy, ssa   ! total, (lon:,lat:,lev:,band:)

  real, dimension(:,:,:), allocatable     :: ext_, sca_, asy_     ! mode,  (lon:,lat:,lev:,band:)
  real, dimension(:),     allocatable     :: density
  integer                                 :: nc

  integer                                 :: i1, j1, i2, j2, km

  integer                                 :: band

  
  integer                                 :: STATUS
  character(len=ESMF_MAXSTR)              :: Iam



                                   Iam = 'MAM::aerosol_optics()'


! Radiation band
! --------------
  call ESMF_AttributeGet(state, name='band_for_aerosol_optics', value=band, __RC__)

! Pressure at layer edges 
! ------------------------
  call ESMF_AttributeGet(state, name='air_pressure_for_aerosol_optics', value=field_name, __RC__)
  call MAPL_GetPointer(state, ple, trim(field_name), __RC__)

  i1 = lbound(ple, 1); i2 = ubound(ple, 1)
  j1 = lbound(ple, 2); j2 = ubound(ple, 2)
                       km = ubound(ple, 3)

! Relative humidity
! -----------------
  call ESMF_AttributeGet(state, name='relative_humidity_for_aerosol_optics', value=field_name, __RC__)
  call MAPL_GetPointer(state, rh, trim(field_name), __RC__)

  i1 = lbound(rh, 1); i2 = ubound(rh, 1)
  j1 = lbound(rh, 2); j2 = ubound(rh, 2)
                      km = ubound(rh, 3)
  
  call ESMF_StateGet(state, 'AEROSOLS', aerosols, __RC__)


  allocate(dp(i1:i2,j1:j2,km), __STAT__)
  dp = (ple(:,:,1:km) - ple(:,:,0:km-1))


  allocate(ext(i1:i2,j1:j2,km), &
           sca(i1:i2,j1:j2,km), &
           ssa(i1:i2,j1:j2,km), &
           asy(i1:i2,j1:j2,km), __STAT__)

  allocate(ext_(i1:i2,j1:j2,km), &
           sca_(i1:i2,j1:j2,km), &
           asy_(i1:i2,j1:j2,km), __STAT__)
  

  ext = 0.0
  sca = 0.0
  ssa = 0.0
  asy = 0.0


  ! compute ext, sca and ssa from aerosols in aitken mode; su, amm, soa, ss
  nc = 4 + 1
  allocate(qa(nc,i1:i2,j1:j2,km), density(nc), __STAT__)
 
  qa = 0.0 
  density = 0.0

  call ESMFL_BundleGetPointerToData(aerosols, 'WTR_A_AIT', q, __RC__)
  qa(1,:,:,:) = q
  density(1)  = MAM_WATER_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'SU_A_AIT',  q, __RC__)
  qa(2,:,:,:) = q
  density(2)  = MAM_SULFATE_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'AMM_A_AIT', q, __RC__)
  qa(3,:,:,:) = q
  density(3)  = MAM_AMMONIUM_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'SOA_A_AIT', q, __RC__)
  qa(4,:,:,:) = q
  density(4)  = MAM_SOA_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'SS_A_AIT',  q, __RC__)
  qa(5,:,:,:) = q
  density(5)  = MAM_SEASALT_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'DGN_WET_AIT', dgn_wet, __RC__)

  ext_ = 0.0 
  sca_ = 0.0 
  asy_ = 0.0

  call MAML_OpticsInterpolate(MAM7_MieTable(1), band, qa, density, dgn_wet, dp, ext_, sca_, asy_, nc, i1, i2, j1, j2, 1, km, rc)

  ext = ext + ext_
  sca = sca + sca_
  asy = asy + sca_*asy_
  deallocate(qa, density, __STAT__)


  ! compute ext, sca and ssa from aerosols in accumulation mode: su, amm, soa, pom, bc, ss
  nc = 6 + 1
  allocate(qa(nc,i1:i2,j1:j2,km), density(nc), __STAT__)

  qa      = 0.0
  density = 0.0

  call ESMFL_BundleGetPointerToData(aerosols, 'WTR_A_ACC', q, __RC__)
  qa(1,:,:,:) = q
  density(1)  = MAM_WATER_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'SU_A_ACC',  q, __RC__)
  qa(2,:,:,:) = q
  density(2)  = MAM_SULFATE_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'AMM_A_ACC', q, __RC__)
  qa(3,:,:,:) = q
  density(3)  = MAM_AMMONIUM_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'SOA_A_ACC', q, __RC__)      
  qa(4,:,:,:) = q
  density(4)  = MAM_SOA_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'POM_A_ACC', q, __RC__)
  qa(5,:,:,:) = q
  density(5)  = MAM_POM_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'BC_A_ACC',  q, __RC__)
  qa(6,:,:,:) = q
  density(6)  = MAM_BLACK_CARBON_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'SS_A_ACC',  q, __RC__)
  qa(7,:,:,:) = q
  density(7)  = MAM_SEASALT_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'DGN_WET_ACC', dgn_wet, __RC__)

  ext_ = 0.0 
  sca_ = 0.0 
  asy_ = 0.0

  call MAML_OpticsInterpolate(MAM7_MieTable(2), band, qa, density, dgn_wet, dp, ext_, sca_, asy_, nc, i1, i2, j1, j2, 1, km, rc)

  ext = ext + ext_
  sca = sca + sca_
  asy = asy + sca_*asy_
  deallocate(qa, density, __STAT__)


  ! compute ext, sca and ssa from aerosols in primary carbon mode: pom, bc
  nc = 2 + 1
  allocate(qa(nc,i1:i2,j1:j2,km), density(nc), __STAT__)

  call ESMFL_BundleGetPointerToData(aerosols, 'WTR_A_PCM', q, __RC__)
  qa(1,:,:,:) = q
  density(1)  = MAM_WATER_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'POM_A_PCM', q, __RC__)
  qa(2,:,:,:) = q
  density(2)  = MAM_POM_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'BC_A_PCM',  q, __RC__)
  qa(3,:,:,:) = q
  density(3)  = MAM_BLACK_CARBON_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'DGN_WET_PCM', dgn_wet, __RC__)

  ext_ = 0.0 
  sca_ = 0.0 
  asy_ = 0.0

  call MAML_OpticsInterpolate(MAM7_MieTable(3), band, qa, density, dgn_wet, dp, ext_, sca_, asy_, nc, i1, i2, j1, j2, 1, km, rc)

  ext = ext + ext_
  sca = sca + sca_
  asy = asy + sca_*asy_
  deallocate(qa, density, __STAT__)


  ! compute ext, sca and ssa from aerosols in fine seasalt mode: su, amm, ss
  nc = 3 + 1
  allocate(qa(nc,i1:i2,j1:j2,km), density(nc), __STAT__)

  call ESMFL_BundleGetPointerToData(aerosols, 'WTR_A_FSS', q, __RC__)
  qa(1,:,:,:) = q
  density(1)  = MAM_WATER_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'SU_A_FSS',  q, __RC__)
  qa(2,:,:,:) = q
  density(2)  = MAM_SULFATE_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'AMM_A_FSS', q, __RC__)
  qa(3,:,:,:) = q
  density(3)  = MAM_AMMONIUM_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'SS_A_FSS',  q, __RC__)
  qa(4,:,:,:) = q
  density(4)  = MAM_SEASALT_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'DGN_WET_FSS', dgn_wet, __RC__)

  ext_ = 0.0 
  sca_ = 0.0 
  asy_ = 0.0

  call MAML_OpticsInterpolate(MAM7_MieTable(4), band, qa, density, dgn_wet, dp, ext_, sca_, asy_, nc, i1, i2, j1, j2, 1, km, rc)

  ext = ext + ext_
  sca = sca + sca_
  asy = asy + sca_*asy_
  deallocate(qa, density, __STAT__)


  ! compute ext, sca and ssa from aerosols in coarse seasalt mode: su, amm, ss; lut = ('water', 'su', 'amm', 'ss')
  nc = 3 + 1
  allocate(qa(nc,i1:i2,j1:j2,km), density(nc), __STAT__)

  call ESMFL_BundleGetPointerToData(aerosols, 'WTR_A_CSS', q, __RC__)
  qa(1,:,:,:) = q
  density(1)  = MAM_WATER_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'SU_A_CSS',  q, __RC__)
  qa(2,:,:,:) = q
  density(2)  = MAM_SULFATE_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'AMM_A_CSS', q, __RC__)
  qa(3,:,:,:) = q
  density(3)  = MAM_AMMONIUM_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'SS_A_CSS',  q, __RC__)
  qa(4,:,:,:) = q
  density(4)  = MAM_SEASALT_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'DGN_WET_CSS', dgn_wet, __RC__)

  ext_ = 0.0 
  sca_ = 0.0 
  asy_ = 0.0

  call MAML_OpticsInterpolate(MAM7_MieTable(5), band, qa, density, dgn_wet, dp, ext_, sca_, asy_, nc, i1, i2, j1, j2, 1, km, rc)

  ext = ext + ext_
  sca = sca + sca_
  asy = asy + sca_*asy_
  deallocate(qa, density, __STAT__)


  ! compute ext, sca and ssa from aerosols in fine dust mode: su, amm, du; lut = ('water', 'su', 'amm', 'du')
  nc = 3 + 1
  allocate(qa(nc,i1:i2,j1:j2,km), density(nc), __STAT__)
 
  call ESMFL_BundleGetPointerToData(aerosols, 'WTR_A_FDU', q, __RC__)
  qa(1,:,:,:) = q
  density(1)  = MAM_WATER_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'SU_A_FDU',  q, __RC__)
  qa(2,:,:,:) = q
  density(2)  = MAM_SULFATE_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'AMM_A_FDU', q, __RC__)
  qa(3,:,:,:) = q
  density(3)  = MAM_AMMONIUM_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'DU_A_FDU',  q, __RC__)
  qa(4,:,:,:) = q
  density(4)  = MAM_DUST_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'DGN_WET_FDU', dgn_wet, __RC__)

  ext_ = 0.0 
  sca_ = 0.0 
  asy_ = 0.0
 
  call MAML_OpticsInterpolate(MAM7_MieTable(6), band, qa, density, dgn_wet, dp, ext_, sca_, asy_, nc, i1, i2, j1, j2, 1, km, rc)

  ext = ext + ext_
  sca = sca + sca_
  asy = asy + sca_*asy_
  deallocate(qa, density, __STAT__)


  ! compute ext, sca and ssa from aerosols in coarse dust mode: su, amm, du; lut = ('water', 'su', 'amm', 'du')
  nc = 3 + 1
  allocate(qa(nc,i1:i2,j1:j2,km), density(nc), __STAT__)

  call ESMFL_BundleGetPointerToData(aerosols, 'WTR_A_CDU', q, __RC__)
  qa(1,:,:,:) = q
  density(1)  = MAM_WATER_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'SU_A_CDU',  q, __RC__)
  qa(2,:,:,:) = q
  density(2)  = MAM_SULFATE_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'AMM_A_CDU', q, __RC__)
  qa(3,:,:,:) = q
  density(3)  = MAM_AMMONIUM_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'DU_A_CDU',  q, __RC__)
  qa(4,:,:,:) = q
  density(4)  = MAM_DUST_COMPONENT_DENSITY

  call ESMFL_BundleGetPointerToData(aerosols, 'DGN_WET_CDU', dgn_wet, __RC__)

  ext_ = 0.0 
  sca_ = 0.0 
  asy_ = 0.0

  call MAML_OpticsInterpolate(MAM7_MieTable(7), band, qa, density, dgn_wet, dp, ext_, sca_, asy_, nc, i1, i2, j1, j2, 1, km, rc)

  ext = ext + ext_
  sca = sca + sca_
  asy = asy + sca_*asy_
  deallocate(qa, density, __STAT__)

 
  ! inputs for radiation:  ext, sca and sca*asy
  ! in the callback     : 'asy' = product of asy and sca
  !                       'ssa' = sca
  ssa = sca


  call ESMF_AttributeGet(state, name='extinction_in_air_due_to_ambient_aerosol', value=field_name, __RC__)
  if (field_name /= '') then 
      call MAPL_GetPointer(state, var, trim(field_name), __RC__)
      var = ext(:,:,:)
  end if

  call ESMF_AttributeGet(state, name='single_scattering_albedo_of_ambient_aerosol', value=field_name, __RC__)
  if (field_name /= '') then 
      call MAPL_GetPointer(state, var, trim(field_name), __RC__)
      var = ssa(:,:,:)
  end if

  call ESMF_AttributeGet(state, name='asymmetry_parameter_of_ambient_aerosol', value=field_name, __RC__)
  if (field_name /= '') then 
      call MAPL_GetPointer(state, var, trim(field_name), __RC__)
      var = asy(:,:,:)
  end if


  deallocate(dp,                    __STAT__) 
  deallocate(ext,  sca,  ssa,  asy, __STAT__)
  deallocate(ext_, sca_, asy_,      __STAT__)

  RETURN_(ESMF_SUCCESS)

 end subroutine aerosol_optics


 end module MAMchem_GridCompMod
