#include "MAPL_Generic.h"
!-------------------------------------------------------------------------
!NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1, GEOS/DAS!
!-------------------------------------------------------------------------
!BOP
!
! !MODULE:  SC_GridCompMod --- SC Grid Component Class
!
! Grid Component class for NASA GSFC Code 614 stratospheric chemistry
!
! !INTERFACE:
!

   MODULE  SC_GridCompMod

! !USES:

   USE ESMF
   USE MAPL_Mod
   USE Chem_Mod 	     ! Chemistry Base Class
   USE Chem_StateMod	     ! Chemistry State
   USE Chem_UtilMod
   USE m_inpak90	     ! Resource file management

   USE ESMF_CFIOFileMOD
   USE MAPL_CFIOMOD

   IMPLICIT none
   INTEGER, PARAMETER :: DBL = KIND(0.00D+00)


! !PUBLIC TYPES:
!
   PRIVATE
   PUBLIC  SC_GridComp       ! The SC object 

!
! !PUBLIC MEMBER FUNCTIONS:
!

   PUBLIC  SC_GridCompInitialize
   PUBLIC  SC_GridCompRun
   PUBLIC  SC_GridCompFinalize

!
! !DESCRIPTION:
!
!  This module implements the Stratospheric Chemistry Grid Component. 
!
! !REVISION HISTORY:
!
!  16Sep2003 da Silva  First crack.
!  24Jan2O05 Nielsen   Implementation of Code 916 chemistry
!  19Dec2005 da Silva  Minor portability mods.
!  11May2012 Nielsen   Capability for FV cubed [Ganymed-1_0_UNSTABLE]
!
!EOP
!-------------------------------------------------------------------------

  TYPE SC_GridComp
    CHARACTER(LEN=ESMF_MAXSTR) :: name = "NASA_GSFC_Stratopheric_Chemistry"

! Dimensions
! ----------
   INTEGER :: i1, i2, im, j1, j2, jm, km

! Surface area [m^{2}] of grid cells
! ----------------------------------
   REAL(KIND=DBL), POINTER :: cellArea(:,:)

! Longitudes and latitudes (radians)
! ----------------------------------
   REAL, POINTER :: lonRad(:,:)
   REAL, POINTER :: latRad(:,:)

! Switches set in SC_GridComp.rc
! ------------------------------
    LOGICAL :: doPSCs
    LOGICAL :: doSediment
    LOGICAL :: importSulfSA   ! bring in GOCART, etc., based sulfate surface area
    LOGICAL :: addSAclim      ! if importSulfSA, possibly add climatological SA to it
    LOGICAL :: verbose
    LOGICAL :: wordy
    LOGICAL :: spinup
    LOGICAL :: useSolCyc

! Largest solar zenith angle (degrees) allowed as daytime.
! --------------------------------------------------------
    REAL :: szaCutoff

! Layer number (1=surface)below which the special
! treatment of the NO photorate will NOT be applied.
! --------------------------------------------------
    REAL :: kNOspec

! Number of years to add or subtract (must be greater than -1000) from the 
! current year for the surface mixing ratios of greenhouse gases and ozone depleting  
! depleting substances. To lock-in a specified year, enter the negative of the year number.
! -----------------------------------------------------------------------------------------
    INTEGER :: GHGYrAdj
    INTEGER :: ODSYrAdj

! Option to prescribe the year for sulface surface area
! -----------------------------------------------------
    INTEGER :: SO4saYr

! Two calendar elements
! ---------------------
    LOGICAL :: isLeapYear
    REAL :: dayOfYear

! Uppermost layer to apply chemistry (1=surface).  Note: This is a *BAD*
! construct attributable to the legacy code that should be changed.
! ----------------------------------------------------------------------
    INTEGER :: levels_cal

! Thermal reaction rates and table lookup parameters.
! Dynamic allocation is now done in rdkrate.F.
! --------------------------------------------------
    INTEGER :: numreacs
    INTEGER :: niters = 3
    INTEGER :: num1d
    INTEGER :: num2d
    INTEGER :: num3d
    INTEGER, POINTER :: indxs1(:), indxs2(:), indxs3(:)
    INTEGER, POINTER :: KRxnOrder(:)
    REAL, POINTER :: cnsttab(:,:)
    LOGICAL, POINTER :: standardKRxn(:)
    CHARACTER(LEN=ESMF_MAXSTR), POINTER :: KRxnName(:)
    LOGICAL :: didKRxnsAlloc = .FALSE.

! Photolysis tables
! -----------------
    INTEGER :: numphoto = 43
    INTEGER :: nxdo
    INTEGER :: nlam
    INTEGER :: nsza
    INTEGER :: numo3
    INTEGER :: nts
    INTEGER :: aqsize

    REAL, POINTER :: sdat(:,:,:,:)
    REAL, POINTER :: o2jdat(:,:,:)
    REAL, POINTER :: sza_tab(:)
    REAL, POINTER :: o3_tab(:,:)
    REAL, POINTER :: xtab(:,:,:)
    REAL, POINTER :: CH2O_aq(:)
    REAL, POINTER :: rlam(:)

! Degrees-to-radians
! ------------------
    REAL :: dtr

! Mixing ratios of gases.  Allow changes to CO2 in FixSSG.
! --------------------------------------------------------
    REAL :: o2 = 0.20946
    REAL :: n2 = 0.7808
    REAL :: NOxTRW
    REAL :: parts = 10.00

! Polar stratospheric cloud parameters
! ------------------------------------
!    PSCtmax - Maximum temperature for calculation of sads
!    PSCpmax - Maximum pressure for calculation of sads. If pmax is 
!              negative, then pmax will be the tropopause pressure.
!    PSCpmin - Minimum pressure for calculation of sads
    REAL :: PSCtmax = 240.
    REAL :: PSCpmax = 150.  !SC_GridComp.rc value overrides in Initialize
    REAL :: PSCpmin = 10.
    REAL :: PSClatlim = 45. !SC_GridComp.rc value overrides in Initialize
                            !Changed to radians in Initialize

! Enforce a maximum allowable condensed HNO3 (ppbv).
! ---------------------------------------------------------------
   INTEGER :: HNO3Ice_MAX

! calcsts - flag to set type 1 aerosols to STS (true) or NAT (false)
! ------------------------------------------------------------------
    LOGICAL :: calcsts = .FALSE.

! ------------------------------------------------------------------
!   Parameters controlling ice PSCs:
!
!    nice - number of ice particles/cm^3
!    densice - max density of ice in grams/cm^3 condensed volume
!    sigice - width of lognormal particle size distribution
!    rice - median radius of ice particle size distribution (microns)
!           used if constantnice is false
!    satratice - saturation ratio required before formation of ice pscs
!    constantnice - if true, set n and sigma, let r float; if false,
!                   set r and sigma, let n float.
! ------------------------------------------------------------------
    LOGICAL :: constantnice=.TRUE.
    REAL :: nice = 1.e-2
    REAL :: sigice = 1.6
    REAL :: rice = 10.
    REAL :: densice = 1.
    REAL :: satratice = 1.

! Parameters controlling nat aerosols:
! Definitions for nat are analogous to those for ice
! --------------------------------------------------
    LOGICAL :: constantnnat=.TRUE.
    REAL :: nnat = 0.1
    REAL :: signat = 1.6
    REAL :: rnat = 0.4
    REAL :: densnat = 1.6
    REAL :: satratnat=1.0

! Parameters controlling sts aerosols:
! Definitions for sts are analogous to those for ice
! No need for saturation ratio because sts are liquid.
! ----------------------------------------------------
    LOGICAL :: constantnsts=.TRUE.
    REAL :: nsts = 10.0
    REAL :: sigsts = 1.6
    REAL :: rsts = 0.4

! Parameters defining assumptions made about lbs aerosols.  The
! subroutine sadsts calls the subroutine calcsulf, which uses these
! parameters and the liquid binary sulfate surface area density to
! estimate the h2so4 mixing ratio needed to calculate the sts surface
! area density
! -------------------------------------------------------------------
    LOGICAL :: constantnlbs=.TRUE.
    REAL :: nlbs = 10.
    REAL :: siglbs = 1.6
    REAL :: rlbs = 0.1

! Species placement in w_c%qa(ic)%data3d 
! --------------------------------------
    INTEGER ::    iAoA
    INTEGER :: 	   iOx,     iNOx,  iHNO3,   iN2O5
    INTEGER :: iHO2NO2,  iClONO2,   iClx,    iHCl
    INTEGER ::   iHOCl,    iH2O2,   iBrx,    iN2O
    INTEGER :: 	  iCl2,    iOClO,  iBrCl,    iHBr
    INTEGER :: iBrONO2,     iCH4,  iHOBr, iCH3OOH
    INTEGER :: 	   iCO,   iHNO3c,   iF11
    INTEGER :: 	  iF12,    iF113,  iHCFC, iHCFC22
    INTEGER ::   iCCl4, iCH3CCl3, iCH3Cl,  iCH3Br
    INTEGER ::  iH1301,  iH12_24,    iO3,    iO3p
    INTEGER :: 	  iO1d,       iN,    iNO,    iNO2
    INTEGER :: 	  iNO3,       iH,    iOH,    iHO2
    INTEGER :: 	   iCl,     iClO,   iBrO,     iBr
    INTEGER ::  iCl2O2,    iCH2O, iCH3O2,  irO3Ox

! Climatologies
! -------------
   INTEGER :: BCnymd
   CHARACTER(LEN=ESMF_MAXSTR) :: climFileName

   INTEGER :: numClimLats
   REAL, POINTER :: climLatRad(:)

   INTEGER :: numGHGs = 3
   INTEGER :: numSSGs = 16
!  CHARACTER(LEN=ESMF_MAXSTR) :: SSGName(16) = (/ "CO2", &
!    "CH4",     "N2O",  "CFC11",    "CFC12",   "CFC113", & 
!   "CCl4", "CH3CCl3", "HCFC22", "HCFC141B", "HCFC142B", & 
!  "H1211",   "H1301",  "H2402",    "CH3BR",    "CH3CL" /)

! CO2 is reset by SSGSfcFlx
! -------------------------
   REAL :: CO2 = 358.00E-06

! 2-D Climatological fields
! -------------------------
   REAL, POINTER :: SSG(:)
   REAL, POINTER :: climRain(:,:)
   REAL, POINTER :: climO3(:,:)
   REAL, POINTER :: climOH(:,:)
   REAL, POINTER :: climH2O(:,:)
   REAL, POINTER :: SO4sa(:,:)

! 3-D interplolated climatological fields
! ---------------------------------------
   REAL, POINTER :: climRain3D(:,:,:)
   REAL, POINTER :: climO33D(:,:,:)
   REAL, POINTER :: climOH3D(:,:,:)
   REAL, POINTER :: climH2O3D(:,:,:)
   REAL, POINTER :: SO4sa3D(:,:,:)

  END TYPE SC_GridComp

CONTAINS

!-------------------------------------------------------------------------
!NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1, GEOS/DAS!
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  SC_GridCompInitialize --- Initialize SC_GridComp
!
! !INTERFACE:
!

   SUBROUTINE SC_GridCompInitialize ( gcSC, w_c, impChem, expChem, &
                                      nymd, nhms, tdt, rc )

! !USES:

  IMPLICIT none

! !INPUT PARAMETERS:

   TYPE(Chem_Bundle), INTENT(in) :: w_c     ! Chemical tracer fields, delp, +
   INTEGER, INTENT(IN) :: nymd, nhms	    ! time
   REAL,    INTENT(IN) :: tdt		    ! chemistry time step (secs)


! !OUTPUT PARAMETERS:

   TYPE(SC_GridComp), INTENT(INOUT) :: gcSC	! Grid Component
   TYPE(ESMF_State), INTENT(INOUT)  :: impChem  ! Import State
   TYPE(ESMF_State), INTENT(INOUT)  :: expChem  ! Export State
   INTEGER, INTENT(OUT) ::  rc                  ! Error return code:
                                                !  0 - all is well
                                                !  1 - 

! !DESCRIPTION: Initializes the SC Grid Component. It primarily sets
!               the import state for each active constituent package.
!
! !REVISION HISTORY:
!
!  18Sep2003 da Silva  First crack.
!  24Jan2O05 Nielsen   Implementation of Code 916 chemistry
!  11May2012 Nielsen   Capability for FV cubed [Ganymed-1_0_UNSTABLE]
!
!EOP
!-------------------------------------------------------------------------
#include "mpif.h"

   CHARACTER(LEN=*), PARAMETER :: Iam = 'SC_GridCompInitialize'
   CHARACTER(LEN=ESMF_MAXSTR) :: rcfilen = 'SC_GridComp.rc'

   TYPE(ESMF_VM) :: vm

   INTEGER :: i, i1, i2, im, j1, j2, jm, k, km
   INTEGER :: n, status

   CHARACTER(LEN=ESMF_MAXSTR) :: fileName
   CHARACTER(LEN=ESMF_MAXSTR) :: fnKRates
   CHARACTER(LEN=ESMF_MAXSTR) :: fnPhoto

   gcSC%name = 'Stratospheric_Chemistry'
   IF(MAPL_AM_I_ROOT()) THEN
    PRINT *,' '
    PRINT *,TRIM(Iam)//':'
   END IF

! Initialize local variables
! NOTE: gcSC%lonRad(i,j) and gcSC%latRad(i,j) are in radians
! ----------------------------------------------------------
   rc = 0
   i1 = gcSC%i1
   i2 = gcSC%i2
   im = gcSC%im
   
   j1 = gcSC%j1
   j2 = gcSC%j2
   jm = gcSC%jm
   
   km = gcSC%km
   gcSC%levels_cal = km
   gcSC%BCnymd = -1

! Grab the virtual machine
! ------------------------
   CALL ESMF_VMGetCurrent(vm, RC=status)
   VERIFY_(status)

! Load resource file
! ------------------
   CALL I90_loadf ( TRIM(rcfilen), status )
   VERIFY_(status)

! Parse resource file
! -------------------
   CALL I90_label ( 'solar_ZA_cutoff:', status )
   VERIFY_(status)
   gcSC%szaCutoff = I90_gfloat( status )
   VERIFY_(status)

   CALL I90_label ( 'NO_cutoff:', status )
   VERIFY_(status)
   gcSC%kNOspec = I90_gfloat( status )
   VERIFY_(status)

   CALL I90_label ( 'TRW_NOx:', status )
   VERIFY_(status)
   gcSC%NOxTRW = I90_gfloat( status )
   VERIFY_(status)

   CALL I90_label ( 'PSCpmax:', status )
   VERIFY_(status)
   gcSC%PSCpmax = I90_gfloat( status )
   VERIFY_(status)

   CALL I90_label ( 'PSClatlim:', status )
   VERIFY_(status)
   gcSC%PSClatlim = I90_gfloat( status )
   VERIFY_(status)

   CALL I90_label ( 'HNO3Ice_MAX:', status )
   VERIFY_(status)
   gcSC%HNO3Ice_MAX = I90_gfloat( status )
   VERIFY_(status)

   CALL I90_label ( 'doPSCs:', status )
   VERIFY_(status)
   i = I90_gint( status )
   VERIFY_(status)
   IF(i == 0) THEN
    gcSC%doPSCs = .FALSE.
   ELSE
    gcSC%doPSCs = .TRUE.
   END IF

   CALL I90_label ( 'importSulfateSA:', status )
   VERIFY_(status)
   i = I90_gint( status )
   VERIFY_(status)
   IF(i == 0) THEN
    gcSC%importSulfSA = .FALSE.
   ELSE
    gcSC%importSulfSA = .TRUE.
   END IF

   CALL I90_label ( 'addSAclim:', status )
   VERIFY_(status)
   i = I90_gint( status )
   VERIFY_(status)
   IF(i == 0) THEN
    gcSC%addSAclim = .FALSE.
   ELSE
    gcSC%addSAclim = .TRUE.
   END IF

   CALL I90_label ( 'spinup:', status )
   VERIFY_(status)
   i = I90_gint( status )
   VERIFY_(status)
   IF(i == 0) THEN
    gcSC%spinup = .FALSE.
   ELSE
    gcSC%spinup = .TRUE.
   END IF

   CALL I90_label ( 'doSediment:', status )
   VERIFY_(status)
   i = I90_gint( status )
   VERIFY_(status)
   IF(i == 0) THEN
    gcSC%doSediment = .FALSE.
   ELSE
    gcSC%doSediment = .TRUE.
   END IF
   
   CALL I90_label ( 'verbose:', status )
   VERIFY_(status)
   i = I90_gint( status )
   VERIFY_(status)
   IF(i == 0) THEN
    gcSC%verbose = .FALSE.
   ELSE
    gcSC%verbose = .TRUE.
   END IF
   
   CALL I90_label ( 'useSolCyc:', status )
   VERIFY_(status)
   i = I90_gint( status )
   VERIFY_(status)
   IF(i == 0) THEN
    gcSC%useSolCyc = .FALSE.
   ELSE
    gcSC%useSolCyc = .TRUE.
   END IF

   CALL I90_Label ( 'GHGYrAdj:', status )
   VERIFY_(status)
   gcSC%GHGYrAdj = I90_Gint( status )
   VERIFY_(status)

   CALL I90_Label ( 'ODSYrAdj:', status )
   VERIFY_(status)
   gcSC%ODSYrAdj = I90_Gint( status )
   VERIFY_(status)

   CALL I90_Label ( 'SO4saYr:', status )
   VERIFY_(status)
   gcSC%SO4saYr = I90_Gint( status )
   VERIFY_(status)

   CALL I90_label ( 'KRateTables:', status )
   VERIFY_(status)
   CALL I90_Gtoken( fnKRates, status )
   VERIFY_(status)

   CALL I90_label ( 'photolysisFile:', status )
   VERIFY_(status)
   CALL I90_Gtoken( fnPhoto, status )
   VERIFY_(status)

   CALL I90_label ( 'climatologiesFile:', status )
   VERIFY_(status)
   CALL I90_Gtoken( gcSC%climFileName, status )
   VERIFY_(status)

! Release resource file
! ---------------------
   CALL I90_release()

! Obtain species order in w_c%qa(ic)%data3d from the registry
! -----------------------------------------------------------
   DO i = w_c%reg%i_SC,w_c%reg%j_xx
    IF(TRIM(w_c%reg%vname(i)) ==      "OX") gcSC%iOx = i
    IF(TRIM(w_c%reg%vname(i)) ==     "NOX") gcSC%iNOx = i
    IF(TRIM(w_c%reg%vname(i)) ==    "HNO3") gcSC%iHNO3 = i
    IF(TRIM(w_c%reg%vname(i)) ==    "N2O5") gcSC%iN2O5 = i
    IF(TRIM(w_c%reg%vname(i)) ==  "HO2NO2") gcSC%iHO2NO2 = i
    IF(TRIM(w_c%reg%vname(i)) ==  "CLONO2") gcSC%iClONO2 = i
    IF(TRIM(w_c%reg%vname(i)) ==     "CLX") gcSC%iClx = i
    IF(TRIM(w_c%reg%vname(i)) ==     "HCL") gcSC%iHCl = i
    IF(TRIM(w_c%reg%vname(i)) ==    "HOCL") gcSC%iHOCl = i
    IF(TRIM(w_c%reg%vname(i)) ==    "H2O2") gcSC%iH2O2 = i
    IF(TRIM(w_c%reg%vname(i)) ==     "BRX") gcSC%iBrx = i
    IF(TRIM(w_c%reg%vname(i)) ==     "N2O") gcSC%iN2O = i
    IF(TRIM(w_c%reg%vname(i)) ==     "CL2") gcSC%iCl2 = i
    IF(TRIM(w_c%reg%vname(i)) ==    "OCLO") gcSC%iOClO = i
    IF(TRIM(w_c%reg%vname(i)) ==    "BRCL") gcSC%iBrCl = i
    IF(TRIM(w_c%reg%vname(i)) ==     "HBR") gcSC%iHBr = i
    IF(TRIM(w_c%reg%vname(i)) ==  "BRONO2") gcSC%iBrONO2 = i
    IF(TRIM(w_c%reg%vname(i)) ==     "CH4") gcSC%iCH4 = i
    IF(TRIM(w_c%reg%vname(i)) ==    "HOBR") gcSC%iHOBr = i
    IF(TRIM(w_c%reg%vname(i)) ==  "CH3OOH") gcSC%iCH3OOH = i
    IF(TRIM(w_c%reg%vname(i)) ==      "CO") gcSC%iCO = i
    IF(TRIM(w_c%reg%vname(i)) =="HNO3COND") gcSC%iHNO3c = i
    IF(TRIM(w_c%reg%vname(i)) ==   "CFC11") gcSC%iF11 = i
    IF(TRIM(w_c%reg%vname(i)) ==   "CFC12") gcSC%iF12 = i
    IF(TRIM(w_c%reg%vname(i)) ==  "CFC113") gcSC%iF113 = i
    IF(TRIM(w_c%reg%vname(i)) ==    "HCFC") gcSC%iHCFC = i
    IF(TRIM(w_c%reg%vname(i)) ==  "HCFC22") gcSC%iHCFC22 = i
    IF(TRIM(w_c%reg%vname(i)) ==    "CCL4") gcSC%iCCl4 = i
    IF(TRIM(w_c%reg%vname(i)) == "CH3CCL3") gcSC%iCH3CCl3 = i
    IF(TRIM(w_c%reg%vname(i)) ==   "CH3CL") gcSC%iCH3Cl = i
    IF(TRIM(w_c%reg%vname(i)) ==   "CH3BR") gcSC%iCH3Br = i
    IF(TRIM(w_c%reg%vname(i)) ==   "H1301") gcSC%iH1301 = i
    IF(TRIM(w_c%reg%vname(i)) ==  "H12_24") gcSC%iH12_24 = i
    IF(TRIM(w_c%reg%vname(i)) == "AOADAYS") gcSC%iAoA = i
    IF(TRIM(w_c%reg%vname(i)) ==  "O3CHEM") gcSC%iO3 = i
    IF(TRIM(w_c%reg%vname(i)) ==     "O3P") gcSC%iO3p = i
    IF(TRIM(w_c%reg%vname(i)) ==     "O1D") gcSC%iO1d = i
    IF(TRIM(w_c%reg%vname(i)) ==       "N") gcSC%iN = i
    IF(TRIM(w_c%reg%vname(i)) ==      "NO") gcSC%iNO = i
    IF(TRIM(w_c%reg%vname(i)) ==     "NO2") gcSC%iNO2 = i
    IF(TRIM(w_c%reg%vname(i)) ==     "NO3") gcSC%iNO3 = i
    IF(TRIM(w_c%reg%vname(i)) == "HATOMIC") gcSC%iH = i
    IF(TRIM(w_c%reg%vname(i)) ==      "OH") gcSC%iOH = i
    IF(TRIM(w_c%reg%vname(i)) ==     "HO2") gcSC%iHO2 = i
    IF(TRIM(w_c%reg%vname(i)) ==      "CL") gcSC%iCl = i
    IF(TRIM(w_c%reg%vname(i)) ==     "CLO") gcSC%iClO = i
    IF(TRIM(w_c%reg%vname(i)) ==     "BRO") gcSC%iBrO = i
    IF(TRIM(w_c%reg%vname(i)) ==      "BR") gcSC%iBr = i
    IF(TRIM(w_c%reg%vname(i)) ==   "CL2O2") gcSC%iCl2O2 = i
    IF(TRIM(w_c%reg%vname(i)) ==    "CH2O") gcSC%iCH2O = i
    IF(TRIM(w_c%reg%vname(i)) ==   "CH3O2") gcSC%iCH3O2 = i
    IF(TRIM(w_c%reg%vname(i)) ==   "RO3OX") gcSC%irO3Ox = i
   END DO

! PSC configuration summary
! -------------------------
   EchoInfo: IF(MAPL_AM_I_ROOT()) THEN

    PRINT *," "
    IF(gcSC%SO4saYr == 0 ) THEN
     n = nymd/10000
    ELSE
     n = gcSC%SO4saYr
    END IF

    IF(gcSC%doPSCs) THEN
     PRINT *,'Reactions for PSCs are enabled'
     IF(gcSC%calcsts) THEN
      PRINT *,' Type 1 aerosols are STS'
      PRINT *,' Sulfate surface area chosen for ',n
     ELSE
      PRINT *,' Type 1 aerosols are NAT'
      PRINT *,' Sulfate surface area ignored'
     END IF
     PRINT *," High-pressure range limit for PSCs: ",gcSC%PSCpmax," hPa"
     PRINT *," Low-latitude range limit for PSCs: ",gcSC%PSClatlim," degrees"
    ELSE
     PRINT *,TRIM(Iam)//': Reactions for PSCs disabled'
    END IF

    PRINT *," "
    PRINT *,"Daytime solar zenith angle cutoff: ",gcSC%szaCutoff," degrees"
    PRINT *,"Lightning NOx source strength: ",gcSC%NOxTRW," cm^{-3} s^{-1}"
    PRINT *," "
   END IF EchoInfo

! Useful constants
! ----------------
   gcSC%dtr = MAPL_PI/180.
   gcSC%PSClatlim = gcSC%PSClatlim*gcSC%dtr

! Thermal reaction rates: Initialize from ASCII file and print summary
! --------------------------------------------------------------------
   fileName = TRIM(fnKRates)
   CALL rdkrate(gcSC, fileName, MAPL_AM_I_ROOT(), status)
   VERIFY_(status)

! Photolysis tables: Initialize from NetCDF file
! ----------------------------------------------
   fileName = TRIM(fnPhoto)
   CALL readPhotTables(fileName, status)
   VERIFY_(status)

! Grab climatological array sizes for subsequent allocation
! ---------------------------------------------------------
   CALL initClimVars(status)
   VERIFY_(status)

! Allocations for climatological fields
! -------------------------------------
   ALLOCATE(gcSC%SSG(gcSC%numSSGs), STAT=status)
   VERIFY_(status)
   ALLOCATE(gcSC%climRain(gcSC%numClimLats,km), STAT=status)
   VERIFY_(status)
   ALLOCATE(gcSC%climO3(gcSC%numClimLats,km), STAT=status)
   VERIFY_(status)
   ALLOCATE(gcSC%climOH(gcSC%numClimLats,km), STAT=status)
   VERIFY_(status)
   ALLOCATE(gcSC%climH2O(gcSC%numClimLats,km), STAT=status)
   VERIFY_(status)

   ALLOCATE(gcSC%climRain3D(i1:i2,j1:j2,km), STAT=status)
   VERIFY_(status)
   ALLOCATE(gcSC%climO33D(i1:i2,j1:j2,km), STAT=status)
   VERIFY_(status)
   ALLOCATE(gcSC%climOH3D(i1:i2,j1:j2,km), STAT=status)
   VERIFY_(status)
   ALLOCATE(gcSC%climH2O3D(i1:i2,j1:j2,km), STAT=status)
   VERIFY_(status)

   ALLOCATE(gcSC%SO4sa(gcSC%numClimLats,km), STAT=status)
   VERIFY_(status)
   ALLOCATE(gcSC%SO4sa3D(i1:i2,j1:j2,km), STAT=status)
   VERIFY_(status)

  RETURN

  CONTAINS

!-------------------------------------------------------------------------
!NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1, GEOS/DAS!
!-------------------------------------------------------------------------
!BOP
!
! !INTERFACE:

 SUBROUTINE readPhotTables(fileName, rc)

! !USES:

  IMPLICIT NONE

! !DESCRIPTION:
!
! Read tables for photolysis in StratChem ... from a NetCDF file
!
! Input parameters:
!
  CHARACTER(LEN=*), INTENT(IN) :: fileName
!
! Output parameters:
!
  INTEGER, INTENT(OUT) :: rc
!
! Restrictions:
!  ASSERT that the number of pressure layers in the dataset equals km.
!
! !REVISION HISTORY:
!  Nielsen     11 May 2012: First crack.
!
!-----------------------------------------------------------------------

  CHARACTER(LEN=ESMF_MAXSTR) :: Iam = "SC::readPhotTables"

  INTEGER :: comm, info, unit, status
  INTEGER :: dimid, i, n

  INTEGER :: length

  INTEGER, PARAMETER :: nD = 7
  CHARACTER(LEN=ESMF_MAXSTR) :: dimName(nD)= (/"nsza  ", "numO3 ", "layers", &
                                               "nlam  ", "nts   ", "nxdo  ", "aqsize" /)

  INTEGER, PARAMETER :: nV = 7
  CHARACTER(LEN=ESMF_MAXSTR) :: varName(nV)= (/"sza    ", &
                        "lambda ", "O3TAB  ",  "SDAT   ", &
                        "O2JDAT ", "XTAB   ",  "CH2O_AQ" /)
  rc = 0

  CALL ESMF_VMGet(vm, MPICOMMUNICATOR=comm, rc=status)
  VERIFY_(status)

#ifdef H5_HAVE_PARALLEL

  CALL MPI_Info_create(info, status)
  VERIFY_(status)
  CALL MPI_Info_set(info, "romio_cb_read", "automatic", status)
  VERIFY_(status)

#ifdef NETCDF_NEED_NF_MPIIO
  status = NF_OPEN_PAR(TRIM(fileName), IOR(NF_NOWRITE,NF_MPIIO), comm, info, unit)
#else
  status = NF_OPEN_PAR(TRIM(fileName), NF_NOWRITE, comm, info, unit)
#endif

#else

  IF(MAPL_AM_I_ROOT(vm)) THEN 
   status = NF_OPEN(TRIM(fileName), NF_NOWRITE, unit)

#endif

   IF(status /= NF_NOERR) THEN
    PRINT *,'Error opening file ',TRIM(fileName), status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

   DO i = 1,nD

    status = NF_INQ_DIMID(unit, TRIM(dimName(i)), dimid)
    IF(status /= NF_NOERR) THEN
     PRINT *,"Error inquiring dimension ID for ", TRIM(dimName(i)), status
     PRINT *, NF_STRERROR(status)
     VERIFY_(status)
    END IF

    status = NF_INQ_DIMLEN(unit, dimid, n)
    IF(status /= NF_NOERR) THEN
     PRINT *,"Error inquiring  dimension length for ", TRIM(dimName(i)), status
     PRINT *, NF_STRERROR(status)
    END IF

    SELECT CASE (i)
     CASE (1)
      gcSC%nsza = n
     CASE (2)
      gcSC%numO3 = n
     CASE (3)
      ASSERT_(n == km)
     CASE (4)
      gcSC%nlam = n
     CASE (5)
      gcSC%nts = n
     CASE (6)
      gcSC%nxdo = n
     CASE (7)
      gcSC%aqsize = n
     CASE DEFAULT
    END SELECT

   END DO

#ifndef H5_HAVE_PARALLEL

  END IF ! MAPL_AM_I_ROOT

  CALL MAPL_CommsBcast(vm, gcSC%nsza, 1, 0, RC=status)
  VERIFY_(status)
  CALL MAPL_CommsBcast(vm, gcSC%numO3, 1, 0, RC=status)
  VERIFY_(status)
  CALL MAPL_CommsBcast(vm, gcSC%nlam, 1, 0, RC=status)
  VERIFY_(status)
  CALL MAPL_CommsBcast(vm, gcSC%nts, 1, 0, RC=status)
  VERIFY_(status)
  CALL MAPL_CommsBcast(vm, gcSC%nxdo, 1, 0, RC=status)
  VERIFY_(status)
  CALL MAPL_CommsBcast(vm, gcSC%aqSize, 1, 0, RC=status)
  VERIFY_(status)

#endif

  ALLOCATE(gcSC%sdat(gcSC%nsza,gcSC%numo3,km,gcSC%nlam), STAT=status)
  VERIFY_(status)
  ALLOCATE(gcSC%o2jdat(gcSC%nsza,gcSC%numo3,km), STAT=status)
  VERIFY_(status)
  ALLOCATE(gcSC%o3_tab(gcSC%numo3,km), STAT=status)
  VERIFY_(status)
  ALLOCATE(gcSC%xtab(gcSC%nlam,gcSC%nxdo,gcSC%nts), STAT=status)
  VERIFY_(status)
  ALLOCATE(gcSC%sza_tab(gcSC%nsza), STAT=status)
  VERIFY_(status)
  ALLOCATE(gcSC%CH2O_aq(gcSC%aqSize), STAT=status)
  VERIFY_(status)
  ALLOCATE(gcSC%rlam(gcSC%nlam), STAT=status)
  VERIFY_(status)

#ifndef H5_HAVE_PARALLEL

  IF(MAPL_AM_I_ROOT()) THEN

#endif

   DO i = 1,nV

    status = NF_INQ_VARID(unit, TRIM(varName(i)), n)
    IF(status /= NF_NOERR) THEN
     PRINT *,"Error getting varid for ", TRIM(varName(i)), status
     PRINT *, NF_STRERROR(status)
     VERIFY_(status)
    END IF

    SELECT CASE (i)
     CASE (1)
      status = NF_GET_VAR_REAL(unit, n, gcSC%sza_tab)
     CASE (2)
      status = NF_GET_VAR_REAL(unit, n, gcSC%rlam)
     CASE (3)
      status = NF_GET_VAR_REAL(unit, n, gcSC%o3_tab)
     CASE (4)
      status = NF_GET_VAR_REAL(unit, n, gcSC%sdat)
     CASE (5)
      status = NF_GET_VAR_REAL(unit, n, gcSC%o2jdat)
     CASE (6)
      status = NF_GET_VAR_REAL(unit, n, gcSC%xtab)
     CASE (7)
      status = NF_GET_VAR_REAL(unit, n, gcSC%CH2O_aq)
     CASE DEFAULT
    END SELECT

    IF(status /= NF_NOERR) THEN
     PRINT *,"Error getting values for ", TRIM(varName(i)), status
     PRINT *, NF_STRERROR(status)
     VERIFY_(status)
    END IF

   END DO

#ifdef H5_HAVE_PARALLEL

   CALL MPI_Info_free(info, status)
   VERIFY_(status)

#else

  END IF ! MAPL_AM_I_ROOT

  length = SIZE(gcSC%sza_tab)
  CALL MPI_Bcast(gcSC%sza_tab, length, MPI_REAL, 0, comm, status)
  VERIFY_(status)

  length = SIZE(gcSC%rlam)
  CALL MPI_Bcast(gcSC%rlam, length, MPI_REAL, 0, comm, status)
  VERIFY_(status)

  length = SIZE(gcSC%o3_tab)
  CALL MPI_Bcast(gcSC%o3_tab, length, MPI_REAL, 0, comm, status)
  VERIFY_(status)

  length = SIZE(gcSC%sdat)
  CALL MPI_Bcast(gcSC%sdat, length, MPI_REAL, 0, comm, status)
  VERIFY_(status)

  length = SIZE(gcSC%o2jdat)
  CALL MPI_Bcast(gcSC%o2jdat, length, MPI_REAL, 0, comm, status)
  VERIFY_(status)

  length = SIZE(gcSC%xtab)
  CALL MPI_Bcast(gcSC%xtab, length, MPI_REAL, 0, comm, status)
  VERIFY_(status)

  CALL MAPL_CommsBcast(vm, gcSC%CH2O_aq, gcSC%aqsize, 0, RC=status)
  VERIFY_(status)

#endif

  status = NF_CLOSE(unit)
  VERIFY_(status)

  RETURN
 END SUBROUTINE readPhotTables

!-------------------------------------------------------------------------
!NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1, GEOS/DAS!
!-------------------------------------------------------------------------
!BOP
!
! !INTERFACE:

 SUBROUTINE initClimVars(rc)

! !USES:

  IMPLICIT NONE

! !DESCRIPTION:
!
! Read and broadcast the latitudes (radians)
! on which the climatological fields are stored. 
!
! Input parameters:
!
! NONE
!
! Output parameters:
!
  INTEGER, INTENT(OUT) :: rc
!
! Restrictions:
!  ASSERT that the number of pressure layers in the dataset equals km.
!
! !REVISION HISTORY:
!  Nielsen     11 May 2012: First crack.
!
!-----------------------------------------------------------------------
 CHARACTER(LEN=ESMF_MAXSTR) :: Iam = "SC::readClimVars"

  INTEGER :: comm, dimid, info, n, status, unit
  CHARACTER(LEN=ESMF_MAXSTR) :: fileName

  rc = 0

  CALL ESMF_VMGet(vm, MPICOMMUNICATOR=comm, rc=status)
  VERIFY_(status)
  fileName = TRIM(gcSC%climFileName)

#ifdef H5_HAVE_PARALLEL

  CALL MPI_Info_create(info, status)
  VERIFY_(status)
  CALL MPI_Info_set(info, "romio_cb_read", "automatic", status)
  VERIFY_(status)

#ifdef NETCDF_NEED_NF_MPIIO
  status = NF_OPEN_PAR(TRIM(fileName), IOR(NF_NOWRITE,NF_MPIIO), comm, info, unit)
#else
  status = NF_OPEN_PAR(TRIM(fileName), NF_NOWRITE, comm, info, unit)
#endif

#else

  IF(MAPL_AM_I_ROOT(vm)) THEN 
   status = NF_OPEN(TRIM(fileName), NF_NOWRITE, unit)

#endif

   IF(status /= NF_NOERR) THEN
    PRINT *,"Error opening file ",TRIM(fileName), status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

   status = NF_INQ_DIMID(unit, 'lat', dimid)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error inquiring dimid for lat", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF
   status = NF_INQ_DIMLEN(unit, dimid, gcSC%numClimLats)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error inquiring dimlen of lat", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

   status = NF_INQ_DIMID(unit, 'lev', dimid)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error inquiring dimid for lev", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF
   status = NF_INQ_DIMLEN(unit, dimid, n)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error inquiring dimlen of lev", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

   ASSERT_(n == gcSC%km)

#ifndef H5_HAVE_PARALLEL

  END IF ! MAPL_AM_I_ROOT

  CALL MAPL_CommsBcast(vm, gcSC%numClimLats, 1, 0, RC=status)
  VERIFY_(status)

#endif

  ALLOCATE(gcSC%climLatRad(gcSC%numClimLats), STAT=status)
  VERIFY_(status)

#ifndef H5_HAVE_PARALLEL

  IF(MAPL_AM_I_ROOT()) THEN

#endif

   status = NF_INQ_VARID(unit, 'lat', n)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error getting varid for lat", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

   status = NF_GET_VAR_REAL(unit, n, gcSC%climLatRad)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error getting values for climLatRad", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

#ifdef H5_HAVE_PARALLEL

   CALL MPI_Info_free(info, status)
   VERIFY_(status)

#else

  END IF ! MAPL_AM_I_ROOT

  CALL MAPL_CommsBcast(vm, gcSC%climLatRad, SIZE(gcSC%climLatRad), 0, RC=status)
  VERIFY_(status)

#endif

  status = NF_CLOSE(unit)
  VERIFY_(status)

  RETURN
 END SUBROUTINE initClimVars

  END SUBROUTINE SC_GridCompInitialize

!-------------------------------------------------------------------------
!NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1, GEOS/DAS!
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  SC_GridCompRun --- Stratospheric chemistry mechanism
!
! !INTERFACE:
!

   SUBROUTINE SC_GridCompRun ( gcSC, w_c, impChem, expChem, nymd, nhms, &
                               tdt, rc )

! !USES:

  IMPLICIT none

! !INPUT/OUTPUT PARAMETERS:

   TYPE(SC_GridComp), INTENT(inout) :: gcSC   ! Grid Component
   TYPE(Chem_Bundle), INTENT(inout) :: w_c    ! Chemical tracer fields   

! !INPUT PARAMETERS:

   TYPE(ESMF_State), INTENT(inout) :: impChem ! Import State
   INTEGER, INTENT(IN) :: nymd, nhms	      ! time
   REAL,    INTENT(IN) :: tdt		      ! chemical timestep (secs)

! !OUTPUT PARAMETERS:

   TYPE(ESMF_State), INTENT(inout) :: expChem   ! Export State
   INTEGER, INTENT(out) ::  rc                  ! Error return code:
                                                !  0 - all is well
                                                !  1 -
 
! !DESCRIPTION: This routine implements the so-called SC Driver. That 
!               is, adds chemical tendencies to each of the constituents
!
! !REVISION HISTORY:
!
!  18Sep2003 da Silva  First crack.
!  24Jan2O05 Nielsen   Implementation of Code 916 chemistry
!  11May2012 Nielsen   Capability for FV cubed [Ganymed-1_0_UNSTABLE]
!
!EOP
!-------------------------------------------------------------------------
#include "mpif.h"

   CHARACTER(LEN=*), PARAMETER :: Iam = 'SC_GridCompRun'
   TYPE(ESMF_VM) :: vm

! Input fields from fvGCM
! -----------------------
   REAL, POINTER, DIMENSION(:,:,:) ::  specHum, T, qctot, rhoa, ple
   REAL, POINTER, DIMENSION(:,:)   ::  cellArea

! Exports not part of internal state
! ----------------------------------
   REAL, POINTER, DIMENSION(:,:,:) ::  natsad,   icesad,   o3ppmv,    ozone
   REAL, POINTER, DIMENSION(:,:,:) ::   dOxdt,     dQdt
   REAL, POINTER, DIMENSION(:,:) ::    n2oflx,   ch4flx,   f11flx,  f113flx
   REAL, POINTER, DIMENSION(:,:) ::    f12flx,  ccl4flx,  hcfcflx,   mcfflx
   REAL, POINTER, DIMENSION(:,:) ::  ch3clflx, ch3brflx, h1301flx, h1224flx
   REAL, POINTER, DIMENSION(:,:) ::    szarad,  scbasek
   REAL, POINTER, DIMENSION(:,:,:) :: SO4SAimp  ! imported sulfate surface area density m2 m-3

! Local
! -----
   INTEGER, PARAMETER :: DBL = KIND(0.00D+00)
   INTEGER, PARAMETER :: FROM_BUNDLE = 1, TO_BUNDLE = -1
   REAL, PARAMETER :: MIN_VALUE = 1.00E-25

   TYPE(ESMF_Clock) :: SCClock
   TYPE(ESMF_Time)  :: timeNow
   TYPE(ESMF_TimeInterval) :: interval
   INTEGER :: ccyy, mm, dd, hh, m, s

   INTEGER :: i, i1, i2, ic, iXj, im
   INTEGER :: j, j1, j2, jm, k, km
   INTEGER :: status

   INTEGER :: examineDt
   INTEGER :: iterCount
   INTEGER :: k1Strat, kRev
   INTEGER :: nStratBegin, nStratEnd, numSpecies
   INTEGER :: passNumber, requiredPasses

   REAL :: dt, qmax, qmin, requiredDt, szan, tropp, tx

   LOGICAL :: dayTime
   LOGICAL :: startOver
   LOGICAL :: tropo(gcSC%km)

! Profiles
! --------
   REAL :: cellDepth(gcSC%km), cellVolume(gcSC%km)
   REAL :: climRainProfile(gcSC%km), climO3Profile(gcSC%km)
   REAL :: climH2OProfile(gcSC%km), climOHProfile(gcSC%km)
   REAL :: denssts(gcSC%km), edgePress(0:gcSC%km)
   REAL :: gcrProfile(gcSC%km), H2OcProfile(gcSC%km), HNO3cProfile(gcSC%km)
   REAL :: kel(gcSC%km), midPress(gcSC%km)
   REAL :: NOColumn(gcSC%km), NOProfile(gcSC%km), numDens(gcSC%km)
   REAL :: O2Column(gcSC%km), O3Column(gcSC%km), O3Profile(gcSC%km)
   REAL :: rmedice(gcSC%km), rmednat(gcSC%km), rmedsts(gcSC%km)
   REAL :: SO4saProfile(gcSC%km), vfall(gcSC%km)

! Reaction rates
! --------------
   REAL(KIND=DBL) :: aj(gcSC%numphoto), ak(gcSC%km,gcSC%numreacs)

! Initial state holding arrays
! ----------------------------
   REAL(KIND=DBL), ALLOCATABLE :: speciesBase(:)
   REAL(KIND=DBL), ALLOCATABLE :: ratesBase(:)

! Species, etc.
! -------------
   REAL(KIND=DBL) ::     h2o,    h2oc,      ox,     nox,    hno3
   REAL(KIND=DBL) ::    n2o5,  ho2no2,  clono2,     clx,     hcl
   REAL(KIND=DBL) ::    hocl,    h2o2,     brx,     n2o,     cl2
   REAL(KIND=DBL) ::    oclo,    brcl,     hbr,  brono2,     ch4
   REAL(KIND=DBL) ::    hobr,  ch3ooh,      co,   hno3c,     f11
   REAL(KIND=DBL) ::     f12,    f113,    hcfc,  hcfc22,    ccl4
   REAL(KIND=DBL) :: ch3ccl3,   ch3cl,   ch3br,   h1301,  h12_24
   REAL(KIND=DBL) ::      o3,     o3p,     o1d,       n,      no
   REAL(KIND=DBL) ::     no2,     no3,       h,      oh,     ho2
   REAL(KIND=DBL) ::      cl,     clo,     bro,      br,   cl2o2
   REAL(KIND=DBL) ::    ch2o,   ch3o2,     ch3,    ch3o,     cho
   REAL(KIND=DBL) ::   saice,   sanat, box_ro3ox

! Estimated species
! -----------------
   REAL(KIND=DBL) ::    o3pe,     o3e,     oxe,      ne,     noe,    no2e
   REAL(KIND=DBL) ::    no3e,    noxe,   n2o5e,   h2o2e,   hno3e, ho2no2e
   REAL(KIND=DBL) ::    hcle,   hocle,     cle,    cloe,   ocloe,  cl2o2e
   REAL(KIND=DBL) :: clono2e,    clxe,     bre,    broe, brono2e,    brxe
   REAL(KIND=DBL) ::    hbre,   hobre

! Ratios
! ------
   REAL(KIND=DBL) ::    ro3po3,    ro1do3,     ro3ox,      rnno
   REAL(KIND=DBL) ::    rnono2,   rno3no2,   rno2nox,    rclclo
   REAL(KIND=DBL) ::   rcloclx,  rocloclo, rcl2o2clo,   rbrobrx
   REAL(KIND=DBL) ::   rbrnbrx

! Production and loss
! -------------------
   REAL(KIND=DBL) ::  	    po3,        lo3,      ph2o2,      lh2o2,      phno3
   REAL(KIND=DBL) ::      lhno3,   phno3het,      pn2o5,      ln2o5,   ln2o5het
   REAL(KIND=DBL) ::    pho2no2,    lho2no2,       pnox,       lnox,     lnoxsq
   REAL(KIND=DBL) ::     pnoxcl,     lnoxcl,       pclx,       lclx,    pclono2
   REAL(KIND=DBL) ::    lclono2, lclono2het,       phcl,       lhcl,      phocl
   REAL(KIND=DBL) ::      lhocl,    pbrono2, lbrono2het,       pbrx,       lbrx
   REAL(KIND=DBL) ::       phbr,       lhbr,      phobr,      lhobr,       pno2
   REAL(KIND=DBL) ::       lno2,       pno3,       lno3,      poclo,      pnoxa
   REAL(KIND=DBL) ::      lnoxa,       pcl2,      pbrcl,     lo3hox,     lo3nox
   REAL(KIND=DBL) ::    lo3oxsq,      lo3cl,     lo3cly,      lo3oh,     lo3brx
   REAL(KIND=DBL) ::     lo3chx

! Grid specs from Chem_Bundle%grid
! --------------------------------
  rc = 0
  i1 = gcSC%i1
  i2 = gcSC%i2
  im = gcSC%im

  j1 = gcSC%j1
  j2 = gcSC%j2
  jm = gcSC%jm

  km = gcSC%km

  iXj = (i2-i1+1)*(j2-j1+1)

! Grab the virtual machine
! ------------------------
  CALL ESMF_VMGetCurrent(vm, RC=status)
  VERIFY_(status)

! Grab some calender elements
! ---------------------------
  ccyy = nymd/10000
  mm = (nymd-ccyy*10000)/100
  dd = nymd-ccyy*10000-mm*100
  hh = nhms/10000
  m = (nhms-hh*10000)/100
  s = nhms-hh*10000-m*100

! Create and set the clock
! ------------------------
  CALL ESMF_TimeSet(timeNow, YY=ccyy, MM=mm, DD=dd, H=hh, M=m, S=s, RC=status)
  VERIFY_(status)
  CALL ESMF_TimeIntervalSet(interval, S=s, RC=status)
  VERIFY_(status)
  SCClock = ESMF_ClockCreate(NAME="SCClock", TIMESTEP=interval, STARTTIME=timeNow, RC=status)
  VERIFY_(status)
  CALL ESMF_ClockSet(SCClock, CURRTIME=timeNow, RC=status)
  VERIFY_(status)

! Import state pointers
! ---------------------
  CALL MAPL_GetPointer(impChem,  specHum,	'Q', RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(impChem,        T,	'T', RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(impChem,      ple,     'PLE', RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(impChem,    qctot,   'QCTOT', RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(impChem,     rhoa, 'AIRDENS', RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(impChem, cellArea,    'AREA', RC=status)
  VERIFY_(status)
   
  IF(gcSC%verbose) THEN
   ic = gcSC%irO3Ox
   CALL pmaxmin('SC: TROPP ', w_c%qa(ic)%data3d(:,:,km), qmin, qmax, iXj, 1, 0.01 )
   CALL pmaxmin('SC:	 T ', T,	qmin, qmax, iXj, km,   1. )
   CALL pmaxmin('SC:	 Q ', specHum,  qmin, qmax, iXj, km,   1. )
   CALL pmaxmin('SC: QCTOT ', qctot,	qmin, qmax, iXj, km,   1. )
   CALL pmaxmin('SC:   RHO ', rhoa,	qmin, qmax, iXj, km,   1. )
   CALL pmaxmin('SC:   PLE ', ple,	qmin, qmax, iXj, km+1, 1. )
   CALL pmaxmin('SC:  AREA ', cellArea, qmin, qmax, iXj, 1,    1. )
  END IF

! If importing SO4 surface area from GOCART::SU then query the import state. Units: m^{2} m^{-3}.
! -----------------------------------------------------------------------------------------------
  IF(gcSC%importSulfSA) THEN
   CALL MAPL_GetPointer(impChem,  SO4SAimp,  'SO4SAREAvolc', RC=status)
   VERIFY_(status)
   IF(gcSC%verbose) THEN
    CALL pmaxmin('SC: SO4SAvolc ', SO4SAimp, qmin, qmax, iXj, km,    1. )
   END IF
  END IF

! Export state pointers
! ---------------------
  CALL MAPL_GetPointer(expChem,  szarad,   'SZARAD', RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem, scbasek,  'SCBASEK', RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem,  n2oflx,   'N2OFLX', ALLOC=.TRUE., RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem,  ch4flx,   'CH4FLX', ALLOC=.TRUE., RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem,  f11flx,   'F11FLX', ALLOC=.TRUE., RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem,  f12flx,   'F12FLX', ALLOC=.TRUE., RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem, f113flx,  'F113FLX', ALLOC=.TRUE., RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem, hcfcflx,  'HCFCFLX', ALLOC=.TRUE., RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem, ccl4flx,  'CCL4FLX', ALLOC=.TRUE., RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem,  mcfflx,   'MCFFLX', ALLOC=.TRUE., RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem,ch3clflx, 'CH3CLFLX', ALLOC=.TRUE., RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem,ch3brflx, 'CH3BRFLX', ALLOC=.TRUE., RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem,h1301flx, 'H1301FLX', ALLOC=.TRUE., RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem,h1224flx, 'H1224FLX', ALLOC=.TRUE., RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem,  o3ppmv,   'O3PPMV', RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem,   ozone,       'O3', RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem,   dOxdt,  'OX_TEND', RC=status)
  VERIFY_(status)
  CALL MAPL_GetPointer(expChem,    dQdt, 'H2O_TEND', RC=status)
  VERIFY_(status)

! PSC preliminaries
! -----------------
  IF(gcSC%doPSCs) THEN
   CALL MAPL_GetPointer(expChem, natsad, 'NATSAD', ALLOC=.TRUE., RC=status)
   VERIFY_(status)
   natsad(:,:,:) = 0.00
   CALL MAPL_GetPointer(expChem, icesad, 'ICESAD', ALLOC=.TRUE., RC=status)
   VERIFY_(status)
   icesad(:,:,:) = 0.00
  END IF

! Initialize tendency diagnostics
! -------------------------------
  IF(ASSOCIATED(dQdt)) dQdt(i1:i2,j1:j2,1:km) = specHum(i1:i2,j1:j2,1:km)
  IF(ASSOCIATED(dOxdt)) dOxdt(i1:i2,j1:j2,1:km) = w_c%qa(gcSC%iOx)%data3d(i1:i2,j1:j2,1:km)

! Periodically update the climatological boundary conditions
! ----------------------------------------------------------
  BCUpdates: IF(gcSC%BCnymd /= nymd) THEN

   IF(MAPL_AM_I_ROOT()) PRINT *,TRIM(Iam)//": Updating climatologies"

   CALL readClimVars(status)
   VERIFY_(status)

   CALL mapToGrid(gcSC%climRain, gcSC%climRain3D, "Rain", status)
   VERIFY_(status)
   CALL mapToGrid(gcSC%climO3, gcSC%climO33D, "O3", status)
   VERIFY_(status)
   CALL mapToGrid(gcSC%climOH, gcSC%climOH3D, "OH", status)
   VERIFY_(status)
   CALL mapToGrid(gcSC%climH2O, gcSC%climH2O3D, "H2O", status)
   VERIFY_(status)
   CALL mapToGrid(gcSC%SO4sa, gcSC%SO4sa3D, "SO4sa", status)
   VERIFY_(status)

   gcSC%BCnymd = nymd
  END IF BCUpdates

! Location of SC species in w_c%qa
! --------------------------------
  nStratBegin =  w_c%reg%i_SC
  nStratEnd =  w_c%reg%j_XX
  numSpecies = nStratEnd-nStratBegin+1

! Except for the ozone-to-odd oxygen ratio, constituents should be non-negative
! -----------------------------------------------------------------------------
  DO i = nStratBegin,nStratEnd-1
   WHERE(w_c%qa(i)%data3d(i1:i2,j1:j2,1:km) <= MIN_VALUE ) &
         w_c%qa(i)%data3d(i1:i2,j1:j2,1:km)  = MIN_VALUE
  END DO
  WHERE(specHum(i1:i2,j1:j2,1:km) <= 1.00E-25) specHum(i1:i2,j1:j2,1:km) = MIN_VALUE

! Cycle though the grid cells on this processor
! ---------------------------------------------
  Latitude: DO j = j1,j2
   Longitude: DO i = i1,i2

! Solar zenith angle (radians) and daytime indicator.
! ---------------------------------------------------
    CALL sza(gcSC%dayOfYear,gcSC%lonRad(i,j),gcSC%latRad(i,j),szan)
    dayTime = (szan <= gcSC%szaCutoff*MAPL_PI/180.00)
    IF(ASSOCIATED(szarad)) szarad(i,j) = szan

! Profiles. Temperature [K] and pressure [hPa]
! --------------------------------------------
    kel(1:km) = T(i,j,km:1:-1)
    edgePress(0:km) = ple(i,j,km:0:-1)*0.01
    midPress(1:km) = 0.50*(edgePress(0:km-1)+edgePress(1:km))

! Number density [cm^-3]
! ----------------------
    numDens(1:km) = rhoa(i,j,km:1:-1)*MAPL_AVOGAD*1.00E-06/MAPL_AIRMW

! Cell depth [cm] and volume [m^3]
! --------------------------------
    cellDepth(1:km) = 100.00*MAPL_RGAS*kel(1:km)*LOG(edgePress(0:km-1)/edgePress(1:km))/MAPL_GRAV
    cellVolume(1:km) = 0.01*cellDepth(1:km)*cellArea(i,j)

! A logical flag to indicate presence in the troposphere.
! -------------------------------------------------------
    tropo(:) = .FALSE.
    ic = gcSC%irO3Ox
    tropp = w_c%qa(ic)%data3d(i,j,km)*0.01
    WHERE(edgePress(1:km) >= tropp) tropo(1:km) = .TRUE.

! Find layer number at base of stratosphere, and verify tropo=.TRUE. at each layer below
! --------------------------------------------------------------------------------------
    k1Strat = km
    DO k = km,1,-1
     IF(tropo(k)) EXIT
     k1Strat = k
    END DO
    tropo(1:k1Strat-1) = .TRUE.
    IF(ASSOCIATED(scbasek)) scbasek(i,j) = km-k1Strat+1

! Ozone and NO number densities
! -----------------------------
    O3profile(1:km) = w_c%qa(gcSC%iO3)%data3d(i,j,km:1:-1)*numDens(1:km)
    NOprofile(1:km) = w_c%qa(gcSC%iNO)%data3d(i,j,km:1:-1)*numDens(1:km)

! Overhead ozone and molecular oxygen
! -----------------------------------
    CALL O3Colum(km,edgePress,midPress,numDens,O3Profile,NOProfile,O3Column,O2Column,NOColumn)

! Condensed water [mole fraction] for SUBROUTINE sediment
! -------------------------------------------------------
    H2OcProfile(1:km) = qctot(i,j,km:1:-1)*MAPL_AIRMW/MAPL_H2OMW

! Profiles of the climatological fields
! -------------------------------------
    climRainProfile(1:km) = gcSC%climRain3D(i,j,km:1:-1)
    climO3Profile(1:km) = gcSC%climO33D(i,j,km:1:-1)
    climOHProfile(1:km) = gcSC%climOH3D(i,j,km:1:-1)
    climH2OProfile(1:km) = gcSC%climH2O3D(i,j,km:1:-1)
    SO4saProfile(1:km) = gcSC%SO4sa3D(i,j,km:1:-1)

! Utilize Sulfate SA imported from GOCART::SU.
! Convert units from  m^{2} m^{-3} to cm^{2} cm^{-3}
! --------------------------------------------------
    IF(gcSC%importSulfSA) THEN
     IF(gcSC%addSAclim) THEN
      SO4saProfile(1:km) = SO4saProfile(1:km) + 0.01 * SO4SAimp(i,j,1:km)
     ELSE
      SO4saProfile(1:km) =                      0.01 * SO4SAimp(i,j,1:km)
     END IF
    END IF

    IF(gcSC%useSolCyc) THEN
     gcrProfile(1:km) = 0.00
!JEN gcrProfile(1:km) = gcSC%climGCR3D(i,j,km:1:-1) --- Awaits implementation updates JEN!
    ELSE
     gcrProfile(1:km) = 0.00
    END IF

! Calculate thermal reaction rates. SUBROUTINE khet3d
! recalculates reactions numbers 110-115 when PSCS are enabled.
! -------------------------------------------------------------
    CALL krates(gcSC,km,numDens,kel,ak)

! At each layer ...
! -----------------
    Layer: DO k=1,km
     kRev = km-k+1

! Place the species in reverse vertical order and convert to number density.
! --------------------------------------------------------------------------
     h2o = specHum(i,j,kRev)*numDens(k)*MAPL_AIRMW/MAPL_H2OMW
     h2oc =  qctot(i,j,kRev)*numDens(k)*MAPL_AIRMW/MAPL_H2OMW
     CALL swapSpecies(FROM_BUNDLE,status)
     VERIFY_(status)

! Initializations to complete a gcSC%spinup (rarely used):
!   1. Constrain HCl to be at least 1.e-11
!   2. Ox and O3 are initialized identically.
!   3. ClO and NO2 are used to determine ak(32).
! --------------------------------------------------------
     IF(gcSC%spinup) THEN
      hcl = MAX(hcl,1.00E-11*numDens(k))
       o3 = MAX(ox,1.00*numDens(k))
      clo = 0.00
      no2 = 0.00
     END IF

! Fix the surface source gases and infer the surface flux [kg s^{-1}]
! -------------------------------------------------------------------
     IF(k == 1) THEN
      CALL SSGSfcFlx(cellVolume(1),numDens(1),n2o,ch4,f11,f12,f113,hcfc, &
                     ccl4,ch3ccl3,ch3cl,ch3br,h1301,h12_24,hcfc22,status)
      VERIFY_(status)
     END IF

! Update photolysis rates.
! ------------------------
     CALL jcalc4(k,szan,O3Column(k),midPress(k),kel(k),aj,gcSC)
     CALL meso_phot(k,km,0,NOColumn,O3Column,O2Column,szan,aj,numDens,gcSC)

! Lightning source of NOx between 20 S and 20 N in troposphere.
! Note: The rate can be adjusted at run time in SC_GridComp.rc.
! -------------------------------------------------------------
     IF(gcSC%latRad(i,j) >= -20.00*gcSC%dtr .AND. gcSC%latRad(i,j) <= 20.00*gcSC%dtr) THEN
      IF(k >= 5 .AND. tropo(k)) nox = nox+gcSC%NOxTRW*tdt
     END IF

! Store initial values in case time step is changed in the main chemistry loop.
! -----------------------------------------------------------------------------
     ALLOCATE(ratesBase(gcSC%numphoto+gcSC%numreacs),STAT=status)
     VERIFY_(status)
     ratesBase(1:gcSC%numphoto) = aj(1:gcSC%numphoto)
     ratesBase(gcSC%numphoto+1:gcSC%numphoto+gcSC%numreacs) = ak(k,1:gcSC%numreacs)
     ALLOCATE(speciesBase(numSpecies+1),STAT=status)
     VERIFY_(status)
     CALL storeBaseChem(ox,nox,hno3,n2o5,ho2no2,clono2,clx,hcl,hocl,h2o2,brx,n2o,cl2, &
            oclo,brcl,hbr,brono2,ch4,hobr,ch3ooh,co,h2o,hno3c,h2oc,f11,f12,f113,hcfc, &
              hcfc22,ccl4,ch3ccl3,ch3cl,ch3br,h1301,h12_24,o3,o3p,o1d,n,no,no2,no3,h, &
              oh,ho2,cl,clo,bro,br,cl2o2,ch2o,ch3o2,box_ro3ox,speciesBase,numSpecies)

! Variables needed for chemistry time step length analysis.  These are changed
! within the while-do loop if the given time step is too long to ensure stability.
! --------------------------------------------------------------------------------
     dt = tdt
     requiredDt = tdt
     requiredPasses = 1
     passNumber = 1
     examineDt = 1
     startOver = .FALSE.

! For the current cell ...
! ------------------------
     ThisCell: DO WHILE(passNumber <= requiredPasses)
      
! Run the mechanism
! -----------------
      Chemistry: DO

       IF(startOver) THEN
        aj(1:gcSC%numphoto) = ratesBase(1:gcSC%numphoto)
        ak(k,1:gcSC%numreacs) = ratesBase(gcSC%numphoto+1:gcSC%numphoto+gcSC%numreacs)
        CALL getBaseChem(ox,nox,hno3,n2o5,ho2no2,clono2,clx,hcl,hocl,h2o2,brx,n2o,cl2, &
             oclo,brcl,hbr,brono2,ch4,hobr,ch3ooh,co,h2o,hno3c,h2oc,f11,f12,f113,hcfc, &
               hcfc22,ccl4,ch3ccl3,ch3cl,ch3br,h1301,h12_24,o3,o3p,o1d,n,no,no2,no3,h, &
               oh,ho2,cl,clo,bro,br,cl2o2,ch2o,ch3o2,box_ro3ox,speciesBase,numSpecies)
        passNumber = 1
       END IF

       startOver = .FALSE.
       CALL chemStage1(status)
       VERIFY_(status)
       IF(startOver) CYCLE

       CALL chemStage2(status)
       VERIFY_(status)
       IF(startOver) CYCLE
       EXIT

      END DO Chemistry

! Parameterized scavenging of species in the troposphere
! ------------------------------------------------------
      IF(k < k1Strat) THEN
           tx = 1.00-climRainProfile(k)*tdt
          hcl =    hcl*tx
       clono2 = clono2*tx
         hno3 =   hno3*tx
         h2o2 =   h2o2*tx
       ho2no2 = ho2no2*tx
          hbr =    hbr*tx
      END IF

! Relax tropospheric ozone to climatology with a time constant of 5 days.
! -----------------------------------------------------------------------
      IF(k < k1Strat) ox = ox+(climO3Profile(k)*numDens(k)-ox)*tdt/(5.00*86400.00)

! End of loop to update mixing ratios in the current cell
! -------------------------------------------------------
      passNumber = passNumber+1

     END DO ThisCell

! Housekeeping
! ------------
     DEALLOCATE(ratesBase,STAT=status)
     VERIFY_(status)
     DEALLOCATE(speciesBase,STAT=status)
     VERIFY_(status)

! Update the internal state bundle
! --------------------------------
     CALL swapSpecies(TO_BUNDLE,status)
     VERIFY_(status)

! Update specific humidity
! ------------------------
     specHum(i,j,kRev) = h2o*MAPL_H2OMW/(numDens(k)*MAPL_AIRMW)

! End of chemistry at layer k of the current (i,j).
! -------------------------------------------------
    END DO Layer

! Sedimentation of HNO3c PSC particles. MOIST handles H2O sedimentation.
! ----------------------------------------------------------------------
    IF(gcSC%doPSCs .AND. gcSC%doSediment) THEN
     HNO3cProfile(1:km) = w_c%qa(gcSC%iHNO3c)%data3d(i,j,km:1:-1)
     CALL sediment(km,tdt,cellDepth,numDens,kel,HNO3CProfile,H2OcProfile, &
                   rmedsts,rmednat,rmedice,denssts,vfall,gcSC)
     w_c%qa(gcSC%iHNO3c)%data3d(i,j,km:1:-1) = HNO3cProfile(1:km)
    END IF

! Move to next cell on this processor
! -----------------------------------
   END DO Longitude
  END DO Latitude

! This switch (rarely used) is set to false after one pass.
! ---------------------------------------------------------
  gcSC%spinup = .FALSE.

!  ------------------
!  Fill export states
!  ------------------

! ---------------------------------------------------------------------
!
!     NAME   STATE    Units             Comments
! -------- --------   ----------------- --------------------------------
!       OX  Internal  mol/mol           Reqired name for ANALYSIS bundle
!       O3  Export    kg/kg             O3CHEM(vmr)*48/28.97
!   03PPMV  Export    ppmv              O3CHEM(vmr)*1.00E+06
!  OX_TEND  Export    mol/(mol s)       Odd-oxygen tendency
! H2O_TEND  Export    kg/(kg s)         Specific humidity tendency
!
! ---------------------------------------------------------------------

!  Ozone mass mixing ratio
!  -----------------------
   IF(ASSOCIATED(ozone)) &
    ozone(i1:i2,j1:j2,1:km) = w_c%qa(gcSC%iO3)%data3d(i1:i2,j1:j2,1:km) * MAPL_O3MW/MAPL_AIRMW 

!  Ozone mole fraction in ppm
!  --------------------------
   IF(ASSOCIATED(o3ppmv)) &
    o3ppmv(i1:i2,j1:j2,1:km) = w_c%qa(gcSC%iO3)%data3d(i1:i2,j1:j2,1:km) * 1.00E+06

!  Odd-oxygen tendency, volume mixing ratio s^{-1}
!  -----------------------------------------------
   IF(ASSOCIATED(dOxdt)) dOxdt(i1:i2,j1:j2,1:km) = &
    (w_c%qa(gcSC%iOx)%data3d(i1:i2,j1:j2,1:km) - dOxdt(i1:i2,j1:j2,1:km))/tdt

! Specific humidity tendency due to chemistry
! -------------------------------------------
   IF(ASSOCIATED(dQdt)) &
    dQdt(i1:i2,j1:j2,1:km) = (specHum(i1:i2,j1:j2,1:km)-dQdt(i1:i2,j1:j2,1:km))/tdt

! Kill the clock
! --------------
   CALL ESMF_ClockDestroy(SCClock, RC=status)
   VERIFY_(status)

   RETURN

 CONTAINS

!-------------------------------------------------------------------------
!NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1, GEOS/DAS!
!-------------------------------------------------------------------------
!BOP
!
! !INTERFACE:

 SUBROUTINE readClimVars(rc)

! !USES:

  IMPLICIT NONE

! !DESCRIPTION:
!
! Read, interpolate in time, and broadcast for time-dependent boundary conditions
!
! Input parameters:
!
! NONE
!
! Output parameters:
!
  INTEGER, INTENT(OUT) :: rc
!
! Restrictions:
!  ASSERT that the number of pressure layers in the dataset equals km.
!
! !REVISION HISTORY:
!  Nielsen     11 May 2012: First crack.
!  Nielsen     15 Jan 2015: Separate year-number shifting for GHG and ODS
!
! Notes
!  MAPL_ClimInterpFac returns m1=12 and m2=1 during the first half of January and
!  the second half of December.
!
!-----------------------------------------------------------------------

  TYPE(ESMF_Time) :: timeNow
  CHARACTER(LEN=ESMF_MAXSTR) :: Iam = "SC::readClimVars"

  INTEGER :: beginYear, ccyy, ccyyAdjusted
  INTEGER :: comm, dimid, info, unit, status, start(3), cnt(3)
  INTEGER :: i, j, k, kLast
  INTEGER :: m1, m2, mo, n, numSSGmonths, numSO4months
  INTEGER :: YrAdj
  
  REAL :: fac, q, r, s

  REAL, ALLOCATABLE :: x(:,:,:)
  REAL, ALLOCATABLE :: zm1(:,:),zm2(:,:)

  CHARACTER(LEN=ESMF_MAXSTR) :: fileName, varName

  CHARACTER(LEN=ESMF_MAXSTR) :: SSGName(16) = (/    "CO2     ", &
    "CH4     ", "N2O     ", "CFC11   ", "CFC12   ", "CFC113  ", & 
    "CCL4    ", "CH3CCL3 ", "HCFC22  ", "HCFC141B", "HCFC142B", & 
    "H1211   ", "H1301   ", "H2402   ", "CH3BR   ", "CH3CL   " /)

  CHARACTER(LEN=ESMF_MAXSTR) :: zmClimName(5) = (/ "RAIN ", "OH   ", &
                                          "H2O  ", "O3   ", "SO4SA"/)

  rc = 0

  CALL ESMF_VMGet(vm, MPICOMMUNICATOR=comm, rc=status)
  VERIFY_(status)

  CALL ESMF_ClockGet(SCClock, CURRTIME=timeNow, RC=STATUS)
  VERIFY_(STATUS)
  CALL ESMF_TimeGet(timeNow, YY=ccyy, MM=mo, RC=STATUS)
  VERIFY_(STATUS)
  CALL MAPL_ClimInterpFac(SCClock, m1, m2, fac, RC=status)
  VERIFY_(status) 

  IF(m1 > m2 .AND. mo ==  1) m1 =  0
  IF(m1 > m2 .AND. mo == 12) m2 = 13

  fileName = TRIM(gcSC%climFileName)

#ifdef H5_HAVE_PARALLEL

  CALL MPI_Info_create(info, status)
  VERIFY_(status)
  CALL MPI_Info_set(info, "romio_cb_read", "automatic", status)
  VERIFY_(status)

#ifdef NETCDF_NEED_NF_MPIIO
  status = NF_OPEN_PAR(TRIM(fileName), IOR(NF_NOWRITE,NF_MPIIO), comm, info, unit)
#else
  status = NF_OPEN_PAR(TRIM(fileName), NF_NOWRITE, comm, info, unit)
#endif

#else

  IF(MAPL_AM_I_ROOT()) THEN 
   status = NF_OPEN(TRIM(fileName), NF_NOWRITE, unit)

#endif

   IF(status /= NF_NOERR) THEN
    PRINT *,"Error opening file ",TRIM(fileName), status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

#ifndef H5_HAVE_PARALLEL

  END IF ! MAPL_AM_I_ROOT

  CALL MAPL_CommsBcast(vm, gcSC%numClimLats, 1, 0, RC=status)
  VERIFY_(status)

#endif

#ifndef H5_HAVE_PARALLEL

  IF(MAPL_AM_I_ROOT()) THEN

#endif

   status = NF_INQ_DIMID(unit, 'time_SSG', dimid)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error inquiring dimid for time_SSG", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

   status = NF_INQ_DIMLEN(unit, dimid, numSSGmonths)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error inquiring dimlen of time_SSG", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

   status = NF_INQ_VARID(unit, 'time_SSG', n)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error getting varid for time_SSG", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

   status = NF_GET_ATT_INT(unit, n, "begin_date", i)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error getting begin_date for time_SSG", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

   beginYear = i/10000

   DO k = 1,gcSC%numSSGs

    IF(k <= gcSC%numGHGs) THEN
     YrAdj = gcSC%GHGYrAdj
    ELSE
     YrAdj = gcSC%ODSYrAdj
    ENDIF

    IF(YrAdj < -1000) THEN
     ccyyAdjusted = ABS(YrAdj)
    ELSE
     ccyyAdjusted = ccyy+YrAdj
    END IF

    i = m1+(ccyyAdjusted-beginYear)*12
    IF(i <            1) i = 1
    IF(i > numSSGmonths) i = numSSGmonths
    j = m2+(ccyyAdjusted-beginYear)*12
    IF(j <            1) j = 1
    IF(j > numSSGmonths) j = numSSGmonths
   
    IF(MAPL_AM_I_ROOT() .AND. k == gcSC%numGHGs ) THEN
     PRINT *,TRIM(Iam)//":"
     PRINT *," GHGs updated to ",ccyyAdjusted,".  Indicies: ",i,j
    END IF
    IF(MAPL_AM_I_ROOT() .AND. k == gcSC%numSSGs ) THEN
     PRINT *," ODSs updated to ",ccyyAdjusted,".  Indicies: ",i,j
    END IF

    varName = TRIM(SSGName(k))

    status = NF_INQ_VARID(unit, TRIM(varName), n)
    IF(status /= NF_NOERR) THEN
     PRINT *,"Error getting varid for ",TRIM(varName), status
     PRINT *, NF_STRERROR(status)
     VERIFY_(status)
    END IF

    status = NF_GET_VARA_REAL(unit, n, i, 1, r)
    IF(status /= NF_NOERR) THEN
     PRINT *,"Error getting first month for ",TRIM(varName), status
     PRINT *, NF_STRERROR(status)
     VERIFY_(status)
    END IF
    
    status = NF_GET_VARA_REAL(unit, n, j, 1, s)
    IF(status /= NF_NOERR) THEN
     PRINT *,"Error getting second month for ",TRIM(varName), status
     PRINT *, NF_STRERROR(status)
     VERIFY_(status)
    END IF

    gcSC%SSG(k) = r*fac+s*(1-fac)

   END DO

   status = NF_INQ_VARID(unit, 'lat', n)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error getting varid for lat", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

   status = NF_GET_VAR_REAL(unit, n, gcSC%climLatRad)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error getting values for climLatRad", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

   start(1) = 1
   start(2) = 1
   cnt(1) = gcSC%numClimLats
   cnt(2) = gcSC%km
   cnt(3) = 1

   ALLOCATE(zm1(gcSC%numClimLats,gcSC%km), STAT=status)
   VERIFY_(status)
   ALLOCATE(zm2(gcSC%numClimLats,gcSC%km), STAT=status)
   VERIFY_(status)

   status = NF_INQ_DIMID(unit, 'time_SO4', dimid)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error inquiring dimid for time_SO4", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

   status = NF_INQ_DIMLEN(unit, dimid, numSO4months)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error inquiring dimlen of time_SO4", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

   status = NF_INQ_VARID(unit, 'time_SO4', n)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error getting varid for time_SO4", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

   status = NF_GET_ATT_INT(unit, n, "begin_date", i)
   IF(status /= NF_NOERR) THEN
    PRINT *,"Error getting begin_date for time_SO4", status
    PRINT *, NF_STRERROR(status)
    VERIFY_(status)
   END IF

   beginYear = i/10000
   
   IF(gcSC%SO4saYr == 0) THEN
    ccyyAdjusted = ccyy
   ELSE
    ccyyAdjusted = gcSC%SO4saYr
   END IF

   i = m1+(ccyyAdjusted-beginYear)*12
   IF(i <	     1) i = 1
   IF(i > numSO4months) i = numSO4months
   j = m2+(ccyyAdjusted-beginYear)*12
   IF(j <	     1) j = 1
   IF(j > numSO4months) j = numSO4months

   IF(MAPL_AM_I_ROOT()) THEN
    PRINT *," SO4  updated to ",ccyyAdjusted,".  Indicies: ",i,j
   END IF

   IF(m1 ==  0) m1 = 12
   IF(m2 == 13) m2 =  1

   DO k = 1,5

    varName = TRIM(zmClimName(k))

    status = NF_INQ_VARID(unit, TRIM(varName), n)
    IF(status /= NF_NOERR) THEN
     PRINT *,"Error getting varid for ",TRIM(varName), status
     PRINT *, NF_STRERROR(status)
     VERIFY_(status)
    END IF

    IF(k <= 4) THEN
     start(3) = m1
    ELSE
     start(3) = i
    END IF

    status = NF_GET_VARA_REAL(unit, n, start, cnt, zm1)
    IF(status /= NF_NOERR) THEN
     PRINT *,"Error getting values for first month of ",TRIM(varName), status
     PRINT *, NF_STRERROR(status)
     VERIFY_(status)
    END IF

    IF(k <= 4) THEN
     start(3) = m2
    ELSE
     start(3) = j
    END IF

    status = NF_GET_VARA_REAL(unit, n, start, cnt, zm2)
    IF(status /= NF_NOERR) THEN
     PRINT *,"Error getting values for second month of ",TRIM(varName), status
     PRINT *, NF_STRERROR(status)
     VERIFY_(status)
    END IF

    SELECT CASE (k)
     CASE (1)
      gcSC%climRain(:,:) = zm1(:,:)*fac+zm2(:,:)*(1.00-fac)
     CASE (2)
      gcSC%climOH(:,:) = zm1(:,:)*fac+zm2(:,:)*(1.00-fac)
     CASE (3)
      gcSC%climH2O(:,:) = zm1(:,:)*fac+zm2(:,:)*(1.00-fac)
     CASE (4)
      gcSC%climO3(:,:) = zm1(:,:)*fac+zm2(:,:)*(1.00-fac)
     CASE (5)
      gcSC%SO4sa(:,:) = zm1(:,:)*fac+zm2(:,:)*(1.00-fac)
     CASE DEFAULT
    END SELECT

   END DO

   DEALLOCATE(zm1, STAT=status)
   VERIFY_(status)
   DEALLOCATE(zm2, STAT=status)
   VERIFY_(status)

   status = NF_CLOSE(unit)
   VERIFY_(status)

#ifdef H5_HAVE_PARALLEL

   CALL MPI_Info_free(info, status)
   VERIFY_(status)

#else

  END IF ! MAPL_AM_I_ROOT

  CALL MAPL_CommsBcast(vm, gcSC%SSG, gcSC%numSSGs, 0, RC=status)
  VERIFY_(status)
  CALL MPI_Bcast(gcSC%climRain, SIZE(gcSC%climRain), MPI_REAL, 0, comm, status)
  VERIFY_(status)
  CALL MPI_Bcast(gcSC%climOH, SIZE(gcSC%climOH), MPI_REAL, 0, comm, status)
  VERIFY_(status)
  CALL MPI_Bcast(gcSC%climH2O, SIZE(gcSC%climH2O), MPI_REAL, 0, comm, status)
  VERIFY_(status)
  CALL MPI_Bcast(gcSC%climO3, SIZE(gcSC%climO3), MPI_REAL, 0, comm, status)
  VERIFY_(status)
  CALL MPI_Bcast(gcSC%SO4sa, SIZE(gcSC%SO4sa), MPI_REAL, 0, comm, status)
  VERIFY_(status)

#endif

 RETURN
 END SUBROUTINE readClimVars

 SUBROUTINE mapToGrid(field2D, field3D, name, rc)
  IMPLICIT NONE
  REAL, INTENT(IN) :: field2D(gcSC%numClimLats,gcSC%km)
  REAL, INTENT(OUT) :: field3D(gcSC%i1:gcSC%i2,gcSC%j1:gcSC%j2,gcSC%km)
  INTEGER, INTENT(OUT) :: rc
  CHARACTER(LEN=*), INTENT(IN) :: name
  CHARACTER(LEN=ESMF_MAXSTR), PARAMETER :: Iam = 'SC::mapToGrid'

  rc = 0

  DO k = 1,gcSC%km
   DO j = gcSC%j1,gcSC%j2
    DO i = gcSC%i1,gcSC%i2
     CALL MAPL_INTERP(field3D(i,j,k), gcSC%latrad(i,j), field2D(:,k), gcSC%climLatRad)
    END DO
   END DO
  END DO

  RETURN
 END SUBROUTINE mapToGrid
  
 SUBROUTINE SSGSfcFlx(dV,m,n2o,ch4,f11,f12,f113,hcfc,ccl4,ch3ccl3, &
                      ch3cl,ch3br,h1301,h12_24,hcfc22,rc)
! ----------------------------------------------------------------------
! Update the surface source gases mixing ratios, and find the surface 
! mass flux per unit time [ kg s^-1 ] that is required to account for 
! the change in mixing ratio.
!
! The species in the argument list are in units of number density.
!
! m is number density in cm^{-3}, dV is the volume of the cell in 
! m^{3}, and tdt is the time step length (s).
! ----------------------------------------------------------------------

  IMPLICIT NONE
  REAL, INTENT(IN) :: dV    ! Cell volume [m^{3}]
  REAL, INTENT(IN) :: m     ! Number density  [cm^{-3}]
  INTEGER, INTENT(OUT) :: rc

  INTEGER, PARAMETER :: DBL=KIND(0.00D+00)
  REAL(KIND=DBL), INTENT(INOUT) :: n2o,ch4,f11,f12,f113,hcfc,ccl4,ch3ccl3
  REAL(KIND=DBL), INTENT(INOUT) :: ch3cl,ch3br,h1301,h12_24,hcfc22

  REAL :: cx

! Molecular weights of 12 species [kg kmol^-1]
! --------------------------------------------
  REAL, PARAMETER ::     mwtN2O =  44.008
  REAL, PARAMETER ::     mwtCH4 =  16.043
  REAL, PARAMETER ::     mwtF11 = 137.36
  REAL, PARAMETER ::     mwtF12 = 120.91
  REAL, PARAMETER ::    mwtF113 = 187.37
  REAL, PARAMETER ::    mwtHCFC = 420.84
  REAL, PARAMETER ::    mwtCCl4 = 153.823
  REAL, PARAMETER :: mwtCH3CCl3 = 133.405
  REAL, PARAMETER ::   mwtCH3Cl =  50.488
  REAL, PARAMETER ::   mwtCH3Br =  94.944
  REAL, PARAMETER ::   mwtH1301 = 148.92
  REAL, PARAMETER ::  mwtH12_24 = 685.05
  
  rc = 0

! CO2 (ppmv) is used only in solverd2.F, and
! multiplication by number density takes place there.
! ---------------------------------------------------
  gcSC%co2     = gcSC%SSG(1) * 1.00E-06

! N20 and CH4 ppbv
! ----------------
  ch4 = gcSC%SSG( 2) * m * 1.00E-09
  n2o = gcSC%SSG( 3) * m * 1.00E-09

! Others pptv
! -----------
  f11     = gcSC%SSG( 4) * m * 1.00E-12
  f12     = gcSC%SSG( 5) * m * 1.00E-12
  f113    = gcSC%SSG( 6) * m * 1.00E-12
  ccl4    = gcSC%SSG( 7) * m * 1.00E-12
  ch3ccl3 = gcSC%SSG( 8) * m * 1.00E-12
  ch3cl   = gcSC%SSG(16) * m * 1.00E-12
  ch3br   = gcSC%SSG(15) * m * 1.00E-12
  h1301   = gcSC%SSG(13) * m * 1.00E-12
  hcfc22  = gcSC%SSG( 9) * m * 1.00E-12

  hcfc    = (gcSC%SSG( 9)+gcSC%SSG(11)+gcSC%SSG(10)*2.00) * m * 1.00E-12
  h12_24  = (gcSC%SSG(12)+gcSC%SSG(14)+gcSC%SSG(14)     ) * m * 1.00E-12

! This conversion constant assumes that 
! the species number densities are in cm^-3.
! ------------------------------------------
  cx = dV*1.00E+06/(tdt*MAPL_AVOGAD)

! Find the flux per unit time
! ---------------------------
    n2oflx(i,j) = (    n2o-w_c%qa(    gcSC%iN2O)%data3d(i,j,km)*m)*cx*mwtN2O
    ch4flx(i,j) = (    ch4-w_c%qa(    gcSC%iCH4)%data3d(i,j,km)*m)*cx*mwtCH4
    f11flx(i,j) = (    f11-w_c%qa(    gcSC%iF11)%data3d(i,j,km)*m)*cx*mwtF11
    f12flx(i,j) = (    f12-w_c%qa(    gcSC%iF12)%data3d(i,j,km)*m)*cx*mwtF12

   f113flx(i,j) = (   f113-w_c%qa(   gcSC%iF113)%data3d(i,j,km)*m)*cx*mwtF113
   hcfcflx(i,j) = (   hcfc-w_c%qa(   gcSC%iHCFC)%data3d(i,j,km)*m)*cx*mwtHCFC
   ccl4flx(i,j) = (   ccl4-w_c%qa(   gcSC%iCCl4)%data3d(i,j,km)*m)*cx*mwtCCl4
    mcfflx(i,j) = (ch3ccl3-w_c%qa(gcSC%iCH3CCl3)%data3d(i,j,km)*m)*cx*mwtCH3CCl3

  ch3clflx(i,j) = (  ch3cl-w_c%qa(  gcSC%iCH3Cl)%data3d(i,j,km)*m)*cx*mwtCH3Cl
  ch3brflx(i,j) = (  ch3br-w_c%qa(  gcSC%iCH3Br)%data3d(i,j,km)*m)*cx*mwtCH3Br
  h1301flx(i,j) = (  h1301-w_c%qa(  gcSC%iH1301)%data3d(i,j,km)*m)*cx*mwtH1301
  h1224flx(i,j) = ( h12_24-w_c%qa( gcSC%iH12_24)%data3d(i,j,km)*m)*cx*mwtH12_24
  
  RETURN
 END SUBROUTINE SSGSfcFlx

 SUBROUTINE swapSpecies(direction,rc)
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: direction
  INTEGER, INTENT(OUT) :: rc
  CHARACTER(LEN=ESMF_MAXSTR) :: Iam = "SC::swapSpecies"
  REAL :: m, rm

  m = numDens(k)
  rm = 1.00/m
  rc = 0

  SELECT CASE (direction)
  
   CASE(1)
           ox = w_c%qa(     gcSC%iOx)%data3d(i,j,kRev)*m
          nox = w_c%qa(    gcSC%iNOx)%data3d(i,j,kRev)*m
         hno3 = w_c%qa(   gcSC%iHNO3)%data3d(i,j,kRev)*m
         n2o5 = w_c%qa(   gcSC%iN2O5)%data3d(i,j,kRev)*m
       ho2no2 = w_c%qa( gcSC%iHO2NO2)%data3d(i,j,kRev)*m
       clono2 = w_c%qa( gcSC%iClONO2)%data3d(i,j,kRev)*m
          clx = w_c%qa(    gcSC%iClx)%data3d(i,j,kRev)*m
          hcl = w_c%qa(    gcSC%iHCl)%data3d(i,j,kRev)*m
         hocl = w_c%qa(   gcSC%iHOCl)%data3d(i,j,kRev)*m
         h2o2 = w_c%qa(   gcSC%iH2O2)%data3d(i,j,kRev)*m
          brx = w_c%qa(    gcSC%iBrx)%data3d(i,j,kRev)*m
          n2o = w_c%qa(    gcSC%iN2O)%data3d(i,j,kRev)*m
          cl2 = w_c%qa(    gcSC%iCl2)%data3d(i,j,kRev)*m
         oclo = w_c%qa(   gcSC%iOClO)%data3d(i,j,kRev)*m
         brcl = w_c%qa(   gcSC%iBrCl)%data3d(i,j,kRev)*m
          hbr = w_c%qa(    gcSC%iHBr)%data3d(i,j,kRev)*m
       brono2 = w_c%qa( gcSC%iBrONO2)%data3d(i,j,kRev)*m
          ch4 = w_c%qa(    gcSC%iCH4)%data3d(i,j,kRev)*m
         hobr = w_c%qa(   gcSC%iHOBr)%data3d(i,j,kRev)*m
       ch3ooh = w_c%qa( gcSC%iCH3OOH)%data3d(i,j,kRev)*m
           co = w_c%qa(     gcSC%iCO)%data3d(i,j,kRev)*m
        hno3c = w_c%qa(  gcSC%iHNO3c)%data3d(i,j,kRev)*m
          f11 = w_c%qa(    gcSC%iF11)%data3d(i,j,kRev)*m
          f12 = w_c%qa(    gcSC%iF12)%data3d(i,j,kRev)*m
         f113 = w_c%qa(   gcSC%iF113)%data3d(i,j,kRev)*m
         hcfc = w_c%qa(   gcSC%iHCFC)%data3d(i,j,kRev)*m
       hcfc22 = w_c%qa( gcSC%iHCFC22)%data3d(i,j,kRev)*m
         ccl4 = w_c%qa(   gcSC%iCCl4)%data3d(i,j,kRev)*m
      ch3ccl3 = w_c%qa(gcSC%iCH3CCl3)%data3d(i,j,kRev)*m
        ch3cl = w_c%qa(  gcSC%iCH3Cl)%data3d(i,j,kRev)*m
        ch3br = w_c%qa(  gcSC%iCH3Br)%data3d(i,j,kRev)*m
        h1301 = w_c%qa(  gcSC%iH1301)%data3d(i,j,kRev)*m
       h12_24 = w_c%qa( gcSC%iH12_24)%data3d(i,j,kRev)*m
           o3 = w_c%qa(     gcSC%iO3)%data3d(i,j,kRev)*m
          o3p = w_c%qa(    gcSC%iO3p)%data3d(i,j,kRev)*m
          o1d = w_c%qa(    gcSC%iO1d)%data3d(i,j,kRev)*m
            n = w_c%qa(      gcSC%iN)%data3d(i,j,kRev)*m
           no = w_c%qa(     gcSC%iNO)%data3d(i,j,kRev)*m
          no2 = w_c%qa(    gcSC%iNO2)%data3d(i,j,kRev)*m
          no3 = w_c%qa(    gcSC%iNO3)%data3d(i,j,kRev)*m
            h = w_c%qa(      gcSC%iH)%data3d(i,j,kRev)*m
           oh = w_c%qa(     gcSC%iOH)%data3d(i,j,kRev)*m
          ho2 = w_c%qa(    gcSC%iHO2)%data3d(i,j,kRev)*m
           cl = w_c%qa(     gcSC%iCl)%data3d(i,j,kRev)*m
          clo = w_c%qa(    gcSC%iClO)%data3d(i,j,kRev)*m
          bro = w_c%qa(    gcSC%iBrO)%data3d(i,j,kRev)*m
           br = w_c%qa(     gcSC%iBr)%data3d(i,j,kRev)*m
        cl2o2 = w_c%qa(  gcSC%iCl2O2)%data3d(i,j,kRev)*m
         ch2o = w_c%qa(   gcSC%iCH2O)%data3d(i,j,kRev)*m
        ch3o2 = w_c%qa(  gcSC%iCH3O2)%data3d(i,j,kRev)*m
    box_ro3ox = w_c%qa(  gcSC%irO3Ox)%data3d(i,j,kRev)

   CASE(-1)

     w_c%qa(     gcSC%iOx)%data3d(i,j,kRev) =      ox*rm
     w_c%qa(    gcSC%iNOx)%data3d(i,j,kRev) =     nox*rm
     w_c%qa(   gcSC%iHNO3)%data3d(i,j,kRev) =    hno3*rm
     w_c%qa(   gcSC%iN2O5)%data3d(i,j,kRev) =    n2o5*rm
     w_c%qa( gcSC%iHO2NO2)%data3d(i,j,kRev) =  ho2no2*rm
     w_c%qa( gcSC%iClONO2)%data3d(i,j,kRev) =  clono2*rm
     w_c%qa(    gcSC%iClx)%data3d(i,j,kRev) =     clx*rm
     w_c%qa(    gcSC%iHCl)%data3d(i,j,kRev) =     hcl*rm
     w_c%qa(   gcSC%iHOCl)%data3d(i,j,kRev) =    hocl*rm
     w_c%qa(   gcSC%iH2O2)%data3d(i,j,kRev) =    h2o2*rm
     w_c%qa(    gcSC%iBrx)%data3d(i,j,kRev) =     brx*rm
     w_c%qa(    gcSC%iN2O)%data3d(i,j,kRev) =     n2o*rm
     w_c%qa(    gcSC%iCl2)%data3d(i,j,kRev) =     cl2*rm
     w_c%qa(   gcSC%iOClO)%data3d(i,j,kRev) =    oclo*rm
     w_c%qa(   gcSC%iBrCl)%data3d(i,j,kRev) =    brcl*rm
     w_c%qa(    gcSC%iHBr)%data3d(i,j,kRev) =     hbr*rm
     w_c%qa( gcSC%iBrONO2)%data3d(i,j,kRev) =  brono2*rm
     w_c%qa(    gcSC%iCH4)%data3d(i,j,kRev) =     ch4*rm
     w_c%qa(   gcSC%iHOBr)%data3d(i,j,kRev) =    hobr*rm
     w_c%qa( gcSC%iCH3OOH)%data3d(i,j,kRev) =  ch3ooh*rm
     w_c%qa(     gcSC%iCO)%data3d(i,j,kRev) =      co*rm
     w_c%qa(  gcSC%iHNO3c)%data3d(i,j,kRev) =   hno3c*rm
     w_c%qa(    gcSC%iF11)%data3d(i,j,kRev) =     f11*rm
     w_c%qa(    gcSC%iF12)%data3d(i,j,kRev) =     f12*rm
     w_c%qa(   gcSC%iF113)%data3d(i,j,kRev) =    f113*rm
     w_c%qa(   gcSC%iHCFC)%data3d(i,j,kRev) =    hcfc*rm
     w_c%qa( gcSC%iHCFC22)%data3d(i,j,kRev) =  hcfc22*rm
     w_c%qa(   gcSC%iCCl4)%data3d(i,j,kRev) =    ccl4*rm
     w_c%qa(gcSC%iCH3CCl3)%data3d(i,j,kRev) = ch3ccl3*rm
     w_c%qa(  gcSC%iCH3Cl)%data3d(i,j,kRev) =   ch3cl*rm
     w_c%qa(  gcSC%iCH3Br)%data3d(i,j,kRev) =   ch3br*rm
     w_c%qa(  gcSC%iH1301)%data3d(i,j,kRev) =   h1301*rm
     w_c%qa( gcSC%iH12_24)%data3d(i,j,kRev) =  h12_24*rm
     w_c%qa(     gcSC%iO3)%data3d(i,j,kRev) =      o3*rm
     w_c%qa(    gcSC%iO3p)%data3d(i,j,kRev) =     o3p*rm
     w_c%qa(    gcSC%iO1d)%data3d(i,j,kRev) =     o1d*rm
     w_c%qa(      gcSC%iN)%data3d(i,j,kRev) =       n*rm
     w_c%qa(     gcSC%iNO)%data3d(i,j,kRev) =      no*rm
     w_c%qa(    gcSC%iNO2)%data3d(i,j,kRev) =     no2*rm
     w_c%qa(    gcSC%iNO3)%data3d(i,j,kRev) =     no3*rm
     w_c%qa(      gcSC%iH)%data3d(i,j,kRev) =       h*rm
     w_c%qa(     gcSC%iOH)%data3d(i,j,kRev) =      oh*rm
     w_c%qa(    gcSC%iHO2)%data3d(i,j,kRev) =     ho2*rm
     w_c%qa(     gcSC%iCl)%data3d(i,j,kRev) =      cl*rm
     w_c%qa(    gcSC%iClO)%data3d(i,j,kRev) =     clo*rm
     w_c%qa(    gcSC%iBrO)%data3d(i,j,kRev) =     bro*rm
     w_c%qa(     gcSC%iBr)%data3d(i,j,kRev) =      br*rm
     w_c%qa(  gcSC%iCl2O2)%data3d(i,j,kRev) =   cl2o2*rm
     w_c%qa(   gcSC%iCH2O)%data3d(i,j,kRev) =    ch2o*rm
     w_c%qa(  gcSC%iCH3O2)%data3d(i,j,kRev) =   ch3o2*rm
     w_c%qa(  gcSC%irO3Ox)%data3d(i,j,kRev) =  box_ro3ox

   CASE DEFAULT

  END SELECT

 RETURN
 END SUBROUTINE swapSpecies

 SUBROUTINE chemStage1(rc)
  INTEGER, INTENT(OUT) :: rc
  CHARACTER(LEN=ESMF_MAXSTR) :: Iam = "SC::chemStage1"
  rc = 0

! Initialize ratios, production and loss rates and estimated species to zero.
! ---------------------------------------------------------------------------
  CALL localZeroSet(ro3po3,ro1do3,ro3ox,rnno,rnono2,rno3no2,rno2nox,rclclo, &
      rcloclx,rocloclo,rcl2o2clo,rbrobrx,rbrnbrx,po3,lo3,ph2o2,lh2o2,phno3, &
      lhno3,phno3het,pn2o5,ln2o5,ln2o5het,pho2no2,lho2no2,pnox,lnox,lnoxsq, &
        pnoxcl,lnoxcl,pclx,lclx,pclono2,lclono2,lclono2het,phcl,lhcl,phocl, &
        lhocl,pbrono2,lbrono2het,pbrx,lbrx,phbr,lhbr,phobr,lhobr,pno2,lno2, &
        pno3,lno3,poclo,pnoxa,lnoxa,pcl2,pbrcl,lo3hox,lo3nox,lo3oxsq,lo3cl, &
       lo3cly,lo3oh,lo3brx,lo3chx,o3pe,o3e,oxe,ne,noe,no2e,no3e,noxe,n2o5e, &
         h2o2e,hno3e,ho2no2e,hcle,hocle,cle,cloe,ocloe,cl2o2e,clono2e,clxe, &
                                          bre,broe,brono2e,brxe,hbre,hobre, &
                      rmedsts(k),rmednat(k),rmedice(k),denssts(k),vfall(k))

! Condensed water number density. QCTOT is mass fraction.
! -------------------------------------------------------
  h2oc = qctot(i,j,kRev)*numDens(k)*MAPL_AIRMW/MAPL_H2OMW

! Polar stratospheric clouds
! --------------------------
  DoingPSCs: IF(gcSC%doPSCs) THEN

   ic = gcSC%irO3Ox
   tropp = w_c%qa(ic)%data3d(i,j,km)*0.01

   CALL pscs(k,km,gcSC%latRad(i,j),tropp,midPress,edgePress, &
  	          kel,climH2OProfile,numDens,SO4saProfile,sanat,saice, &
  	     rmedsts,rmednat,rmedice,denssts,hno3,hno3c,h2o,h2oc,gcSC)

! Check for HNO3COND above chosen limit
! -------------------------------------
   qmax = 1.00E+09*hno3c/numDens(k)
   IF(qmax > gcSC%HNO3Ice_MAX) THEN
    PRINT *,TRIM(Iam)//": Found HNO3COND above limit: ",qmax," ppbv"
    status = 1
    VERIFY_(status)
   END IF

! Export states for NAT and ice surface area density [m^{-1}]
! -----------------------------------------------------------
   natsad(i,j,kRev) = 1.00E+02*sanat
   icesad(i,j,kRev) = 1.00E+02*saice

! Khet3d resets heterogeneous reaction rates 110-115.
! ---------------------------------------------------
   IF(k >= k1Strat) THEN
    CALL khet3d(k,km,gcSC%numreacs,gcSC%parts,SO4saProfile,sanat, &
		saice,numDens,kel,h2o,hcl,clono2,ak,status)
    VERIFY_(status)
   END IF

! Ad-hoc reaction rate adjustments.
! ---------------------------------
   IF(ak(k,111)* clono2 > 5.00E-04) ak(k,111) = 5.00E-04/clono2
   IF(ak(k,114)*   hocl > 5.00E-04) ak(k,114) = 5.00E-04/hocl
   IF(hcl/numDens(k) < 1.00E-12) THEN
    ak(k,111) = 0.00
    ak(k,114) = 0.00
   END IF

  END IF DoingPSCs

  IF(ak(k,31)*numDens(k)     > 5.00E-04) ak(k,31) = 5.00E-04/ numDens(k)
  IF(ak(k,32)*numDens(k)*clo > 5.00E-04) ak(k,32) = 5.00E-04/(numDens(k)*clo)
  IF(ak(k,32)*numDens(k)*no2 > 5.00E-04) ak(k,32) = 5.00E-04/(numDens(k)*no2)

! Constrain tropospheric OH
! -------------------------
  IF(k < k1Strat) oh = climOHProfile(k)

  IF(gcSC%spinup .AND. passNumber == 1) CALL sproic(k,km,midPress,numDens,daytime, &
            aj,ak,br,bro,brono2,brx,cl,cl2o2,clo,clx,h,h2o,hno3,ho2,n,n2o5,no,no2, &
                no3,nox,o1d,o3,o3p,oclo,oh,ox,gcSC%numphoto,gcSC%numreacs,gcSC%o2)

  CALL part(k,k1Strat,km,daytime,numDens,aj,ak,bro,brx,ch2o,ch3o2,ch4,clo,clono2, &
             h,hbr,hno3,ho2,hobr,n,n2o5,no,no2,no3,nox,o3,o3p,oh,rbrnbrx,rbrobrx, &
                   rcl2o2clo,rclclo,rcloclx,rnno,rno2nox,rno3no2,rnono2,rocloclo, &
                             gcSC%levels_cal,gcSC%numreacs,gcSC%numphoto,gcSC%o2) 

  CALL chxpart(k,k1Strat,km,daytime,numDens,aj,ak,brx,ch2o,ch3,ch3br,ch3o,ch3o2, &
           ch3ooh,ch4,cho,clx,ho2,nox,o1d,o3p,oh,rbrnbrx,rbrobrx,rclclo,rcloclx, &
             rno2nox,rnono2,gcSC%levels_cal,gcSC%numreacs,gcSC%numphoto,gcSC%o2)

  CALL hoxpart(k,k1Strat,km,daytime,numDens,tropo,climOHProfile,ak,aj,brx,ch2o,ch3o, &
              ch3o2,ch3ooh,ch4,cho,clono2,clx,co,h,hbr,hcl,h2o,h2o2,hno3,ho2,ho2no2, &
            hobr,hocl,nox,o1d,o3,o3p,oh,rbrnbrx,rbrobrx,rclclo,rcloclx,rnno,rno2nox, &
             rnono2,gcSC%levels_cal,gcSC%niters,gcSC%numreacs,gcSC%numphoto,gcSC%o2)

  CALL oxpandl(k,k1Strat,km,tropo,daytime,numDens,aj,ak,brx,ch2o,ch3,ch3ooh,ch4, &
      cl2o2,clono2,clx,h,h2o,h2o2,hbr,hcl,hno3,ho2,ho2no2,hocl,lh2o2,lo3,lo3brx, &
      lo3chx,lo3cl,lo3cly,lo3hox,lo3nox,lo3oh,lo3oxsq,n2o,n2o5,nox,o1d,o3,oh,ox, &
   ph2o2,po3,climRainProfile,rbrobrx,rclclo,rcloclx,rnno,rno2nox,rno3no2,rnono2, &
        ro1do3,ro3ox,ro3po3,gcSC%levels_cal,gcSC%numreacs,gcSC%numphoto,gcSC%o2)

  CALL brpandl(dt,k,k1Strat,km,daytime,ak,aj,brono2,brx,ch2o,ch3br,h12_24,h1301,hbr, &
         ho2,hobr,lbrono2het,lbrx,lhbr,lhobr,o1d,oh,pbrx,phbr,phobr,climRainProfile, &
                rbrnbrx,rbrobrx,gcSC%levels_cal,gcSC%numreacs,gcSC%numphoto,gcSC%o2)

  CALL ncpandl(dt,k,k1Strat,km,daytime,numDens,ak,aj,ccl4,ch3ccl3,ch3cl,ch4,clono2, &
        clx,f11,f113,f12,h2o,h2o2,hcfc,hcl,hno3,ho2,ho2no2,hocl,lbrono2het,lclono2, &
        lclono2het,lclx,lhcl,lhno3,lho2no2,lhocl,ln2o5,ln2o5het,lnox,lnoxcl,lnoxsq, &
         n,n2o,n2o5,no,no2,nox,o1d,o3p,oh,pbrono2,pclono2,pclx,phcl,phno3,phno3het, &
            pho2no2,phocl,pn2o5,pnox,pnoxcl,climRainProfile,rclclo,rcloclx,rno2nox, &
            rno3no2,gcSC%levels_cal,gcSC%numreacs,gcSC%numphoto,gcSC%o2,gcrProfile)

  CALL solverest(dt,requiredDt,examineDt,k,k1Strat,km,daytime,tropo,numDens,aj,ak,br, &
       brcl,bre,bro,broe,brono2,brono2e,brx,brxe,ccl4,ch3ccl3,ch3cl,cl,cle,cl2,cl2o2, &
       cl2o2e,clo,cloe,clono2,clono2e,clx,clxe,f11,f113,f12,h2o2,h2o2e,hbr,hbre,hcfc, &
             hcl,hcle,hno3,hno3e,ho2,ho2no2,ho2no2e,hobr,hobre,hocl,hocle,lbrono2het, &
        lbrx,lclono2,lclono2het,lclx,lh2o2,lhbr,lhcl,lhno3,lho2no2,lhobr,lhocl,ln2o5, &
      ln2o5het,lno2,lno3,lnox,lnoxa,lnoxcl,lnoxsq,lo3,n2o,n2o5,n2o5e,ne,no2,no2e,no3, &
    no3e,no,noe,nox,noxe,o1d,o3,o3e,o3p,o3pe,oclo,ocloe,oh,ox,oxe,pbrcl,pbrono2,pbrx, &
          pcl2,pclono2,pclx,ph2o2,phbr,phcl,pho2no2,phocl,phno3,phno3het,phobr,pn2o5, &
     pno2,pno3,pnox,pnoxa,pnoxcl,po3,poclo,climRainProfile,rbrnbrx,rbrobrx,rcl2o2clo, &
          rclclo,rcloclx,rnno,rnono2,rno2nox,rno3no2,ro3ox,ro3po3,rocloclo,box_ro3ox, &
                      gcSC%levels_cal,gcSC%numreacs,gcSC%numphoto,gcrProfile,gcSC%o2)

  IF(examineDt <= 3) CALL verifyChemDt(examineDt,requiredPasses,tdt,dt,requiredDt,startOver)

  RETURN
 END SUBROUTINE chemStage1

 SUBROUTINE chemStage2(rc)
  INTEGER, INTENT(OUT) :: rc
  CHARACTER(LEN=ESMF_MAXSTR) :: Iam = "SC::chemStage2"
  rc = 0

  CALL partest(k,k1Strat,km,daytime,numDens,aj,ak,broe,brxe,ch2o,ch3o2,ch4,cloe,clono2e, &
         h,hbre,hno3e,ho2,hobre,ne,n2o5e,noe,no2e,no3e,noxe,o3e,o3pe,oh,rbrnbrx,rbrobrx, &
     		          rcl2o2clo,rclclo,rcloclx,rnno,rno2nox,rno3no2,rnono2,rocloclo, &
                                    gcSC%levels_cal,gcSC%numreacs,gcSC%numphoto,gcSC%o2)

  CALL hoxpartest(k,k1Strat,km,daytime,numDens,tropo,climOHProfile,ak,aj,brxe,ch2o,ch3o, &
           ch3o2,ch3ooh,ch4,cho,clono2e,clxe,co,h,hbre,hcle,h2o,h2o2e,hno3e,ho2,ho2no2e, &
             hobre,hocle,noxe,o1d,o3,o3p,oh,rbrnbrx,rbrobrx,rclclo,rcloclx,rnno,rno2nox, &
                 rnono2,gcSC%levels_cal,gcSC%niters,gcSC%numreacs,gcSC%numphoto,gcSC%o2)

  CALL oxpandlest(k,k1Strat,km,tropo,daytime,numDens,aj,ak,brxe,ch2o,ch3,ch3ooh,ch4, &
 cl2o2e,clono2e,clxe,h,h2o,h2o2e,hbre,hcle,hno3e,ho2,ho2no2e,hocle,lh2o2,lo3,lo3brx, &
           lo3chx,lo3cl,lo3cly,lo3hox,lo3nox,lo3oh,lo3oxsq,n2o,n2o5e,noxe,oh,ox,oxe, &
       ph2o2,po3,climRainProfile,rbrobrx,rclclo,rcloclx,rnno,rno2nox,rno3no2,rnono2, &
            ro1do3,ro3ox,ro3po3,gcSC%levels_cal,gcSC%numreacs,gcSC%numphoto,gcSC%o2)

  CALL brpandlest(dt,k,k1Strat,km,daytime,ak,aj,brono2,brx,brxe,ch2o,ch3br,h12_24,h1301,hbre, &
            ho2,hobr,hobre,lbrono2het,lbrx,lhbr,lhobr,o3e,oh,pbrx,phbr,phobr,climRainProfile, &
                  rbrnbrx,rbrobrx,ro1do3,gcSC%levels_cal,gcSC%numreacs,gcSC%numphoto,gcSC%o2)

  CALL ncpandlest(dt,k,k1Strat,km,daytime,numDens,ak,aj,brxe,ccl4,ch3ccl3,ch3cl,ch4,clono2e, &
           clxe,f11,f113,f12,h2o,h2o2e,hcfc,hcl,hcle,hno3e,ho2,ho2no2,ho2no2e,hocle,lclono2, &
                 lclono2het,lclx,lhcl,lhno3,lho2no2,lhocl,ln2o5,ln2o5het,lnox,lnoxcl,lnoxsq, &
               n2o,n2o5,n2o5e,ne,noe,no2e,noxe,o3e,o3pe,oh,pclono2,pclx,phcl,phno3,phno3het, &
             pho2no2,phocl,pn2o5,pnox,pnoxcl,climRainProfile,rbrnbrx,rclclo,rcloclx,rno2nox, &
              rno3no2,ro1do3,gcSC%levels_cal,gcSC%numreacs,gcSC%numphoto,gcSC%o2,gcrProfile)
 
       CALL solverd2(dt,requiredDt,examineDt,k,k1Strat,km,daytime,tropo,numDens,aj,ak,br, &
     brcl,bro,brono2,brx,ccl4,ch2o,ch3br,ch3ccl3,ch3cl,ch3o2,ch3ooh,ch4,cho,cl,cl2,cl2o2, &
            clo,clono2,clono2e,clx,clxe,co,f11,f113,f12,h,h12_24,h1301,h2o,h2o2,hbr,hcfc, &
                              hcfc22,hcl,hcle,hno3,ho2,ho2no2,hobr,hocl,hocle,lbrono2het, &
            lbrx,lclono2,lclono2het,lclx,lh2o2,lhbr,lhcl,lhno3,lho2no2,lhobr,lhocl,ln2o5, &
                ln2o5het,lno2,lno3,lnox,lnoxa,lnoxcl,lnoxsq,lo3,n,n2o,n2o5,n2o5e,no2,no3, &
                                         no,nox,o1d,o3,o3p,oclo,oh,ox,pbrcl,pbrono2,pbrx, &
              pcl2,pclono2,pclx,ph2o2,phbr,phcl,pho2no2,phocl,phno3,phno3het,phobr,pn2o5, &
         pno2,pno3,pnox,pnoxa,pnoxcl,po3,poclo,climRainProfile,rbrnbrx,rbrobrx,rcl2o2clo, &
       rclclo,rcloclx,rnno,rnono2,rno2nox,rno3no2,ro1do3,ro3ox,ro3po3,rocloclo,box_ro3ox, &
            gcSC%levels_cal,gcSC%numreacs,gcSC%numphoto,gcrProfile,gcSC%o2,gcSC%co2,szan)

  IF(examineDt <= 3) CALL verifyChemDt(examineDt,requiredPasses,tdt,dt,requiredDt,startOver)

  RETURN
 END SUBROUTINE chemStage2

 END SUBROUTINE SC_GridCompRun

!-------------------------------------------------------------------------
!NASA/GSFC, Global Modeling and Assimilation Office, Code 610.1, GEOS/DAS!
!-------------------------------------------------------------------------
!BOP
!
! !IROUTINE:  SC_GridCompFinalize --- The Chem Driver 
!
! !INTERFACE:
!

   SUBROUTINE SC_GridCompFinalize ( gcSC, w_c, impChem, expChem, &
                                    nymd, nhms, cdt, rc )

! !USES:

  IMPLICIT NONE

! !INPUT/OUTPUT PARAMETERS:

   TYPE(SC_GridComp), INTENT(inout) :: gcSC   ! Grid Component

! !INPUT PARAMETERS:

   TYPE(Chem_Bundle), INTENT(in)  :: w_c      ! Chemical tracer fields   
   INTEGER, INTENT(in) :: nymd, nhms	      ! time
   REAL,    INTENT(in) :: cdt  	              ! chemical timestep (secs)


! !OUTPUT PARAMETERS:

   TYPE(ESMF_State), INTENT(inout) :: impChem	! Import State
   TYPE(ESMF_State), INTENT(inout) :: expChem	! Import State
   INTEGER, INTENT(out) ::  rc                  ! Error return code:
                                                !  0 - all is well
                                                !  1 -
 
! !DESCRIPTION: This routine finalizes this Grid Component.
!
! !REVISION HISTORY:
!
!  18Sep2003 da Silva  First crack.
!  11May2012 Nielsen   Capability for FV cubed [Ganymed-1_0_UNSTABLE]
!
!EOP
!-------------------------------------------------------------------------

   CHARACTER(LEN=*), PARAMETER :: Iam = 'SC_GridCompFinalize'
   INTEGER :: status
   
   rc = 0

   DEALLOCATE(gcSC%climLatRad, STAT=status)
   VERIFY_(status)

   DEALLOCATE(gcSC%sdat, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%o2jdat, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%o3_tab, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%xtab, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%sza_tab, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%CH2O_aq, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%rlam, STAT=status)
   VERIFY_(status)

   DEALLOCATE(gcSC%SSG, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%climRain, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%climO3, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%climOH, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%climH2O, STAT=status)
   VERIFY_(status)

   DEALLOCATE(gcSC%climRain3D, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%climO33D, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%climOH3D, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%climH2O3D, STAT=status)
   VERIFY_(status)

   DEALLOCATE(gcSC%SO4sa, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%SO4sa3D, STAT=status)
   VERIFY_(status)

   DEALLOCATE(gcSC%KRxnOrder, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%standardKRxn, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%KRxnName, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%cnsttab, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%indxs1, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%indxs2, STAT=status)
   VERIFY_(status)
   DEALLOCATE(gcSC%indxs3, STAT=status)
   VERIFY_(status)

   RETURN

 END SUBROUTINE SC_GridCompFinalize

 END MODULE SC_GridCompMod

