! $Id: GEOS_LandGridComp.F90,v 1.39.24.2.8.2.6.1.8.1.8.1.2.2 2014-04-15 20:28:59 ltakacs Exp $

#include "MAPL_Generic.h"

!=============================================================================
module GEOS_LandGridCompMod

!BOP
! !MODULE: GEOS_LandGridCompMod -- A Module to combine VegDyn and Catch Gridded Components

! !DESCRIPTION: This gridded component operates on the land tiles as
! as child of GEOS\_SurfaceGridComp.  The core functionality is the
! calculation of energy and water fluxes impacting the lowest layer
! of the atmospheric grid.  In order to operate on the tilespace
! specified by its parent, GEOS\_LandGridComp runs its child VegdynGridComp
! to determine relevant time-dependent land-surface characteristics.
! All parameters calculated in VegdynGridComp are required by CatchGridComp.
! Furthermore, several exports of the Vegdyn routines are also exports
! from the Land composite, for use in other modules, such as the case
! for lai and grn needed in radiation.  Vegdyn will be updated first.
! Then the catchment call will be issued.  The composite exports
! consist of the union of the catchment exports with a subset of the 
! vegdyn exports.  All imports and exports are on the prescribed tile
! grid in the (IM, JM)=(NTILES, 1) convention.  

!
! !USES:

  use ESMF
  use MAPL_Mod

  use GEOS_VegdynGridCompMod,  only : VegdynSetServices   => SetServices
  use GEOS_CatchGridCompMod,   only : CatchSetServices    => SetServices

  implicit none
  private

! !PUBLIC MEMBER FUNCTIONS:

  public SetServices


!EOP


  integer                                 :: VEGDYN
  integer, allocatable                    :: CATCH(:)

contains

!BOP

! !IROUTINE: SetServices -- Sets ESMF services for this component

! !INTERFACE:

  subroutine SetServices ( GC, RC )

! !ARGUMENTS:

    type(ESMF_GridComp), intent(INOUT) :: GC  ! gridded component
    integer, optional                  :: RC  ! return code

! !DESCRIPTION:  The SetServices for the Physics GC needs to register its
!   Initialize and Run.  It uses the MAPL\_Generic construct for defining 
!   state specs and couplings among its children.  In addition, it creates the   
!   children GCs (VegDyn, Catch) and runs their respective SetServices.

!EOP

!=============================================================================
!
! ErrLog Variables

    character(len=ESMF_MAXSTR)              :: IAm
    integer                                 :: STATUS
    character(len=ESMF_MAXSTR)              :: COMP_NAME

! Locals

    character(len=ESMF_MAXSTR)              :: GCName
    type(ESMF_Config)                       :: CF
    integer                                 :: NUM_CATCH
    integer                                 :: I
    character(len=ESMF_MAXSTR)              :: TMP


!=============================================================================

! Begin...

!------------------------------------------------------------
! Get my name and set-up traceback handle
!------------------------------------------------------------

    call ESMF_GridCompGet(GC                                 ,&
                          NAME=COMP_NAME	             ,&
                          CONFIG=CF                          ,&
                                                    RC=STATUS )
    VERIFY_(STATUS)

    Iam = trim(COMP_NAME) // 'SetServices'

!------------------------------------------------------------
! Register services for this component
!------------------------------------------------------------

    call MAPL_GridCompSetEntryPoint ( GC, ESMF_METHOD_INITIALIZE, Initialize, RC=STATUS )
    VERIFY_(STATUS)
    call MAPL_GridCompSetEntryPoint ( GC, ESMF_METHOD_RUN, Run1, RC=STATUS )
    VERIFY_(STATUS)
    call MAPL_GridCompSetEntryPoint ( GC, ESMF_METHOD_RUN, Run2, RC=STATUS )
    VERIFY_(STATUS)

    call ESMF_ConfigGetAttribute ( CF, NUM_CATCH, Label="NUM_CATCH_ENSEMBLES:", default=1, RC=STATUS)
    VERIFY_(STATUS)

!------------------------------------------------------------
! Create children's gridded components and invoke their 
! SetServices
!------------------------------------------------------------

    VEGDYN  = MAPL_AddChild(GC, NAME='VEGDYN', SS=VegdynSetServices, RC=STATUS)
    VERIFY_(STATUS)

    allocate (CATCH(NUM_CATCH), stat=status)
    VERIFY_(STATUS)
    if (NUM_CATCH == 1) then
       CATCH(1) = MAPL_AddChild(GC, NAME='CATCH', SS=CatchSetServices, RC=STATUS)
       VERIFY_(STATUS)
    else
       do I = 1, NUM_CATCH
          WRITE(TMP,'(I3.3)') I
          GCName  = 'ens' // trim(TMP) // ':CATCH'
          CATCH(I) = MAPL_AddChild(GC, NAME=GCName, SS=CatchSetServices, RC=STATUS)
          VERIFY_(STATUS)
       end do
    end if

!BOS

!------------------------------------------------------------
! Set UNIQUE IMPORT or EXPORT specs for the Composite GC
! There are none of these at the moment...
!------------------------------------------------------------

! !IMPORT STATE:

! There are no explicit imports

! !EXPORT STATE:

! Selected exports of the children are made explicit exports of land

! These are from RUN2 of the first catchment instance

    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'LST',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TST',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'QST',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'DELTS',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'DELQS',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'ALBVR',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'ALBVF',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'ALBNR',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'ALBNF',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'EMIS',     CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SNOWMASS', CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SNOWDP',   CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'GHFLX',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TPUNST',   CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TPSURF',   CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TPSNOW',   CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TPWLT',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TPSAT',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'ASNOW',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SHSNOW',   CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'AVETSNOW', CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'FRSAT',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'FRUST',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'FRWLT',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'WET1',     CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'WET2',     CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'WET3',     CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'WCSF',     CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'WCRZ',     CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'WCPR',     CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TP1',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TP2',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TP3',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TP4',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TP5',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TP6',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'EVAPOUT',  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SUBLIM' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SHOUT'  ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'RUNOFF' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'EVPINT' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'EVPSOI' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'EVPVEG' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'EVPICE' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'WAT10CM',  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'WATSOI' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'ICESOI' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'EVPSNO' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'BASEFLOW', CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'RUNSURF',  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SMELT'  ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'HLWUP'  ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SWNDSRF',  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'LWNDSRF',  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)                                                                         
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'HLATN'  ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'QINFIL' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'ACCUM'  ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'EVLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'PRLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SNOLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'DRPARLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'DFPARLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'LHSNOW' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SWNETSNOW' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'LWUPSNOW' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'LWDNSNOW' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TCSORIG' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TPSN1IN' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TPSN1OUT' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'LHLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SHLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SWLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SWDOWNLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'LWLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'GHLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'GHSNOW' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'GHTSKIN' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SMLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TWLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TELAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TSLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'DWLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'DHLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SPLAND' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SPWATR' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'SPSNOW' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'WESNN1' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'WESNN2' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'WESNN3' ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'CAPAC'  ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'CDCR2'  ,  CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)

! These are from RUN1 of vegdyn and the first catchment instance

    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'TH',       CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'QH',       CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'CHT',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'CQT',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'CMT',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'CNT',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'RIT',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'Z0',       CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'D0',       CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'Z0H',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'VENT',     CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'GUST',     CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'LAI',      CHILD_ID = VEGDYN,RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'GRN',      CHILD_ID = VEGDYN,RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'ROOTL',    CHILD_ID = VEGDYN,RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'Z2CH',     CHILD_ID = VEGDYN,RC=STATUS  )
    VERIFY_(STATUS)   
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'MOU50M',   CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'MOV50M',   CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'MOT10M',   CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'MOQ10M',   CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'MOU10M',   CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'MOV10M',   CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'MOT2M',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'MOQ2M',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'MOU2M',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'MOV2M',    CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)
    call MAPL_AddExportSpec ( GC, SHORT_NAME = 'ITY',      CHILD_ID = CATCH(1), RC=STATUS  )
    VERIFY_(STATUS)

!EOS
    
!------------------------------------------------------------
! Set internal connections between the children's IMPORTS 
! and EXPORTS... this will be a one-way street between the
! vegdyn routine and the catchment model
!------------------------------------------------------------

! !CONNECTIONS:

    DO I = 1, NUM_CATCH
       call MAPL_AddConnectivity (                              &
            GC                                                 ,&
            SHORT_NAME  = (/'LAI  ','GRN  ','ROOTL','Z2CH ', 'ITY  ' /) ,&
            DST_ID =  CATCH(I)                                 ,&
            SRC_ID =  VEGDYN                                   ,&
                                                      RC=STATUS )
       VERIFY_(STATUS)
    END DO


    call MAPL_TimerAdd(GC, name="INITIALIZE"    ,RC=STATUS)
    VERIFY_(STATUS)
    call MAPL_TimerAdd(GC, name="RUN1"          ,RC=STATUS)
    VERIFY_(STATUS)
    call MAPL_TimerAdd(GC, name="RUN2"          ,RC=STATUS)
    VERIFY_(STATUS)


    call MAPL_GenericSetServices(GC, RC=STATUS )
    VERIFY_(STATUS)

    deallocate(CATCH)

    RETURN_(ESMF_SUCCESS)
  
  end subroutine SetServices


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


! !IROUTINE: Initialize -- Initialize method for the composite Surface Gridded Component

! !INTERFACE:

  subroutine Initialize ( GC, IMPORT, EXPORT, CLOCK, RC )

! !ARGUMENTS:

    type(ESMF_GridComp), intent(inout) :: GC     ! Gridded component 
    type(ESMF_State),    intent(inout) :: IMPORT ! Import state
    type(ESMF_State),    intent(inout) :: EXPORT ! Export state
    type(ESMF_Clock),    intent(inout) :: CLOCK  ! The clock
    integer, optional,   intent(  out) :: RC     ! Error code

! !DESCRIPTION: The Initialize method of the Land Composite Gridded Component.
 

!EOP

! ErrLog Variables

    character(len=ESMF_MAXSTR)           :: IAm 
    integer                              :: STATUS
    character(len=ESMF_MAXSTR)           :: COMP_NAME
    
! Local derived type aliases

    type (MAPL_MetaComp    ), pointer       :: MAPL
    type (MAPL_MetaComp    ), pointer       :: CHILD_MAPL 
    type (MAPL_LocStream       )            :: LOCSTREAM
    type (ESMF_DELayout        )            :: LAYOUT
    type (ESMF_Config          )            :: CF
    type (ESMF_GridComp        ), pointer   :: GCS(:)
  
    integer                                 :: I

!=============================================================================

! Begin... 

! Get the target components name and set-up traceback handle.
! -----------------------------------------------------------

    call ESMF_GridCompGet ( GC, name=COMP_NAME, RC=STATUS )
    VERIFY_(STATUS)
    Iam = trim(COMP_NAME) // "Initialize"

! Get my internal MAPL_Generic state
!-----------------------------------

    call MAPL_GetObjectFromGC ( GC, MAPL, RC=STATUS)
    VERIFY_(STATUS)

    call MAPL_TimerOn(MAPL,"INITIALIZE", RC=STATUS ); VERIFY_(STATUS)
    call MAPL_TimerOn(MAPL,"TOTAL", RC=STATUS ); VERIFY_(STATUS)

! Get the land tilegrid and the child components
!----------------------------------------------- 

    call MAPL_Get (MAPL, LOCSTREAM=LOCSTREAM, GCS=GCS, RC=STATUS )
    VERIFY_(STATUS)

! Place the land tilegrid in the generic state of each child component
!---------------------------------------------------------------------

    do I = 1, SIZE(GCS)
       call MAPL_GetObjectFromGC( GCS(I), CHILD_MAPL, RC=STATUS )
       VERIFY_(STATUS)
       call MAPL_Set (CHILD_MAPL, LOCSTREAM=LOCSTREAM, RC=STATUS )
       VERIFY_(STATUS)
    end do

    call MAPL_TimerOff(MAPL,"TOTAL", RC=STATUS ); VERIFY_(STATUS)

! Call Initialize for every Child
!--------------------------------

    call MAPL_GenericInitialize ( GC, IMPORT, EXPORT, CLOCK,  RC=STATUS)
    VERIFY_(STATUS)

    call MAPL_TimerOff(MAPL,"INITIALIZE", RC=STATUS ); VERIFY_(STATUS)

    RETURN_(ESMF_SUCCESS)
  end subroutine Initialize


!=============================================================================

!BOP
! !IROUTINE: Run1 -- First Run method for the composite Land Gridded Component
! !INTERFACE:
  subroutine Run1(GC, IMPORT, EXPORT, CLOCK, RC )
! !ARGUMENTS:

  type(ESMF_GridComp), intent(inout) :: GC     ! Gridded component 
  type(ESMF_State),    intent(inout) :: IMPORT ! Import state
  type(ESMF_State),    intent(inout) :: EXPORT ! Export state
  type(ESMF_Clock),    intent(inout) :: CLOCK  ! The clock
  integer, optional,   intent(  out) :: RC     ! Error code

! !DESCRIPTION: This first run method calls the children's
!   first run methods. VEGDYN has only one, and it is called here.
!EOP

! ErrLog Variables

   character(len=ESMF_MAXSTR)          :: IAm 
   integer                             :: STATUS
   character(len=ESMF_MAXSTR)          :: COMP_NAME

! Local derived type aliases

   type (MAPL_MetaComp),      pointer  :: MAPL
   type (ESMF_GridComp),      pointer  :: GCS(:)
   type (ESMF_State),         pointer  :: GIM(:)
   type (ESMF_State),         pointer  :: GEX(:)
   character(len=ESMF_MAXSTR),pointer  :: GCnames(:)
   integer                             :: I 

!=============================================================================

! Begin... 

! Get the target components name and set-up traceback handle.
! -----------------------------------------------------------

    call ESMF_GridCompGet ( GC, name=COMP_NAME, RC=STATUS )
    VERIFY_(STATUS)
    Iam = trim(COMP_NAME) // "Run1"

! Get my internal MAPL_Generic state
!-----------------------------------

    call MAPL_GetObjectFromGC ( GC, MAPL, RC=STATUS)
    VERIFY_(STATUS)

    call MAPL_TimerOn(MAPL,"TOTAL", RC=STATUS ); VERIFY_(STATUS)
    call MAPL_TimerOn(MAPL,"RUN1", RC=STATUS ); VERIFY_(STATUS)

    call MAPL_Get (MAPL, GCS=GCS, GIM=GIM, GEX=GEX, GCnames=GCnames,rc=STATUS)
    VERIFY_(STATUS)

! Call the children's RUN methods
!--------------------------------

    DO I = 1, size(GCS)
       call MAPL_TimerOn(MAPL,trim(GCnames(i)), RC=STATUS ); VERIFY_(STATUS)
       call ESMF_GridCompRun(GCS(I), importState=GIM(I), exportState=GEX(I), &
                             CLOCK=CLOCK, PHASE=1, userRC=STATUS)
       VERIFY_(STATUS)
       call MAPL_TimerOff(MAPL,trim(GCnames(i)), RC=STATUS ); VERIFY_(STATUS)
    END DO

    call MAPL_TimerOff(MAPL,"RUN1", RC=STATUS ); VERIFY_(STATUS)
    call MAPL_TimerOff(MAPL,"TOTAL", RC=STATUS ); VERIFY_(STATUS)

    RETURN_(ESMF_SUCCESS)

  end subroutine Run1

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!BOP
! !IROUTINE: Run2 -- Second Run method for the composite Land Gridded Component
! !INTERFACE:
  subroutine Run2(GC, IMPORT, EXPORT, CLOCK, RC )

! !ARGUMENTS:

  type(ESMF_GridComp), intent(inout) :: GC     ! Gridded component 
  type(ESMF_State),    intent(inout) :: IMPORT ! Import state
  type(ESMF_State),    intent(inout) :: EXPORT ! Export state
  type(ESMF_Clock),    intent(inout) :: CLOCK  ! The clock
  integer, optional,   intent(  out) :: RC     ! Error code

! !DESCRIPTION: This second run method call only the catchments second method.
!EOP

! ErrLog Variables

   character(len=ESMF_MAXSTR)          :: IAm 
   integer                             :: STATUS
   character(len=ESMF_MAXSTR)          :: COMP_NAME

! Local derived type aliases

   type (MAPL_MetaComp),  pointer  :: MAPL
   type (ESMF_GridComp),  pointer  :: GCS(:)
   type (ESMF_State),     pointer  :: GIM(:)
   type (ESMF_State),     pointer  :: GEX(:)
   character(len=ESMF_MAXSTR),pointer  :: GCnames(:)

   integer :: I
  
!=============================================================================

! Begin... 

! Get the target components name and set-up traceback handle.
! -----------------------------------------------------------

    call ESMF_GridCompGet ( GC, name=COMP_NAME, RC=STATUS )
    VERIFY_(STATUS)
    Iam = trim(COMP_NAME) // "Run2"

! Get my internal MAPL_Generic state
!-----------------------------------

    call MAPL_GetObjectFromGC ( GC, MAPL, RC=STATUS)
    VERIFY_(STATUS)

    call MAPL_TimerOn(MAPL,"TOTAL", RC=STATUS ); VERIFY_(STATUS)
    call MAPL_TimerOn(MAPL,"RUN2", RC=STATUS ); VERIFY_(STATUS)

    call MAPL_Get (MAPL, GCS=GCS, GIM=GIM, GEX=GEX, GCnames=GCnames,rc=STATUS)
    VERIFY_(STATUS)

! Call the children's RUN methods
!--------------------------------
    DO I=1,size(GCS)
       if (I == VEGDYN) cycle
       call MAPL_TimerOn(MAPL,trim(GCnames(i)), RC=STATUS ); VERIFY_(STATUS)
       call ESMF_GridCompRun(GCS(I), importState=GIM(I), exportState=GEX(I), &
                             CLOCK=CLOCK, PHASE=2, userRC=STATUS)
       VERIFY_(STATUS)
       call MAPL_TimerOff(MAPL,trim(GCnames(i)), RC=STATUS ); VERIFY_(STATUS)
    END DO

    call MAPL_TimerOff(MAPL,"RUN2", RC=STATUS ); VERIFY_(STATUS)
    call MAPL_TimerOff(MAPL,"TOTAL", RC=STATUS ); VERIFY_(STATUS)
    
    RETURN_(ESMF_SUCCESS)

  end subroutine Run2

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

end module GEOS_LandGridCompMod
